#!/usr/bin/perl
#Chris Kline; chris.kline@ca.com
#Copyright CA Technologies, Inc. 2011-2012, All Rights Reserved
#------------------------------------------------------------------------------
#
#TODO:
#Grab SSL info from viewsslservers?
# ---------------------------------------------------------------------------
#
# Script to pull in Performance Stats from TIM on MTP - Stateless EPA Plugin
# This script uses wget to retrieve TIM statistics from MTP (multiprocessor),
# which has different formatting than traditional TIM.  See the config-file
# template for full details on how to set things up. 
# Based on original work by Pat Shueh.
#
# ---------------------------------------------------------------------------
# REVISION HISTORY
# 1.0    XX/XX/10 Initial by Pat Shueh.
# 2.0    11/06/12 Complete rewrite to account for MTP; using new parsing module.
# 2.0.2  12/13/12 Small update for Solaris compatibility. Also automatically finds wget for Windows
#
#-------------------------------------------------------------------------------
# Source Control Information
# $Revision: 67 $
# $LastChangedDate: 2012-12-03 17:17:00 -0700 (Mon, 03 Dec 2012) $
# $Author: cnkline $
#
# ---------------------------------------------------------------------------
# INITIALIZE THE ENVIRONMENT

use warnings;
use strict;

use FindBin;
use lib "$FindBin::Bin/../lib";
use lib "$FindBin::Bin/../lib/public";
use lib "$FindBin::Bin/../lib/public/XS";  #location of XS modules for HTML::Parser and deps

use File::Spec::Functions qw(catfile);
use File::Basename;
use File::Path qw(mkpath);
use HTML::TableParser;  #depends on HTML::Parser, which is XS; used for reading tables
use HTML::TreeBuilder;  #used to parse output HTML

use Kline::Config;
use Kline::Server;
use Kline::Display;
use Kline::Wily;
use Kline::Utils::HTML;

my $title="Wily CEM TIM Monitor EPAgent";
my ($progname,undef,undef)=fileparse($0, qr/\.[^.]*/); #strip the extension
my $defaultconfig="$FindBin::Bin/config/$progname.properties";

my $scriptVer="2.0.2";
($scriptVer .= '$Revision: 67 $') =~ s/\$Revision: (\d+)\s\$/\.$1/; #auto update upon commit
my $scriptDate; #backward compatibility
($scriptDate = '$LastChangedDate: 2012-12-03 17:17:00 -0700 (Mon, 03 Dec 2012) $') =~ s/.*\(\D+(\d+\s\w+\s\d+)\).+/$1/; #auto update upon commit

my @TPtitles; #global TableParser column titles
#----------------------------------------------------------------------------
#FUNCTIONS
#----------------------------------------------------------------------------


#----------------------------------------------
sub loadConfig {
#Load the config file and set all the defaults for unspecified properties
#Receives $1 as optional config file
#----------------------------------------------

  my $file=shift || $defaultconfig;
  my $cfg=&Kline::Config::getConfig($file); #will die if not exist
  $cfg->{file}=$file; #print after the title

  #defaults
  if ($^O=~/mswin/i) {
    $cfg->{wgetlocation}="$FindBin::Bin/prereqs/wget.exe" unless ($cfg->{wgetlocation});
  } else {
    $cfg->{wgetlocation}="/usr/bin/wget" unless ($cfg->{wgetlocation});
  }
  
  $cfg->{loglocation}="$FindBin::Bin/logs" unless ($cfg->{loglocation}); 

  #if only a single TIM, config won't show an array, update it
  unless (ref $cfg->{TIM} eq "ARRAY") {
    &showDebug("Single TIM found; forcing array reference",1);
    $cfg->{TIM}=[$cfg->{TIM}];
  }
  print "Processing " . @{$cfg->{TIM}} . " TIM(s) from $file\n";
  
  foreach my $h (@{$cfg->{TIM}}) { #loop through the TIM stanzas in config file and add in defaults
    die "FATAL: No IP Address found for TIM $h->{TIMname}\n" unless $h->{TIM_IP};
    $h->{TIMname}="TIM_$h->{TIM_IP}" unless ($h->{TIMname}); 
    $h->{user}="admin" unless ($h->{user}); 
    $h->{password}="quality" unless ($h->{password});
    
    if ($h->{active} && $h->{active}=~/false/i) {
      $h->{active}=0;
    } else {
      $h->{active}=1;
    }

    if ($h->{testData} && $h->{testData}!~/false/i) {
      print "TIM $h->{TIMname} will use test data.\n";
      $h->{testData}=1;
      $h->{viewstats}="$FindBin::Bin/$h->{viewstats}" if $h->{viewstats}=~/^\./; #keeps us from working off the wrong starting dir
      $h->{viewstatus}="$FindBin::Bin/$h->{viewstatus}" if $h->{viewstatus}=~/^\./; #keeps us from working off the wrong starting dir
      $h->{watchdog}="$FindBin::Bin/$h->{watchdog}" if $h->{watchdog}=~/^\./; #keeps us from working off the wrong starting dir
    } else {
      $h->{testData}=0;
    }
  }

  &Kline::Display::dumpHref($cfg,2);
  return $cfg;

} #loadConfig

#----------------------------------------------
sub checkTIMStatus() {
#main method for viewstatus; calls getViewstatusHTML which calls parseStatsData  
#----------------------------------------------
#ooo packets from viewstatus (page has worker info but no cpu info)
#	number of workers
#	total packets (since last restart; viewstats shows 
#	If MTP and OOO packets, then problem because napatech card is overrun
#	If std TIM and OOO packets, then it may recover.  Check over time.
  my $h=shift || die "incorrect args: $!";

  #get the metrics
  my $outfile;
  if ($h->{testData}) { #use stored data
    -f $h->{viewstatus} || die "FATAL: Cannot find test viewstatus data at $h->{viewstatus}\n";
    $outfile=$h->{viewstatus};
  } else { #normal
    $outfile=&getViewstatusHTML($h);
  }
  
  &showDebug("Parsing viewstatus output",1);
  #clean up the file (remove HTML)
  my $tree = HTML::TreeBuilder->new();
  open my $fh, '<:encoding(UTF-8)', $outfile || die "Cannot open $outfile: $!";  # auto decoding on read
  $tree->parse_file($fh);
  my $content = $tree->look_down( '_tag' , 'title' );
  die "FATAL: Unable to retrieve metrics; you need to enable access via settings in \"Configure Tim Web Protect Options\" settings\n" if ($content->as_text=~/Web Request Validation/i);
  die "FATAL: Unexpected viewstatus file content found at $outfile\n" unless ($content->as_text=~/Tim Status/i);
  #$content = $tree->look_down( '_tag' , 'pre' );
  #my @d=split(/\n+/,$content->as_text);
  $tree->delete;
  
  #use the new method to parse stats data
  $h->{folder}="Tim Status";
  $h->{outfile}=$outfile;
  #These are the columns we should ignore in the status table
  $h->{ignore}=[
    qr/\s+Status$/,
  ];
  $h->{noclean}=[
    "Updated",
    "feed0",
    
  ];
  &parseStatus_TableParser($h); #will write metrics from viewstats
  undef @TPtitles;
  
  #report where we got the metrics from
  $h->{key}="URL";
  $h->{value}=$h->{url} || "test data only";
  &Kline::Wily::printMetric($h);

  undef @TPtitles;
  &showDebug("Completed processing of viewstatus",1);
  
} #checkTIMStatus

#----------------------------------------------
sub getViewstatusHTML() {
#retrieves the HTML file for viewstatus
#----------------------------------------------
  my $h=shift || die "incorrect args: $!";
  my $wget=$h->{wgetlocation};
  die "FATAL: Cannot find wget at $wget\n" unless -f $wget;
  unless (-d $h->{loglocation}) {
    mkpath $h->{loglocation} || die "FATAL: Cannot create log path $h->{loglocation}: $!\n";
    print "Creating log directory $h->{loglocation}\n";
  }
  my $outfileprefix=&_getOutfilePrefix($h);

  #make sure the post text is clean
  $h->{user}=&Kline::Utils::HTML::encodeURI($h->{user});
  $h->{password}=&Kline::Utils::HTML::encodeURI($h->{password});

  my $outfile="${outfileprefix}_viewstatus.htm";
  $h->{outfile}=$outfile;
  $h->{url}= "\"http://$h->{TIM_IP}/cgi-bin/wily/packages/cem/tim/viewstatus?&unsupported=1&webRequestId=id_\""; #the webRequestId is required for 9.1.x and will not hurt previous versions
  
  #specs for standard TIM 
  $h->{type}="standard" unless $h->{type};
  if ($h->{type}=~/standard/i) {
    $h->{wget_args}="--http-user=\"$h->{user}\" --http-password=\"$h->{password}\" -O \"$outfile\"";
  #specs for MTP
  } elsif ($h->{type}=~/mtp/i) {
    $h->{cookies}=&setMTPSession($h) unless $h->{cookies};  #Login and get a session going if not already present
    $h->{wget_args}="--load-cookies $h->{cookies} -O $outfile";
  } #type
    
  #run the command
  my ($cmd,$rc,$output);
  $cmd={}; #5.8 on Sun gets mad if you don't predeclare.
  $cmd->{command} = "$wget ".$h->{wget_args}." ".$h->{url};

  ($rc,$output)=Kline::Server::cmdLine($cmd);
  $rc=$rc>>8; #shifted
  #&showDebug("Command output: \n@$output\n",1); #not needed

  #watch return codes
  if ($rc==0) { #it's a standalone TIM
    print "Successfully retrieved $outfile\n";
  } elsif ($rc==6) { #bad auth
    die "FATAL AUTHORIZATION ERROR: Unexpected return code $rc on command $cmd->{command}\n@$output\n";
  } else { #something went wrong
    die "FATAL: Unexpected return code $rc on command $cmd->{command}\n@$output\n";
  }
  
  return $outfile;
  
  } #getViewstatusHTML


#----------------------------------------------
sub checkTIMStats() {
#main method for viewstats; calls getViewstatsHTML 
#----------------------------------------------
  my $h=shift || die "incorrect args: $!";

  #get the metrics
  my $outfile;
  if ($h->{testData}) { #use stored data
    -f $h->{viewstats} || die "FATAL: Cannot find test viewstats data at $h->{viewstats}\n";
    $outfile=$h->{viewstats};
  } else { #normal
    $outfile=&getViewstatsHTML($h);
  }
  
  &showDebug("Parsing viewstats output",1);
  #clean up the file (remove HTML)
  my $tree = HTML::TreeBuilder->new();
  open my $fh, '<:encoding(UTF-8)', $outfile || die "Cannot open $outfile: $!";  # auto decoding on read
  $tree->parse_file($fh);
  my $content = $tree->look_down( '_tag' , 'title' );
  die "FATAL: Unable to retrieve metrics; you need to enable access via settings in \"Configure Tim Web Protect Options\" settings\n" if ($content->as_text=~/Web Request Validation/i);
  die "FATAL: Unexpected viewstats file content found at $outfile\n" unless ($content->as_text=~/Tim Packet Statistics/i);
  #$content = $tree->look_down( '_tag' , 'pre' );
  #my @d=split(/\n+/,$content->as_text);
  $tree->delete;
  
  #use the new method to parse stats data
  $h->{folder}="Tim Packet Statistics";
  $h->{outfile}=$outfile;
  &parseStats_TableParser($h); #will write metrics from viewstats
  
  #report where we got the metrics from
  $h->{key}="URL";
  $h->{value}=$h->{url} || "test data only";
  &Kline::Wily::printMetric($h);

  &showDebug("Completed processing of viewstats",1);
  
} #checkTIMStats

#----------------------------------------------
sub getViewstatsHTML() {
#retrieves the HTML file for viewstats
#returns filename
#----------------------------------------------
  my $h=shift || die "incorrect args: $!";
  my $wget=$h->{wgetlocation};
  die "FATAL: Cannot find wget at $wget\n" unless -f $wget;
  unless (-d $h->{loglocation}) {
    mkpath $h->{loglocation} || die "FATAL: Cannot create log path $h->{loglocation}: $!\n";
    print "Creating log directory $h->{loglocation}\n";
  }
  my $outfileprefix=&_getOutfilePrefix($h);

  #make sure the post text is clean
  $h->{user}=&Kline::Utils::HTML::encodeURI($h->{user});
  $h->{password}=&Kline::Utils::HTML::encodeURI($h->{password});

  my $outfile="${outfileprefix}_viewstats.htm";
  $h->{url}= "\"http://$h->{TIM_IP}/cgi-bin/wily/packages/cem/tim/viewstats?file=5minutes&unsupported=1&lines=1&webRequestId=id_\""; #the webRequestId is required for 9.1.x and will not hurt previous versions
  
  #specs for standard TIM 
  $h->{type}="standard" unless $h->{type};
  if ($h->{type}=~/standard/i) {
    $h->{wget_args}="--http-user=\"$h->{user}\" --http-password=\"$h->{password}\" -O \"$outfile\"";
  #specs for MTP
  } elsif ($h->{type}=~/mtp/i) {
    $h->{cookies}=&setMTPSession($h) unless $h->{cookies};  #Login and get a session going if not already present
    $h->{wget_args}="--load-cookies $h->{cookies} -O $outfile";
  } #type
    
  #run the command
  my ($cmd,$rc,$output);
  $cmd={}; #5.8 on Sun gets mad if you don't predeclare.
  $cmd->{command} = "$wget ".$h->{wget_args}." ".$h->{url};

  ($rc,$output)=Kline::Server::cmdLine($cmd);
  $rc=$rc>>8; #shifted
  #&showDebug("Command output: \n@$output\n",1); #not needed

  #watch return codes
  if ($rc==0) { #it's a standalone TIM
    print "Successfully retrieved $outfile\n";
  } elsif ($rc==6) { #bad auth
    die "FATAL AUTHORIZATION ERROR: Unexpected return code $rc on command $cmd->{command}\n@$output\n";
  } else { #something went wrong
    die "FATAL: Unexpected return code $rc on command $cmd->{command}\n@$output\n";
  }
  
  return $outfile;
  
} #getViewstatsHTML

#----------------------------------------------
sub checkWatchdogStats() {
#main method for watchdog; calls getWatchdogHTML which calls parseWatchdogData  
#----------------------------------------------
  my $h=shift || die "incorrect args: $!";
  
  #get the metrics
  my $outfile;
  if ($h->{testData}) { #use stored data
    -f $h->{watchdog} || die "FATAL: Cannot find test watchdog data at $h->{watchdog}\n";
    $outfile=$h->{watchdog};
  } else { #normal
    $outfile=&getWatchdogHTML($h);
  }
  
  #-----
  #regardless of MTP or standard we should now have watchdog output ready to parse
  &showDebug("Parsing watchdog output",1);
  #clean up the file (remove HTML)
  my $tree = HTML::TreeBuilder->new();
  open my $fh, '<:encoding(UTF-8)', $outfile || die "Cannot open $outfile: $!";  # auto decoding on read
  $tree->parse_file($fh);
  my $content = $tree->look_down( '_tag' , 'title' );
  die "FATAL: Unexpected watchdog file content found at $outfile\n" unless ($content->as_text=~/Tim watchdog log/i);
  $content = $tree->look_down( '_tag' , 'pre' );
  my @d=split(/\n+/,$content->as_text);
  $tree->delete;
  my $m=&parseWatchdogData(\@d); #returns hashref of metrics
  
  #-----
  #check if has restarted since the last review
  if ($m->{running}==1) { #OK
    my $outfileprefix=&_getOutfilePrefix($h);
    my $persist="${outfileprefix}_persist";
    if (-f $persist) {
      &showDebug("Comparing last restart time to previous review",1);
      my $last=&Kline::Utils::getFileContents($persist);
      if ($last eq $m->{starttime}) {
        print "TIM has NOT been restarted since last review\n";
	$m->{restart}=0;
      } else {
        print "TIM has been restarted since last review ($m->{starttime})\n";
	$m->{restart}=1;
      }
      $m->{restart_check}=localtime();
    } else {
      &showDebug("No persistence file exists at $persist; no restart check can be completed",1);
    }
  
    #persist the current start time string
    &showDebug("Persisting last TIM start time at $persist",1);
    open PERSIST,">$persist" || die "FATAL: Cannot write persistence file $persist: $!\n";
    print PERSIST $m->{starttime};
    close PERSIST;
  
  } #run
  
  $m->{type}=$h->{type} || "unknown"; #transfer over
  
  #print the metrics
  foreach (sort keys %$m) {
    my $i={};
    $i->{TIMname}=$h->{TIMname}; #move it to the new href
    $i->{metricType}="stringEvent" if /pid/i;
    $i->{folder}="Watchdog" unless $m=~/type/;
    $i->{key}=$_;
    $i->{value}=$m->{$_};
    &Kline::Wily::printMetric($i);
  } #foreach
  
  #report where we got the metrics from
  $h->{folder}="Watchdog";
  $h->{key}="URL";
  $h->{value}=$h->{url} || "test data only";
  &Kline::Wily::printMetric($h);
  
  undef @TPtitles;
  &showDebug("Completed processing of watchdog stats",1);
} #checkWatchdogStats

#----------------------------------------------
sub _getOutfilePrefix() {
#----------------------------------------------
  my $h=shift || die "FATAL: incorrect args!";
  my $outfileprefix="$h->{TIMname}";
  $outfileprefix=~s/\./_/g;
  $outfileprefix="$h->{loglocation}/$outfileprefix";
  return $outfileprefix;
  
} #_getOutfilePrefix

#----------------------------------------------
sub getWatchdogHTML(){
# Retrieves the watchdog HTML but also confirms whether this is standard or MTP TIM (different auth mechanisms)
#is it up (just because :80 is up doesn't mean TIM is running)
#check watchdog log, if get RC=8 then it's a MTP, if get RC=6 bad auth, RC=0 old tim good
#is it old/new tim (remember this)
#Returns the file where the HTML is located
#----------------------------------------------
  my $h=shift || die "incorrect args: $!";
  my $wget=$h->{wgetlocation};
  die "FATAL: Cannot find wget at $wget\n" unless -f $wget;
  unless (-d $h->{loglocation}) {
    mkpath $h->{loglocation} || die "FATAL: Cannot create log path $h->{loglocation}: $!\n";
    print "Creating log directory $h->{loglocation}\n";
  }
  my $outfileprefix=&_getOutfilePrefix($h);

  #make sure the post text is clean
  $h->{user}=&Kline::Utils::HTML::encodeURI($h->{user});
  $h->{password}=&Kline::Utils::HTML::encodeURI($h->{password});

  my $outfile="${outfileprefix}_watchdog.htm";
  my ($cmd,$rc,$output);
  
  #specs for standard TIM (try first)
  $h->{type}="standard" unless $h->{type};
  if ($h->{type}=~/standard/i) {
    $h->{url}= "\"http://$h->{TIM_IP}/cgi-bin/wily/packages/cem/tim/viewfile?file=watchdoglog\"";
    $h->{wget_args}="--http-user=\"$h->{user}\" --http-password=\"$h->{password}\" -O \"$outfile\" --max-redirect=2";

    $cmd={}; #5.8 on Sun gets mad if you don't predeclare.
    $cmd->{command} = "$wget ".$h->{wget_args}." ".$h->{url};
    $cmd->{scalar}=1; #return as scalar string not aref

    ($rc,$output)=Kline::Server::cmdLine($cmd);
    $rc=$rc>>8; #shifted
    #&showDebug("Command output: \n@$output\n",1); #already included 

    #determine what type of MTP and whether we got the file yet
    if ($rc==0) { #it's a standalone TIM
      $h->{type}="standard";
      print "TIM $h->{TIM_IP} is $h->{type} type\n";
      print "Successfully retrieved $outfile\n";
    } elsif ($rc==6) { #bad auth
      die "FATAL AUTHORIZATION ERROR: Unexpected return code $rc on command $cmd->{command}\n@$output\n";
    } elsif ($rc==8) { #max redirects, must be a MTP
      $h->{type}="mtp";
      print "TIM $h->{TIM_IP} is $h->{type} type.\n";
      #scrape the output of wget to pull the login URL
      if ($output=~m<Location: (http://[^/]+)/sso.+ \[following\]\s+2 redirections exceeded>) {
        $h->{login_server}=$1;
	&showDebug("Using SSO Login Server \"$h->{login_server}\"",1);
      } else {
        die "ERROR: Cannot find login URL for MTP\n";
      }
    } else { #something went wrong
      die "FATAL: Unexpected return code $rc on command $cmd->{command}\n@$output\n";
    }
    #&showDebug("Command output: \n@$output\n",1); #not needed
  } #standard

  if ($h->{type}=~/mtp/) {
    #Login and get a session going
    $h->{cookies}=&setMTPSession($h) unless $h->{cookies};
    
    #now retrieve the watchdog file
    &showDebug("Retrieving watchdog file for MTP TIM",1);
    $outfile="${outfileprefix}_watchdog.htm";
    $h->{url}= "\"http://$h->{TIM_IP}/cgi-bin/wily/packages/cem/tim/viewfile?file=watchdoglog\"";
    $h->{wget_args}="--load-cookies=\"$h->{cookies}\" -O \"$outfile\"";

    $cmd->{command} = "$wget ".$h->{wget_args}." ".$h->{url};
    ($rc,$output)=Kline::Server::cmdLine($cmd);
    $rc=$rc>>8; #shifted
    #&showDebug("Command output: \n@$output\n",1); #not needed
    die "FATAL: Unexpected return code $rc on command $cmd->{command}\n@$output\n" if $rc;
    print "Successfully retrieved $outfile\n";
  } #mtp

  return $outfile;
  
} #getWatchdogHTML

#----------------------------------------------
sub setMTPSession() {
#Logs into the MTP and returns a session cookie
#----------------------------------------------
  my $h=shift || die "incorrect args: $!";
  my $wget=$h->{wgetlocation};
  die "FATAL: Cannot find wget at $wget\n" unless -f $wget;
  unless (-d $h->{loglocation}) {
    mkpath $h->{loglocation} || die "FATAL: Cannot create log path $h->{loglocation}: $!\n";
    print "Creating log directory $h->{loglocation}\n";
  }
  my $outfileprefix=&_getOutfilePrefix($h);

  my ($cmd,$rc,$output,$outfile);
  die "FATAL: No MTP Login Server Found" unless $h->{login_server};
  &showDebug("Setting login session cookie for TIM",1);
  $outfile="${outfileprefix}_login.htm";
  my $redirect="http://$h->{TIM_IP}/admin/signin.php";
  $redirect=&Kline::Utils::HTML::encodeURI($redirect);
  $h->{url}= "\"$h->{login_server}/sso/sign-in-process.jsp?txtSsoProductCode=mtp&txtSsoRedirectUrl=$redirect&txtUsername=$h->{user}&txtPassword=$h->{password}\"";
  $h->{cookies}="${outfileprefix}_cookies.txt";
  $h->{wget_args}="--save-cookies=\"$h->{cookies}\" --keep-session-cookies -O \"$outfile\"";

  #run the login
  $cmd->{command} = "$wget ".$h->{wget_args}." ".$h->{url};
  ($rc,$output)=Kline::Server::cmdLine($cmd);
  $rc=$rc>>8; #shifted
  #&showDebug("Command output: \n@$output\n",1); #not needed
  die "FATAL: Unexpected return code $rc on command $cmd->{command}\n@$output\n" if $rc;

  #RC=0 even if wrong password.  Need to scrape for <title>Log In</title> to see if fails
  #success = <title>Multi-Port Collector: Analysis</title>
  $output=&Kline::Utils::getFileContents($outfile);
  if ($output=~m{<TITLE>Multi-Port Collector: Analysis</TITLE>}i) { #yay
    #&showDebug("Successfully logged in to MTP $h->{TIMname}",1);
    print "Successfully logged in to MTP $h->{TIMname}\n";
  } elsif ($output=~m{<TITLE>Log In</TITLE>}i) { #auth error
    die "FATAL AUTHORIZATION ERROR: Unable to login to SSO for MTP at $outfile\n";
  } else {
    die "FATAL: Unexpected TITLE tag in resultant HTML for MTP at $outfile\n";
  }
  
  return $h->{cookies};
} #setMTPSession

#----------------------------------------------
sub parseWatchdogData() {
#Receives copy of the logfile content (cleaned), as aref
#Returns hashref of metrics
#----------------------------------------------
#pick TIM last start time from watchdog log and compare to previous to see if restart
#	grab each section and only pay attn to the final one
#	if last line is "tim pid = #####" then tim is still running
#	if last line is "timwatchdog exiting at DATESTAMP" then tim is down'
#	if down, then next line up will say "tim exited with signal # after ### seconds at DATESTAMP"
#	RETURN PID, Start Time, UP/DOWN, Down Time
  my $a=shift || die "incorrect args: $!";
  my $h={};
  
  foreach (@$a) {
    next if (/^\s+$/); 
    &showDebug("line is \"$_\"",3);
    
    if (/Running tim at (.+)/) { #tim started
      undef $h; #get rid of old sections
      $h->{starttime}=$1;
      $h->{running}++;
      &showDebug("TIM was found running: $_",2);
      next;
    }
    if ($h->{running} && /tim pid = (\d+)/) {
      $h->{pid}=$1;
      next;
    }
    if ($h->{running} && /^(\d+).+tim exited with status (\d+) after (\d+) seconds at (.+)/) { #tim stopped
      ($h->{rc},$h->{seconds},$h->{endtime})=($2,$3,$4);
      $h->{running}=3; #0=no data, 1=ok, 2=caution, 3=alert
      delete $h->{starttime};
      delete $h->{pid};
      &showDebug("TIM was stopped: $_",2);
      next;
    }
  } #foreach
  
  #the last entries should be most current
  if ($h->{running}) {
    print "TIM is presently running with pid $h->{pid}\n";
  } else {
    print "TIM is stopped\n";
  }
  return $h;

} #checkWatchdogData

#----------------------------------------------
sub parseStatus_TableParser() {
#Parses TIM Status Table
#Uses callback functions; see _TP functions
#----------------------------------------------

  my $h=shift || die "FATAL: incorrect args: $!\n";
  my $file=$h->{outfile} || die "FATAL: no file to parse";
  &showDebug("Processing table data",1);
  my @reqs = (
           {
            id => '2.1',                     # table id
            #cols => [ 'Bytes Analyzed',
            #          qr/throughput/ ],     # column name matches
            #obj => $obj,                    # method callbacks
 	    hdr => \&_TPheader,              # function callback
            row => \&_TProw_TIMStatus,                 # function callback
            start => \&_TPstart,             # function callback
            end => \&_TPend,                 # function callback
            udata => $h,		     #user data to pass in whatever I like
           },
          );

  # create parser object
  my $t = HTML::TableParser->new( \@reqs, 
                   { Decode => 1, Trim => 1, Chomp => 1 } );
  open my $fh, '<:encoding(UTF-8)', $file || die "Cannot open $file: $!";  # auto decoding on read
  $t->parse_file( $fh );

} #parseStatus_TableParser

#----------------------------------------------
sub parseStats_TableParser() {
#Parses TIM Stats table
#Uses callback functions; see _TP functions
#----------------------------------------------

  my $h=shift || die "FATAL: incorrect args: $!\n";
  my $file=$h->{outfile} || die "FATAL: no file to parse";
  &showDebug("Processing table data",1);
  my @reqs = (
           {
            id => '2.1',                     # table id
            #cols => [ 'Bytes Analyzed',
            #          qr/throughput/ ],     # column name matches
            #obj => $obj,                    # method callbacks
 	    hdr => \&_TPheader,              # function callback
            row => \&_TProw_TIMStats,                 # function callback
            start => \&_TPstart,             # function callback
            end => \&_TPend,                 # function callback
            udata => $h,		     #user data to pass in whatever I like
           },
          );

  # create parser object
  my $t = HTML::TableParser->new( \@reqs, 
                   { Decode => 1, Trim => 1, Chomp => 1 } );
  open my $fh, '<:encoding(UTF-8)', $file || die "Cannot open $file: $!";  # auto decoding on read
  $t->parse_file( $fh );

} #parseStats_TableParser

### function callbacks (for TableParser)
#----------------------------------------------
sub _TPstart() {
#----------------------------------------------
  my ( $id, $line, $udata ) = @_;
  &showDebug("STARTING table $id, line $line",1);

} #_TPstart

#----------------------------------------------
sub _TPend() {
#----------------------------------------------
  my ( $id, $line, $udata ) = @_;
  &showDebug("ENDING table $id, line $line",1);

} #_TPend

#----------------------------------------------
sub _TPheader() {
#----------------------------------------------
  my ( $id, $line, $cols, $udata ) = @_;
  &showDebug("Found HEADER row in table $id, line $line",1);
  foreach (@$cols) {
    &showDebug("HEADER Column: $_",2);
  }
  @TPtitles=@$cols;
  
} #_TPheader

#----------------------------------------------
sub _TProw_TIMStats()  {
#handles the format of the TIM Packets table
#----------------------------------------------
  my ( $id, $line, $cols, $udata ) = @_;
  &showDebug("Found ROW data in table $id, line $line",1);
  my $i=0;
  my $d={};
  
  #list of column names that should not be cleaned
  my $noclean={}; #must predeclare on Solaris
  $noclean->{"Date/time"}=1;
  $noclean->{"Updated"}=1;
  $noclean->{"feed1"}=1;
  $noclean->{"Total CPU usage"}=1;
  
  foreach (@$cols) {
    &showDebug("ROW $TPtitles[$i]: $_",2);
    $_=&Kline::Wily::cleanMetric($_) unless $noclean->{$TPtitles[$i]}; #ensure it's a properly formatted data point
    $d->{$TPtitles[$i]}="$_"; #load up the output href
    $i++;
  }
  &dumpHref($d,1);

  #send the metrics out
  foreach my $metric (sort keys %$d) {
    my $i=$udata;
    $i->{key}=$metric;
    $i->{value}=$d->{$metric};
    $i->{folder}=$udata->{folder};
    $i->{metricType}=undef;
    &Kline::Wily::printMetric($i);
  }

} #_TProw_TIMStats

#----------------------------------------------
sub _TProw_TIMStatus()  {
#handles the format of the Tim Status table
#----------------------------------------------
  my ( $id, $line, $cols, $h) = @_;
  &showDebug("Found ROW data in table $id, line $line",1);
  if ($h->{ignore}) {
    foreach my $col (@$cols) {
      foreach my $i (@{$h->{ignore}}) {
        if ($col=~/$i/) {
            &showDebug("Skipping table row $col (matches $i)",1);
            return;
        }
      }
    }
  }
  
  my ($metric,$value);
  
  #MTP block
  if ($h->{type}=~/mtp/) {
    #list of column names that should not be cleaned
    my $i=0;
    my $d={};
    my @titles=@TPtitles; #localize it because we're decorating it and will need to trash when we're done
    foreach (@$cols) {
      $titles[$i]=&Kline::Utils::trim($titles[$i]);
      @$cols[0]=&Kline::Utils::trim(@$cols[0]);
      $_=&Kline::Utils::trim($_);
      $titles[$i].=" @$cols[0]";
      if ($i==0 || ($_ eq "")) { #skip empties
	&showDebug("Value for $titles[$i] is empty.  Skipping",1) if $i;
	$i++;
	next;
      }
      &showDebug("Populating $titles[$i]: $_",2);
      $_=&Kline::Wily::cleanMetric($_) unless @$cols[0]=~/Date\/time|Updated|feed\d|Total CPU usage/i; #ensure it's a properly formatted data point
      $d->{$titles[$i]}="$_"; #load up the output href
      $i++;
    }
    &dumpHref($d,1);
 
    #send the metrics out (looped, so has to be separate from the one below)
    foreach my $metric (sort keys %$d) {
      my $i=$h;
      $i->{key}=$metric;
      $i->{value}=$d->{$metric};
      $i->{folder}=$h->{folder};
      $i->{metricType}=undef;
      &Kline::Wily::printMetric($i);
      
    } #foreach

  } else { #standard TIM block
  
    print "WARN: Table row has more than expected two columns: @$cols\n" if (@$cols > 2);
    $metric=@$cols[0];
    $value=@$cols[1];
  
    if ($h->{noclean}) {#ensure it's a properly formatted data point
      my $clean=1;
      foreach (@{$h->{noclean}}) {
	if (/$metric/) {
	  &showDebug("Skipping cleaning metric $metric (matches $_)",1);
	  $clean=0;
	}
      }
      $value=&Kline::Wily::cleanMetric($value) if $clean;
    } else {
      $value=&Kline::Wily::cleanMetric($value);
    }
  
    #send the metrics out
    if (! defined $value) {
     &showDebug("WARN: No value found for metric $metric.  Skipping...",1);
     return;
    }
    $h->{key}=$metric;
    $h->{value}=$value;
    $h->{metricType}=undef;
    &Kline::Wily::printMetric($h);
  
  } #MTP/Std
  

} #_TProw_TIMStatus

#----------------------------------------------
sub logError() {
#----------------------------------------------
  my $errorLogFile = catfile($FindBin::Bin,"logs","$progname.err");
  mkpath(dirname($errorLogFile)) unless (-d dirname($errorLogFile));
  open ERR,">>$errorLogFile" || die "Main program failed; error logging cannot be opened: $!\n";
  print "\n\nFailed with the following error.  See $errorLogFile for detail and history.\n\n$@\n";  
  print ERR "\n\n$progname failed with the following at ". localtime() .":\n@_\n";
  close ERR;
  #die "$@\n";
} #logError

#--------------------------------------------------------------------------
#MAIN PROGRAM; START HERE
#--------------------------------------------------------------------------

eval { #capture any fatal errors from the actual module and log so we'll get notified when run via wrapper

  select STDERR; #EPA needs all logging on STDERR

  &printTitleC("$title ($scriptVer, $scriptDate)");
  my $h=&loadConfig(@ARGV);
  print "Using config file: $h->{file}\n\n";

  foreach my $tim (@{$h->{TIM}}) { #iterate for each TIM
    print localtime() . " Polling $tim->{TIMname} at $tim->{TIM_IP}\n";
    unless ($tim->{active}) {
      print "Skipping TIM $tim->{TIMname}; deactivated\n";
      next;
    }
    
    # Send basic info first
    $tim->{key}="TIM IP";
    $tim->{value}=$tim->{TIM_IP};
    $tim->{metricType}="stringEvent";
    &Kline::Wily::printMetric($tim);
    $tim->{key}="Last Check";
    $tim->{value}=localtime();
    &Kline::Wily::printMetric($tim);
    
    $tim->{loglocation}=$h->{loglocation}; #nicer to just have it in the local context
    $tim->{wgetlocation}=$h->{wgetlocation}; #nicer to just have it in the local context
    
    #watchdog (up/down) data
    #-----
    eval { #capture if dies so we can report status
      print "Gathering watchdog status details for $tim->{TIMname}\n";
      &checkWatchdogStats($tim); #will write metrics from watchdog
      $tim->{key}="status";
      $tim->{value}=1; #0=no data, 1=ok, 2=caution, 3=alert
      $tim->{metricType}="intCounter";
      $tim->{folder}="Watchdog";
      &Kline::Wily::printMetric($tim);
    }; #eval
    if ($@) {
      my $text=$@;
      &showDebug("Failed to retrieve watchdog status; reporting to EPA",1);
      &logError($text) ;
      #print $@;
      $tim->{key}="status";
      $tim->{value}=0; #0=no data, 1=ok, 2=caution, 3=alert
      $tim->{metricType}="intCounter";
      $tim->{folder}="Watchdog";
      &Kline::Wily::printMetric($tim);
      #$tim->{key}="Error Text"; #didn't work due to crummy characters even with encoding
      #$tim->{value}=&Kline::Utils::HTML::encodeURI($text);
      #&Kline::Wily::printMetric($tim);
      next; #next TIM
    }

    #pick primary stats data from viewstats
    #-----
    eval { #capture if dies so we can report status
      print "Gathering viewstats details for $tim->{TIMname}\n";
      &checkTIMStats($tim); #scrapes the viewstats file
      $tim->{key}="status";
      $tim->{value}=1; #0=no data, 1=ok, 2=caution, 3=alert
      $tim->{metricType}="intCounter";
      $tim->{folder}="Tim Packet Statistics";
      &Kline::Wily::printMetric($tim);
    }; #eval
     if ($@) {
      my $text=$@;
      &showDebug("Failed to retrieve viewstats; reporting to EPA",1);
      &logError($text) ;
      #print $@;
      $tim->{key}="status";
      $tim->{value}=0; #0=no data, 1=ok, 2=caution, 3=alert
      $tim->{metricType}="intCounter";
      $tim->{folder}="Tim Packet Statistics";
      &Kline::Wily::printMetric($tim);
      #$tim->{key}="Error Text";  #didn't work due to crummy characters even with encoding
      #$tim->{value}=&Kline::Utils::HTML::encodeURI($text);
      #&Kline::Wily::printMetric($tim);
      next; #next TIM
    }

    #supplementary status data (ooo packets, etc.)
    #-----
    eval { #capture if dies so we can report status
      print "Gathering viewstatus details for $tim->{TIMname}\n";
      &checkTIMStatus($tim); #scrapes the viewstats file
      $tim->{key}="status";
      $tim->{value}=1; #0=no data, 1=ok, 2=caution, 3=alert
      $tim->{metricType}="intCounter";
      $tim->{folder}="Tim Status";
      &Kline::Wily::printMetric($tim);
    }; #eval
     if ($@) {
      my $text=$@;
      &showDebug("Failed to retrieve viewstatus; reporting to EPA",1);
      &logError($text) ;
      $tim->{key}="status";
      $tim->{value}=0; #0=no data, 1=ok, 2=caution, 3=alert
      $tim->{metricType}="intCounter";
      $tim->{folder}="Tim Status";
      &Kline::Wily::printMetric($tim);
      #$tim->{key}="Error Text"; #didn't work due to crummy characters even with encoding
      #$tim->{value}=&Kline::Utils::HTML::encodeURI($text);
      #&Kline::Wily::printMetric($tim);
      next; #next TIM
    }
    print "$tim->{TIMname} done\n";
  } #loop tim

}; #eval

if ($@) {
  &logError($@) ; #main eval
}

1;
