#!/usr/bin/perl
package Kline::All;
# Chris Kline; cnkline@us.ibm.com
# Copyright IBM Corp. 2003,2008 All Rights Reserved
# ---------------------------------------------------------------------------
#
# Kline all-import module, which simply imports all other Kline modules and
# makes their functions available (via a flat namespace).  This module is for
# backwards compatibility.  New Kline programs should explicitly import
# the correct Kline module and not rely on bulk import as below.
#
# ---------------------------------------------------------------------------
# REVISION HISTORY
# 2.0   -BT50563: Split Kline.pl according to genre.  First ed of this file.
#       Removed noStanzaErr and instead added optional 4th parm to pass so no
#       global var is needed. Removed toolsConfigGetValue because it was
#       redundant.  Added 5th arg to getPropertyValue.  Added dumpHref func.
# 2.0.1 -Fixed checkConfigFile to work with Windows (since we are using this
#       for health checks
# 2.1   -BT55480, loadCFG is not persisting and reloads everything each time
#       Appears previous work persistent getPropertyValue.  Moved that to
#       load CFG
# 2.1.1 -Updated updateToolsVersion to call loadCFG with 1 so that it keeps
#       case intact (the toolsVersionKSH was failing cuz of case mixup)
# 2.1.2 -Updated showInstalledPatches to match with Kline.setup 3.0.2
#       -changed stat cmd in getScriptVersions to pull from common fileStat
# 2.1.3 -BT59118: toolsVersion debug statements to help know where stuff goes
#       Updated to handle pre-8.5 versions (2nd edition)
#       Fixed loadCFG to allow for $3 to force refresh of requested info
# 2.1.4 BT59674: Updated loadCFG to work with older PERLs
# 2.2   Added missing skipPattern to getScriptVersions so pkg make works right
#       with svn filter, now follows links; general cleanup to allow subs used
#       during package creation to work within svn repos
# 2.2.1 BT63578 for querying installed patches
# 2.2.2 toolsVersion now returns a third element in array, instDate
# 3.0   BT66191; redo into a true perl module, formerly base.pl
# 3.1   12/26/07 Removed dependency on KlineDIR env var and just look in 
#       ../lib/Kline dir compared to the current binary
# 3.1.1 02/06/08 Added POD
#
# ---------------------------------------------------------------------------
# INITIALIZE THE ENVIRONMENT
#
#* * * * * * * * * * * *
# This file just sources in all the other files and is for backwards compat
#* * * * * * * * * * * *

use strict;
require Exporter;
our @ISA=qw(Exporter);
our @EXPORT;

my $scriptVer="3.1.1"; #backwads compat
my $scriptDate; #backward compatibility
($scriptDate = '$LastChangedDate: 2012-01-10 20:43:31 -0700 (Tue, 10 Jan 2012) $') =~ s/.*\(\D+(\d+\s\w+\s\d+)\).+/$1/; #auto update upon commit
our $VERSION=$scriptVer;
our $VERDATE=$scriptDate;

=head1 NAME

Kline::All - Import all Kline legacy functions.

=head1 SYNOPSIS

 use Kline::All;
 trim($foobar);		#comes from Kline::Utils and is imported

=head1 DESCRIPTION

This module is created for backward compatibility only and will import all exported symbols from C<<Kline::<x>> modules.  See the export lists of each of those modules for details.  Newly written subs are no longer added to the export list in order to make code easier to read (will full package name on each reference).


This module has no subs.

To see a list of all imported symbols, use this module with the C<DEBUG> environment variable set.

Use DEBUG=1 to see what modules are being imported (and from which dir).

Use DEBUG=2 to see a list of all the subs that are exported at end.

Use DEBUG=3 to see a list of all subs each module is exporting.

Uncomment the Exporter::Verbose line in this module to see what is actually imported.

=cut

#BEGIN { $Exporter::Verbose=1 } #DEBUG ONLY

BEGIN {
use File::Basename qw(fileparse);	#D:\perl_oreilly\prog\ch07_027.htm
use FindBin;
use Cwd qw(abs_path);
my $toolsdir=abs_path("$FindBin::Bin/../lib");
die "Cannot find tools directory!" unless $toolsdir;
eval "use lib \"$toolsdir\"";
  
select STDERR;

if (exists $ENV{DEBUG} && $ENV{DEBUG} >= 1) {
  print "Using Kline directory $toolsdir\n"; 
}

my $count=0;
foreach my $x (glob("$toolsdir/Kline/*.pm")) {
  my ($file,$dir,$ext)=fileparse($x, qr/\.[^.[^.]*/);
  next if $file=~/All/;
  $count++;
  eval "use Kline::$file qw(:DEFAULT)";
  eval 'push( @EXPORT, @Kline::' . $file . '::EXPORT)';
  
  if (exists $ENV{DEBUG} && $ENV{DEBUG} >= 1) {
    print "IMPORTING Kline::$file\n"; 
    eval ' print "  NOTICE: NO EXPORTS FOR Kline::$file\n" unless @Kline::'. $file .'::EXPORT';
  }
  if (exists $ENV{DEBUG} && $ENV{DEBUG} >= 3) {
    print "\nKline::$file EXPORT:\n"; 
    eval 'print join(", ", @Kline::' . $file . '::EXPORT)'; 
    print "\nMAIN EXPORT now:\n"; 
    print join ", ",sort @EXPORT; 
    print "\n";
  }
} #end foreach

if (exists $ENV{DEBUG} && $ENV{DEBUG} == 2) {
  print "\nFINAL LIST OF IMPORTED FUNCTIONS:\n"; 
  print join "\n",sort @EXPORT; 
  print "\n";
}

die "Could not find any Kline modules at $toolsdir!\n" unless $count;

} #BEGIN

#----------------------------------------------
1; #tell the sourcing program we are true

__END__

#ADD'L DOCUMENTATION-----------------------------------------------------------

=head1 AUTHOR

Chris Kline <cnkline@us.ibm.com>

=head1 SEE ALSO

N/A.

=head1 COPYRIGHT

(C) Copyright IBM Corp. 2003,2008 All Rights Reserved

=cut

