#!/usr/bin/perl
package Kline::Display;
# Display.pm
# Chris Kline; Chris.Kline@ca.com
# ---------------------------------------------------------------------------
#
# This script is a library of commonly used functions for Kline programs that
# use Perl.  It can be "sourced in" using the "require" pragma.
# This library contains only functions that perform display-related activities.
# These functions may call functions from other libraries.
#
# ---------------------------------------------------------------------------
# REVISION HISTORY
# 2.0   -BT50563: Split Kline.pl according to genre.  First ed of this file.
# 2.0.1 Updated printSection to allow for wantarray
# 2.0.2 -BT63311: Was improperly selecting STDIN, breaking installer
# 3.0   -BT66191 to move to a true PERL module
# 3.0.1  12/07/07 Updated dumpHref to work with AoH refs
# 3.0.2  04/22/08 Update for dumpHref when the ref is empty
# 3.1    07/02/08 Added trimString
# 3.1.1  12/05/11 Fixed printFunctionNames;
# 3.1.2  11/13/12 Fixed dumpHref because it was showing other hrefs
#
#-------------------------------------------------------------------------------
# Source Control Information
# $Revision: 55 $
# $LastChangedDate: 2012-11-13 14:58:28 -0700 (Tue, 13 Nov 2012) $
# $Author: cnkline $
#
# ---------------------------------------------------------------------------
# INITIALIZE THE ENVIRONMENT

use strict;
use English; 		#D:\perl_oreilly\prog\ch07_013.htm
require Exporter;
our @ISA=qw(Exporter);
our @EXPORT=qw(clear drawLine pressEnter printArrowSection printError printItem printItemCR printSection printTitle printTitleC dumpHref showDebug printFunctionNames);

use File::Basename qw(basename);
use Text::Wrap qw(wrap);
use Data::Dumper;

use lib '..'; #so we can run these direct to make sure they compile and not fail for missing mods
use Kline::Utils;

my $scriptVer="3.1.2";
my $scriptDate; #backward compatibility
($scriptDate = '$LastChangedDate: 2012-11-13 14:58:28 -0700 (Tue, 13 Nov 2012) $') =~ s/.*\(\D+(\d+\s\w+\s\d+)\).+/$1/; #auto update upon commit

$Text::Wrap::columns=80 unless $Text::Wrap::columns; #D:/perl_oreilly/prog/ch07_071.htm

#----------------------------------------------------------------------------

our $debug=$ENV{DEBUG} || 0;     #debug level for showing errors; 0 is off

#----------------------------------------------
sub dumpHref {
#receive any href and dump it for debug help
#Optionally receives debug value at which point data should be printed
#Also works with AoH references now
#If the ref is undef, returns quietly
#----------------------------------------------

  my $ref=shift(@_) || &showDebug("CALLED INCORRECTLY!",0);
  my $this=&Kline::Utils::thisFunc();
  my $parent=&Kline::Utils::parentFunc();
  
  my $level=shift(@_) || 0; #show dump only if debug is this or higher
  if ($debug ge $level) { #proceed
    &showDebug("HASH DUMPING ENABLED AT DEBUG LEVEL $level",$level);
  } else { #debug not high enough
    &showDebug("DEBUG LEVEL NOT HIGH ENOUGH ($debug) TO ENABLE DUMPING ($level)!",$debug) if $debug;
    return;
  }
  
  #if it's an array, assume elements are hashes and loop
  my @a; #list of hashes to dump
  return unless defined $ref; #if undef then just return quietly so we don't get "expected ref, not ''" messages
  if (ref($ref) eq "ARRAY") {
    &showDebug("Assume array ref hash hash elements; looping",2);
    $a=$ref; #send it straight over
  } elsif (ref($ref) eq "HASH") { #stick into a single item array
    push @{$a},$ref; 
  } else { #not array and not loop, die
    die "$this expected a hash or array reference from $parent, not \"$ref\"\n";
  }
  
  my $count;
  foreach my $href (@{$a} ) {
    $count++;
    if (ref($href) ne "HASH") { #make sure the passed val is a href
      die "$this expected a hash reference from $parent, not \"$href\"\n";
    }
  
    &showDebug("HASH DUMP #$count FOLLOWS:",$level);
    #this will be unsorted output, but is way faster than recursing
    print STDERR Dumper(\$href);
  } #end loop

} #end dumpHref

#----------------------------------------------
sub showDebug {
#print DEBUG statements to STDERR if DEBUG variable is on
#Shows debug level followed by current function and line number.
#Then shows caller's parent line number and file, if other than Kline.pl
#If Kline.pl, then the caller's parent function is shown
#STDERR statements are for building this function only
#----------------------------------------------
  my ($statement,$level)=@_;
  return unless ($debug && $debug >= $level); #for performance, don't do anything if no debug
  my $parent_func; #function of the caller's parent, if any
  my ($calling_package,$parent_file,$parent_line,$calling_func,$calling_hasargs,$calling_wantarray,$calling_evaltext,$calling_is_required,$calling_hints,$calling_bitmask)=&Kline::Utils::parentFunc();
  #print STDERR "\n  calling_package: $calling_package\n  parent_file: $parent_file\n  parent_line: $parent_line\n  calling_func: $calling_func\n  calling_hasargs: $calling_hasargs\n  calling_wantarray: $calling_wantarray\n  calling_evaltext: $calling_evaltext\n  calling_is_required: $calling_is_required\n  calling_hints: $calling_hints\n  calling_bitmask: $calling_bitmask\n";
  $calling_func=$1 if ($calling_func && $calling_func =~ /::(.+)/); #trim calling func name if main pkg
  if ($parent_file) { $parent_file=basename($parent_file) } else { $parent_file="" }
  if ($parent_file eq "Kline.pl") { #only show filename if other than Kline.pl
    $parent_func=(caller(2))[3];
    $parent_func=$1 if ($parent_func && $parent_func =~ /::(.+)/); #trim caller's parent func name if main pkg
    #print STDERR "  parent func: $parent_func\n";
    if ($parent_func) { #parent is Kline.pl and has a func name, so show that
      $parent_file="$parent_func ";
    } else { #parent is Kline.pl but no parent func, so show nuttin
      $parent_file="";
    }
  } else {  #calling file isn't Kline.pl, show it
    $parent_file.=" ";
  }
  $level=-1 unless ($level); #don't let it be undef if I forgot to pass
  my ($package,$calling_file,$calling_line)=&Kline::Utils::thisFunc(); #get the line number of calling func
  #print STDERR "  package: $package\n  calling_file: $calling_file\n  calling_line: $calling_line\n";
  $calling_func=basename("$calling_file") unless $calling_func; #show the file calling debug if no parent function
  $parent_file="noparent" unless ($parent_line); #
  $parent_line="" unless $parent_line; #will be empty when called from direct file (no functions or parent)
  print STDERR "DEBUG($level)--$calling_func $calling_line ($parent_file$parent_line)--$statement\n";

} #end showDebug

#----------------------------------------------
sub printFunctionNames {
#receive a filename and return sorted list of all functions for either perl or ksh
#line numbers are included in output.
#----------------------------------------------
  my ($filename)=@_;
  my $type;
  my @result;
  my @contents=&Kline::Utils::getFileContents($filename);
  &showDebug("File has " . @contents . " lines.\n",1);
  foreach (@contents) {
    $type=$1 if (m{^#\!\S*/bin/(ksh|perl)});
    $type="perl" if (/^package/);
    last if $type;
  }
  unless ($type) {
    warn "$filename is not a perl or ksh file\n";
    return 0;
  }
  &showDebug("$filename is $type\n",1);
  my $line;
  foreach (@contents) {
    chomp;
    $line++;
    #if ((/\(\)\s*{/ && $type eq "ksh") || (/^sub / && $type eq "perl")) {
    my $function;
    if ($type=~/perl/ && /^sub (\w+).+\{/) {
      $function=$1;
      &showDebug("perl function: $_\n",2);
    }
    if ($type=~/ksh/ && /^(\w+)\(\) \{/) {
      $function=$1;
      &showDebug("ksh function: $_\n",2);
    }

    if ($function) {
      &showDebug("$type $function ($line)\n",1);
      push @result,"$function ($line)\n";
    }
  }  #foreach line

  if (@result) {#functions found
    @result=sort { lc($a) cmp lc($b) } @result; #sort alphabetical
    foreach (@result) { print }
  } else {
    print "No functions found in this file.\n";
  }

} #end printFunctionNames

#----------------------------------------------
sub drawLine {
# Draw a line the length of the screen
# If pass any arg, then return the val as an array
#----------------------------------------------
  my $style=shift @_ || "-";
  my $length=$Text::Wrap::columns;
  my $count=0;
  my @output;
  if ($style=~/num/i) { #print line markers
    $style="-";
    push @output,"1234567890" x 8;
    push @output,"\n";
  } #continue normal

  while ($count < $length) {
    push @output, "$style";
    $count ++;
  }
  push @output,"\n";
  if (wantarray()) {
    return @output ;
  } else { #ws called in scalar context, which is likely kshtoperl
    print @output;
    return 0;
  }
	
} #end drawLine


#----------------------------------------------
sub printTitle {
# Print a Title bar
#----------------------------------------------
  &clear();
  drawLine "-";
  print "@_\n";
  drawLine "-";
  print "\n";

} #end printTitle

#----------------------------------------------
sub clear {
#clear the screen
#----------------------------------------------
  system("clear") unless ($OSNAME=~/win/i);

} #end clear

#----------------------------------------------
sub printTitleC {
# Print a Title bar, same as above, but with copyright 
#----------------------------------------------
  &clear();
  drawLine "-";
  print "@_\n";
  print "Copyright CA, Inc. 2008-2012 All Rights Reserved.\n";
  drawLine "-";
  print "\n";

} #end printTitleC

#----------------------------------------------
sub printArrowSection {
#simply print an section header with a standard looking arrow
#----------------------------------------------
	print "\n";
	print "==> @_\n";

} #end printSection

#----------------------------------------------
sub printSection {
#write a new section entry to the STDOUT
#----------------------------------------------
  my $sectionChar="=";
  my $totalLength=$Text::Wrap::columns;
  &showDebug("section length is $totalLength",2);
  my $section=shift @_;
  my $length=length($section)+2 ;
  my $position=0;;
  my @out;
  push @out,"\n$sectionChar$sectionChar$section";
  my $line;
  while ($position < ($totalLength - $length)) {
    $line.=$sectionChar;
    $position++
  }
  push @out,"$line\n" if $line;
  
  if (wantarray) {
    return @out;
  } else {
    print @out;
    return 0;
  }

} #end printSection

#----------------------------------------------
sub printItem {
#simply print an item with a standard look
#this is used by mngmqlogs and should be any longer than it is now or formatting
#will get messed up.
#----------------------------------------------
	print "  @_\n";

} #end printItem

#----------------------------------------------
sub printItemCR {
#same as previous but adds a CR before the item
#----------------------------------------------

	print "\n";
	&printItem(@_);

} #end printItemCR

#----------------------------------------------
sub pressEnter {
# pause the screen until the user hits enter
# will append any parms to the output, if passed
#----------------------------------------------
  select STDERR;
  print "\nPress <ENTER> to continue";
  print @_ if (@_);
  print "\n";
  <STDIN>;
  #the below was added as part of SVN revision 129 but caused lots of things to
  #break.  I can't even tell why it was added, so commented back in rev 328. -ck
  #select STDIN;

} #end pressEnter

#----------------------------------------------
sub printError {
#receive an array of strings and print them under an error title
#----------------------------------------------
  select STDERR;
  printTitle("Kline Error");
  my $parent=&parentFunc;
  &showDebug("no error output provided by $parent",1) unless @_;
  foreach (@_) { print }
  select STDOUT;

} #end printError

#----------------------------------------------
sub trimString {
#Receive a string ($1) and a length ($2) and truncate, replacing last 3 chars with ... if needed
#----------------------------------------------
  my $string=shift;
  my $max=shift;
  return $string unless $max;
  my $trim=length($string) - $max;
  $trim=0 unless $trim > 0;
  my $display=$string; #display format of name, in case it needs to be truncated
  if ($trim) {
    $display=substr($display, 0,-1 * $trim);
    $display=~s/(...)$/\.\.\./; #now replace last 3 chars with ... to show we truncated it
    &Kline::Display::showDebug("Trimming $trim chars (max $max) from $string to yield $display for display",1);
  } #end trim
  return $display;

} #trimString

#----------------------------------------------
1; #tell the sourcing program we are true
