#!/usr/bin/perl
package Kline::Message;
#Chris Kline; Chris.Kline@ca.com
#
#-------------------------------------------------------------------------------
#REVISION HISTORY
#
# 1.0   07/28/08 Initial; taken from Lucas Theisen's Notification::Message mod.
#
#SOURCE-CONTROL INFORMATION-----------------------------------------------------
#
# $Revision: 22 $
# $LastChangedDate: 2012-01-10 20:43:31 -0700 (Tue, 10 Jan 2012) $
# $Author: cnkline $
#
# ---------------------------------------------------------------------------
# INITIALIZE THE ENVIRONMENT

use strict;
use warnings;

use English;
use Sys::Hostname;

BEGIN { #if it croaks due to missing sendmail, we want to move on since we may be using smtp
  eval {
    require MIME::Lite;
    import MIME::Lite;
  };
  if ($@) {
    warn "\n\n". "*" x 70 ."\nWARNING: MIME::Lite did not load properly:\n\n$@\n\nMail features may fail!  Continuing in 5 seconds...\n". "*" x 70 ."\n\n";
    sleep 5;
  }
};

BEGIN { #only load windows mods if we expect to need them
  if ($^O=~/win/i) {
    require Win32;
    import Win32;
  }
};

my $WINDOWS = ($^O=~/win/i);

# ---------------------------------------------------------------------------
=head1 NAME

Kline::Message - Wrapper to MIME::Lite

=head1 SYNOPSIS

 use Kline::Message;
 Kline::Message->new(
   to => \@to_addr,
   from => $from_addr,
   cc => \@cc_addr,
   bcc => \@bcc_addr,
   subject => "something",
   bodyText => "your data here",
   method => "smtp", #or "sendmail"
 )->send;

=head1 DESCRIPTION

This module provides a wrapper onto the CPAN (pure-perl) MIME::Lite module, which can send email using the Net::SMTMP (builtin) where there is not sendmail installed.  It works on both Windows and UNIX.

=cut
# ---------------------------------------------------------------------------
# GLOBALS

(our $VERDATE = '$LastChangedDate: 2012-01-10 20:43:31 -0700 (Tue, 10 Jan 2012) $') =~ s/.*\(\D+(\d+\s\w+\s\d+)\).+/$1/; #auto update upon commit

# ---------------------------------------------------------------------------

#---------------------------------
sub new {
#Originally designed with the intent of pulling a large list of addresses from a
#database, so it ensures that people are only sent the message once.  Now ensures
#no required items are missing and allows for SMTP or sendmail, and wraps to
#MIME::Lite.  Willl use user@hostname if no {from} addr is passed
#
#Takes following arguments in a hash:
#{from} Optional.  Default to user@hostname if null.
#{to} One of these is required.
#{cc}
#{bcc}
#{subject} Optional
#{bodyText}  #one of these is required.  This is acual data
#{dataFile}  #this is data in a file 
#{method}.  Optional.  (sendmail or smtp). Defaults to sendmail if undef
#{smtpRelays} Required if method = smtp.  Aref of SMTP relays, one of which will be randomly chosen
#{smtpTimeout}. Optional.  Defaults to 15 sec
#{smtpUser}. Optional
#{smtpPassword}. Optional.
#{sendmailProgram}. Optional to force use of specific prog;  MIME::Lite uses /usr/lib/sendmail, /usr/sbin, or search on PATH by default
#{type} Optional.  Tries to determine HTML vs Text if null.  Defaults to text/plain
#{debugAddr} Optional.  Overrides the destination addrs and sends the mail here for testing
#{options} Optional.  Passed in properly, but not yet implemented within.
#
#TODO: Add support for attachments, which MIME::Lite already supports fully
#---------------------------------
  my $prototype = shift;
  my $class = ref( $prototype ) || $prototype;
  my $self = { };
  bless( $self, $class );

  my $arguments = shift;

  # ensure that each recipient is in only one of the three recipient categories
  my @toArray = ( );
  my @ccArray = ( );
  my @bccArray = ( );
  my %recipientHash = ( );
  foreach my $element ( (ref $arguments->{to} eq "ARRAY") ? @{$arguments->{to}} : ($arguments->{to}) || "") {
    $recipientHash{$element} = 1;
  }
  foreach my $element ( (ref $arguments->{cc} eq "ARRAY") ? @{$arguments->{cc}} : ($arguments->{cc}) || "") {
    $recipientHash{$element} |= 2;
  }
  foreach my $element ( (ref $arguments->{bcc} eq "ARRAY") ? @{$arguments->{bcc}} : ($arguments->{bcc} || "") ) {
    $recipientHash{$element} |= 4;
  }
  foreach my $element ( sort { $a cmp $b } keys %recipientHash ) {
    if ( $recipientHash{$element} & 1 ) {
      push( @toArray, $element );
    }
    elsif ( $recipientHash{$element} & 2 ) {
      push( @ccArray, $element );
    }
    elsif ( $recipientHash{$element} & 4 ) {
      push( @bccArray, $element );
    }
  }

  # load the data from a file
  if ( $arguments->{dataFile} ) {
    open( DATAFILE, $arguments->{dataFile} );
    $arguments->{bodyText} = join( '', <DATAFILE> );
    close( DATAFILE );
  }

  $self->{from} = $arguments->{from} if $arguments->{from};
  $self->{to} = \@toArray;
  $self->{cc} = \@ccArray;
  $self->{bcc} = \@bccArray;
  unless ($self->{from}) {
    my $user=($WINDOWS) ? &Win32::LoginName : getpwuid($EUID);
    $self->{from} = "$user\@". hostname();
  }
  
  unless (@{$self->{to}} || @{$self->{cc}} || @{$self->{bcc}}) {
    die "ERROR: No destination addresses were specified!\n";
  }
    
  $self->{subject} = $arguments->{subject} || "";
  $self->{type} = $arguments->{type} || "text/plain";
  $self->{bodyText} = $arguments->{bodyText} || "";
  $self->{method}=$arguments->{method} || "sendmail";
  $self->{debugAddr} = $arguments->{debugAddr} || undef;
  
  if ($self->{method}=~/sendmail/i) {
    &Kline::Display::showDebug("Setting mail-send default using $self->{method} method",1);
    my $prog = $arguments->{sendmailProgram} || undef;
    my $origprog=$prog;
    if ($prog) {
      $prog=$1 if ($prog=~/(^[^\s-]+)/); #strip off any args that were included
      if (! -x $prog) {
        die "ERROR: Using sendmail as transfer method but the manual location at $prog is not found!\n";
      }
      die "No sendmail program was specified and no default exists on Windows!  Try the SMTP method instead.\n" if ($WINDOWS && ! $prog);
      if ($origprog) {
        $origprog.=" -t -oi -oem" unless $prog=~/\s-/; #add the default switches if none already there
        &Kline::Display::showDebug("Mail will be sent using manual sendmail at $origprog",1);
        MIME::Lite->send("sendmail","$origprog"); #set the send default, but only if we have args to provide it.  Otherwise, sendmail is already the default on unix and doesn't need this, and on win we'll need args anyway.
      }
    }
  
  } elsif ($self->{method}=~/smtp/i) {
    &Kline::Display::showDebug("Setting mail-send default using $self->{method} method",1);
    #check for required parms
    $self->{smtpTimeout} = $arguments->{smtpTimeout} || 15;
    $self->{smtpUser} = $arguments->{smtpUser} || undef;
    $self->{smtpPassword} = $arguments->{smtpPassword} || undef;
    $self->{smtpRelays} = $arguments->{smtpRelays} ;
    die "FATAL: No SMTP relays were specified, but the send method is SMTP.\n" unless $self->{smtpRelays};
    my $args={
            Debug=>$ENV{DEBUG},
           };
    $args->{Timeout} = $self->{smtpTimeout} if $self->{smtpTimeout};
    if ($self->{smtpUser}) {
      &Kline::Display::showDebug("Using SMTP authentication user $self->{smtpUser}",1);
      $args->{AuthUser} = $self->{smtpUser};
      if ($self->{smtpPassword}) {
        &Kline::Display::showDebug("Using SMTP authentication password $self->{smtpUser}",1);
        $args->{AuthPass} = $self->{smtpPassword};
      }
    } else {
      &Kline::Display::showDebug("Using no SMTP authentication",1);
    }
    $self->{howargs}=$args;
    MIME::Lite->send("smtp",$args); #set the send default
    
  } else {
    die "Invalid transfer method \"$self->{method}\" was specified\n";
  }
  
  if ( $self->{debugAddr} ) {
    print STDERR "\nDEBUG OVERRIDE; SENDING MAIL TO $self->{debugAddr}\n";
    $self->{msg} = MIME::Lite->new(
          From => $self->{from},
          To => $self->{debugAddr},
          Subject => $self->{subject},
          Type => $self->{type},
          Data => "DEBUG MODE DUMPED MESSAGE\n\nto: ".
            join( ',', @{$self->{to}} ) ."\ncc: ".
            join( ',', @{$self->{cc}} ) ."\nbcc: ".
            join( ',', @{$self->{bcc}} ) ."\nbody: \n$self->{bodyText}");
  }
  else {
        $self->{msg} = MIME::Lite->new(
          From => $self->{from},
          To => join( ',', @{$self->{to}} ),
          Cc => join( ',', @{$self->{cc}} ),
          Bcc => join( ',', @{$self->{bcc}} ),
          Subject => $self->{subject},
          Type => $self->{type},
          Data => $self->{bodyText} );
  }
  
  return $self;
} #new

#---------------------------------
sub send {
#Sends mail according to the setup in new()
#If DEBUG env var is 2+, mail is printed to screen and not really sent
#If smtp, loop through relays until we find one that works
#Does not currenly support attachments
#---------------------------------
  my $self = shift;

  if ($self->{method}=~/sendmail/i) {
    if ( $ENV{DEBUG} && $ENV{DEBUG} >= 2 ) {
      print STDERR "Mail not actually sending\n";
      print STDERR ( $self->toString( ) );
    }
    else {
      eval {
        $self->{msg}->send;
      };
      if ( $@ || ! $self->{msg}->last_send_successful ) {
        die "sendmail error - $@\n";
      }
      else {
        return 1;
      }
    }
  }
  
  #if we get here, then it's smtp, so try until we get a valid relay
  # try the relays in random order
  my $error   = '';
  my $success = 0;
  foreach my $relay ( sort { rand() <=> rand() } @{$self->{smtpRelays}} ) {
    &Kline::Display::showDebug("Attempting send using SMTP relay $relay",1);
    eval {
      if ( $ENV{DEBUG} && $ENV{DEBUG} >= 2 ) {
        print STDERR "Mail not actually sending\n";
        print STDERR ( $self->toString( ) );
      }
      else {
        $self->{msg}->send_by_smtp( $relay, $self->{howargs} );
      }
    };
    
    if ( $@ || ! $self->{msg}->last_send_successful ) {
      # failed to send via this relay, try another.
      $error .= "$relay - $@\n";
    }
    else {
      $success = 1;
      last;
    }
  }

  if ( ! $success ) {
    die( $error );
  }
  
} #send

#---------------------------------
sub toString {
  #for debug
#---------------------------------
    my $self = shift;
    return "To: " . ( defined( $self->{to} ) ? join( ',', @{$self->{to}} ) : "" ) . "\n" .
           "Cc: " . ( defined( $self->{cc} ) ? join( ',', @{$self->{cc}} ) : "" ) . "\n" .
           "Bcc: " . ( defined( $self->{bcc} ) ? join( ',', @{$self->{bcc}} ) : "" ) . "\n" .
           "From: " . ( defined( $self->{from} ) ? $self->{from} : "" ) . "\n" .
           "Subject: " . ( defined( $self->{subject} ) ? $self->{subject} : "" ) . "\n" .
           "Type: " . ( defined( $self->{type} ) ? $self->{type} : "" ) . "\n" .
           "Body:\n" . ( defined( $self->{bodyText} ) ? $self->{bodyText} : "" ) . "\n";
} #toString

#---------------------------------
# need to return 1 for use statement
1;
