#!/usr/bin/perl
package Kline::Server;
# Server.pm
# Chris Kline; Chris.Kline@ca.com
# ---------------------------------------------------------------------------
#
# This script is a library of commonly used functions for Kline programs that
# use Perl.  It can be "sourced in" using the "require" pragma.
# This library contains only functions that perform server-related activities.
# These functions may call functions from other libraries.
#
# ---------------------------------------------------------------------------
# REVISION HISTORY
# 2.0   -BT50563: Split Kline.pl according to genre.  First ed of this file.
#       Added fileSize function
# 2.0.1 -BT55477: /etc/lsb-release messes up determination of linux type.
# 2.0.2 -BT58145: Rewrote checkHA to include better checks for Linux
# 2.0.3 Moved fileSize to utils.pl
# 2.0.4 -Added jpackage-release to the grep -v for linux distrib
# 2.1   Added englishLocale BT59383 to check locale for English, return T/F
# 2.2   -BT56331; added lastLogin to query the "last" cmd for user login 
#       info.
# 3.0   -BT66191; moving to a true perl module
# 3.0.1 12/11/07 Updated showLastLogin formatting
# 3.0.2 12/26/07 Updated linuxDistrib to handle multiple release files and use
#       perl File::find instead of shelling out
# 3.1   01/21/08 Moved portIsListening from WMQ.pm and updated to allow both aoh and scalar
#       Added getUsersGroups to return a list of all groups a users is a member of
#       03/06/08 added cmdLine, isInGroup
# 3.1.1 05/12/08 Added diskFree support for HP and Win32
# 4.0   06/09/08 Updates to support Windows.  Includes updates to following subs:
# 		 checkPlatform
#		 isInGroup
#		 portIsListening
#		 cmdLine (now handles dirs with spaces on win)
#		 filetime (windows only; convert timestamp to usable format)
#		 findSysCmd (new, to locate system-level cmds on UNIX/Win)
#       06/13/08 Added findSysCmd2 (new, supplemental to findSysCmd)
#       06/20/08 Updated cmdLine for timeout vals and incorporated SystemCall from Utils
#       06/27/08 Added dumpRegistrySubtree for use with OAMDisabled
# 4.0.1 12/15/11 Moved the require for Win32::AdminMisc down into getWinUserOrGroupInfo since it's harder to load and won't often be needed.
#                Also found newer version of Win32::AdminMisc that seems to work on Perl 5.10+ fine
#
#-------------------------------------------------------------------------------
# Source Control Information
# $Revision: 22 $
# $LastChangedDate: 2012-01-10 20:43:31 -0700 (Tue, 10 Jan 2012) $
# $Author: cnkline $
#
# ---------------------------------------------------------------------------
# INITIALIZE THE ENVIRONMENT

use strict;
use English; 		#D:\perl_oreilly\prog\ch07_013.htm
require Exporter;
our @ISA=qw(Exporter);
our @EXPORT=qw(checkHA checkPlatform checkUser convertTilde diskFree englishLocale linuxDistrib portIsListening getUsersGroups isInGroup cmdLine $winquote);  #if exports fail, try to compile manually first

BEGIN { #only load windows mods if we expect to need them
  if ($^O=~/mswin/i) { #be careful to not match on darwin
    require Win32; #builtin
    import Win32;
    require Win32::TieRegistry; #builtin, used for all reg calls
    import Win32::TieRegistry;
    require Win32::DriveInfo; #CPAN: used for diskFree
    import Win32::DriveInfo;
    require Win32::API; #CPAN: used for filetime (copied in from CPAN)
    import Win32::API;
    require Win32::NetAdmin; #builtin, used for isInGroup
    import Win32::NetAdmin;
    require Win32API::Net; #builtin
    import Win32API::Net;
    require FindBin;
    #Win32::AdminMisc is added below in getWinUserOrGroupInfo since it's harder to load
  #} else {
    #warn "$^O detected\n";
  }
};
$Win32::TieRegistry::Registry->Delimiter("/") if $^O=~/mswin/i; #from Win32::Registry

use lib '..'; #so we can run these direct to make sure they compile and not fail for missing mods
use Kline::Utils;
use Kline::Display;
use Text::Wrap; #not used here other than for the columns variable
use File::Find;	#used for linuxDistrib
use File::Spec::Functions; #catfile
use File::Basename qw(dirname);
use FindBin;
use Carp qw(croak);
use Sys::Hostname;

my $scriptVer="4.0.1";
my $scriptDate; #backward compatibility
($scriptDate = '$LastChangedDate: 2012-01-10 20:43:31 -0700 (Tue, 10 Jan 2012) $') =~ s/.*\(\D+(\d+\s\w+\s\d+)\).+/$1/; #auto update upon commit

our $winquote="\""; #for windows, in case of spaces in the command string; they don't like single quotes; gnosis has shared quote val, but server doesn't know about it

my $WINDOWS = ( $^O eq 'MSWin32' );

#----------------------------------------------------------------------------

#----------------------------------------------
sub diskFree {
# receives a filesystem name and queries for the free space
# returns a statement about mount points and free space in free text
# returns hash with all info if requested (wantarray), with space in kb's
# Supports HP, Sun, Linux, AIX, and Win32
#----------------------------------------------

  my $filesys=shift(@_);
  &showDebug("Searching disk free for $filesys",1);
  my $thisFunction = &thisFunc();
  unless (-d $filesys) {
    warn "WARNING: ($thisFunction) Directory $filesys doesn't exist!\n";
    return -1;
  }
  my (
    %info,	#hash of items from the df cmd output
    @array,
    $name, 	#name of actual filesystem that is mounted
    $space,	#free space in kb on the mounted filesystem
    $percent,	#free space in percent on the mounted filesys
  );
	
  if ($WINDOWS) { #WINDOWS
    eval { require Win32::DriveInfo; };
    if ( $@) { #not installed
     warn "WARNING: Cannot find Win32::DriveInfo; cannot check disk space on windows\n";
     return undef;
    } else { #mod is working
      #field 5 is total space, 6 is free space
      my $drive=$1 if $filesys=~/^(\w:)/; #just pull the drive letter
      my @ds=Win32::DriveInfo::DriveSpace($drive);
      $info{space}= sprintf "%.0f",$ds[6] /1000 ; #change bytes to kb's, round to nearest KB
      $info{percent}=sprintf "%.0f%",($ds[6] / $ds[5]) * 100 if ($ds[5] && $ds[6]); #don't divide by zero, divide and show percent free, rounding to nearest percent
      $info{name}=$drive if $drive;
    } #end $@ check
    
  } else { #UNIX/Linux
    #run df cmd and read from a pipe
    my $cmd="df -k";
    $cmd="df -Pk" if $^O=~/hpux/i; #conventional format; otherwise, HP is ugly output to parse
    
    open(DF,"$cmd $filesys|") or die "cannot execute df: $!\n";
    while (<DF>) {
      push @array,split() unless (/^Filesystem/);
      #last if @array; #exit the loop if match found; should be only 1
    }

    close DF
      or warn $! ? "system error closing df pipe: $!"
      : "child error: wait status $? from df pipe";
  
    #make sure we have the data we tried to retrieve
    unless (@array) {
      warn "($thisFunction) Could not find any filesystem info for $filesys\n";
      return -1;
    }
	
    #set the field numbers (indices) to use from the array
    #remember that all these indices are actually plus 1 on df
    if ($^O =~ /solaris/) { #field 5 is used percent
      $info{space}=$array[3];
      $info{percent}=$array[4];
      $info{name}=$array[5];
    } elsif ($^O =~ /linux/) { #field 5 is used percent
      $info{space}=$array[3];
      $info{percent}=$array[4];
      $info{name}=$array[5];
    } elsif ($^O =~ /aix/) { #field 4 is used percent
      $info{space}=$array[2];
      $info{percent}=$array[3];
      $info{name}=$array[6];
    } elsif ($^O =~ /hpux/) { #field 4 is used percent 
      #HP has a different format, but we catch that above when set cmd
      $info{space}=$array[3];
      $info{percent}=$array[4];
      $info{name}=$array[5];
    }
  } #end OS branch
  
  unless (keys %info) { #no info was found
    warn "($thisFunction) $^O is not a supported platform for diskFree\n";
    return undef;
  }

  #return undef unless defined wantarray; #void context
  if (defined wantarray()) {
    if (wantarray) { #array (hash)
      return %info;
    } else { #scalar
      return \%info;
    }
  } else { #ws called in void context, which is likely kshtoperl
    for (sort keys %info) { print "$_: $info{$_}\n"; }
    return 0;
  }
	
} #end diskFree

#----------------------------------------------
sub checkHA {
# Check the system for the presence of HA products so that a warning can be
# displayed.  Warning will display if any arg is passed, otherwise, just returns
# true or false.
# Rewritten at 2.0.2 to check all packages against all platforms
#
# CURRENTLY CALLED FROM managemq.sh profile via profile.mqm on login
#----------------------------------------------
  &showDebug("Checking for HA software",1);
  my $HAref= {
    "cluster" => "HACMP Server", #AIX
    "VRTSvcs" => "Veritas Cluster Server", #AIX, Sun, Linux
    "SUNWsc" => "Sun Cluster Utilities", #Sun
    "clusmon" => "Linux Cluster Monitor", #Linux
    "ksh-93r" => "TEST DEBUG PACKAGE.  TURN ME OFF",
  };

  foreach my $pkg (sort keys %{$HAref}) {
    SWITCH: {
      if ($^O =~/solaris/i) { system("pkginfo $pkg >/dev/null 2>&1"); last SWITCH; }
      if ($^O =~/aix/i) { system("lslpp -l $pkg\\\* >/dev/null 2>&1"); last SWITCH; }
      if ($^O =~/linux/i) { system("rpm -q $pkg >/dev/null 2>&1"); last SWITCH; }
      return 0; #unsupported OS
    }
	  my $HAtype = "$HAref->{$pkg}" if $? == 0;;
    $HAtype ? &showDebug("HA type ($pkg) $HAtype detected",2) : &showDebug("No HA software detected for $pkg",2);
    print STDERR "\n***********THIS SERVER IS CONFIGURED FOR HIGH AVAILABILITY CLUSTERS*************\n     This server appears to have $HAtype installed.\n     Queue managers may be configured for failover and may not currently be \n     mounted on this node.  If HA monitoring is enabled, do not arbitrarily\n     start or stop a queue manager, as this could trigger an automatic\n     failover to occur.  Use caution and properly cascade all configuration \n     changes to both nodes, as necessary!\n********************************************************************************\n\n" if (@_ && $HAtype) ;
    return $HAtype if $HAtype;
  } #end foreach
  return 0;

} #end checkHA

#----------------------------------------------
sub checkPlatform {
#check whether this platforms is supported and needs any syntax changes
#If called with any args, it returns any warnings instead of printing them
#sets the installed binary path (WITHOUT bin SUBDIR)
#----------------------------------------------

  my @result;
  &showDebug("$^O detected",3);

  PLATFORM: { #platform "Win" for Win32
    if ($^O =~ /linux/) {
      $Kline::WMQ::mqBinaries="/opt/mqm";
      &linuxDistrib(); #not really needed, but will help in debug to show type
      last PLATFORM;
    }
    if ($^O =~ /aix/) {
      $Kline::WMQ::mqBinaries="/usr/mqm";
      last PLATFORM;
    }
    if ($^O =~ /solaris/) {
      $Kline::WMQ::mqBinaries="/opt/mqm";
      $ENV{PATH}="/usr/xpg4/bin/:$ENV{PATH}";
      last PLATFORM;
    }
    if ($^O =~ /tru64|hpux/) { #allow continue with warning
      $Kline::WMQ::mqBinaries="/opt/mqm"; 
      push @result,"\nWARNING:  This is not a completely supported platform for Kline ($^O).\n";
      push @result,"Kline has not been tested and may not function correctly.  Proceed at your \nown risk.  See http://wmq.boulder.ibm.com/Kline for more information.\n\n";
      last PLATFORM;
    }
    if ($WINDOWS) { #must get from registry
      $Kline::WMQ::mqBinaries=$Win32::TieRegistry::Registry->{"$Kline::WMQ::registryBaseKey/FilePath"};
      &Kline::Display::showDebug("Extracting MQ install location from registry",1);
      last PLATFORM;
    }
    
    #if we get this far, don't continue at all; not supported at all
    die "\nERROR:  This is not a supported platform for Kline ($^O). Cannot continue.\n";
  } #end Platform

  if (wantarray()) {
    return @result;
  } else { #ws called in scalar context, which is likely kshtoperl
    &printError(@result) if @result;
    return 0;
  }

} #end checkPlatform

#----------------------------------------------
sub checkUser {
#make sure the current user id is valid for this operation
#will print warning if any scalar, return it if list context
#----------------------------------------------
  my $callingProg=shift @_;
  my $requiredUser="mqm";
  my @result;
  my $current=getpwuid $EFFECTIVE_USER_ID; #$> is EFFECTIVE_USER_ID
  if ("$current" ne "$requiredUser") {  
    push @result,"You must run this program ($callingProg) as user $requiredUser.  You are currently\nrunning as user $current.  Switch to user mqm and run this program again.\n\n";
  }

  if (wantarray()) {
    return @result;
  } else { #ws called in scalar context, which is likely kshtoperl
    &printError(@result);
    return 0;
  }

} #end checkUser

#----------------------------------------------
sub convertTilde {
# Receive a directory name with possible tilde notation and convert
# to proper name.  If dir doesn't exist, then return as is
#----------------------------------------------
  my $dirname=shift(@_);
  &showDebug("dirname start $dirname",1);
  $dirname =~ s{ ^ ~ ( [^/]* ) }
      { $1
              ? (getpwnam($1))[7]
              : ( $ENV{HOME} || $ENV{LOGDIR} || (getpwuid($>))[7] )
      }ex;
  &showDebug("dirname now $dirname",1);
  return $dirname;

} #end convertTilde

#----------------------------------------------
sub linuxDistrib {
#Returns the name(s) of the linux distribution, either as array or string
#If more than one version and scalar, return them all as a long string with 
#| separators.  Return only top line if mult lines
#----------------------------------------------
  die "Not a linux distribution" unless ($^O=~/linux/i);
  my @files=glob "/etc/*-release";
  my $skip=("lsb-release|jpackage"); #add any release files to be skipped here
  @files=grep !/$skip/,@files; #remove any skipped files
  &showDebug("The following linux distribution files were found: @files",1);
  return "No linux distribution files were found" unless (@files);
  my @distrib;
  foreach (@files) { #go through each file and keep only the first line
    my $topline=(&getFileContents($_))[0];
    chomp $topline;
    push @distrib,$topline;
  }
  &showDebug("This is \"@distrib\"",1);
  return wantarray ? @distrib : join " | ",@distrib;

} #end linuxDistrib

#----------------------------------------------
sub englishLocale {
#Uses system "locale" command to determine whether
#English is active and returns true/false
#Used http://www.opengroup.org/onlinepubs/007908799/xbd/envvar.html
#for a descr on the env vars
#----------------------------------------------
  my $cmd="locale";
  my $out=`$cmd 2>&1`;
  my @env_vars=qw/LC_ALL LC_MESSAGES LC_CTYPE LANG/;
  foreach my $var (@env_vars) {
    if ($out=~/$var=(.+)/) {
      my $val=$1;
      $val=~s/\"//g; #trim any quotes
      &showDebug("Locale Variable $var = \"$val\"",1);
      if ($val=~/en_|^C$|POSIX/i) { #looking for english or C
        &showDebug("$var=$val indicates ENGLISH locale.  OK",1);
        return 1;
      }
    }
  }#end loop
  #if we get here, not English
  &showDebug("Could not validate that English is the locale\n$out",1);
  return 0;

} #end englishLocale

#----------------------------------------------
sub lastLogin {
#get info from "last" (wtmp) on login times and durations
#returns href with users' info, $2 as the begin date on wtmp file
#DOES NOT HANDLE ID'S STILL LOGGED IN YET!
#----------------------------------------------
  croak "Cannot run on windows" if $WINDOWS;

  my $href;
  my $time="";
  my $lastcmd=&findSysCmd("last");
  die "Cannot locate \"last\" on this system.\n" unless $lastcmd;

  open CMD,"$lastcmd |";
  while (<CMD>) {   # While we have input...   
    # Find lines and save username, login time 
    if (/^(\S+)\s+(\S+)\s+(\S+)\s+(\S+\s+\S+\s+\S+\s+\S+).+\(([^\(]+):([^\)]+)\)/) {   
      # Increment total hours, minutes, and logins 
      my ($id,$last,$hr,$min)=($1,$4,$5,$6);
      $hr=($1 * 24) + $2 if ($hr=~/(\d+)\+(\d+)/);
      $href->{$id}{hours} += $hr; 
      $href->{$id}{minutes} += $min; 
      $href->{$id}{logins}++; 
      $href->{$id}{last}=$last unless $href->{$id}{last};
    } elsif (/^(\S+)\s+(\S+)\s+(\S+)\s+(\S+\s+\S+\s+\S+\s+\S+).*still logged in/) {   
      $href->{$1}{logins}++; 
      $href->{$1}{last}=$4 unless $href->{$1}{last};
      $href->{$1}{still}++; #user is still logged in
    } elsif (/^wtmp begins\s+(.+)/) {
      chomp($time=$1);
    } else {
      &showDebug("Could not match line $_",2);
    } 
  } 
  close CMD;
 
  # For each user in the array...        
  foreach my $user (sort(keys %{ $href })) { 
    # Calculate hours from total minutes 
    $href->{$user}{hours} += int($href->{$user}{minutes} / 60); 
    $href->{$user}{minutes} %= 60; 
  }

  return ($href,$time);

} #end lastLogin

#----------------------------------------------
sub showLastLogin {
#print info from lastLogin pretty
#optionally receive list of names and show only those
#updated 07/08 to support Windows
#----------------------------------------------
  my $glength=29; #gecos max len for printing
  my $idlength=10;
  if ($WINDOWS) {
    $glength=19;
    $idlength=20;
  }
  my @names=@_ if @_;
  my ($href,$time);
  ($href,$time)=&lastLogin() unless $WINDOWS; #does an open on the wtmp file
  my $still;
  my $format="%-${idlength}s%-${glength}s%8s%10s  %-20s\n"; #(LOGIN,NAME,LOGINS,TOT_HRS,LAST)
  my @out; #buffer output in case it gets returned
  my $header="LOGIN SUMMARY";
  $header.=" SINCE $time" if $time;
  push @out,$header ."\n";
  push @out,sprintf $format,"LOGIN","NAME","LOGINS","TOT_HRS","LAST";
  push @out,"-" x ($Text::Wrap::columns - 1)."\n";
  if (@names) { #if received names, print table with those names only
    foreach my $id (@names) {
      my ($ts,$name);
      if ($WINDOWS) {
        my $info=&getWinUserOrGroupInfo($id); #will already be cached if called previously
        $name=substr($info->{_fullname},0,$glength) || "";
        #populate the stuff that would have come from wtmp into the same hash for below block
        $href->{$id}{logins} = $info->{_numlogons} if $info->{_numlogons};
        $href->{$id}{last} = localtime($info->{_lastlogon}) if $info->{_lastlogon};
        $ts="N/A";
      } else { #UNIX
        $name=substr((getpwnam($id))[5,6],0,$glength) || "";
        $ts=sprintf("%d:%.2d",$href->{$id}{hours},$href->{$id}{minutes}) || "";
      }
      if (exists($href->{$id})) {
        push @out,sprintf "$format",$id,$name,$href->{$id}{logins},$ts,$href->{$id}{last};
      } else { #id hasn't logged in
        push @out,sprintf "$format",$id,$name ,0,"","NONE";
      }
    } #end foreach
  } else { #print all of them
    die "Can only run this sub if a list of users is passed on windows" if ($WINDOWS);
    
    foreach my $id2 (sort keys %{ $href }) {
      next if $id2=~/reboot/;
      my ($comments,$gecos)=(getpwnam($id2))[5,6];
      $gecos.="($comments)" if $comments; #content of this field varies
      $gecos=substr($gecos,0,$glength) || ""; #get the GECOS info and chop is to glength
      my $ts2=sprintf("%d:%.2d",$href->{$id2}{hours},$href->{$id2}{minutes}) || "";
      if (exists($href->{$id2}{still})) {
        $href->{$id2}{logins}="**$href->{$id2}{logins}";
        $still++;
      }
      push @out,sprintf($format,$id2,$gecos,$href->{$id2}{logins},$ts2,$href->{$id2}{last});
    }
  }
  push @out,"\nNOTE: (**) above indicates the id is still logged in as of " . localtime() . "\n" if $still;

  if (wantarray) {
    return @out;
  } else { #called in void context
    print @out;
  }

} #end showLastLogin

#----------------------------------------------
sub portIsListening {
#receives AoH with "port" and adds "status", then returns AoH
#otherwise, receives a scalar port number and returns true/false
#used by WMQ::listenerInfo and Gnosis::userLogin_SSHLoginDisabled
#updated 06/06/08 for Win
#----------------------------------------------
  my @AoH=@_;
  my $status;
    
  my $cmd=&findSysCmd("netstat"); #find the valid netstat command to exec
  my $verb="LISTEN"; #what we have to grep for in the lines
  die "Could not find netstat command!" unless $cmd;
    
  #UNIX/Linux /bin/netstat on SLES, /usr/bin/netstat on AIX, Sun, and HP
  #netstat -an works on XP, 2003, and 2000 via %WINDOWS%\system32\netstat.exe
  #netstat output varies by platform but port follows either dot or :, and then space
      #WIN:  TCP    0.0.0.0:135            0.0.0.0:0              LISTENING
      #SLES: tcp        0      0 :::1414                 :::*                    LISTEN
      #SUN10: *.22                 *.*                0      0 49152      0 LISTEN
      #HP:   tcp        0      0  *.23                   *.*                     LISTEN
      #AIX:  tcp        0      0  *.1414                 *.*                    LISTEN
  if ($WINDOWS) {
    $verb="LISTENING"; #different on windoze
    $cmd="$winquote$cmd$winquote";
  }
  my $rawout=&cmdLine("$cmd -an");
  
  foreach my $entry (@AoH) {
    my $port;
    if (ref $entry eq 'HASH') {
      $port=$entry->{port} || 0;
      &showDebug("Found an AoH entry looking for status on port $port",2);
    } elsif (! ref $entry) { #must be a straight port number (not a reference)
      $port=$entry;
      &showDebug("Found a scalar request looking for status on port $port",2);
    } else {
      my $this=&thisFunc;
      die "Cannot execute $this against object type \"". (ref $entry) ."\".  Expected AoH (hash) or scalar.";
    }
    
    my @out=grep /\b$verb\b/,@$rawout; #keep only the listening lines
    $status=(grep /(:|\.)$port\b/,@out) ? 1 : 0; #search for our port, with proper notation
    $status=0 if ($port == 0); #don't do a port check if 0
    &showDebug("Port $port listening status: $status",1);
    $entry->{status}=$status if (ref $entry eq 'HASH');
  }
  return wantarray ? @AoH : $status;

} #end portIsListening

#----------------------------------------------
sub getUsersGroups {
#receive name of user and return a list of all his groups (primary and secondary).  Reverse of what's done in the middle of Gnosis::Plugins::Common::GroupMembership module
#Returns either array or aref of groups
#Return undef if user does not exist
#----------------------------------------------
  my $user=shift || die "No user provided";
  my $groups;

  my $id=getpwnam($user);
  return undef unless defined $id; #undef if user doesn't exist

  #get the user's primary group
  push @$groups,scalar(getgrgid((getpwuid($id))[3]));
  

  #get secondary groups
  while ( my ($name,$members)=(getgrent)[0,3] ) {
    if ($members=~/\b$user\b/) {
      &Kline::Display::showDebug("$user secondary membership to $name",3);
      push @$groups,$name;
    }
  }
  endgrent;
  &showDebug("$user is a member of the following groups (primary first): @$groups",1);
  return wantarray ? @$groups : $groups;

} #end getUsersGroups

#----------------------------------------------
sub isInGroup {
#checks the current real UID to see if in the $1group.
#Optionally receives $2 as name of user to check instead of current
#Returns true/false, or undef if group doesn't exist
#updated 05/2008 to support Win32; checks local groups first and then net groups on Win
#----------------------------------------------

  my $group=shift || die "ERROR: No group name to check membership was passed";
  my $user=shift || ($WINDOWS ? &Win32::LoginName() : getpwuid($UID));
  &showDebug("Checking $user for $group group membership",1);
  if ($WINDOWS) {
    my $server="";
    #make sure user and group both exist
    unless (Win32::NetAdmin::UsersExist($server,$user)) {
      &showDebug("user $user does not exist",1);
      return undef;
    }
    #TODO: Add check whether group actually exists
    
    if (Win32::NetAdmin::LocalGroupIsMember($server,$group,$user)) {
      &showDebug("$user is in the local $group group",1);
      return 1;
    } else {
      &showDebug("user is not a member of local group $group",2);
      if (Win32::NetAdmin::GroupIsMember($server,$group,$user)) {
        &showDebug("$user is in the network $group group",1);
        return 1;
      } else {
        &showDebug("user is not a member of network group $group",1);
        return 0;
      }
    }
  } else { #UNIX
    unless (defined getgrnam($group)) {
      &showDebug("group $group does not exist",1);
      return undef;
    }
    my $groups=&getUsersGroups($user); #Kline::Server, checks all groups
    unless ($groups) { #will be undef if the user doesn't exist
      &showDebug("user $user does not exist",1);
      return undef;
    }
    if (grep /\b$group\b/,@$groups) {
      &showDebug("$user is in the $group group",1);
      return 1;
    } else {
      &showDebug("$user is NOT in the $group  group",1);
      return 0;
    }
  } #end UNIX vs Win

} #end isInGroup

#----------------------------------------------
sub cmdLine {
#Run a command externally and capture its return code ($1) and output (aref, $2).
#Default if called scalar is to return output in an aref with nothing added
#If an optional $2 argument is passed as a scalar (DEPRECATED), run as "user" hashref key below
#If the $2 is a hashref (preferred), the following vals are allowed:
#if key "scalar" is true, then return the cmd output as a scalar, not an aref.
#if key "user" is true, run the command (UNIX) as that user id  with su -c (ignored unless currently running as root).
#if key "shift_rc" is true, the returned RC will be shifted prior
#if key "no_redirect" is true, then no command redirection is appended
#if key "no_rc" is true, then no return code is returned; only the output is returned
#if key "command" and the hashref is the first argument, then only a single argument is needed
#if key "timeout" included, then use that value as command timeout, else 30 sec
#if key "verbose" true, then cmd output has errors and debug appended (default: add nothing)
#stderr and stdout are combined by default
#exit code is not shifted by default
#Return only the aref unless wantarray; use debug stmts to see behavior
#When run on windows, if spaces are found in the cmd, quotes are added if not already found
#Warning is thrown if called with redirection on windows and not already quoted.
#06/08 Combined with SystemCall to include timeouts
# Based on O'Reilly's Perl Cookbook recipe 16.21.2
# and http://perlmonks.org/?node_id=77188
#----------------------------------------------
  my $cmd=shift;
  my $arg={};
  if (ref($cmd) eq "HASH") {
    &showDebug("Single hashref",1);
    $arg=$cmd;
    $cmd=$arg->{command};
  } else {
    &showDebug("Optional 2nd arg hashref",1);
    $arg=shift if @_;
  }
  die "No command provided" unless $cmd;
  my $user;
  my $redirect=" 2>&1";
  my $time=$arg->{timeout} || 30; #assume 30 sec timeout if not stated
  my $verbose=$arg->{suppress} || 0;
  &Kline::Display::showDebug("Verbose output enabled",1) if $verbose;
  
  #parse input values
  if (ref $arg ne "HASH") { #new format, backwards compat in case they pass a scalar for user name
    $arg->{user}=$arg;
  }
  $redirect="" if $arg->{no_redirect};
  
  my $su="";
  if ($UID == 0 && $arg->{user}) {
    $su="su $arg->{user} -c ";
  } elsif ($arg->{user}) {
    &showDebug("WARNING: Cannot switch to user $arg->{user}",1);
  }
  $cmd="$su\"$cmd\"" if $su;
  
  #if win, allow for spaces in dir names by quoting, but only if spaces are found and it includes directory markers and not already quoted somewhere in the string
  if ($WINDOWS && $cmd=~/\s/ && $cmd=~/\\/ && $cmd!~/["']/) {
    $cmd="$winquote$cmd$winquote";
    &showDebug("NOTE: command includes spaces and was quoted for execution to $cmd",1);
    warn "WARNING: just quoted a line with redirection; it will break!  Instead, " if $cmd=~/[<>]/;
  }

  #run it
  my @out;
  $cmd.=$redirect;
  &showDebug("Running cmd {{$cmd}}",1);
  undef $!;
  if ($WINDOWS) { #no alarms work on windows #####################
    &showDebug("No alarm timeouts can be set on Windows",1);
    @out=`$cmd`;
  } else { #UNIX, Use alarms #############################
    &showDebug("Running cmd with timeout $time seconds",1);
    my $pid = open(PIPE, "$cmd|") or die "Couldn't fork comand $cmd: $!\n";
    eval {
      local $SIG{ALRM} = sub {
        #my ($caller,$line)=(caller(4))[3,2]; #got to go one more up since in eval
        croak "FATAL: $time-second timeout exceeded on command {{ $cmd }} \n"; #(cmdLine called by $caller:$line)\n";
        };
      alarm($time);                   # schedule alarm in 10 seconds
      #eval {
      while (<PIPE>) {
        push @out,$_;
      };
      #};
      close PIPE;
      alarm(0);
    }; #end eval
    alarm(0);                      # race condition protection
    if ($@) {
      kill 9, $pid;
      close PIPE;
      die $@ ;#unless $@=~/TIMEOUT/i;
    };
  } #end UNIX vs Win
  
  my $rc=$? >> 8; #not returned!
  push @out,"\n$!\n" if ($! && $verbose); #reminder: we cannot set $! ourselves and have it stick
  &showDebug("\$?=$?, shifted rc=$rc",1);
  &showDebug("output:\n @out",2);
  push @out,"Command \"$cmd\" exited normally and produced no output.\n" if ($rc == 0 && ! @out && $verbose);
  my $adj="raw"; #just adjective for debug below
  unless ($arg->{shift_rc}) {
    &showDebug("Returning unshifted return code",2);
    $rc=$?; #return the raw RC
    $adj="shifted";
  }
  
  #set the error msg if not success
  $!=join "\n",@out if $rc;
  
  my $out=join "\n",@out if $arg->{scalar}; #scalar equivalent
  if ($arg->{no_rc}) {
    if ($arg->{scalar}) { #return the actual array or the string equiv
      &showDebug("Returning only either full output array or string equiv",2);
      return wantarray ? @out : $out;
    } else { #return the actual array or aref
      &showDebug("Returning only either full output array or aref",2);
      return wantarray ? @out : \@out;
    }
  } 
  #if we get here, return the rc and the output
  if ($arg->{scalar}) { #return the rc and string equiv
    &showDebug("Returning string equiv and (wantarray) possibly $adj rc",2);
    return wantarray ? ($rc,$out) : $out;
  } else { #return the rc and aref
    &showDebug("Returning aref and (wantarray) possibly $adj rc",2);
    return wantarray ? ($rc,\@out) : \@out;
  }

} #end cmdLine

#----------------------------------------------
sub filetime {
#Converts a FILETIME binary format to string text (Win32);
#CPAN module copied directly from http://cpan.uwinnipeg.ca/htdocs/Win32-UrlCache/Win32/UrlCache/FileTime.pm.html
#Tests show it's returning in GMT, not localtime
#----------------------------------------------
# BOOL FileTimeToSystemTime(
#   CONST FILETIME *lpFileTime,
#   LPSYSTEMTIME    lpSystemTime,
# )

  die "filetime only runs on windows" unless $WINDOWS;
  
  my $FileTimeToSystemTime = Win32::API->new(
    kernel32 => FileTimeToSystemTime => 'PP', 'I'
  );

  die "Can only use filetime on Windows\n" unless $WINDOWS;
  my $filetime = shift;

# FILETIME
#   DWORD dwLowDateTime;
#   DWORD dwHighDateTime;

  if ( substr( $filetime, 0, 2 ) eq '0x' && length( $filetime ) == 18 ) {
    my $low   = substr $filetime, 2, 8;
    my $high  = substr $filetime, 10;
    $filetime = pack 'H8H8', $low, $high;
  }

# SYSTEMTIME
#   WORD wYear;
#   WORD wMonth;
#   WORD wDayOfWeek;
#   WORD wDay;
#   WORD wHour;
#   WORD wMinute;
#   WORD wSecond;
#   WORD wMilliseconds;

  my $systemtime = pack( 'S8', (0) x 8 );

  $FileTimeToSystemTime->Call( $filetime, $systemtime );

  my @systimes = unpack( 'S8', $systemtime );

  splice @systimes, 2, 1;  # remove wDayOfWeek

  return sprintf "%04d-%02d-%02d %02d:%02d:%02d", @systimes;

} #end filetime

#----------------------------------------------
sub findSysCmd {
#receive name of a system command (that would normally be in the path) and
#find its absolute location on this box;  return undef if not found
#works for both Win and UNIX in std locations and adds .exe to end on win
#----------------------------------------------
  my @dirs= ($WINDOWS) ? ("$ENV{SystemRoot}","$ENV{SystemRoot}\\system32") : ("/bin","/usr/bin/","/usr/sbin","/usr/local/bin");
  my $cmd=shift;
  $cmd.=".exe" if $WINDOWS && $cmd!~/exe$/; #add exe if not already there
  my $found;
  foreach (@dirs) {
    if (-f catfile($_,$cmd)) {
      $cmd=(catfile($_,$cmd));
      $found++;
      last;
    }
  }
  &showDebug("Located command at $cmd",2) if $found;
  return ($found) ? $cmd : undef;

} #end findSysCmd

#----------------------------------------------
sub _thisDir {
#Internal.  Will return the caller's dir (must be in the 2nd sub in order for
#correct dir name to be returned.  Otherwise, it returns FindBin)
#----------------------------------------------
  return (&Kline::Utils::parentFunc())[1]; #find the caller's file's dir
}

#----------------------------------------------
sub findSysCmd2 {
#Very similar to findSysCmd above but it will first look for the given command
#in the current prog's dir or caller's file directory, and then call findSysCmd if nothing there
#Created for Server::Schedule to look for execs that might not be found on the OS
#Given an exec name, look for it first in the same path as this library, and
#then in system directories, return undef if not found
#----------------------------------------------
  my $cmd=shift;
  my $thisDir=dirname(_thisDir()); #find the caller's file's dir
  $cmd.=".exe" if $WINDOWS && $cmd!~/exe$/; #add exe if not already there
  
  my $file=catfile($FindBin::Bin,$cmd);
  &showDebug("Looking for supplemental exec at $file",1);
  return $file if -f $file; #look for it in the calling file's dir
  
  $file=catfile($thisDir,$cmd); #now look for it in the calling lib's dir
  &showDebug("Looking for supplemental exec at $file",1);
  return $file if -f $file; #look for it in the calling file's dir
  
  &showDebug("Did not find it there; continuing to normal findSysCmd",1);
  return &Kline::Server::findSysCmd($cmd); #if we get here, then just  look in normal dirs

} #findSysCmd2

#----------------------------------------------
sub dumpRegistrySubtree {
#Windows only; given a Win32::TieRegistry pointer, dump the entire contents
#to an array for printing.  calls itself recursively for every subkey
#----------------------------------------------
  croak "Can only run on windows" unless $WINDOWS;
  my $pointer=shift || die "No pointer";
  my $indent=shift || 0;
  my @dump; #contents
  push @dump," " x $indent . $pointer->Path ."\n";
  $indent+=2;
  foreach ($pointer->ValueNames) {
    push @dump," " x $indent . "$_ = " . $pointer->GetValue($_) ."\n";
  }
  foreach ($pointer->SubKeyNames) {
    push @dump,&dumpRegistrySubtree($pointer->{"$_/"},$indent);
  }

  return wantarray ? @dump : \@dump;

} #end dumpRegistrySubtree

{ #naked block
my $winAcctInfo={}; #persist so we don't have to hit the OS over and over for same id's
#----------------------------------------------
sub getWinUserOrGroupInfo {
#Windows only; given an account, pull the account info, first from local and then from domain
#Returns undef if the user or group is not found.  Will not return group members, just attrs
#uses Win32::AdminMisc, which will only run if msvcr70.dll is found (visual studio v7, which you can get at http://www.dll-files.com/dllindex/dll-files.shtml?msvcr70; should be found if .NET is installed)
#returns a hashref with all the user's info included a manual tag at $info->{source} showing where it came from
#Some fields (not all)
      #For AD USER: USER_NAME, USER_FULL_NAME, USER_COMMENT, USER_LAST_LOGON, USER_FLAGS, USER_USER_ID, USER_NUM_LOGONS, USER_PRIV and _source/_type
      #For AD Group: GROUP_ATTRIBUTES, GROUP_COMMENT, GROUP_ID, GROUP_NAME
      #For Local User: comment, flags, fullName, lastLogon, name, numLogons, primaryGroupId, priv, userId,
#Receives optional values in $parms:
# {domain}.  If true, only look for the account on the domain and not locally
#ID's are case-insensitive
#----------------------------------------------
  croak "Can only run on windows" unless $WINDOWS;
  my $acct=shift || die "No user/group";
  my $parms=shift if @_;
  
  #cached info from previous calls
  return $winAcctInfo->{$acct} if exists $winAcctInfo->{$acct};
  
  my $domainonly=0; #if a domain was prefixed, look only for domain objects
  
  #if we are passed a prefix that is not this server, only check the domain
  my $id=$acct; #we will store on the cache with full name, but have to use trimmed version for queries
  my $scope="local"; #assume the local hostname is prepended already (we don't do that here if it's local)
  if ($id=~/([^\\]+)\\(.+)/) { #if there is a domain, take it off and put in _scope
    $scope=$1;
    $id=$2;
    $domainonly++ unless (lc($scope) eq lc(hostname));
  }
  
  my $info={};
  if (!$domainonly && Win32API::Net::UserGetInfo("",$id,3,$info)) { #checks for a local user only
    &showDebug ("Found local user info using Win32API::Net::UserGetInfo",2);
    $info->{_source}="Win32API::Net::UserGetInfo";
    $info->{_type}="local user";
    
  } elsif (!$domainonly && Win32API::Net::LocalGroupGetInfo("",$id,1,$info)) { #get local group info
    &showDebug ("Found local group info using Win32API::Net::LocalGroupGetInfo",2);
    $info->{_source}="Win32API::Net::LocalGroupGetInfo";
    $info->{_type}="local group";
  } else { #doesn't exist locally, try on the domain
    my $domain=($domainonly) ? $info->{_scope} : ""; #see if the username has a specific domain attached
    
    #load Win32::AdminMisc
    $ENV{PATH}.="$FindBin::Bin"; #this is to allow Win32::AdminMisc to find the msvcr70.dll
    require Win32::AdminMisc; #used in getWinUserInfo via require, not use, http://www.roth.net/perl/adminmisc
    import Win32::AdminMisc; #must also have msvcr70.dll somewhere in the path or it will fail (we put a copy in our bin/ dir)
    
    if (&Win32::AdminMisc::UserGetMiscAttributes($domain,$id,$info)) { #this doesn't return the domain itself
      &showDebug("Found domain user info using Win32::AdminMisc::UserGetMiscAttributes",2);
      $info->{_source}="Win32::AdminMisc::UserGetMiscAttributes";
      $info->{_type}=(exists $info->{USER_NAME}) ? "domain user" : "global group";
    } else {
      &showDebug("Could not find account info via (local) Win32API::Net::UserGetInfo or (domain) Win32::AdminMisc::UserGetMiscAttributes",1);
    }
  }
  #store the common fields so we don't have to worry about the source
  $info->{_scope}=$scope; #couldn't set earlier because would get overwritten
  $info->{_idname}=$id;
  $info->{_id}=$info->{USER_USER_ID} || $info->{GROUP_ID} || $info->{userId} || "";
  $info->{_primarygroup}=$info->{USER_PRIMARY_GROUP_ID} || $info->{GROUP_ID} || $info->{primaryGroupId} || "";
  $info->{_fullname}=$info->{USER_FULL_NAME} || $info->{GROUP_NAME} || $info->{fullName} || "";
  $info->{_comment}=$info->{USER_COMMENT} || $info->{GROUP_COMMENT} || $info->{comment} || "";
  $info->{_lastlogon}=$info->{USER_LAST_LOGON} || $info->{lastLogon} || undef;
  $info->{_numlogons}=$info->{USER_NUM_LOGONS} || $info->{numLogons} || undef;
  $info->{_flags} = $info->{USER_FLAGS} || $info->{flags} || undef;
  
  $winAcctInfo->{$acct}=$info; #cache for next time
  
  return $info || undef;

} #end getWinUserInfo
} #end naked block

#----------------------------------------------
1; #tell the sourcing program we are true
