#!/usr/bin/perl
package Kline::Server::Schedule;
# Chris Kline; cnkline@us.ibm.com
# Copyright IBM Corp. 2008 All Rights Reserved
# ---------------------------------------------------------------------------
#
# Library to handle the scheduling of automated tasks on both UNIX and Windows
#
# ---------------------------------------------------------------------------
# REVISION HISTORY
# 1.0   Initial
#
#----------------------------------------------------------------------------
# Source Control Information
# $Revision: 3057 $
# $LastChangedDate: 2008-07-29 14:32:33 -0600 (Tue, 29 Jul 2008) $
# $Author: cnkline $
#
# ---------------------------------------------------------------------------
# INITIALIZE THE ENVIRONMENT

use warnings;
use strict;
use English; 		#D:\perl_oreilly\prog\ch07_013.htm
use Text::Wrap;   #for printing output when querying current jobs

use lib '../..'; #so we can run these direct to make sure they compile and not fail for missing mods
use Kline::Utils;
use Kline::Server;
use Kline::Display;

BEGIN { #only load windows mods if we expect to need them
  if ($^O=~/win/i) {
    require Win32;
    import Win32;
    require Win32::Service;
    import Win32::Service;
  }
};

use File::Basename qw(dirname fileparse);

my $scriptVer="1.0";
my $scriptDate; #backward compatibility
($scriptDate = '$LastChangedDate: 2008-07-29 14:32:33 -0600 (Tue, 29 Jul 2008) $') =~ s/.*\(\D+(\d+\s\w+\s\d+)\).+/$1/; #auto update upon commit
our $VERSION=$scriptVer;
our $VERDATE=$scriptDate;

=head1 NAME

Kline::Server::Schedule - Scheduler libraries for UNIX and Windows

=head1 SYNOPSIS

 use Kline::Server::Schedule;

=head1 DESCRIPTION

This module provides helper functions to automate the scheduling of jobs on UNIX and Windows.  It can create new scheduled jobs, remove them, and also tracks which jobs have been scheduled so we know which ones to remove if we uninstall our products.


=head2 GLOBAL VARIABLES

=over 4

=item $trackerdelim.  Symbol we use in the tracker file for field separation.

=item $tokenprefix.  Set of symbols used as prefix in crontab file comments to identify "our" comments

=back

=cut
our $trackerdelim=";"; #Symbol we use in the tracker file for fields (used by gnosis.pl to split)
my $tokenprefix='~!@'; #crontab token prefix to identify "our" comments
my $key_prefix="KEYS"; #for the tracker file header; used in parsing after it's written
my ($progname,undef,undef)=fileparse($0, qr/\.[^.]*/); #strip the extension of the current prog as a taskname
my $WINDOWS = ( $^O=~/win/i );

#----------------------------------------------------------------------------

#----------------------------------------------
sub createScheduledJob {
#----------------------------------------------
=over 4

=item createScheduledJob()

 sample code

Add a scheduled job using input parms, whether Windows or UNIX.  On Windows XP and newer, use the schtasks.exe command-line program to interact with scheduled tasks.  Since this utility does not exist on Windows 2000, Microsoft provided a jt.exe program that behaves in a similar fashion.  If that program is found to be included either in the system path or in the same directory as this library, jobs will alternatively be scheduled using that utility.  Otherwise, jobs cannot be scheduled on Windows 2000 and earlier systems.  Jobs will be scheduled with a task name equial to the exeucting program.username if not manually specified.

Will confirm that the specified parameters do not already exist as a job, based upon task names, not job parameters.

All schedules will be attempted as the current user (no user-switching options will be used, even if available on the scheduler).

Returns 1 for success, 0 for already scheduled, undef for error (see $!).  In list context, returns code and aref containing log information.

Input values correspond as follows:

B<(Required) $var  type.>  Descr
exec.  Full command that should be executed
schtasks.  Location of the schtasks executable on windows.  Will default to %WINDOWS%\system32 if null.
jt.  
taskname.  Name of the job.  In cron, comment will be added to identify the job.  In windows, the task will be named as such.  If null, the task will be named after the current program basename(extensions removed).  On UNIX, job names are automatically unique across users.  On Windows, they are unique across systems and so the task name will have the user's name appended if not specified.
wintimestring.  The appropriate time switches for Windows task scheduler.  Does not include the executable string or the "/create" switch, just the time switches.  If this is included, then none of the defaults or single-field entries below will be used.
winuser.  The user that the scheduled task should run as.  If not specified, will run as SYSTEM.
winpassword.  The password for the user specified above.  Required if winuser is specified.  When the password for the user changes, the task must also be updated in order for the job to execute. 
unixtimestring.  The five cron fields that determine when the job should execute.  If this is included, then none of the defaults or single-field entries below will be used.
trackerfile.  If not null, then upon completion of a new schedule setup, the trackScheduledJob sub is called and passed this value.  Otherwise, call it on your own.
comments.  Optional comments to place in the crontab file to explain the job since it may not be self explanatory.

THE FOLOWING ARE NOT YET IMPLEMENTED
The following values are not adjusted for Windows; use {DAILY} instead of these to set a job to execute once daily at given time.
minute.  First value from crontab.  Assumed * (or windows equiv) if null.
hour.  Second value from crontab.  Assumed * (or windows equiv) if null.
day.  Third value from crontab.  Assume * (or windows equiv) if null.
month.  Fourth value from crontab.  Assume * (or windows equiv) if null.
dayweek.  Fifth value from crontab.  Assume * (or windows equiv) if null.

WINDOWS: Information on using the Task Scheduler and user security: http://technet2.microsoft.com/windowsserver/en/library/f60ea8d6-d2af-4dd0-b050-c09951826f5d1033.mspx?mfr=true
By default, to schedule a task, you must be a member of the Administrators, Backup Operators, or Server Operators group on the local computer.
http://technet2.microsoft.com/windowsserver/en/library/f60ea8d6-d2af-4dd0-b050-c09951826f5d1033.mspx?mfr=true (view a log of past scheduled actions)
http://technet2.microsoft.com/windowsserver/en/library/f60ea8d6-d2af-4dd0-b050-c09951826f5d1033.mspx?mfr=true (scheduler best practices)
If you must run the task as a specific user, the password for that user will need to be provided, and must be updated whenever changed.  The alternative is to run as the SYSTEM account (default), but your task may not have necessarily authority.


=back

=cut

  my $parms=shift;
  my ($status,@log); #stuff we'll return
  $status=1; #assume good until proven wrong
  my ($cmd,$rc,$out); #the command used to set the job
  my $taskexec=$parms->{exec} || die "Cannot continue without a command to schedule for execution"; #complete command that scheduler should execute
  my $user=$parms->{user} || ($WINDOWS) ? &Win32::LoginName : getpwuid($EUID);
  my $taskname=$parms->{taskname} || "$progname.$user";
  my $tmpfile="/tmp/crontab.$user.new";
  
  push @log,"\nBegin adding scheduled task at ". localtime() ."\n";
  push @log,"Received the following execution parameters\n";
  foreach my $parm (sort keys %$parms) {
      push @log,"  $parm = $parms->{$parm}\n";
  }
  
  ################################
  if ($WINDOWS) { #WIN, check for schtasks or jt
    my ($util,$llog)=&_findSchedUtil($parms);
    push @log,@$llog if @$llog;
    import Kline::Server qw($winquote);
    die "Cannot find an appropriate scheduler program (using either schtasks.exe or jt.exe)\n" unless (-f $util);    
    push @log,&_schtasksSvcRunning(); #make sure the service is running; will add log stuff and die if not running (svc is same regardless of tool)
        
    my $task;
    ($task,$llog)=&_queryWinTasks($util,$taskname);
    push @log,@$llog if @$llog;
    if ($task) { #already scheduled
        push @log,"Requested task is already scheduled.  \nNo action necessary.\n";
        $status=0;
        &Kline::Display::showDebug("Spooled log output:\n@log",1);
        return wantarray ? ($status,\@log): $status;
    }
        
    #get here then we need to set a new task
    push @log,"Requested task $taskname is not yet configured.  Configuring now.\n";
    if ($util=~/schtasks/i) { #XP and newer
        #instructions for using task scheduler via cmdline at http://support.microsoft.com/?kbid=814596
        #use schtasks /create /? for help
        $parms->{exec}=~s/"/\\"/g; #on windows, the exec string for schtasks must have any quotes in args escaped
        $cmd="$winquote$util$winquote /create /tn $winquote$taskname$winquote /tr $winquote$parms->{exec}$winquote";
        $parms->{user}="SYSTEM" unless exists $parms->{user};
        $cmd.=" /ru $parms->{user}";
        $cmd.=" /rp ". $parms->{password} if (exists $parms->{password});
        die "Must have a password specified in order to schedule a task for user $parms->{user}\n" if ($parms->{user}!~/^system$/i && ! $parms->{password});
        if (exists $parms->{wintimestring} ) { #create task using exact user-specified time vals
          $cmd.=" $parms->{wintimestring}";
        } else { #try for defaults
          die "Sorry, not yet coded for default or single-field time values.  Try using the wintimestring instead until this stub is completed.";
        }
    } elsif ($util=~/jt/i) { #2K and older
      die "Not yet coded to support $util.  Sorry.\n";
    }
    
  ################################
  } else { #UNIX, do cron
    my $util=&Kline::Server::findSysCmd("crontab") || die "Cannot find crontab command\n";
    my ($task,$llog)=&_queryUNIXTasks({
                                       util=>$util,
                                       taskname=>$taskname,
                                      });
    push @log,@$llog if @$llog;
    if ($task) { #already scheduled
        push @log,"Requested task is already scheduled.  \nNo action necessary.\n";
        $status=0;
        &Kline::Display::showDebug("Spooled log output:\n@log",1);
        return wantarray ? ($status,\@log): $status;
    }
        
    #get here then we need to set a new task
    push @log,"Requested task $taskname is not yet configured.  Configuring now.\n";
    my @details;
    push @details,"#\n","#${tokenprefix}$taskname TASK". "#" x 60 ."\n";
    push @details,"#NOTE: REMOVING THE ABOVE TOKEN WILL BREAK $taskname AUTOMATIC REMOVAL ABILITY\n";
    push @details,"#Cronjob(s) configured by $progname at ". localtime() ."\n";
    push @details,"#$parms->{comments}\n" if exists $parms->{comments};
    
    #add the actual job content
    if (exists $parms->{unixtimestring} ) { #create task using exact user-specified time vals
      push @details,"$parms->{unixtimestring} $parms->{exec}\n";
    } else { #try for defaults
      die "Sorry, not yet coded for default or single-field time values.  Try using the unixtimestring instead until this stub is completed.";
    }
    
    push @details,"#${tokenprefix}END $taskname CRONJOBS". "#" x 50 ."\n";
    
    #write the new file
    ($rc,$out)=&Kline::Server::cmdLine({
                                          command=>"$util -l",
                                          shift_rc=>1,
                                          verbose=>1,
                                          });
    open OUT,">$tmpfile" || die "Cannot open temp crontab file $tmpfile: $!\n";
    print OUT @$out if $rc==0 && @$out; #add current jobs first
    print OUT @details; #add new job
    close OUT;
    
    #set the new file
    $cmd="$util $tmpfile";
  } #end WIN/UNIX

  push @log,"Attempting to add new scheduled task with command: \"$cmd\".\n";
  ($rc,$out)=&Kline::Server::cmdLine({
                                        command=>$cmd,
                                        shift_rc=>1,
                                        verbose=>1,
                                        });
  if ($rc) {
    my @msg=("ERROR: Cannot create new scheduled task using {{ $cmd }} (RC=$rc): @$out");
    if (-f $tmpfile) {
      push @msg,"\nwhere temp-file content is as follows:\n". &Kline::Utils::getFileContents($tmpfile) ."\n";
      unlink $tmpfile;
    }
    die "@msg\n";
  }
  unlink $tmpfile if (-f $tmpfile);  
  push @log,"\nSuccessfully scheduled new task \"$taskname\":\n",@$out;
        
  #if we get here and there's something in the trackerfile, call that
  if (exists $parms->{trackerfile} && &trackScheduledJob( { action=>"create",
                            taskname=>$taskname,
                            trackerfile=>$parms->{trackerfile},
                            "exec" => $parms->{exec},
                            time => ($WINDOWS) ? $parms->{wintimestring} : $parms->{unixtimestring},
                            })) {
    push @log,"Stored job-tracking information at $parms->{trackerfile}.\n";
  } else {
    push @log,"No updates were needed/made in job-tracking information at $parms->{trackerfile}.\n";
  }
  
  &Kline::Display::showDebug("Spooled log output:\n@log",1);
  return wantarray ? ($status,\@log): $status;
  
} #end createScheduledJob

#----------------------------------------------
sub _schtasksSvcRunning {
#Checks whether the scheduler service is running.  Dies if no, returns log if yes
#----------------------------------------------
  my @log;
  my $status={};
  #&Win32::Service::GetServices("",$status); #gets long name as key, short name as val
  my $svc="schedule"; #short name schedule, long name "Task Scheduler"
  &Win32::Service::GetStatus("","schedule",$status); #must use the service short name
  #CurrentState=1 not running, 4=running http://msdn.microsoft.com/en-us/library/ms685996(VS.85).aspx
  if ($status->{CurrentState}==4) { #running
    push @log,"Verified that the scheduler service is running.\n";
  } elsif ($status->{CurrentState}==1) { #not running
    &Win32::Service::StartService("",$svc) || die "Could not start service \"$svc\": $!\n"; #try to start if not already started
    push @log,"Successfully started service \"$svc\" (short name)\n";
  } else { #unknown status
      die "Service \"$svc\" is not stopped and is not running.  Cannot continue\n";
  }
  return @log;

} #end _schtasksSvcRunning
  
#----------------------------------------------
sub _findSchedUtil {
#Internal sub to find the correct scheduler util to be used (really only needed on Win)
#----------------------------------------------
  my @log;
  my $parms=shift;
  my $cmd=$parms->{schtasks} || &Kline::Server::findSysCmd2("schtasks.exe"); #will look in bindir and libdir, else sys dir
  $cmd=$parms->{jt} || &Kline::Server::findSysCmd2("jt.exe") unless $cmd; #will look in bindir, libdir, sys dirs
  push @log,"Using scheduling utility at $cmd.\n";
  return ($cmd,\@log);
  
} #end _findSchedUtil

#----------------------------------------------
sub _queryWinTasks {
#INTERNAL sub to query whether a schtask/jt is already setup or not
#returns the scheduler line (task + next exec date) if found, else undef
#----------------------------------------------
  my $util=shift || die "Need a utility location";
  my $taskname=shift || die "Need a taskname";
  my @log;
  import Kline::Server qw($winquote);
  my $cmd="$winquote$util$winquote";
  $cmd.=($util=~/schtasks/i) ? "/query /nh" : "/se"; #switches for either schtasks or jt.exe
  my ($rc,$out)=&Kline::Server::cmdLine({
                                           command=>$cmd,
                                           shift_rc=>1,
                                           verbose=>1,
                                           });
  die "ERROR: Cannot check current scheduled tasks using $cmd (RC=$rc): @$out" if $rc;
  push @log,"\nFound the currently scheduled tasks:\n",@$out;
    
  push @log,"\nSearching for task named \"$taskname\".\n";
  foreach (@$out) {
    next if /^\s*$|^\[TRACE\]/; #whitespace
    if (/^($taskname)(\.job)?\b/i) {
        push @log,"Found task at $_\n";
        my $line=chomp($_);
        return ($line,\@log);
    }
  }
  push @log,"Found no task named \"$taskname\".\n";
  return (undef,\@log);
  
} #end _queryWinTasks

#----------------------------------------------
sub _queryUNIXTasks {
#INTERNAL sub to check crontab entry for current user (only!).  Receive a taskname (token)
#to search for in comments preceeding a job.  Return the job if found, else undef
#CAVEAT: Jobs must have appropriate token commenting
#----------------------------------------------
  my $parms=shift;
  my $util=$parms->{util} || die "Need a crontab location";
  my $taskname=$parms->{taskname} || die "Need a taskname";
  my $user=$parms->{user} || "";
  my ($entry,@log); #entry is the crontab entry that matches, if any
  my $cmd="$util -l";
  my ($rc,$out)=&Kline::Server::cmdLine( { command => $cmd,
                                            shift_rc=>1,
                                            user => $user,
                                            verbose=>1,
                                            } );
  if ($rc) { #means we couldn't see cron jobs, which likely means nothing is setup, but regardless means we're done here
    push @log,"No current cron jobs were found.\n";
    return ($entry,\@log);
  }
  push @log,"\nSearching for task named \"$taskname\".\n";
  my $line;
  foreach (@$out) { #loop through crontab and look for token
    $line++;
    if (/^#$tokenprefix$taskname/ .. /^#${tokenprefix}END $taskname/) {
      push @log,"Found task \"$taskname\" in the following entry (beginning at crontab line $line):\n\n" unless $entry;
      push @log,$_;
      $entry.=$_; #incl CRLF
    }
    
  }
  push @log,"No matching jobs were found for task $taskname.\n" unless $entry;
  return ($entry,\@log);

} #end _queryUNIXTasks

#----------------------------------------------
sub trackScheduledJob {
#----------------------------------------------
=over 4

=item trackScheduledJob()

 sample code

Add or remove a scheduled job to a config file so it can later be tracked and removed if needed.  If hashref contains "action", it may either be "create" or "remove".  If null, assume create.  Return true for success, false for fail (no change needed due to job already there or already removed).

Input values correspond as follows:
action. Optional.  Can be "query, "create" or "remove".  Assume query if null.  Query returns array of task names
trackerfile.  Required.  Tracker file where we keep a list of all scheduled tasks under our control.
taskname. Required if action is "create" or "remove".  The name of the task to add/remove
exec.  Optional.  Show what will execute at the time (so that a query can detail it).  Default is empty string.
time.  Optional.  Show what time the job executes.  Default is empty string.
format.  Optional.  Can be either "parsed" or "printed".  If "parsed", then results are passed back as a AoH for parsed use.  Else default "printed" where results come back as a paragraph for pretty printing.

File is formatted as TASK_NAME;USER;EXEC OPTS,EXEC TIME,TIME_ADDED, where TASK_NAME is either the task name (Win) or token (UNIX) by which the job is known.  USER is (UNIX) the user under which the task exists (not needed on windows).

=back

=cut

  my $change; #flag to show whether we need to write new file
  my $user=($WINDOWS) ? &Win32::LoginName : getpwuid($EUID);
  my $parms=shift || die "Need parms";
  my $file=$parms->{trackerfile} || die "Need a tracker file";
  my $format;
  if (exists $parms->{format} && $parms->{format}=~/parse/i) {
    $format=0; #return AoH, parsed
  } else { 
    $format=1; #1=return pretty format, not parsed for use
  }
    
  $parms->{action}="query" unless exists $parms->{action};
  
  my @content=&Kline::Utils::parseDelimited( { file=>$file,
                                                delim=>$trackerdelim,
                                                print=>$format, #1=send in text, not parsed
                                                key_prefix=>$key_prefix,
                                                } );
  
  if ($parms->{action}=~/query/i) { #parse the current stuff and return as href
    return wantarray ? @content : \@content;
  }
  
  @content=&Kline::Utils::getFileContents($file); #the parsed format is actually less easy to deal with at this point
  if ($parms->{action}=~/create/i) { 
    my @keys=("TASK NAME","OWNER","EXEC STRING","EXEC TIME","TIME CREATED");
    my $header="#This file was written by $progname to track installed scheduled tasks.  Do not edit.\n";
    $header.="#$key_prefix$trackerdelim". join($trackerdelim,@keys) . "\n"; #The header of the CSV needs to include a key for use when parsing, so the fields aren't hard coded into anything.  The parseDelimited function will parse out whatever it finds
    $parms->{exec}="" unless $parms->{exec};
    $parms->{time}="" unless $parms->{time};
    my $addstr="$parms->{taskname}$trackerdelim$user$trackerdelim$parms->{exec}$trackerdelim$parms->{time}$trackerdelim".localtime()."\n";
    if (@content && grep /\S/,@content) { #add additionals after checking not there already
      if (grep /^$parms->{taskname}$trackerdelim/i,@content) {
        &Kline::Display::showDebug("Entry already exists at tracker file $file for $parms->{taskname}",1);
      } else {
        $change++;
        &Kline::Display::showDebug("Adding entry to existing tracker file $file for $parms->{taskname}",1);
        push @content,$addstr;
      }
    } else { #no current content in file
      $change++;
      &Kline::Display::showDebug("Adding first entry to tracker file $file for task $parms->{taskname}",1);
      push @content,$header,$addstr;
    }
  } elsif ($parms->{action}=~/remove/i) { #delete
    if (@content) { #existing file
      if (grep /^$parms->{taskname}$trackerdelim/i,@content) {
        &Kline::Display::showDebug("Removing entry from existing tracker file $file for $parms->{taskname}",1);
        $change++;
        @content=grep !/^$parms->{taskname}$trackerdelim/i,@content;
      } else {
        &Kline::Display::showDebug("Entry does not exist at tracker file $file for $parms->{taskname}",1);
      }
    } else {
      &Kline::Display::showDebug("No tracker file exists at $file; nothing to delete",1);
    }
  } else {
    die "Invalid action type.  Must be query, create or remove";
  }
    
  if ($change) {
    open FILE,">$file" || die "Cannot open tracker file for output $file: $!\n";
    print FILE @content;
    close FILE;
    return 1; #change made
  }
  return 0;  #no change made

}#end trackScheduledJob

#----------------------------------------------
sub deleteScheduledJob {
#----------------------------------------------
=over 4

=item createScheduledJob()

 sample code

Remove the appropriate scheduled job (by task name, not other characteristics).  Passes parms through to findSchedUtil (Win).  If a schedule-tracking file is provided, the job will be removed from there as well.  Returns true for success, false for fail, followed by aref with log

Input values correspond as follows:
as_user.  Optional.  If populated, Call the system command(s) on UNIX as this user (only used if root).
taskname.  Optional.  Defaults to user.progname
trackerfile.  Optional.  Remove the job from the schedule file provided.


=back

=cut

  my $parms=shift;
  my ($status,@log); #stuff we'll return
  $status=1; #assume all good
  my ($cmd,$rc,$out); #the command used to set the job
  my ($progname,undef,undef)=fileparse($0, qr/\.[^.]*/); #strip the extension of the current prog as a taskname
  my $user=($WINDOWS) ? &Win32::LoginName : getpwuid($EUID);
  my $taskname=$parms->{taskname} || "$progname.$user";
  my $tmpfile="/tmp/crontab.$user.new";
  my $as_user=$parms->{as_user} || undef;
  
  push @log,"\nReceived the following execution parameters\n";
  foreach my $parm (sort keys %$parms) {
      push @log,"  $parm = $parms->{$parm}\n";
  }
  push @log,"\nBegin deleting scheduled task at ". localtime() ."\n";
  
  #############################
  if ($WINDOWS) { #WIN, check for schtasks or jt
    my ($util,$llog)=&_findSchedUtil($parms);
    push @log,@$llog if @$llog;
    import Kline::Server qw($winquote);
    die "Cannot find scheduler utility $util\n" unless -f $util;
    if ($util=~/schtasks/i) { #XP and newer
        push @log,&_schtasksSvcRunning(); #make sure the service is running; will add log stuff and die if not running
        
        #instructions for using task scheduler via cmdline at http://support.microsoft.com/?kbid=814596
        my $task;
        ($task,$llog)=&_queryWinTasks($util,$taskname);
        push @log,@$llog if @$llog;
        unless ($task) { #doesn't exist
            push @log,"Requested task does not exist.  \nNo action necessary.\n";
            $status=0;
            &Kline::Display::showDebug("Spooled log output:\n@log",1);
            return wantarray ? ($status,\@log): $status;
        }
        
        #get here then we need to delete a task
        #use schtasks /delete /? for help
        $cmd="$winquote$util$winquote /delete /tn $winquote$taskname$winquote /f";
    } elsif ($util=~/jt/i) {
         die "Sorry, not yet coded for jt.exe\n"; 
    } else {
        die "Could not find scheduling utility to use";
    }
  #############################
  } else { #UNIX
    my $util=&Kline::Server::findSysCmd("crontab") || die "Cannot find crontab command\n";
    my ($task,$llog)=&_queryUNIXTasks( {
                                        util =>$util,
                                        taskname => $taskname,
                                        user => $as_user,
                                       } );
    push @log,@$llog if @$llog;
    unless ($task) { #already scheduled
        push @log,"Requested task does not exist.  \nNo action necessary.\n";
        $status=0;
        &Kline::Display::showDebug("Spooled log output:\n@log",1);
        return wantarray ? ($status,\@log): $status;
    }
        
    #get here then we need to delete task
    ($rc,$out)=&Kline::Server::cmdLine({
                                          command=>"$util -l",
                                          shift_rc=>1,
                                          user => $as_user,
                                          verbose=>1,
                                          }); #no reason for this to fail since we added jobs previously
    
    my @current=split /\n/,$task;
    foreach (@current) { $_.="\n"; };
    $out=&Kline::Utils::cmpArray($out,\@current,"SIMPDIFF");  #find all the stuff in out not in task
    
    #write the new file
    open OUT,">$tmpfile" || die "Cannot open temp crontab file $tmpfile: $!\n";
    print OUT @$out if $rc==0 && @$out; #add current jobs first
    close OUT;
    
    #set the new file
    $cmd="$util $tmpfile";
  } #end UNIX vs Win
  
  push @log,"\nAttempting to delete a scheduled task with command: \"$cmd\".\n";
  ($rc,$out)=&Kline::Server::cmdLine({ command=>$cmd,
                                         shift_rc=>1,
                                         user=>$as_user, #ignored for windows
                                         verbose=>1,
                                        });
  die "ERROR: Cannot delete scheduled task using $cmd (RC=$rc): @$out" if $rc;
  push @log,"\nSuccessfully deleted scheduled task \"$taskname\":\n",@$out;
  unlink $tmpfile if (-f $tmpfile);  
        
  #if we get here and there's something in the trackerfile, call that
  if (exists $parms->{trackerfile}) {
    if (&trackScheduledJob( { action=>"remove",
                             taskname=>$taskname,
                             trackerfile=>$parms->{trackerfile},
                             })) {
      push @log,"Removed job-tracking information at $parms->{trackerfile}.\n";
    } else {
      push @log,"No job-tracking information was found/needed update at $parms->{trackerfile}.\n";
    }
  } else {
    &Kline::Display::showDebug("No tracker file provided; no tracking updates will be done.",1);
  }
  
  &Kline::Display::showDebug("Spooled log output:\n@log",1);
  return wantarray ? ($status,\@log): $status;
  
}#end deleteScheduledJob

#----------------------------------------------
sub deleteAllScheduledJobs {
#----------------------------------------------
=over 4

=item createScheduledJob()

 sample code

Receieve a file that contains a list of scheduled tasks and removes them all by calling deleteScheduledJob for each one.  Returns true for success, false for problems, followed by aref with log info.

=back

=cut

  my @log;
  my $file=shift || die "Need tracker file";
  push @log,"\nRemoving ALL scheduled jobs as contained at $file.\n";
  my @jobs=&trackScheduledJob( { action=>"query",
                                 trackerfile=>$file,
                                 format=>"parsed", #will return as AoH
                                 });
  my $user=($WINDOWS) ? $Win32::LoginName : getpwuid($EUID);
  push @log,"\nFound " .@jobs. " scheduled jobs for which to attempt removal.\n";
  foreach my $job (@jobs) { #href
    { #naked block to ensure we have the parms we expect
      my @found=keys %$job;
      my @reqd=("TASK NAME","OWNER","TIME CREATED","EXEC TIME");
      my @missing=&Kline::Utils::cmpArray(\@reqd,\@found,"SIMPDIFF"); #return those reqd that are missing
      die "file $file is missing required parms \"@missing\"" if @missing;
    } #naked block
    
    my $su="";
    if ($user!~/\b$job->{OWNER}\b/i) { #can only remove my own jobs if not root
      my $nogo=0;
      if ($WINDOWS) { #WIN, cur user doesn't match job owner
        $nogo++ unless &Win32::isAdminUser(); #can't remove unless in admin group
      } else { #UNIX, cur user doesn't match job owner and is not root
        #on unix, need to su -c
        $su=$job->{OWNER};
        $nogo++ unless $EUID==0; #must be root
      }
      return (0,["ERROR: Cannot remove scheduled task \"$job->{'TASK NAME'}\" for user $job->{OWNER}.  Must be done either as that user or root/Administrator.\n"]) if $nogo;
    } #end user-auth check to remove
    push @log,"\nRemoving job for task name \"$job->{'TASK NAME'}\" (scheduled by user: $job->{OWNER}), scheduled on $job->{'TIME CREATED'} to run at \"$job->{'EXEC TIME'}\".\n";
    my ($status,$llog)=&deleteScheduledJob( { taskname=>$job->{'TASK NAME'},
                                              trackerfile=>$file,
                                              as_user=>$su,
                                            });
    push @log,@$llog if @$llog;
    return ($status,\@log) unless $status;
  } #foreach job
  return (1,\@log); #get here then success even if no jobs needed removing
  
}#end deleteAllScheduledJobs

#----------------------------------------------
1; #return true
__END__

#ADD'L DOCUMENTATION-----------------------------------------------------------

=head1 AUTHOR

Chris Kline <cnkline@us.ibm.com>

=head1 SEE ALSO

N/A.

=head1 COPYRIGHT

(C) Copyright IBM Corp. 2008 All Rights Reserved

=cut
