#!/usr/bin/perl
package Kline::Test;
# Chris Kline; cnkline@us.ibm.com
# Copyright IBM Corp. 2008 All Rights Reserved
# ---------------------------------------------------------------------------
#
# Shared test libraries for Kline
#
# ---------------------------------------------------------------------------
# REVISION HISTORY
# 1.0    02/06/08 Initial.
#
#-------------------------------------------------------------------------------
# Source Control Information
# $Revision: 22 $
# $LastChangedDate: 2012-01-10 20:43:31 -0700 (Tue, 10 Jan 2012) $
# $Author: cnkline $
#

# ---------------------------------------------------------------------------
# INITIALIZE THE ENVIRONMENT

use strict;
use warnings;
use English; 		#D:\perl_oreilly\prog\ch07_013.htm

use File::Basename qw(basename);
use FindBin;
use Cwd qw(abs_path);
use File::Spec::Functions;

use Kline::Inventory::WMQ;

my $scriptVer="1.0";
my $scriptDate; #backward compatibility
($scriptDate = '$LastChangedDate: 2012-01-10 20:43:31 -0700 (Tue, 10 Jan 2012) $') =~ s/.*\(\D+(\d+\s\w+\s\d+)\).+/$1/; #auto update upon commit
our $VERSION=$scriptVer;
our $VERDATE=$scriptDate;

=head1 NAME

Kline::Test - Helper test functions for Kline modules.

=head1 SYNOPSIS

 use Kline::Test;

=head1 DESCRIPTION

This module provides helper functions that are used for testing other Kline libraries.  Many functions require certain system or WMQ-configuration settings in order to be properly tested.  Therefore, this suite can be used to enable and disable certain test conditions.

=cut

#VARIABLES-------------------------------------------------------------------

=head2 GLOBAL VARIABLES

The following variables are set for global use in this module.

=over 4

=item $testcfgdir

We will always assume that the test-configuration directory, where temp files are stored and where control files (such as pre/post-test markers) may be found, will be found relative to the executing test program using these libraries.  That is, if the test is run from t/test1.t, then we assume that t/config/ is the config-file directory.

=back

=cut

our $testcfgdir;
$testcfgdir=catfile($FindBin::Bin,"config");
$testcfgdir=abs_path($testcfgdir) || die "testcfg directory does not exist";


=head2 METHODS

The following methods are available in this module.

=cut

#METHODS---------------------------------------------------------------------

{ #naked block for $exposed
my $exposed;
#----------------------------------------------
sub getExposedInventory {
#----------------------------------------------

=over 4

=item getExposedInventory()

Looks for cached MQ inventory exposed results within the C<$testcfgdir>.  If found, it loads them.  If not found, it runs the MQ inventory (without writing to XML) and writes to the cache file for future speed reuse.  Remove the cache file to force a reload.  Always returns the exposed results (array ref of hash refs, one for each queue manager).

If called more than once in a single session, return the already loaded information instead of going to the cache file again.

Receives no arguments.  Returns the inventory results ref.

=back

=cut

  return $exposed if $exposed; #if we already have results, just return them

  my $cachefile=catfile($testcfgdir,"WMQ_inventory.cache");
  if (-f $cachefile) {
    &Kline::Display::showDebug("Loading cached results from $cachefile",1);
    $exposed=&Kline::Utils::loadDataDumper($cachefile);
  } else { #no file, write one
    &Kline::Display::showDebug("Saving inventory results to cache file",1);
    ($exposed,undef)=&Kline::Inventory::WMQ::main(); #passing no args means no XML to write, returns log as well, but we don't care
    &Kline::Utils::saveDataDumper($exposed,$cachefile,"Exposed-inventory results from Kline::Inventory::WMQ::main");
  }

  return $exposed;

} #end getExposedInventory
} #end exposed naked block

{# naked block for $changes
my $changes; #href to track whether changes were made right now
#----------------------------------------------
sub pretestConfig {
#----------------------------------------------

=over 4

=item pretestConfig()

 &Kline::Test::pretestConfig($scriptname,$parms);

Runs a pretest-configuration script if a marker file doesn't yet exist, indicating that this was already done.  If the marker file exists, the configuration is skipped (assumed already complete) and the C<<$changes->{scriptname}>> value is set to false (showing no changes were made).  If the marker is not found and changes are then made, the C<<$changes->{scriptname}>> value is set to true, so that C<posttestConfig()> can undo the changes later.  

If you desire to persist a pretest condition beyond the actual test itself, run the pretest-configuration script in a separate execution cycle, where the marker file will be created but the $change value will not persist, so that C<posttestConfig()> will not undo any changes.  In such a case, changes will need to be manually undone as well.

Returns true if changes are successfully made, false if changes are attempted but external script exits nonzero, and undef if no changes are attempted.  If changes fail, the C<<$changes->{script}>> value will be reset to zero so no further changes will be attempted.

Input values follow:

B<(Required) $scriptname. String file name.>  This is the name of the file that will make the changes (both the "do" and "undo" changes should be in this same file.  To make changes, a "do" argument will be passed as the first argument when the script is called.

B<(Optional) $parms. String.>  This represents any additional parameters that you may wish to pass into $scriptname.

=back

=cut

  my $script=shift || die "No script name was passed";
  die "Script $script does not exist or is not executable\n" unless -x $script;
  my $parms=shift if @_;
  my $marker=catfile($testcfgdir,basename($script) .".marker");
  my $called_by=(caller(1))[3] || (caller(0))[1]; #get current stack if no parent
  print "\n$called_by PRE-TEST: \n";

  if (-f $marker) {
    print "  Marker already exists.  No pre-test changes need to be made.\n";
    $changes->{$script}=0;
    return undef;
  }
  my $cmd="$script do";
  $cmd.=" $parms" if $parms;
  my $outfile=catfile($testcfgdir,basename($script) .".out");
  $cmd.=" 2>&1 >$outfile"; #save the output
  if (system($cmd)==0) {
    print "  Pre-test changes successful with \"$script\"\n";
    $changes->{$script}=1; #signal that we made pretest changes
    #write the marker
    open MARKER,">$marker" || die "Cannot create marker at $marker: $!\n";
    print MARKER "$cmd\n";
    close MARKER;

    return 1;
  } else { #nonzero return code
    warn "  ERROR: \"$cmd\" did not exit with 0: ($?) $!\n";
    $changes->{$script}=0;  #don't undo anything later
    return 0;
  }

} #end 

#----------------------------------------------
sub posttestConfig {
#----------------------------------------------

=over 4

=item posttestConfig()

 &Kline::Test::posttestConfig($scriptname,$parms);

Runs a posttest-configuration script if C<<$changes->{scriptname}>> is set to true, indicating that the C<prerestConfig()> made changes at the beginning of this test.  In other words, this undoes any changes made by the previous sub.  Removes the marker if it undoes any configuration.  If no changes were made at the beginning, this sub simply returns without making any further changes.

Returns true if changes are successfully made, false if changes are attempted but external script exits nonzero, and undef if no changes are attempted.

Input values follow:

B<(Required) $scriptname. String file name.>  This is the name of the file that will make the changes (both the "do" and "undo" changes should be in this same file.  To unmake changes, an "undo" argument will be passed as the first argument when the script is called.

B<(Optional) $parms. String.>  This represents any additional parameters that you may wish to pass into $scriptname.

=back

=cut

  my $script=shift || die "No script name was passed";
  die "Script $script does not exist or is not executable\n" unless -x $script;
  my $parms=shift if @_;
  my $marker=catfile($testcfgdir,basename($script) .".marker");
  my $called_by=(caller(1))[3] || (caller(0))[1]; #get current stack if no parent
  print "\n$called_by POST-TEST: \n";

  unless (exists $changes->{$script} && $changes->{$script} && -f $marker) {
    print "  No pre-test changes were made.  No post-test changes need to be made.\n";
    return undef;
  }
  my $cmd="$script undo";
  $cmd.=" $parms" if $parms;
  my $outfile=catfile($testcfgdir,basename($script) .".out");
  $cmd.=" 2>&1 >>$outfile"; #save the output (with append because of above)
  if (system($cmd)==0) {
    print "  Post-test changes successful with \"$script\"\n";
    $changes->{$script}=1; #signal that we made pretest changes
    #remove the marker
    unlink $marker || die "Cannot remove marker $marker: $!\n";
    delete $changes->{$script};

    return 1;
  } else { #nonzero return code
    warn "  ERROR: \"$cmd\" did not exit with 0: ($?) $!\n";
    delete $changes->{$script};
    return 0;
  }


} #end posttestConfig
} #end $changes naked block

#----------------------------------------------
1; #tell the sourcing program we are true

__END__

#ADD'L DOCUMENTATION-----------------------------------------------------------

=head1 AUTHOR

Chris Kline <cnkline@us.ibm.com>

=head1 SEE ALSO

N/A.

=head1 COPYRIGHT

(C) Copyright IBM Corp. 2008 All Rights Reserved

=cut

