#!/usr/bin/perl
package Kline::Utils;
# Utils.pm
# Chris Kline; Chris.Kline@ca.com
# ---------------------------------------------------------------------------
#
# This script is a library of commonly used functions for Kline programs that
# use Perl.  It can be "sourced in" using the "require" pragma.
# This library contains only functions that perform utility-related activities.
# These functions perform generic activities that can be used for many reasons.
# These functions may call functions from other libraries.
#
# ---------------------------------------------------------------------------
# REVISION HISTORY
# 2.0   -BT50563: Split Kline.pl according to genre.  First ed of this file.
# 2.1   Added saveDataDumper and loadDataDumper to easily save hashes to a file 
#       and back.
# 2.1.1 Updated saveDataDumper and loadDataDumper for better syntax
# 2.1.2 Added by_last_mod to sort files by last mod date, fixed saveDataDumper 
#       to handle arrays
# 2.2   -BT56509: Added fileStat to get pretty much any stat off a file
#       Changed fileModDate to use fileStat so we always use the OO File::stat
#       Changed by_last_mod to use fileStat
#2.2.1 -BT59353 Updated mailiing to also look at /bin/mail
#2.2.2 -BT59674 problems with stat and fileStat function in old perl versions
#2.3   -Added convertWeekdayNum and covertWeekdayName and weekOfMonth
#      -Added convertRange
#2.3.1 -Added getFileContents (first used by qmgrStatus for MQInventory)
#2.3.2 -BT67328, convertRange now takes an array
#3.0   -BT66191, move to true perl module
#3.1   -BT69832, Added changeMode,systemCall for Gnosis SCP
#3.1.1  Updates in conjunction with Gnosis and Fusion for the trim function to
#       not return an empty value
#3.1.2  Updated trim function to not remove empty elements
#3.2    06/20/08 Added parseCSVtoAoH, removed SystemCall and combined into
#                Kline::Server::cmdLine
#       07/31/08 Added random key generator
#3.2.1  12/05/11 Added by_alphabetical
#3.2.2  01/10/12 Added CC option to sendMail
#3.2.3  06/06/12 Fixes in SendMail for non-array email addresses (like for singles)
#3.2.4  07/05/12 Fixes for SendMail for cc email addresses (typo before)
#3.2.5  11/06/12 Added roundInteger (used by Wily metrics)
#3.2.6  11/19/12 Fixed trim to account for Unicode whitespace
#
#-------------------------------------------------------------------------------
# Source Control Information
# $Revision: 58 $
# $LastChangedDate: 2012-11-20 15:47:21 -0700 (Tue, 20 Nov 2012) $
# $Author: cnkline $
#
# ---------------------------------------------------------------------------
# INITIALIZE THE ENVIRONMENT

use strict;
use English; 		#D:\perl_oreilly\prog\ch07_013.htm
require Exporter;
our @ISA=qw(Exporter);
our @EXPORT=qw(by_alphabetical by_last_mod by_mostly_numeric cmpArray convertMonthName
convertMonthNum convertRange convertWeekdayName convertWeekdayNum fileModDate
fileSize fileStat flushBuffer generateMailMsg getFileContents getUnique
loadDataDumper lowerCase parentFunc saveDataDumper showTime thisFunc trim
weekofMonth touchFile);

use File::stat;         #used to find date stamp on binary files for versions, changeMode
use Time::Local;
use POSIX qw(strftime); #for converting timestamps
use Data::Dumper;
use Net::SMTP;		#builtin 5.8+
use MIME::Lite;         #lib/public
use Carp;

use lib '..'; #so we can run these direct to make sure they compile and not fail for missing mods
use lib '../public'; #so we can run these direct to make sure they compile and not fail for missing mods
use Kline::Display;
#BEGIN { $Exporter::Verbose=1 }

my $scriptVer="3.2.6";
my $scriptDate; #backward compatibility
($scriptDate = '$LastChangedDate: 2012-11-20 15:47:21 -0700 (Tue, 20 Nov 2012) $') =~ s/.*\(\D+(\d+\s\w+\s\d+)\).+/$1/; #auto update upon commit

#----------------------------------------------------------------------------

#----------------------------------------------
sub by_alphabetical {
# sort alphabetic (default is asciibetical)
#----------------------------------------------

  lc($a) cmp lc($b);

} #end by_alphabetical

#----------------------------------------------
sub by_mostly_numeric {
# sort given numeric filenames (*.LOG files) by numbers
# for use when managing MQ linear log files
# Only works if there are some numbers in the array
#----------------------------------------------

  ($a <=> $b) || ($a cmp $b);

} #end by_mostly_numeric

# -----------------------------------------------
sub showTime {
# gets local time and displays dd/mm hh:mm format, which fits well into
# log files and is wrapped by () for easy parsing
# -----------------------------------------------
    #not getting this method to work correctly
    #use Time::localtime;
    #my $tm = localtime;
    #return "($tm->mday/$tm->mon $tm->hours:$tm->min)";

    #another method, but returning empty values right now
    #return sprintf("(%02d/%02d/%02d %02d:%02d)",
    #  sub { ($_[4]+1, $_[3], $_[5]+1900, $_[2], $_[1])}->(localtime));

    #yet another method using POSIX; more reqmts, but seems to work
    return strftime("(%m/%d %H:%M:%S)",localtime);

} #end showTime

# -----------------------------------------------
sub generateMailMsgOLD {
# Receives $1 as either "email","pager", or "custom", followed by $2 as the subject
# and $3 as the message contents
# If custom, then $4 should be addr list (comma separated)
# Function will find the appropriate addresses and send the message
# returns true if success
# -----------------------------------------------

    my ($type,$subj,$body) = @_;
    my @addr; #space-separated list of mail addresses
    my $mailcmd;

    if ($type && $type =~ /email/i) {
        @addr = &getEmails("email");
    } elsif ($type && $type =~ /pager/i) {
        @addr = &getEmails("pager email");
    } elsif ($type && $type =~/custom/i) {
        @addr = split /,/,$_[3]; #4th arg
        &Kline::Display::showDebug("Using custom email addrs: @addr",1);
    } else { #something went wrong
        die "generateMsg was called incorrectly\n";
    }
  
    #make sure we have valid addrs
    if (grep(/your_name/i,@addr) || !@addr) {
      #name is either default or empty
      warn "No valid mail address found: \"@addr\". Cannot send mail!\n";
      return 0;
    }
    #find the right mail command
    if ( -f "/usr/bin/mailx" ) {
      $mailcmd="/usr/bin/mailx";
    } elsif ( -f "/usr/bin/mail" ) {
      $mailcmd="/usr/bin/mail";
    } elsif ( -f "/bin/mail" ) {
      $mailcmd="/bin/mail";
    } else { #no mail command found, we're toast
      die "Cannot find valid mail command!  No mail can be sent!\n";
    }
    &Kline::Display::showDebug("Using mail command $mailcmd",1);

    #now generate the actual message
    open(MAIL, "|$mailcmd -s \"$subj\" \"@addr\"") or die "can't fork mail: $!";
    #why does sendmail directly not work
    #open(MAIL, "|/usr/lib/sendmail -oi -s $subj $addr") or die "can't fork sendmail: $!";
    print MAIL $body;
    close(MAIL);
    return 1;

} #end generateMailMsgOLD

#----------------------------------------------
sub sendMail {
#CONSIDER USING KLINE::MESSAGE INSTEAD
#Send mail, platform-portable, using Net::SMTP, included in Perl 5.8+
#supports attachments.  Receives href with following
#@{address} #scalar or aref
#{from} defaults to noreply@nowhere.com
#{subject} may be empty, but must be defined
#{body}
#@{attachment} scalar or aref OPTIONAL
#{smtprelay} REQUIRED
#{timeout} default 60 sec
#{authuser} OPTIONAL
#{authpassword} OPTIONAL
#{port} OPTIONAL
#----------------------------------------------
  &showDebug("Deprecated.  Consider Using Kline::Message Instead!!!",1);  
  my $h=shift || die "Missing href argument";
  &Kline::Display::dumpHref($h,2);
  my %mimetypes=('html'=>'text/html','pdf'=>'application/pdf','xls'=>'application/x-excel','txt'=>'text/plain','xml'=>'text/xml','csv'=>'text/csv');
  $h->{timeout}=60 unless $h->{timeout};

  #convert input addr array if needed
  die "Missing TO: address" unless $h->{address};
  my $mailToHeader = "";
  my @smtpAddresses;
  my $counter = 0;
  if (ref $h->{address} eq 'ARRAY') {
    @smtpAddresses=@{$h->{address}};
	
    foreach (@smtpAddresses) {
      if ($counter == 0) {
        $mailToHeader=$_;
      } else {
        $mailToHeader=$mailToHeader . "," . $_;
      }
      $counter++;
    }
    &showDebug("Sending mail TO $mailToHeader",1);
  } else {
    &showDebug("Input addresses were scalar and will not be converted; sending mail TO $h->{address}",1);
    $mailToHeader=$h->{address};
  }
 
  $h->{from}='noreply@nowhere.com' unless $h->{from};
  die "Missing SUBJECT:" unless defined($h->{subject});
  die "Missing SMTP Relay" unless $h->{smtprelay};
  my $msg = MIME::Lite->new(
    From    => $h->{from},
    To      => $mailToHeader,
    Subject => $h->{subject},
    Type    =>'multipart/mixed',
  );

  #if cc, add them separately
  if ($h->{cc}) {
    my ($mailCCHeader, @smtpCCAddresses);
    &showDebug("Adding CC addresses $h->{cc}",1);
    if (ref $h->{cc} eq 'ARRAY') {
      @smtpCCAddresses=@{$h->{cc}};
      my $counter = 0;
      foreach (@smtpCCAddresses) {
        if ($counter == 0) {
          $mailCCHeader=$_;
        } else {
          $mailCCHeader=$mailCCHeader . "," . $_;
        }
        $counter++;
      }
      &showDebug("Sending mail CC to $mailCCHeader",1);
    } else {
      &showDebug("Input CC addresses were scalar and will not be converted: $h->{cc}",1);
      $mailCCHeader=$h->{cc};
    }    
    &showDebug("Sending CC mail to $mailCCHeader",1);
    $msg->add("Cc" => $mailCCHeader);
  }
  
  #attachments need a mime type, and the body needs to be attached itself
  if ($h->{attachment}) {
    $msg->attach(
	Type     =>'TEXT',
	Data     =>$h->{body},
    );
	
    unless (ref $h->{attachment} eq 'ARRAY') {
      $h->{attachment}=[$h->{attachment}];
    }
    &showDebug(@{$h->{attachment}} . " Attachments requested.",1);
	
    foreach (@{$h->{attachment}}) {
      die "Cannot find attachment $_" unless (-f $_);
      my $attachmentFileName=$_;
      $attachmentFileName=~s/\W/_/g; #no spaces
      if (/\.(\w+)$/) { #at this point, all attachments need extensions to match mime type
        my $extension = $1;
        $msg->attach(
          Type        => $mimetypes{$extension},
          Path        => $_,
          #Filename    =>"$attachmentFileName.$extension", #this is messing up names, don't use
          Disposition => 'attachment'
        );
      }
    } #foreach
  } else {
    &showDebug("No attachments requested.",1);
  }
	
  &showDebug("Sending Mail",1);
  if ($h->{authuser}) { #with auth
    $h->{port}=25 unless $h->{port};
    $msg->send('smtp',$h->{smtprelay},Port => $h->{port},Timeout=>$h->{timeout},SkipBad => 1,User => $h->{authuser},Password => $h->{authpassword});
  } else { #no authentication
    $msg->send('smtp',$h->{smtprelay},Timeout=>$h->{timeout},SkipBad => 1);
  }
	
} #sendMail

#----------------------------------------------
sub trim {
#cut any leading or trailing whitespace (including \n) and return it
#can accept an array and will pass back another array
#from perl cookbook D:\perl_oreilly\cookbook\ch01_15.htm
#does not remove elements that are empty
#----------------------------------------------
    my @in=@_;
    my @out;
    foreach my $val (@in) {
      next unless (defined $val);
      #http://www.effectiveperlprogramming.com/blog/991
      #Found that whitespace wasn't always matched by \s because of unicode
      #replaced to \p{space} which matches them all, but only works on perl 5.12+, otherwise you get "can't find Unicode property definition 'space'"
      #Instead, open the file and decode it like this:  open my $fh, '<:encoding(UTF-8)', $filename;  # auto decoding on read
      $val=~s/^\s+//; #remove leading whitespace
      $val=~s/\s+$//; #remove trailing whitespace
      push @out,$val; #unless $val=~/^$/; (do not remove the line just because it's empty
    }
    return wantarray ? @out : $out[0];

} #end trim

#----------------------------------------------
sub lowerCase {
#take the input and return it all lowercase
#YOU SHOULD INSTEAD USE lc(), WHICH IS BUILTIN
#----------------------------------------------
    my @out = @_;
    for (@out) {
	s/([^\W0-9_])/\l$1/g
    }
    return wantarray ? @out : $out[0]
} #end lowerCase

#----------------------------------------------
sub getUnique {
#Receive an array and return only unique values
#Perl Cookbook D:\perl_oreilly\cookbook\ch04_07.htm
#return just the num uniq vals if scalar context
#----------------------------------------------
  my @list=@_; #input list
  my @uniq; #return list
  my %seen = ();
  foreach my $item (@list) {
    push(@uniq, $item) unless $seen{$item}++;
  }
  return wantarray ? @uniq : scalar(@uniq);

} #end getUnique

#----------------------------------------------
sub convertMonthNum {
#receives a month num and returns the name 
#http://perlmonks.thepen.com/95456.html
#----------------------------------------------
  my $month=shift@_;
  my %num2mon = qw(
    1 Jan  2 Feb  3 Mar  4 Apr  5 May  6 Jun
    7 Jul  8 Aug  9 Sep 10 Oct 11 Nov 12 Dec
  );

  return $num2mon{ lc substr($month, 0, 3) };

} #end convertMonthName

#----------------------------------------------
sub convertMonthName {
#receives a month name (case not important) and returns the number
#http://perlmonks.thepen.com/95456.html
#----------------------------------------------
  my $month=shift@_;
  my %mon2num = qw(
    jan 1  feb 2  mar 3  apr 4  may 5  jun 6
    jul 7  aug 8  sep 9  oct 10 nov 11 dec 12
  );

  return $mon2num{ lc substr($month, 0, 3) };

} #end convertMonthName

#----------------------------------------------
sub convertWeekdayNum {
#receives a weekday num and returns the name 
#http://perlmonks.thepen.com/95456.html
#----------------------------------------------
  my $weekday=shift@_;
  my %num2weekday = qw(
    0 Sun  1 Mon  2 Tue  3 Wed  4 Thu  5 Fri  6 Sat
  );

  return $num2weekday{ lc substr($weekday, 0, 3) };

} #end convertWeekdayNum

#----------------------------------------------
sub convertWeekdayName {
#receives a weekday name (case not important) and returns the number
#http://perlmonks.thepen.com/95456.html
#----------------------------------------------
  my $weekday=shift@_;
  my %week2num = qw(
    sun 0  mon 1  tue 2  wed 3  thu 4  fri  5  sat 6
  );

  return $week2num{ lc substr($weekday, 0, 3) };

} #end convertWeekdayName

#----------------------------------------------
sub thisFunc {
#returns name of the function (not this one, but the one calling this
#----------------------------------------------
  if (wantarray) {
    return caller(1);
  } else {
    my $called_by=(caller(1))[3];
    return $called_by || "NO_NAME";
  }
} #end thisFunc

#----------------------------------------------
sub parentFunc {
#returns the caller of the function calling this one; if no caller, show the
#filename that this was called from
#----------------------------------------------
  if (wantarray) {
    return caller(2);
  } else { 
    #keep going if only a scalar was wanted
    my $called_by=(caller(2))[3] || (caller(1))[1]; #get current stack if no parent
    $called_by=$1 if ($called_by && $called_by =~ /main::(.+)/); #keep the line short
    return $called_by || "NO_PARENT";
  }
} #end parentFunc

#----------------------------------------------
sub cmpArray {
# file:///D:/perl_oreilly/cookbook/ch04_09.htm
# Find the intersection, difference, or union between two lists.  
#
# REQUIRES reference to two arrays,
# followed by ISECT, UNION, SYMDIFF or SIMPDIFF parm.
# You have a pair of lists, each having unduplicated items. You'd like to find
# out which items are in both lists (intersection), one but not the other 
# (difference), or either (union). 
#
# We compute the symmetric difference here, not the simple difference. These 
# are set theoretic terms. A symmetric difference is the set of all the 
# elements that are members of either @A or @B, but not of both. A simple 
# difference is the set of members of @A but not of @B, which we calculated in
# the SIMPDIFF parm.
#
#----------------------------------------------
  #file:///D:/perl_oreilly/cookbook/ch10_06.htm
  my $this=&thisFunc;
  my ($a,$b,$parm)=@_;
  my (@a,@b,@result);
  #first check that we are receiving two array refs
  unless (@_ == 3 && ref($a) eq 'ARRAY' && ref($b) eq 'ARRAY') {
    warn "usage: $this ARRAYREF1 ARRAYREF2 PARM";
    return 0;
  }

  my (@union, @isect, @diff, %union, %isect, %count);
  &Kline::Display::showDebug("First array:\n@$a",2);
  &Kline::Display::showDebug("Second array:\n@$b",2);
  &Kline::Display::showDebug("converting arrays to unique vals",2);
  @a=&getUnique(@$a);
  @b=&getUnique(@$b);
  &Kline::Display::showDebug("Sorting arrays",2);
  @a=sort(@a);
  @b=sort(@b);
  @union = @isect = @diff = ();
  %union = %isect = ();
  %count = ();

  foreach my $e (@a, @b) {
    $count{$e}++;
  }

  foreach my $e (keys %count) {
    push(@union, $e);
    if ($count{$e} == 2) {
        push @isect, $e;
    } else {
        push @diff, $e;
    }
  }

  if ($parm =~/ISECT/i) {
    @result=sort(@isect);
    &Kline::Display::showDebug("Items found in BOTH arrays (INTERSECTION):\n@result",1);
  } elsif ($parm =~/UNION/i) {
    @result=sort(@union);
    &Kline::Display::showDebug("Items found in EITHER or BOTH array (UNION):\n@result",1);
  } elsif ($parm =~/SYMDIFF/i) {
    @result=sort(@diff);
    &Kline::Display::showDebug("Items found in only ONE or OTHER array (SYMMETRIC DIFF):\n@result",1);
  } elsif ($parm =~/SIMPDIFF/i) { #items in a but not in b
    #file:///D:/perl_oreilly/cookbook/ch04_08.htm
    my %seen = ();                  # lookup table to test membership of B
    my @aonly = ();                 # answer

    # build lookup table
    foreach my $item (@b) { $seen{$item} = 1 }

    # find only elements in @A and not in @B
    foreach my $item (@a) {
      unless ($seen{$item}) {
        # it's not in %seen, so add to @aonly
        push(@aonly, $item);
      }
    }
    @result=sort(@aonly);
    &Kline::Display::showDebug("Items found in first but NOT second array (SIMPLE DIFF):\n@result",1);
  } else { #invalid
    warn "invalid parm \"$parm\" passed to $this\n";
  }

  return ( wantarray ) ? @result : \@result;

} #end cmpArray

#----------------------------------------------
sub flushBuffer {
#receives filehandle name and sets the buffer to flush for it
#It should be called with \*FILEHANDLE
#----------------------------------------------

  #cookbook 7.19, flush buffer immediately for this FH
  my $old_fh = select(shift @_);
  $| = 1;
  &Kline::Display::showDebug("Filehandle " . select . " is now set to flush",1);
  select($old_fh);

} #end flushBuffer

#----------------------------------------------
sub saveDataDumper {
#receives object hash ref ($1), filename ($2)
#Optional comments for the header of the file can be passed as $3.
#NOTE: this will save with VAR1, which loadDataDumper will always assume
#----------------------------------------------
  my @args=@_;
  my $ref=shift(@_);
  my $file=shift(@_);
  unless ($ref && $file) {
    &Kline::Display::showDebug("IMPROPER ARGUMENTS!! (@args), RETURNING",0);
    return 0;
  }
  my $comments=shift(@_) || "No comments were provided with this file";
  my $fileext=0;
  if (-f "$file" ) {
    while (-f "$file.$fileext") {
      $fileext++;
    }
    &Kline::Display::showDebug("$file already exists, renaming old to $file.$fileext",0);
    rename "$file","$file.$fileext" or die "Cannot rename $file to $file.$fileext: $!\n";
  }
  &Kline::Display::printItem("Saving data to $file");
  open SAVE,">$file" or die "Cannot open $file: $!\n";
  print SAVE "#$comments\n";
  my $parent=&parentFunc();
  print SAVE "#Generated by $parent at ".localtime().".\n";
  if (ref($ref) eq "HASH") {
    print SAVE "#NOTE: ". scalar(keys(%$ref))." 1st level elements were identified in the following object.\n";
  } elsif (ref($ref) eq "ARRAY") {
    print SAVE "#NOTE: ". scalar(@$ref)." array elements were identified in the following object.\n";
  } else {
    print SAVE "#NOTE: The below object is not an array or hash.\n";
  }
  print SAVE Data::Dumper->Dump([$ref]);
  close SAVE;
  &Kline::Display::showDebug("Object save completed",2);

} #end saveDataDumper

#----------------------------------------------
sub loadDataDumper {
#receives filename and loads contents (from saveDataDumper) into memory for use
#NOTE: this will load with VAR1, which saveDataDumper will assume, returning ref
#----------------------------------------------
  my ($VAR1);
  my $file=shift(@_);
  die "Cannot find $file to load\n" unless ( -f "$file");
  &Kline::Display::showDebug("Loading data from perl hash file $file",1);
  open LOAD,"<$file" or die "Cannot open $file for object load: $!\n";
  my $olddelim=$/;
  undef $/;
  eval <LOAD>;
  die "cannot recreate object from $file: $! $@" if $@;
  close LOAD;
  &Kline::Display::showDebug("Data load from $file complete",1);
  $/=$olddelim;
  &Kline::Display::printItem("Found ".scalar(keys %$VAR1 )." top-level elements in $file");
  return $VAR1;

} #end loadDataDumper

#----------------------------------------------
sub by_last_mod {
#sort a list of files by the last modification date
#http://www.tek-tips.com/faqs.cfm?fid=5095
#----------------------------------------------
  # vars $a and $b automatically passed in
  # perl function 'stat' returns array of info on a file
  # 10th element of the stat array is last modified date,
  # returned as number of seconds since 1/1/1970.
  my $adate = &fileStat($a,"mtime"); # get last modified date
  my $bdate = &fileStat($b,"mtime"); # get last modified date
  return $adate <=> $bdate;

} #end by_last_mod

#----------------------------------------------
sub fileModDate {
# Query the qmstatus.ini file and use it for the time the qmgr most likely
# was taken down.  Can also be used to query mod date on any given file.
# Call with no args to return scalar for other perl.  Call with any add'l
# args to print the value for pickup in ksh or whatever.
# If file not found, likely because it isn't mounted (HA), so don't throw error
# calls fileStat, which uses File::stat
#----------------------------------------------
  my $file = &trim(shift(@_)) || &Kline::Display::showDebug("ERROR: No file was provided",0);
  my $modDate;
  if ($file && -f "$file") {
    &Kline::Display::showDebug("Checking mtime on \"$file\"",1);
    my $time=&fileStat($file,"mtime");
    &Kline::Display::showDebug("stat shows time as $time",2);
    $modDate = strftime '%d-%b-%g @ %R', localtime $time;
  } else { #file not there, perhaps cuz of HA
    &Kline::Display::showDebug("File not found: \"$file\"",1);
    $modDate = "N/A";
  }
  &Kline::Display::showDebug("$file moddate $modDate",1);
  return $modDate unless @_; #if no add'l args, return now (to perl)
  print "$modDate\n";  #if we get here, than other args mean to print for ksh

} #end fileModDate

#----------------------------------------------
sub fileStat {
#receives name of a file (or dir), and an inode field and returns its value
#For mode, returns the mode like 0755
#used by SSLSecurity health module, by_mod_date, fileModDate
#uses File::stat
#accepts dev, ino, mode, nlink, uid, gid, rdev, size, atime, mtime, ctime,
#blksize, blocks; see pg 801/894 in Camel
#----------------------------------------------
  my $file=shift(@_) || &Kline::Display::showDebug("CALLED INCORRECTLY",0);
  my $stat=shift(@_) || &Kline::Display::showDebug("CALLED INCORRECTLY",0);
  $stat=lc($stat);
  if (-e "$file" ) { #it can exist in any flavor
    my $st=stat($file) || die "Can't stat $file: $!";
    my $eval=eval  '$st->'.$stat; #perl 5.005 can't understand $st->stat;
    if ($stat=~/mode/) {
      return (sprintf "%04o",$eval & 0777) if $stat=~/mode/;
    } else {
      return $eval;
    }
  } else { #don't throw loud errors in case caller expects nothing
    &Kline::Display::showDebug("$file does not exist",1);
    return 0;
  }

} #end fileStat

#----------------------------------------------
sub fileSize {
#receives name of a file and returns the size in KB
#could also call the fileStat (above) with size parm
#----------------------------------------------
  my $file=shift(@_) || &Kline::Display::showDebug("CALLED INCORRECTLY",0);
  if ( -f "$file" ) {
    my $size= -s "$file";
    $size=sprintf "%.1f",($size / 1024);
    &Kline::Display::showDebug("$file is $size KB",1);
    return $size;
  } else {
    &Kline::Display::showDebug("$file does not exist",1);
    return 0;
  }

} #end fileSize

#----------------------------------------------
sub weekOfMonth {
#Loosely based on CPAN DateTime module
# Takes as arguments:
#  - The date (epoch timestamp) or it will pull current
#  - The day that we want to call the start of the week (7 is Sunday (optional)
#http://datetime.perl.org/index.cgi?FAQProjectAdministrivia
#----------------------------------------------
  &Kline::Display::showDebug("This function still needs some work; it's not done yet",0);
  &pressEnter();
  my $dt            = shift || timelocal(localtime);
  my $start_of_week = shift || 7;
  &Kline::Display::showDebug("Using $dt as the baseline timestamp (date: ".localtime($dt).")",2);

  # Work out what day the first of the month falls on
  my ($cday,$cmonth,$cyear)=((localtime($dt))[3,4,5]);
  my $first=timelocal(0,0,0,1,$cmonth,$cyear); #epoch time for first of this month
  my $wday = ((localtime $first)[6]); #get the day of week for 1st of month
  &Kline::Display::showDebug("The first of month $cmonth fell on day $wday (" . localtime($first) . ")",3);

  # And adjust the day to the start of the week
  #$wday = ($wday - $start_of_week + 7) % 7;

  my $num=int($cday / 7);
  my $remainder=$cday % 7;
  $num++ if $remainder; #if we're anywhere in the next week then add one

  # Then do the calculation to work out the week
  my $week  = 0;
  if ($wday < 5) { #first of month fell on Sun-Thurs, then counts as week 0
      $week++ unless $remainder;
  } else {
      $week-- if $remainder;
  }
  
  #my $num=sprintf("%.0f",$decimal); #round up to the number of weeks we're in
  $num=$num + $week;#then go back one week depending on when the month started
  
  &Kline::Display::showDebug(localtime($dt) ." falls in week $num of the month",1);
  return $num;

} #end weekOfMonth

#----------------------------------------------
sub convertRange {
#receives a number of strings with a range of numbers like 1,3,4-6
#and returns array with 1,3,4,5,6 in elements
#REMOVED:optional second arg is the range delimiter (default comma)
#----------------------------------------------
  my @outrange;
  my $delim=",";
  my $hasnums; #only do the by_mostly_numeric if there are numbers in the array somewhere
  &Kline::Display::showDebug("Received \"@_\" to convert to a range, iterating...",1);
  foreach my $string (@_) {
	  &Kline::Display::showDebug("Converting \"$string\" to a range",1);
	  my @range=split(/$delim/,$string);
	  foreach my $val (sort @range) {
	    if ($val=~/(\d+)\-(\d+)/) {
	      push @outrange,$1..$2;
	    } else {
	      push @outrange,$val;
	    }
	    $hasnums++ if $val=~/\d/;
	  }
  }
  @outrange=sort by_mostly_numeric(@outrange) if $hasnums;
  &Kline::Display::showDebug("Range @_ converted to @outrange",2);
  return @outrange;

} #convertRange

#----------------------------------------------
sub getFileContents {
#receive a file name and return its contents either as a string or array
#return undef if file not found or cannot open
#----------------------------------------------
  my $file=shift(@_);
  if (-f $file) {
    my @contents;
    open (FILE,"<$file") || return undef;
    while (<FILE>) {
        push @contents,$_;
    }
    close FILE;
    &Kline::Display::showDebug("FILE $file CONTENTS:\n@contents",3);
    &Kline::Display::showDebug("FILE $file LINES: " . @contents,2);
    if (wantarray) {
      return @contents;
    } else {
      return join "",@contents;
    }
  } else {
    &Kline::Display::showDebug("File $file is not found",1);
    return undef;
  }

} #end getFileContents

#----------------------------------------------
sub touchFile {
#receive filenames and touches them
#----------------------------------------------
  my $now=time();
  my $bad=0;
  foreach my $file (@_) {
    utime($now,$now,$file) ||
      open TMP, ">>$file" ||
    $bad++;
    if ($bad) {
      warn "Could not touch $file: $!\n";
      return 0 ;
    }
  }
  close TMP;
  return 1;

} #end touchFile

#----------------------------------------------
sub changeMode {
#changes the file mode the same as chmod but logs
#the before and after settings for helpful debugging
#Created due to issues with oct that I didn't want to remember over and over
#this works for directories and files
#receives $1 as new mode (string) and $2 as filename
#returns chmod true false, undef if error
#----------------------------------------------
  my ($newMode,$file)=@_;
  &Kline::Display::showDebug("NOT CALLED WITH CORRECT ARGUMENTS",0) unless ($newMode && $file);
  my $newModeOct=$newMode; #like 0600, etc octal mode; works fine for setting but
  #may confuse people so showing the normal output instead
  $newModeOct=oct($newMode); #unless ($newMode=~/^0/); #convert to oct if not already

  #check that action is actually needed
  my $currentMode=&fileStat($file,"mode");
  if ($newMode eq $currentMode) {
    &Kline::Display::showDebug("No file-mode change needed; $file already set to $newMode",1);
    return 1;
  }

  &Kline::Display::showDebug("WARNING: $file does not exist",0) unless (-f "$file" || -d "$file");
  my $noun= (-f "$file") ? "file" : "directory";
  &Kline::Display::showDebug("Changing $file mode from $currentMode to $newMode",1);
  undef $!;
  my $RC=chmod ($newModeOct, $file);
  $!="Could not change mode to $newMode for $file ($newModeOct): $!\n" unless $RC;
  $currentMode=&fileStat($file,"mode");
  &Kline::Display::showDebug("$noun mode is now $currentMode",1);
  return ($!) ? undef : $RC; #return code of chmod cmd

} #end changeMode

#----------------------------------------------
sub parseDelimited {
#Receive either a filename {file} or unparsed array ref {content} in href and parse it, returning a AoH of the data
#if {print} is true or called void, it will return (or print if void) all the data in pretty table
#expects a KEY header row in data, as commented row beginning with {key_prefix}, using delims to separate cols; defaults to "KEYS"
#expects {delim} as CSV delimiter, else default to comma (,)
#Tracks max length of fields and prints only wide as necessary
#Created out of a need in Schedule.pm to parse a CSV, but the need is very generic, so placed here.
#----------------------------------------------
  my $parms=shift;
  my @content=&getFileContents($parms->{file}) if (exists $parms->{file});
  @content=@{$parms->{content}} if (exists $parms->{content});
  my $delim=$parms->{delim} || ",";
  &Kline::Display::showDebug("Parsing content using delimiter \"$delim\"",1);
  my $key_prefix=$parms->{key_prefix} || "KEYS";
  &Kline::Display::showDebug("Parsing content using key_prefix\"$key_prefix\"",1);
  my @out; #aoh
  my @keys;
  my %max;
  foreach (@content) {
    my $href;
    next if /(^#[^($key_prefix$delim)])|(^\s*$)/;
    chomp;
    my @row=split /$delim/,$_;
    if (/^#?$key_prefix$delim/) {
      @keys=@row;
      shift @keys; #get rid of the #KEYS first entry
      next;
    }
    next unless @keys;
    my $i=0;
    foreach my $field (@row) {
      $href->{$keys[$i]}=$row[$i];
      $max{$keys[$i]}=length($row[$i]) if (length($row[$i]) > $max{$keys[$i]} || ! exists $max{$keys[$i]});
      $i++;
    }
    push @out,$href;
    
  }
  #make sure the header has enough cols
  foreach (keys %max) {
    $max{$_}=length($_) if length($_) > $max{$_};
  }
  
  if ((exists $parms->{print} && $parms->{print}) || ! defined wantarray) { #print the stuff
    &Kline::Display::showDebug("Returning printable output, not parsed hash refs",2);
    my @print;
    if (@out) {
      my $format; #set the format by max lengths
      foreach (@keys) {
        my $l=$max{$_};
        $format.="%-${l}s  ";
      }
      $format.="\n";
      
      push @print,sprintf $format,@keys;
      push @print,"-" x ($Text::Wrap::columns - 1) . "\n";
      foreach my $item (@out) {
        my @line;
        foreach (@keys) { push @line,$item->{$_}; }
        push @print,sprintf $format,@line;
      }
    } else {
      &Kline::Display::showDebug("No output was found",1);
    }
    (defined wantarray) ? return (wantarray) ? @print : \@print : print @print;
  }
  
  return (wantarray) ? @out : \@out;
  
} #end parseDelimited
    
#----------------------------------------------
sub generate_random_string {
#Initially created for DB crypt keys
#----------------------------------------------
	my $length_of_randomstring=shift;
	my @chars=('a'..'z','A'..'Z','0'..'9','_');
	my $random_string;
	foreach (1..$length_of_randomstring) 
	{
		# rand @chars will generate a random 
		# number between 0 and scalar @chars
		$random_string.=$chars[rand @chars];
	}
	return $random_string;
}

#----------------------------------------------
sub roundInteger {
#receives a floating point and returns rounded integer value
#----------------------------------------------
  my $m=shift;
  my $n=sprintf("%.0f",$m);  
  &Kline::Display::showDebug("Rounding $m to $n",2);
  return $n;

} #roundInteger

#----------------------------------------------
1; #tell the sourcing program we are true
