#!/usr/bin/perl
package Kline::Version;
# Version.pm (formerly base.pl)
# Chris Kline; cnkline@us.ibm.com
# Copyright IBM Corp. 2003,2007 All Rights Reserved
# ---------------------------------------------------------------------------
#
# This script is a library of commonly used functions for Kline programs that
# use Perl.  It can be "sourced in" using the "require" pragma.
# This library contains only functions that perform Kline-base activities.
# These functions provide the methods used to connect Kline together, such
# as for config and log files, and package writing.
# These functions may call functions from other libraries.
#
# ---------------------------------------------------------------------------
# REVISION HISTORY
# 2.0   -BT50563: Split Kline.pl according to genre.  First ed of this file.
#       Removed noStanzaErr and instead added optional 4th parm to pass so no
#       global var is needed. Removed toolsConfigGetValue because it was
#       redundant.  Added 5th arg to getPropertyValue.  Added dumpHref func.
# 2.0.1 -Fixed checkConfigFile to work with Windows (since we are using this
#       for health checks
# 2.1   -BT55480, loadCFG is not persisting and reloads everything each time
#       Appears previous work persistent getPropertyValue.  Moved that to
#       load CFG
# 2.1.1 -Updated updateToolsVersion to call loadCFG with 1 so that it keeps
#       case intact (the toolsVersionKSH was failing cuz of case mixup)
# 2.1.2 -Updated showInstalledPatches to match with Kline.setup 3.0.2
#       -changed stat cmd in getScriptVersions to pull from common fileStat
# 2.1.3 -BT59118: toolsVersion debug statements to help know where stuff goes
#       Updated to handle pre-8.5 versions (2nd edition)
#       Fixed loadCFG to allow for $3 to force refresh of requested info
# 2.1.4 BT59674: Updated loadCFG to work with older PERLs
# 2.2   Added missing skipPattern to getScriptVersions so pkg make works right
#       with svn filter, now follows links; general cleanup to allow subs used
#       during package creation to work within svn repos
# 2.2.1 BT63578 for querying installed patches
# 2.2.2 toolsVersion now returns a third element in array, instDate
# 3.0   BT66191; redo into a true perl module, formerly base.pl
# 3.0.1 12/05/11 Fix getVersions to use getFileContents
#
#-------------------------------------------------------------------------------
# Source Control Information
# $Revision: 22 $
# $LastChangedDate: 2012-01-10 20:43:31 -0700 (Tue, 10 Jan 2012) $
# $Author: cnkline $
#

# ---------------------------------------------------------------------------
# INITIALIZE THE ENVIRONMENT

use strict;
use English; 		#D:\perl_oreilly\prog\ch07_013.htm
require Exporter;
our @ISA=qw(Exporter);
our @EXPORT=qw(getScriptVersions getVersionChanges installedPatches packageChecksum resetScriptModDates scriptVer showInstalledPatches showScriptVersions showVersionChanges toolsVersion toolsVersionInfo toolsVersionKSH updateToolsVersion updateToolsVersionKSH writeFinalChecksum);

use File::Basename qw(basename dirname);	#D:\perl_oreilly\prog\ch07_027.htm
use File::Find;
use POSIX qw(strftime); #for converting timestamps

use lib '..'; #so we can run these direct to make sure they compile and not fail for missing mods
use Kline;
use Kline::Config;
use Kline::Display;
use Kline::Server;
use Kline::Utils;
#use Kline::WMQ;

my $scriptVer="3.0.1";
my $scriptDate; #backward compatibility
($scriptDate = '$LastChangedDate: 2012-01-10 20:43:31 -0700 (Tue, 10 Jan 2012) $') =~ s/.*\(\D+(\d+\s\w+\s\d+)\).+/$1/; #auto update upon commit

#----------------------------------------------------------------------------
die "KlineDIR environment variable must be set" unless $Kline::KlineHome;

#----------------------------------------------
{
my @files;
sub findAllToolsFiles {
#internal sub called by find in getScriptVersions
#required in order to tell find to follow symlinks
#----------------------------------------------
  if ($File::Find::name=~/\.svn/) {
    &showDebug("Skipping $File::Find::name: SVN temp file",4);
  } else {
    push @files,$File::Find::name;
  }
} #end findAllToolsFiles

#----------------------------------------------
sub getScriptVersions{
#retrieve all the script versions included in this package of tools, along with
#modification dates.  Returns an anon hash with all the info, in case needs to
#be parsed.  Otherwise, call showScriptVersions, which just prints it STDOUT.
#Will automatically do a 'find' on the Kline directory
#If @_ is passed, list of filenames (regexpr) that should be skipped.
#----------------------------------------------
  my $skipPattern=shift @_;
  my ($file,$version,$date,$type);
  &showDebug("looking for files at $Kline::KlineHome",2);
  &showDebug("SKIPPING FILES CONTAINING \"$skipPattern\"",1) if $skipPattern;
  find { wanted => \&findAllToolsFiles, follow => 1 }, "$Kline::KlineHome"; #follow links
  &showDebug("found ".scalar @files." files in find sub",1);
  my @AoH;
  my $rec; #anon hash
  for my $file (sort @files) {
    my $dateMatch=1; #if true, then the script date and mod date do not match
    next if ($file =~ /^$Kline::KlineHome$/); #this will remove the toolsdir itself
    next if ($skipPattern && $file =~ /$skipPattern/);
    &showDebug("Found Kline file \"$file\"",3);
    $rec={};

    #get the info for all text files, ver is 0 if not set
    if ( -T $file) { #file is a text file
      ($version,$date,$type)=&scriptVer($file);
      $type="N/A" unless $type;
    } else { #file is not a text file
      $version="N/A";
      $date=0;
      $type="binary";
    } #end text file check

    #get the file modification date
    my ($mtime,$day,$month,$year);
    my $modDate;
    $mtime=&fileStat($file,"mtime"); #uses OO File::stat
    &showDebug("$file mtime is $mtime",3);
    ($day, $month, $year)=(localtime($mtime))[3,4,5];
    $day=sprintf "%02d",$day; #pad with zeroes to match
    $month=&convertMonthNum($month+1) || "ERROR";
    $year+=1900;
    $modDate="$day $month $year";
    $rec->{mtime}=$mtime;

    #see whether the date in script matches the mod date; use ! on report if no
    if ($date) { #must be a script file, test if matches mod
      unless ("$date" eq "$modDate") {
        $dateMatch=0;
        &showDebug("$file: $date == $modDate",2);
      }
    } else { #could be text or binary, use modDate
      $dateMatch=-1; #no match attempted
      $date=$modDate;
    }

    #get checksum for the file
    my $checksum;
    if (-d "$file") {
      $checksum="directory";
    } elsif ($skipPattern && "$file" =~ /$skipPattern/) {
      $checksum="skipped";
    } else {
      $checksum=`cksum "$file" | awk '{print \$1}'`;
      chomp $checksum;
    }

    #keep this here so name is only switched after it's been used
    $file =~ s/^$Kline::KlineHome\///g;  #drop toolsdir name from start of each line

    #load the data
    &showDebug("Loading $file into hash array",3);
    $rec->{script}="$file";
    $rec->{version}=$version;
    $rec->{type}="$type";
    $rec->{date}=$date;
    $rec->{datematch}=$dateMatch;
    $rec->{checksum}=$checksum;
    push @AoH, $rec;

  } #end loop of files

  &showDebug("Total of ".scalar @AoH." files were loaded into hash array",2);
  return @AoH;

} #end getScriptVersions
} #end naked block for @files

#----------------------------------------------
sub getVersionChanges {
#Used when making a package.  Compares versions from the current versions file
#to actual current versions and writes a list of files that have been modified.
#Meant to create a list of files that updated since the last version, for a
#changelog.  Output is returned in a hash for further use.  If formatted
#output is desired, then use showVersionChanges.
#----------------------------------------------
  my $count=0;
  my @AoH=&getScriptVersions(); #all current versions
  my @newAoH; #scripts not found in old version
  my @oldAoH; #scripts not found in new version
  my @updatedAoH; #scripts that changed but were present in both
  #these are just the possible values from the generated hash
  #script, version, type, date, datematch, checksum

  chdir $Kline::KlineHome;
  &showDebug("NOTICE: This function does not yet show scripts that were removed; only new or changed",0);
  my ($oldVer,undef)=&toolsVersion();
  $oldVer=~tr/\./-/;
  &showDebug("Comparing versions against previous at $Kline::KlineHome/Kline.v$oldVer.versions",0);
  open PREVIOUS, "<$Kline::KlineHome/Kline.v$oldVer.versions"
    or die "Cannot open versions file $Kline::KlineHome/Kline.v$oldVer.versions: $!\n";
  my @old;
  while (<PREVIOUS>) {  chomp; push (@old,$_);}
  close PREVIOUS;

  for my $href (@AoH) {
    &showDebug("Now comparing $href->{script} (ver: $href->{version}, cksum: $href->{checksum}",2);
    for my $line (@old) { #each line in the old version file
      if ($line =~ /^($href->{script})\s+(\S+)\s+(\S+)\s+(\d+ \w+ \d+).\s+(\S+)/) {
        &showDebug("Loading $href->{script} with previous info (ver: $3, cksum, $5",2);
        ($href->{o_type},$href->{o_version},$href->{o_date},$href->{o_checksum})= ($2,$3,$4,$5);
        last;
      } else {
        next;
      }
    }
  } #end new version loop
  undef @old;

  #add all the stuff from oldAoH onto AoH
  push (@AoH,@oldAoH);

  #now loop through the hash again and organize
  my %HoH;
  for my $href (@AoH) {
    if (exists($href->{o_version}) && exists($href->{version})) {#is from both versions, check if modified
      if ("$href->{version}" ne "$href->{o_version}") { #ver has changed, add to CHANGED hash
        &showDebug("$href->{script} has changed (old: \"$href->{o_version}\", new: \"$href->{version}\")",1);
        $HoH{CHANGED}{$href->{script}}{version}=$href->{version};
        $HoH{CHANGED}{$href->{script}}{old_version}=$href->{o_version};
      } elsif ("$href->{o_checksum}" ne "$href->{checksum}") { #ver didn't change, but checksum did
        &showDebug("$href->{script} has a changed checksum ($href->{checksum} vs $href->{o_checksum})",1);
        $HoH{CHANGED}{$href->{script}}{checksum}=$href->{checksum};
        $HoH{CHANGED}{$href->{script}}{old_checksum}=$href->{o_checksum};
      } else { #ver has not changed, so can discard
        &showDebug("$href->{script} has not changed (version $href->{version})",2);
      }
    } elsif (exists $href->{o_version}) { #add to REMOVED hash  #####NOTHING HERE YET BECAUSE NO DATA
      &showDebug("$href->{script} has been removed",1);
      $HoH{REMOVED}{$href->{script}}{old_version}=$href->{o_version};
    } elsif (exists $href->{version}) { #add to NEW hash
      &showDebug("$href->{script} is a new file (version $href->{version} $href->{date})",1);
      $HoH{NEW}{$href->{script}}{version}=$href->{version};
      $HoH{NEW}{$href->{script}}{file_type}=$href->{type};
    } else {
      warn "We should not get here (press enter)!! ";
      <STDIN>;
    }
  }
  return %HoH

} #end getVersionChanges

#----------------------------------------------
sub showInstalledPatches {
#build upon installedPatches and show the output in a pretty table
#with a synopsis and date installed.  Call with args to cause it
#to pass back an array with the output string.
#----------------------------------------------
  my $href=&installedPatches(); #href with all version/patch info
  my $format="%-12s%-42.40s%-10s%-11s\n"; 
  my @output; #buffer output and return if specified
  push @output,sprintf("$format","PATCH","DESCRIPTION","VERSION","INST DATE");
  push @output,"-"x75,"\n";
  my $patches=0;
  foreach my $stanza (sort keys %{ $href } ) {
    &showDebug("Searching stanza $stanza for patches",1);
    next unless $stanza=~/PATCH/;
    $patches++;
    my $num=$href->{$stanza}{patchNumber} || "NOT_FOUND";
    my $date="NOT_FOUND";
    my $ver=$href->{$stanza}{outputVersion} || "NOT_FOUND";
    my $descr= $href->{$stanza}{description} || "NOT_FOUND";
    if ($href->{$stanza}{installDateEpoch}) {
      $date=strftime '%Y-%b-%d', localtime($href->{$stanza}{installDateEpoch});
    }
    push @output,sprintf("$format","$num","$descr","$ver","$date");
  }
  if ($patches) {
    unshift @output,"A total of $patches patches have been installed.\n\n";
  } else {
    push @output,"No patches are installed.\n";
  }

  if (wantarray) { #return the output, don't print it
    return @output;
  } else { #likely ksh
    print @output; #if we get here, then print it here
    return 0;
  }

} #end showInstalledPatches

#----------------------------------------------
sub installedPatches {
#retrieve a list of currently applied patches, 0 if none, otherwise return a
#array with each patch
#if we pass a file name, it'll check that; otherwise it will look for config
#file, and if that's a prob, it'll even check against whatever toolshome is.
#This should catch all the options.
#NOT designed to be called externally.  Use showInstalledPatches for pretty results.
#returns href with all info on each patch
#----------------------------------------------
  my ($verFile,$version);
  my $thisFunction = &thisFunc();
  if (@_) {
    $verFile=shift @_;
  } else {
    $verFile=&getPropertyValue("Kline","KlineVersionDoc","",1); #pass 4th arg as 1 to disable err
  }
  if ($verFile) {
    $verFile="$Kline::KlineHome/$verFile" unless ($verFile =~ /^\//);
  } else {
    $verFile="$Kline::KlineHome/.Kline.version";
  }
  my $href=&loadCFG("$verFile",1,1); #keep case mixed, force reload
  return $href->{"$verFile"};

} #end installedPatches

#----------------------------------------------
sub packageChecksum {
#runs showScriptVersions, filters Kline.setup.cfg, writes to file
#and then reads the cksum from that file and returns.  For use when
#creating packages
#----------------------------------------------
  my $file="/tmp/Kline.cksum";
  open CKSUM,">$file" or die "can't open checksum file for write at $file\n";
  select CKSUM;
  &showScriptVersions("Kline.setup.cfg|Kline.v.+versions");
  close CKSUM;
  select STDOUT;
  my $cksum=`cksum $file | awk '{print \$1}'`;
  unlink $file;
  chomp $cksum;
  return $cksum;

} #end packageChecksum

#----------------------------------------------
sub resetScriptModDates {
#in conjunction with showScriptVersions, files that show with ! have a
#scriptDate that doesn't match the modDate of the file itself.  THis will
#touch the file back to the scriptDate at 00:00:00
#----------------------------------------------
  my $count=0;
  my @AoH=&getScriptVersions();
  chdir $Kline::KlineHome;
  for my $href (@AoH) {
    unless ($href->{datematch}) { #if yes, then nothing to do
      &showDebug("Resetting $href->{script} to $href->{date}",0);
      if (system("touch $href->{script} -md \"$href->{date}\"")) {
        &showDebug("FAILED to reset $href->{script}",1);
      } else { #success
        $count++;
      }
    }
  }
  &showDebug("Reset $count files successfully",0);

} #end resetScriptModDates

#----------------------------------------------
sub scriptVer {
#returns the date and version of a script, as well as whether ksh or perl
#TODO: See printFunctionNames.pl for updated version of this thing
#----------------------------------------------
  my $file=shift @_;
  my $ver=0;
  my $type=0;
  my $date=0;
  my $thisFunction = &thisFunc();
  my @contents=&Kline::Utils::getFileContents($file);
  foreach (@contents) {
    $type="ksh" if (m{#\!.+ksh});
    $type="perl" if (m{#\!.+perl});
    $ver=$1 if (/scriptVer=(.+)/);
    $ver=$1 if (/SCRIPTVER=(.+)/);
    $date=$1 if (/scriptDate=(.+)/);
    $date=$1 if (/SCRIPTDATE=(.+)/);
    last if ($type && $ver && $date);
  }
  $ver=$1 if ($ver =~ /"(.+)"/);
  $date=$1 if ($date =~ /"(.+)"/);
  &showDebug("file $file v$ver, type:$type, date:$date",3);

  print "$ver  $date\n" if @_;
  return ($ver,$date,$type);

} #end scriptVer

#----------------------------------------------
sub showScriptVersions {
#see getScriptVersions.  This just takes that output and prints it pretty
#if scalar or void, prints it to STDOUT.  If array context, returns
#receives files to filter and passes to getScriptVersiosn
#needs to be a full screen to avoid wrapping
#----------------------------------------------
  #cksum is 10 chars
  my $format="%-52s%-7s%7s  %-13s%-11s\n";

  my @AoH=&getScriptVersions(@_);  #pass any args straight through
  my @output;
  push @output,&drawLine("~");
  push @output,"NOTE: N/A is a binary file that could not be read; 0 is a file with no ver info\n";
  push @output,"      Dates with (!) mean that file mod date doesn't match scriptDate in file.\n";
  push @output,"      Maximize the screen to align columns properly.\n";
  push @output,sprintf "$format","FILE","TYPE","VERSION","SCRIPT DATE","CHECKSUM";
  push @output,&drawLine("~");
  my $count=0;
  for my $href (@AoH) {
    my $warn="";
    $warn="!" unless ($href->{datematch});
    push @output,sprintf "$format","$href->{script}","$href->{type}","$href->{version}","$href->{date}$warn","$href->{checksum}";
    if ($warn) {
      &showDebug("Hash dump follows",3);
      for (keys %$href) {
        &showDebug("$_=$href->{$_}",3);
      }
      &showDebug("".basename($href->{script})." mod date: ".localtime($href->{mtime})." ",1);
      $count++;
    }
  }
  &showDebug("$count files were not matched for modification dates",1) if $count;
  #return the array if array, count if scalar, and nothing if void
  #return undef unless defined wantarray; #void context
  if (wantarray()) {
    return @output;
  } else { #ws called in scalar context, which is likely kshtoperl
    print @output;
    return 0;
  }

} #end showScriptVersions

#----------------------------------------------
sub showVersionChanges {
#runs getVersionChanges and formats output for use in changelog
#----------------------------------------------
  #cksum is 10 chars
  my $format="%-52s%-7s%7s  %7s%-13s\n";

  my %HoH=&getVersionChanges(@_);  #pass any args straight through
  my $old=&toolsVersion();
  my $time=&showTime(); #pulls time in parens
  &printArrowSection("Listing change log from $old to now $time:");
  my $i; my $j;
  for my $type (sort keys %HoH) {
    for my $script (sort keys %{ $HoH{$type} } ) {
      $i++;
      print "  $type file ($i): $script\n";
      print "    UPDATES:";
      for my $props (sort keys %{ $HoH{$type}{$script} } ) {
        $j++;
        print " $props=$HoH{$type}{$script}{$props}  ";
      }
      print "\n";
      undef $j;
    }
    undef $i;
    #print "\n"; #line between each type
  }

} #end showVersionChanges

#----------------------------------------------
sub toolsVersion {
#retrieve current tools version, 0 if none (which we'll never get)
#if we pass a file name, it'll check that; otherwise it will look for config
#file, and if that's a prob, it'll even check against whatever toolshome is.
#This should catch all the options.
#returns version number and installation directory
#----------------------------------------------
  my ($verFile,$version);
  my $dir=0;
  my $date="";
  my $thisFunction = &thisFunc();
  if (@_) {
    &showDebug("Called with args: @_ (should be filename)",1);
    $verFile=shift @_;
  } else {
    &showDebug("Called WITHOUT file name, getting file name from Kline.cfg (getPropertyValue)",1);
    $verFile=&getPropertyValue("Kline","KlineVersionDoc","",1); #pass 4th arg as 1 to disable err
    &showDebug("Version file DEFAULT from Kline.cfg set to $verFile",2);
  }
  $verFile=&trim($verFile);
  if ($verFile) {  #append toolsdir if just a relative name
    unless ($verFile =~ /^\//) {
      &showDebug("Prepending tools dir $Kline::KlineHome to $verFile",2);
      $verFile="$Kline::KlineHome/$verFile" ;
    }
  } else {  #no file found in Kline.cfg, so take a shot here
    &showDebug("Still no version file identified, trying manually at $Kline::KlineHome/.Kline.version",2);
    $verFile="$Kline::KlineHome/.Kline.version";
  }
  &showDebug("Version file FINAL set to $verFile",2);
  unless (-f "$verFile") {
    &showDebug("Version file $verFile does not exist, returning 0 quietly",1);
    return (0,0);
  }

  #will only get here if the file actually exists
  &showDebug("Loading config from version file $verFile",2);
  my $href=&loadCFG("$verFile");
  &showDebug("Examining loaded config for VERSION element",2);
  if (exists($href->{"$verFile"}{VERSION}{version})) { #is v8.5 format, 3rd edition
    $version=$href->{"$verFile"}{VERSION}{version};
    $date=$href->{"$verFile"}{VERSION}{installdate} || 0;
    &showDebug("version: $version identified via location: $dir",1);
  } else { #pre 8.5 format, likely only used during initial install, which could be an upgrade
    my $tmpver=`cat $verFile | grep "Kline Version Control 2nd Ed"`;
    if ($tmpver) { #is 2nd ed file
      &showDebug("Found 2nd edition version file at $verFile",1);
      $version=$1 if ($tmpver=~/Version=(\S+)/); #line looks like Kline Version Control 2nd Ed;VersionDate=05/06/03 20:03:25;Version=8.01
    } else {
      &showDebug("ERROR: Version file $verFile is not a recognized Kline version file (2nd or 3rd ed)!",0);
    } #end check 2nd ed ver file
  } #end check 3rd ed ver file
  $dir=dirname("$verFile") if $version; #return the install location too

  #return what we found
  if ($version) {
    if (wantarray) {
      return ($version,$dir,$date);
    } else { #scalar context, so ver only
      return $version;
    }
  } else {
    return (0,0);
  }

} #end toolsVersion

#----------------------------------------------
sub toolsVersionInfo {
#like toolsVersion but shows more info (didn't want to change the other
#copy since stuff might rely on it;  this takes advantage of the new format
#version file
#----------------------------------------------
  my @out;
  my $verFile=&getPropertyValue("Kline","KlineVersionDoc","",1); #pass 4th arg as 1 to disable err
  $verFile="$Kline::KlineHome/$verFile";
  my $href=&loadCFG("$verFile",1);
  push @out,&printSection("Kline VERSION");
  push @out,"Kline version $href->{$verFile}{VERSION}{version} is installed.  Following are version details.\n\n";
  my $format="%-25s%-50s\n";
  push @out,sprintf("$format","Parameter","Value");
  push @out,&drawLine();
  foreach my $key (sort keys %{ $href->{"$verFile"}{VERSION} } ) {
    push @out,sprintf("$format","$key","$href->{$verFile}{VERSION}{$key}");
  }
  
  #now show the patch information
  push @out,"\n";
  push @out,&printSection("INSTALLED PATCHES");
  push @out,&showInstalledPatches();
  push @out,"\n";

  if (wantarray) { #return all the output for use elsewhere
    return @out;
  } else { #likely kshtoperl 
    print @out;
    return 0;
  }
} #end toolsVersionInfo

#----------------------------------------------
sub toolsVersionKSH {
#only for Kline.functions for backwards compat.  returns ver,date
#----------------------------------------------

  my $verFile=&getPropertyValue("Kline","KlineVersionDoc","",1); #pass 4th arg as 1 to disable err
  $verFile="$Kline::KlineHome/$verFile";
  my $href=&loadCFG("$verFile",1);
  return "$href->{$verFile}{VERSION}{version},$href->{$verFile}{VERSION}{compiledDate}" || "0,0";

  } #end toolsVersionKSH

#----------------------------------------------
sub writeFinalChecksum {
#called during package creation.  Take the package checksum created above and
#write it to the filename in @_.  Pulls the version via the toolsVersion func.
#Needs to receive an existing file with <CHECKSUM> stanza already formed.
#----------------------------------------------
  my @file;
  &printSection("Calculate and Insert Kline Package Checksum");
  my $setupFile=shift @_ || "Kline.setup.cfg";
  $setupFile="$Kline::KlineHome/$setupFile" unless ($setupFile =~ /^\//);
  my ($version,undef)=&toolsVersion();
  &printItem("Calculating checksum for version $version");
  my $cksum=&packageChecksum;
  &printItem("Checksum $cksum will now be added to $setupFile");
  my $result=open FILE,"<$setupFile" or warn "Cannot open $setupFile for cksum read:$!\n";
  if ($result) {
    while (<FILE>) {
      if (/^$version=/){
        chomp;
        &printItem("$version was already in file and will be removed: $_");
      } else { #not in file, push for new file
        push @file,$_ }
      }
    close FILE;
  }

  my $added=0;
  my @newfile;
  for (@file) {
    push @newfile,$_;
    next if (/^#|^\s*$/);
    unless ($added) {
      if (/<CHECKSUM>/) {
        push @newfile,"$version=$cksum\n";
        $added++;
      }
    }
  }
  open FILE,">$setupFile" or die "Cannot open $setupFile for cksum update:$!\n";
  for (@newfile) { print FILE; }
  close FILE;

  &printItem("Checksum add complete");
  return 0;

} #end writeFinalChecksum

#----------------------------------------------
sub updateToolsVersion {
#update the version file with passed information
#$1 is href with all the junk to append, 1st level should be VERSION or PATCH*
#$2 if passed causes it to append, not overwrite the stanza (ident vals are overwritten)
#$3 if passed will override the install dir where ver file is written
#----------------------------------------------
  #Kline Version Control 2nd Ed;VersionDate=10/26/04 17:06:24;Version=8.50r

  my $vals=shift(@_); #href with new VERSION and/or PATCH entries 
  my $href; #current config file values
  my $append=shift(@_) || 0;
  my $installDir=shift(@_) || ""; #used during install since the KlineHome is set to temp location
  my $verFile=&getPropertyValue("Kline","KlineVersionDoc","",1); #pass 4th arg as 1 to disable err
  if ($installDir) {
    &showDebug("Altering the install dir since it would have been the tempdir during tools installation",1);
    $verFile="$installDir/$verFile"; #used only on install, see above
  } else {
    $verFile="$Kline::KlineHome/$verFile"; #normal method
  }
  $href=&loadCFG("$verFile",1,1); #load existing version settings, keep case intact, force fresh load and no cached vals
 
  foreach my $stanza ( keys %{ $vals } ) { #loop through the new vals we're passing to update
    if (exists $href->{"$verFile"}{$stanza} && !($append) ) { #then overwrite; if $2 is passed, then append existing (such as when installing and adding install info to the xisting compile info)
      &showDebug("The $stanza stanza will be completely overwritten in $verFile",1);
      undef $href->{$verFile}{$stanza};
    } else { #then append
      &showDebug("The $stanza stanza will be appended in $verFile; current values follow:",1);
      &dumpHref($href,2);
    }
    foreach my $key ( sort keys %{ $vals->{$stanza} } ) { #write the new values to memory
      &showDebug("adding $key, $vals->{$stanza}{$key}",2);
      $href->{"$verFile"}{$stanza}{$key}=$vals->{$stanza}{$key};
    }
  } #end foreach

  #write the new file, with warnings and in order
  open VERFILE,">$verFile" or die "Cannot open $verFile for rewrite:$!\n";
  &showDebug("Writing new version file $verFile",2);
  &dumpHref($vals,3);
  select VERFILE;
  print "#Kline Version Control 3rd Ed.  Do not edit this file manually.  \n#Any comments in this file will be overwritten automatically when upgrading or patching.\n\n";
  foreach my $stanza ( reverse sort keys %{ $href->{"$verFile"} } ) { #version at top, followed by patches
  print "<$stanza>\n";
    foreach my $key ( sort keys %{ $href->{$verFile}{$stanza} } ) { 
      my $val=$href->{$verFile}{$stanza}{$key};
      print "  $key=$href->{$verFile}{$stanza}{$key}\n";
    }
    print "</$stanza>\n\n";
  }

  select STDOUT;
  close VERFILE;
  

} #end updateToolsVersion

#----------------------------------------------
sub updateToolsVersionKSH {
#shim so this can be called w/out href from ksh
#until Kline.setup.ksh is ported for making packages
#----------------------------------------------
  my $href;
  my $version=shift(@_);
  $href->{VERSION}{version}=$version;
  $href->{VERSION}{compiledDate}=localtime;
  $href->{VERSION}{compiledDateEpoch}=time;
  &showDebug("Updating version from ksh callout",1);
  &updateToolsVersion($href);

} #end updateToolsVersionKSH

#----------------------------------------------
1; #tell the sourcing program we are true
