#!/usr/bin/perl
package Kline::WMQ;
# WMQ.pm
# Chris Kline; cnkline@us.ibm.com
# Copyright IBM Corp. 2003,2008 All Rights Reserved
# ---------------------------------------------------------------------------
#
# This script is a library of commonly used functions for Kline programs that
# use Perl.  It can be "sourced in" using the "require" pragma.
# This library contains only functions that perform WMQ-related activities.
# These functions may call functions from other libraries.
#
# ---------------------------------------------------------------------------
# REVISION HISTORY
# 2.0   -BT50563: Split Kline.pl according to genre.  First ed of this file.
#       Changed allObjTypes to be a cfg value and not global variable in 
#       loadQmgrObjects.  Also updated the algorithm for speed.
# 2.0.1 -Fixed client/server detect to not show stderr
# 2.1   -BT55594.  loadQmgrObjects updated for multiple objects with same name
#       use anon hash instead (should only be issue in clustering- clusqmgr)
# 2.1.1 -Updated log gathering to skip if no files found, such as for @SYSTEM
# 2.1.2 -Updated getMQVersion to also check cfg file for a max version
# 2.2   -BT38453. Added checkChanSecurity for new health module ChanSecurity
#       Added DLQSummary to use q to show the unique reason codes for dead msgs
# 2.3   Added UNIXQmgrList which simplifies results of getQmgrs and returns only an
#       array with the qmgr names.  Used by gnosis.
# 2.4   Added add'l attrs to qmgrStatus to support Gnosis MQInventory module.
#       Updated getMQSCAttr to handle QMGR object (with no name)
#       Updated WMQConfig to get install pkg names and locations
# 2.5   Added listener functions from mngmqlsr for gnosis use
#       Separated getQmgrs from showQmgrs (for ksh use) and made getQmgrs accept
#       multiple qmgrs names (used to be just one)
#       getQmgrs now caches; added reloadQmgrs to reload the cache
# 2.5.1 BT65828 to update getMQVersion for latest formatting
# 2.5.2 Tiny change to make sure reloadQmgrs returns the hash val
# 2.5.3 BT67328 fixed AIX WMQConfig values
# 3.0   BT66191; move to true perl modules
# 3.0.1 BT66191/66193; Only calling cfg-file derived funcs in MQVersion when full
#       Kline pkg is installed, so that it can run in Gnosis without errs
# 3.0.2 12/10/07 Update for Gnosis in loadQmgrObjects to not query Kline.cfg
# 3.1   12/20/07 BT66171 Updated checkChanSecurity so that we can return the
#                stored info for other perl use.  Added loadChanSecurity which
#                gets all the info and checkChanSecurity just prints for ksh
#		 Added getOAMPermissions for health checks, which is a subset
#                of mqauth52, parsing amqoamd for perl use
#	01/09/08 Added getObjTypeList to return list of all types for this level
#                Used by WMQHUOAMAltusrAuth health check to check all types
#       01/14/08 Rewrite and streamline of loadQmgrObjects; now uses regexpr
#                name pattern instead of runmqsc direct pass. Removed portisListening
#                to Server.pm so it can be used outside of MQ.
#       03/06/08 Added isInMQMGroup.  Then added dies for all subs that require 
#                mqm group membership.  Made a few changes to allow for key
#                inventory subs to be run without mqm group membership
# 4.0   06/09/08 Updates to support Windows for inventory.  Replaced many backtick
#                with cmdLine so debug will show us more output. All RUNMQSC calls
#                were updated for proper quoting on UNIX vs Win.  See corresponding
#                updates in Kline::Server.  Should only load Win mods when needed.
#                Also includes updates to the following subs for Windows:
#                    	dspmqver
#			                WMQConfig
#               			listenerInfo
#	                		qmgrRunmqlsrPorts
#               			qmgrStatus
#			                getQmgrs
#			                getDefQmgr
#			                isQmgrRunning
#                     isInMQMGroup (checks also for admin group on Win)
#       06/24/08 Updates for getOAMPermissions to handle windows; also switched from
#                native command-exec to using cmdLine; involved much rework
#       09/04/08 Workaround in getOAMPermissions for dmpmqaut rc=71 in some versions
#
#-------------------------------------------------------------------------------
# Source Control Information
# $Revision: 22 $
# $LastChangedDate: 2012-01-10 20:43:31 -0700 (Tue, 10 Jan 2012) $
# $Author
#

# ---------------------------------------------------------------------------
# INITIALIZE THE ENVIRONMENT

use strict;
use English; 		#D:\perl_oreilly\prog\ch07_013.htm
require Exporter;
our @ISA=qw(Exporter);
our @EXPORT=qw(DLQSummary UNIXQmgrList WMQConfig alterGetPut alterGetPutMenu checkChanSecurity convertObjType convertQmgrFS convertReturnCode debugInetAoH dumpLogs dumpMQClusterRepos dumpObjectHash getObjTypeList getAttrList getDefQmgr getMQSCAttr getMQVersion getQmgrs isQmgrRunning listenerInfo loadQmgrObjects loadQueues qmgrInetdPorts qmgrRunmqlsrPorts qmgrStatus queueSize reloadQmgrs selectQmgr showQmgrs triggerCheck isInMQMGroup);

BEGIN { #only load windows mods if we expect to need them
  if ($^O=~/win/i) {
    require Win32;
    import Win32;
    require Win32::TieRegistry;
    import Win32::TieRegistry;
    require Win32::Process::Info;
    import Win32::Process::Info;
  }
};
$Win32::TieRegistry::Registry->Delimiter("/") if $^O=~/win/i; #from Win32::Registry

use File::Basename qw(basename dirname);	#D:\perl_oreilly\prog\ch07_027.htm
use File::Spec::Functions; #catfile
use Text::Wrap; #used only for the columns variable
use Carp;
    
use lib '..'; #so we can run these direct to make sure they compile and not fail for missing mods
use Kline; #sets KlineHome, used by dumpLogs, clusterqsummary, and others
use Kline::Config;
use Kline::Display;
use Kline::Server;
use Kline::Utils;

my $scriptVer="4.0";
my $scriptDate; #backward compatibility
($scriptDate = '$LastChangedDate: 2012-01-10 20:43:31 -0700 (Tue, 10 Jan 2012) $') =~ s/.*\(\D+(\d+\s\w+\s\d+)\).+/$1/; #auto update upon commit

#SHARED VARIABLES
our $mqBinaries="";
our $registryBaseKey="/HKEY_LOCAL_MACHINE/SOFTWARE/IBM/MQSeries/CurrentVersion/"; #windows registry base key location

#PRIVATE VARIABLES
my $mqs_ini="/var/mqm/mqs.ini";
my $WINDOWS = ( $^O=~/win/i );
&Kline::Server::checkPlatform unless $mqBinaries;
my $runmqsc=catfile("$mqBinaries","bin","runmqsc");
$runmqsc.=".exe" if $WINDOWS;
#die "Cannot locate runmqsc at $runmqsc\n" unless (-f $runmqsc); #cannot do this since this mod gets sourced on boxes with no WMQ too
$runmqsc="$winquote$runmqsc$winquote" if $runmqsc=~/\s/; #have to be after the file test, and only if there's spaces

my $mqscquote=($WINDOWS) ? "" : "\""; #for unix, need quotes on runmqsc cmds.  For windows, they must not be there

#----------------------------------------------------------------------------
die "KlineDIR environment variable must be set" unless $Kline::KlineHome; #used only to locate external files for Kline (dumpLogs, dumpMQClusterRepos DLQSummary)

#----------------------------------------------
sub getDefQmgr {
# Look for a default or single queue manager; will return undef if
# no default was found or if more than one qmgr is on the machine
# TESTED: Win32 (using a local mqs.ini file/dir)
# Doese not require mqm group membership
# Updated 05/2008 to support Win32
#----------------------------------------------
  my $defQmgrName;
  my @qmgrs; #array to hold all qmgr names; we'll use single name if no default 
  
  ###############################  
  if ($WINDOWS) {
    my $pointer=$Win32::TieRegistry::Registry->{$registryBaseKey};
    $defQmgrName=$pointer->{"Configuration/DefaultQueueManager//Name"};
    if ($defQmgrName) {
      &showDebug("Default Queue Manager: $defQmgrName",1);
      return $defQmgrName;
    }
    #get a list of qmgrs so we can check if only one at bottom
    @qmgrs=$pointer->{"Configuration/QueueManager/"}->SubKeyNames;
    
  ###############################  
  } else { #UNIX/Linux 
    open MQSINI, "<$mqs_ini" #globally readable
      or die "Cannot open MQ config file $mqs_ini: $!\n";
    while (<MQSINI>) { #search for a true default qmgr
      if (/^DefaultQueueManager:/ .. /^\s+Name=/) {
        $defQmgrName = (split /=/)[1] if (/^\s+Name=/);
        chomp ($defQmgrName) if ($defQmgrName);
      }
      if ($defQmgrName) {
        &showDebug("Default Queue Manager: $defQmgrName",1);
        close MQSINI;
        return $defQmgrName;
      }
    }
    #keep searching for a single qmgr if we came up empty already
    seek (MQSINI,0,0); #reset the position in the file to beginning
    while (<MQSINI>) { #search for a single qmgr only
      if (/^QueueManager:/ .. /^\s+Name=/) {
        push (@qmgrs, $1) if (/^\s+Name=(\w+)/);
        #found a qmgr name
      }
    }
    close MQSINI;
  } #end UNIX vs Win
  
  #only return a found name if single qmgr present
  if (@qmgrs == 1) {
    $defQmgrName = pop(@qmgrs);
    &showDebug("Single qmgr found: $defQmgrName",1);
    return $defQmgrName;
  }

  &showDebug("No default or single qmgr",1);
  return undef;
	
} #end getDefQmgr

{ #naked block for static qmgrs
  my %qmgrs;

#----------------------------------------------
sub reloadQmgrs{
#unset the Qmgrs hash so that it will reload
#----------------------------------------------
  undef %qmgrs;
  &getQmgrs(@_);
  return %qmgrs;

} #end reloadQmgrs 

#----------------------------------------------
sub getQmgrs {
# Load a hash with a list of all qmgr names and the Prefix for its data dir
# followed by log directory (space delimited).  Can receive qmgr name(s), which
# will look only for info on qmgr(s) and load in vars. 
# hash contents is returned at end.
# will return previously loaded data if called again, unless reloadQmgrs is called
# If a qmgr ini file is unavailable, the log value will be null
# will DIE if no queue managers are found, so call with an eval if you don't want that
#
# updated 3/08 to always get all qmgrs, but return only the portion requested; caching now works better too
#----------------------------------------------
  my $parent=&parentFunc();
  my @manual;#get only specific qmgrs
  if (@_ && $_[0]) { 
    &showDebug("specific qmgr(s) retrieval only: \"@_\"",1);
    @manual=split /\s+/,join(" ",@_); #take all args and whitespace and put back together in array
  }
  
  if (%qmgrs) { #return cached data for speed
    &showDebug("Returning existing cached QMGR hash for speed; use reloadQmgrs to force",1);
    if (@manual) {
      my %lqmgrs;
      my $notfound;
      foreach my $manqmgr (@manual) {
        if (exists $qmgrs{$manqmgr}) {
          $lqmgrs{$manqmgr}=$qmgrs{$manqmgr};
        } else { #qmgr either doesn't exist or wasn't loaded previously
          $notfound++;
        }
      } #end foreach
      &showDebug("Not all requested qmgrs were found.  Reloading from scratch",1) if $notfound;
      return %lqmgrs unless $notfound;
    } else { #return all of them
      return %qmgrs;
    } 
  }
  
  #if we get here, load all the qmgr info from scratch
  #############################
  if ($WINDOWS) {
    my $pointer=$Win32::TieRegistry::Registry->{$registryBaseKey ."Configuration/QueueManager/"}; #base already has a slash, and can't add another for keys
    
    #get qmgrs
    foreach my $qmgr ($pointer->SubKeyNames) {
      &showDebug("Found qmgr $qmgr",1);
      my $prefix=$pointer->{"$qmgr//Prefix"};
      my $logpath=$pointer->{"$qmgr/Log//LogPath"};
      $qmgrs{$qmgr}="\"$prefix\" \"$logpath\"";
      die "Found qmgr $qmgr, but couldn't locate prefix and/or log in registry" unless $prefix && $logpath;
    }
  #############################
  } else { #UNIX/Linux
    { #naked block to scope the vars
    my ($qmgr,$prefix,$directory);
    open MQSINI, "<$mqs_ini" #should be globally readable
     or die "Cannot open MQ Config File $mqs_ini: $!\n";
    while (<MQSINI>) { #search for a qmgr name
      if (/^QueueManager:/ .. /^\s+Directory=/) {
        $qmgr = (split /=/)[1] if (/^\s+Name=/);
        $prefix = (split /=/)[1] if (/^\s+Prefix=/);
        $directory = (split /=/)[1] if (/^\s+Directory=/);
        chomp ($qmgr) if ($qmgr);
        chomp ($prefix) if ($prefix);
        chomp ($directory) if ($directory);
        #tricky; if all found and either not a single qmgr or single 
        #is a match, then load it
        if ($qmgr && $prefix && $directory) {
          $prefix.="/qmgrs/$directory";
          &showDebug("Adding qmgr: $qmgr, prefix: $prefix",1);
          $qmgrs{$qmgr}="$prefix";
          $qmgr=undef;
          $prefix=undef;
          $directory=undef;
        }
      } #end if
    } #end while
    close MQSINI; #all qmgrs should now be loaded
    } #end naked block for scope

    #This part will add the qmgr log dir to the hash for each qmgr.
    #If not found (HA, not mounted), then leave blank
    { #new naked block for scope
    my ($logDir,$qmgr,$prefix);
    &showDebug("Populating qmgr log directories",1);
    while ( ($qmgr,$prefix) = each %qmgrs ) {
      my $qmini="$prefix/qm.ini";
      if ( -f $qmini ) { #ini file exists
        open QMINI, "<$qmini" #should be globally readable
         or die "Cannot open $qmgr qm.ini file: $!\n";
        &showDebug("$qmgr: Searching $qmini",2);
        while (<QMINI>) { #search for log stanza
          if (/^Log:/ .. /^\s+LogPath=/) {	
            $logDir = (split /=/)[1] if (/^\s+LogPath=/);
            if ($logDir) {
              chomp ($logDir);
	      &showDebug("$qmgr: found Log Dir:$logDir",2);
	      #need to remove trailing slash for consistency
	      $logDir = $1 if ($logDir =~ m{(.+)/$});
	      $qmgrs{$qmgr}.=" $logDir";
              $logDir=undef;
            }
	  } #end ranger operator search for stanza (p103)
        } #end file loop
        close QMINI;
      } else { 
        &showDebug("$qmgr ini file doesn't exist: $qmini",1);
      } #end check for ini file
    }  #end loop of qmini files
    } #end naked block
  } #end UNIX/WIn
    
  die "No queue managers were found\n" unless keys(%qmgrs);
  #if this dies in WMQIUINV.pm, it will cause the entire health check to fail
  if (@manual) {
    my %lqmgrs;
    my $notfound;
    foreach my $manqmgr (@manual) {
       if (exists $qmgrs{$manqmgr}) {
        $lqmgrs{$manqmgr}=$qmgrs{$manqmgr};
      } else { #qmgr either doesn't exist or wasn't loaded previously
        $notfound++;
      }
    } #end foreach
    ($notfound) ? die "Queue manager(s) \"@manual\" could not be found!\n" : return %lqmgrs;
  } else { #return all of them
    return %qmgrs;
  } 

} #end getQmgrs
} #end naked block
  
#----------------------------------------------
sub showQmgrs {
#prints the output retrieved in getQmgrs for ksh consumption
# it will print a report with qmgrs, qmfs, and logfs, which can be used 
# externally by ksh progs
# CAUTION with the caching of qmgrs above, but not likely a worry since this
# will be called from ksh which won't persist anyhow
#----------------------------------------------
  my %qmgrs=&getQmgrs(@_);
  foreach my $key (sort keys %qmgrs) {
    my ($qmfs,$logfs)= (split / /,$qmgrs{$key});
    $logfs="" unless ($logfs); #so we don't print uninited value
    printf "%s %s %s\n",$key,$qmfs,$logfs; # GetQmgrs needs space-delimited values
  }

} #end showQmgrs

#----------------------------------------------
sub convertReturnCode {
# Receives an integer return code and returns the reason
#----------------------------------------------

  my $code = pop @_;
  # RETURN CODES FROM 5.2.0.4
  # 3   Queue manager being created	*endmqm,strmqm,dltmqm
  # 5   Queue manager running		*strmqm,dltmqm
  # 8   Queue manager already exists	*crtmqm
  # 10  Command completed with unexpected results	*strmqcsv,endmqcsv
  # 					*dspmqfls,runmqtmc
  # 16  Queue manager does not exist	*endmqm,strmqm,dltmqm,runmqtrm
  # 20  An error occurred during processing	*strmqcsv,endmqcsv,dspmqfls
  # 					*runmqtrm,runmqtmc
  # 23  Log not available		*strmqm
  # 36  Invalid arguments supplied	*rcdmqimg,rcrmqobj,dspmqaut,setmqaut
  # 40  Queue manager not available	*rcdmqimg,rcrmqobj,endmqm,dspmqaut
  # 					*setmqaut
  # 49  Queue manager stopping	*rcdmqimg,rcrmqobj,endmqm,strmqm,crtmqm,dltmqm
  # 				*dltmqm,dspmqaut,setmqaut
  # 68  Media recovery is not supported	*rcdmqimg, rcrmqobj
  # 69  Storage not available	*rcdmqimg,rcrmqobj,endmqm,strmqm,crtmqm,dltmqm
  # 				*dspmqaut,setmqaut
  # 71  Unexpected error	*rcdmqimg,rcrmqobj,endmqm,strmqm,crtmqm,dltmqm
  # 				*dspmqaut,setmqaut
  # 72  Queue manager name error	*rcdmqimg,rcrmqobj,endmqm,strmqm,crtmqm
  # 					*dltmqm,dspmqaut,setmqaut
  # 100 Log location invalid	*strmqm,crtmqm,dltmqm
  # 111 Created but probs	*crtmqm
  # 112 Deleted but probs	*endmqm
  # 115 Invalid log size	*crtmqm
  # 119 User not authorized	*rcdmqimg, rcrmqobj
  # 128 No objects processed	*rcdmqimg, rcrmqobj
  # 131 Resource problem        *rcdmqimg
  # 132 Object damaged          *rcdmqimg
  # 133 Unknown object name	*dspmqaut,setmqaut
  # 135 Temporary object cannot be recovered	*rcdmqimg, rcrmqobj
  # 136 Object in use             *rcrmqobj
  # 145 Unexpected object name	*dspmqaut,setmqaut
  # 146 Object name missing	*dspmqaut,setmqaut
  # 147 Object type missing 	*dspmqaut,setmqaut
  # 148 Invalid object type	*dspmqaut,setmqaut
  # 149 entity name missing	*dspmqaut,setmqaut
  # 150 Authorization specification missing	*setmqaut
  # 151 Invalid authorization specification	*setmqaut

  my %returnCodes = (
	  3	=>	"Queue manager being created",
	  5	=>	"Queue manager running",
	  8	=>	"Queue manager already exists",
	  10	=>	"Command completed with unexpected results",
	  16	=>	"Queue manager does not exist",
	  20	=>	"An error occurred during processing",
	  23	=>	"Log not available",
	  36	=>	"Invalid arguments supplied",
	  40	=>	"Queue manager not available",
	  49	=>	"Queue manager stopping",
	  68	=>	"Media recovery is not supported",
	  69	=>	"Storage not available",
	  71	=>	"Unexpected error (Media recovery not allowed)",
	  72	=>	"Queue manager name error",
	  100	=>	"Log location invalid",
	  111	=>	"Queue manager created, but with problems",
	  112	=>	"Queue manager deleted, but with problems",
	  115	=>	"Invalid log size",
	  119	=>	"user not authorized",
	  128	=>	"No objects processed",
	  131	=>	"Resource problem",
	  132	=>	"Object damaged",
	  133	=>	"Unknown object name",
	  135	=>	"Temporary object cannot be recovered",
	  136	=>	"Object in use",
	  145	=>	"Unexpected object name",
	  146	=>	"Object name missing",
	  147	=>	"Object type missing",
	  148	=>	"Invalid object type",
	  149	=>	"Entity name missing",
	  150	=>	"Authorization specification missing",
	  151	=>	"Invalid authorization specification",
	  9216	=>	"AMQ7965 Incorrect number of arguments",
  );
  return $returnCodes{$code} || "Return code not found";
} #end ConvertReturnCode

#----------------------------------------------
sub checkReturnCode {
# check a given return code and print an error if it isn't zero
#----------------------------------------------
  my ($cmd,$code) = @_;
  $code = $code >> 8; #shift right one byte (p812 in book)
  if ( $code != 0 ) {
	  my $message = &convertReturnCode($code);
	  warn "  ERROR: $cmd: $message ($code)\n\n";
  }
	
} #end checkReturnCode

#----------------------------------------------
sub convertQmgrFS {
# Will receive the name of a qmgr (or anything, really), and convert
# according to the WMQ guidelines, which are that dots become !, 
#The rules governing the transformation of a queue manager name are as follows: 
#Transform individual characters: 
#. becomes ! 
#/ becomes & 
#If the name is still not valid: 
#Truncate it to eight characters 
#Append a three-character numeric suffix 
#----------------------------------------------
	my @filesys=@_; #This can accept any number of filesys names
	&showDebug("Converting qmgr path names",1);
	#See page 157 in manual
	tr/\./\!/ for @filesys;
	tr{/}{&} for @filesys;
    	return wantarray ? @filesys : $filesys[0];
} #end convertQmgrFS

#---------------------------------------
sub dspmqver {
#return the location of the mqversion cmd, if any
#Updated 05/13/08 to support Windows
#---------------------------------------
  &checkPlatform unless ($mqBinaries); #find binary location
  my @try;
  push @try,catfile($ENV{HOME},"dspmqver") if $ENV{HOME}; #for dev testing, it's nice to be able to place a dummy file in homedir
  push @try,catfile($mqBinaries,"bin","dspmqver");
  push @try,catfile($mqBinaries,"bin","mqver");
  
  foreach (@try) {
    &Kline::Display::showDebug("Checking for mq-version command at $_",1);
    if ($WINDOWS) {
      $_.=".exe";
    }
    if (-x $_) {
      &showDebug("found mq version command at $_",1);
      return $_;
    }
  }
  
  &showDebug("Niether mqver nor dspmqver were found at @try.  Can't find version!",1);
  return ""; #if we get here, then no MQ installed
} #end dspmqver

#----------------------------------------------
sub getMQVersion {
#returns the version and fixpack levels
#if passed with "print" arg, it will verify min version and throw error if lower
#otherwise, just returns level
#Updated to look for dspmqver (supported in v6.0) first
#Updated for x.y major and a.b.c.d complete to match v6 stuff
#Does not require mqm group membership, has not been updated for windows (that code sits in WMQConfig)
#----------------------------------------------
#[mqm@mqslinux mqm]$ mqver              **TESTED OK NOW**
#MQSeries Version: 520
#CMVC level:       p520-CSD04G
#BuildType:        IKAP - (Production)
#
#[mqm@mqslinux bin]$ mqver              **TESTED OK NOW**
#MQSeries Version: 520
#CMVC level:       p520-08-031028
#BuildType:        IKAP - (Production)
#
#cnkline@nytids003[/home/cnkline]mqver  **TESTED OK NOW**
#Name:        WebSphere MQ
#Version:     530
#CMVC level:  p000-L021011
#BuildType:   IKAP - (Production)
#
#mqm@d03ha101b1_boota[/u/mqm]mqver
#Name:        WebSphere MQ
#Version:     530.3  CSD03
#CMVC level:  p530-CSD03J
#BuildType:   IKAP - (Production)
#
#cnkline@tusnws011[/home/cnkline]mqver
#Name:        WebSphere MQ
#Version:     530.5  CSD05
#CMVC level:  p530-05-L030926
#BuildType:   IKAP - (Production)
#
#mqm@pogi[tools]dspmqver
#Name:        WebSphere MQ
#Version:     530.8  CSD08
#CMVC level:  p530-08-L040921
#BuildType:   IKAP - (Production)
#
#mqm@pogi[/home/mqm]dspmqver
#Name:        WebSphere MQ
#Version:     530.11  CSD11
#CMVC level:  p530-11-L050802
#BuildType:   IKAP - (Production)
#
#mqm@pinatubo:~> dspmqver
#Name:        WebSphere MQ
#Version:     6.0.0.0
#CMVC level:  p000-L050106
#BuildType:   IKAP - (Production)
#
#---------------------------------------

  my $thisFunction = &thisFunc();
  &showDebug("Looking for binaries at $mqBinaries",2);
  my($major,$complete);
  my $dspmqver=&dspmqver();
  my $mqver="";
  $mqver = &Kline::Server::cmdLine("$dspmqver") if $dspmqver;

  if (@$mqver) { #mqver was found
    &showDebug("Version Output:\n@$mqver",3);
    foreach (@$mqver) {
      if (/^Version:\s+(\d\.\d)\.\d\.\d/) { #6.0+ version
        $major=$1;
        $complete=$1 if (/^Version:\s+(\d\.\d\.\d\.\d)/); #6.0.0.0 fixpack
      } elsif (/^Version:\s+(\d+)/) { #5.3 version
        #know we have 530, check whether 530.5 (base has no suffix)
        $major=$1;
        $major="5.3" if $major == "530";
        if (/^Version:\s+(\d+)\.(\d+)/) { #check for a CSD
          $complete="$major.0.$2";
        } else {
          $complete="$major.0.0"; #base 5.3 code
        }
      } elsif (/^MQSeries Version:\s+(\d+)/) { #5.2 version
        $major =$1; #should be 520
        $major="5.2" if $major == "520";
        #can't get the CSD in the same line, but the next line should have it
      } elsif (/CMVC level:\s+p\d\d\d-CSD(\d+)|CMVC level:\s+p\d\d\d-(\d\d)-/) { #5.2 CSD, incl goofy CSD08
        $complete=$1||$2;
        $complete=sprintf("%d",$complete); #trim leading zeroes in case of 07/08/etc
        $complete="$major.0.$complete";
      }
    } #end foreach

    $major="0.0" unless $major;
    $complete="$major.0.0" unless $complete;
    print STDERR "$thisFunction could not determine MQ version level\n" if ($major == 0.0);
    &showDebug("MQ Version $major ($complete) was found",1);
  } else { #5.1 or earlier, need old method
    &showDebug("mqver not found; version 000/00 set",1);
    $major=0.0;
    $complete=0.0.0.0;
  }
 
  #check whether the version meets Kline minimums, but only if full Kline installed
  my @result;
  if (defined $Kline::full && $Kline::full) {
    my ($minVer,$minCSD)=split /,/,&getPropertyValue("GLOBAL","minmqversion");
    my ($maxVer,$maxCSD)=split /,/,&getPropertyValue("GLOBAL","maxmqversion");
    &showDebug("Minimum Version: $minVer; Minimum complete: $minCSD",1);
    &showDebug("Maximum Version: $maxVer; Maximum complete: $maxCSD",1);
    $minCSD="base" if ($minCSD=~/N\/A/i);
    $maxCSD="all" if ($maxCSD=~/N\/A/i);
    if ($major < $minVer || ($major == $minVer && ($minCSD=~/\d+/ && $complete < $minCSD))) {
      push @result,"WARNING: WMQ Version $major ($complete) is not supported for Kline.\nThe minimum version supported is $minVer ($minCSD). \n";
    }
    if ($major > $maxVer || ($major == $maxVer && ($maxCSD=~/\d+/ && $complete > $maxCSD))) {
      push @result,"WARNING: WMQ Version $major ($complete) is not supported for Kline.\nThe maximum version supported is $maxVer ($maxCSD). \n";
    }
    &printError(@result) if (@result && @_ && (shift(@_) =~ /print/i));
  } else { #Not full Kline install
    &showDebug("Full Kline package is not installed, cannot check for config file-derived min/max WMQ version",1);
  }

  if (wantarray()) {
    return ($major,$complete,@result);
  } else { #ws called in scalar context, which is likely kshtoperl
    &printError(@result) if @result;
    print "MQ Version $major, CSD $complete\n";
    return 0;
  }

} #end getMQVersion

#----------------------------------------------
sub isQmgrRunning {
#Tests whether the given queue manager is running and returns true or
#false.  
#----------------------------------------------
  my $qmgr=shift(@_);
  &Kline::Server::checkPlatform unless $mqBinaries;
  my $dspmq=catfile($mqBinaries,"bin","dspmq");
  if ($WINDOWS) {
    $dspmq.=".exe";
    $dspmq="\"$dspmq\""; #quote it so windows knows where the spaced-cmd ends and the args begin; the Server::cmdLine won't know where to break it.
  }
  my $status=&Kline::Server::cmdLine("$dspmq -m $qmgr");
  if (grep /STATUS\(Running\)/i,@$status) {
    return 1
  } else {
    return 0
  }

} #end isQmgrRunning

#----------------------------------------------
sub queueSize {
# Used (first) by lsmqdp to show queue size on disk
# Expects name of qmgr and queue, will return queue
# size in kb.  Rounding is done after math
#----------------------------------------------
  my ($qmgr,$queue)=@_;
  my $size;
  my $thisFunction = &thisFunc();
  my $parent= &parentFunc();
  &showDebug("Called by $parent",1);
  &checkPlatform unless ($mqBinaries); #find binary location
  my $dspmqfls=catfile($mqBinaries,"bin","dspmqfls");
  $dspmqfls.=".exe" if $WINDOWS;
  &showDebug("check size for queue $queue on qmgr $qmgr",1);

  #get the location of the queue file via dspmqfls
  my $output=&Kline::Server::cmdLine("$dspmqfls -m $qmgr -t ql $queue 2>&1");
  #normal output looks like this (with some vertical spacing too)
  #WebSphere MQ Display MQ Files
  #QLOCAL    SYSTEM.ADMIN.COMMAND.QUEUE
  #/var/mqm/qmgrs/AMY/queues/SYSTEM!ADMIN!COMMAND!QUEUE
  #shows "No objects processed." on STDERR if doesn't exist, rc still 0

  foreach my $qdir (@$output) {
    if ($qdir=~/No objects processed/i) {
      warn "WARNING: (dspmqfls) Queue $queue could not be found on the filesystem for qmgr $qmgr\n";
      return 0;
    }
    next unless ($qdir =~ /^\//); #take only the actual filesystem row
    $qdir=&trim($qdir); #net the dir where q file is at
    #5th column is dize on aix, sun, and linux
    #bytes by default (OS can convert, but easier to do here)
    $size=&trim(`ls -l $qdir/q | awk '{print \$5}'`); #needs update for Win
    $size=sprintf "%.1f",($size / 1024);
    &showDebug("$queue size is $size kb",1);
  } #end foreach

  return $size;
}#end queueSize

#----------------------------------------------
sub dumpLogs {
#get qmgr filesys and copy all AMQ and FDC to file named
#used by tarmqlog
#BT38359 MAJOR REWRITE updated to receive hash with various args so
#calling program can control behavior.
#if key qmgr exists, single qmgr is run, else all qmgrs
#if key assemble exist, files are only assembed at that location
#output directory goes to key outputdir, fails if not declared
#if key FDC exists, only FDC files are gathered, not logs
#if key tarfile exists, output file name is overridden
#Also added so @SYSTEM is gathered whenever FDC's are gathered.
#----------------------------------------------
  use Sys::Hostname;
  my $hostname = hostname();
  my %args=@_; #should receive a hash

  #check if output dir
  unless ($args{outputdir} || $args{assemble}) { #if not found, exit
    warn "WARNING: outputdir key was not found in the hash argument.  Call me correctly!\n";
    return 0;
  }
  
  #check for filename override take only the basename
  my $tarFile=$args{tarfile} || "AMQERRLogs_${hostname}.tar";
  $tarFile=basename($tarFile);
  &showDebug("Using tar file name $tarFile",1);

  my $singleQmgr; #if empty, do them all
  $singleQmgr=$args{qmgr} if exists($args{qmgr});
  &showDebug("Logs to be gathered for single qmgr $singleQmgr only",1) if $singleQmgr;

  my $outputDir=$args{outputdir} || $args{assemble};
  my $tempDir="/tmp/AMQERRLogs.${hostname}";
  &showDebug("Package destination is $outputDir",1);
  &showDebug("Creating files at temp dir $tempDir",1);
  (-d $tempDir) ? &showDebug("Temp Dir $tempDir already exists",1) : system("mkdir -p $tempDir");
  (-d $outputDir) ? &showDebug("Output Dir $outputDir already exists",1) : system("mkdir -p $outputDir");

  #check whether FDC only
  if (exists($args{FDC})) { #don't get qmgr logs
    &showDebug("Only FDC logs will be gathered; no qmgr logs",1);
  } else { #get qmgr logs
    my %qmgrHash=&getQmgrs($singleQmgr);
    #add @SYSTEM to the hash
    $qmgrHash{'@SYSTEM'}='/var/mqm/qmgrs/@SYSTEM';
    foreach my $qmgr (sort keys %qmgrHash) {
      my $prefix;
      if ($qmgrHash{$qmgr}=~/([^\s]+)/) { #get only the first "word"
        $prefix=$1;
      }
      &printArrowSection("Gathering AMQERR Logs for $qmgr");
      &showDebug("qmgr file prefix is $prefix",1);
      my @files=qw(AMQERR01.LOG AMQERR02.LOG AMQERR03.LOG);
      if (-d "$prefix/errors") { 
        &printItem("${qmgr}'s files were found at $prefix and will be used");

        my $filecount=&Kline::Server::cmdLine("ls -1 $prefix/errors | wc -l"); #aref,if empty, then no files found
        unless (@$filecount) {
          &printItem("No error log files were found at $prefix/errors.  Skipping collection");
          next;
        }
        #get the last mod times for STDOUT
        my ($oldest, $newest);
        chomp($oldest=`$Kline::KlineHome/parsemqerrs -sdm $qmgr | grep ^Oldest | awk -F';' '{print \$4}'`);
        chomp($newest=`$Kline::KlineHome/parsemqerrs -sdm $qmgr | grep ^Newest | awk -F';' '{print \$4}'`) if $oldest;
        if ($oldest) { #files were found
          &printItem("Log files span from $oldest to $newest");
        }
        foreach (@files) {
          if (-r "$prefix/errors/$_") {
            &showDebug("Copying $prefix/errors/$_ $tempDir/${qmgr}_$_",1);
            ((system "cp -p $prefix/errors/$_ $tempDir/${qmgr}_$_") == 0) || &printItem("ERROR: $_ did not copy successfully: $!");
          } else { #not readable
            &printItem("ERROR: ${qmgr}'s $_ is not found or not readable (permissions?)");
          }
        } #end loop AMQERR files
      } else {
        &printItem("${qmgr}'s filesystem $prefix is not found or mounted (perhaps HA?)");
        &printItem("This is an expected condition for the \@SYSTEM directory") if ($qmgr eq '@SYSTEM');
      }
    }
  } #end FDConly check

  my $numFDC=&trim(`ls -1 /var/mqm/errors | wc -l | awk '{print \$NF}'`);
  &printArrowSection("Gathering $numFDC FDC files and logs");
    mkdir "$tempDir/FDC",0777 unless (-d "$tempDir/FDC");
    ((system "cp -p /var/mqm/errors/* $tempDir/FDC/") == 0) || &printItem("ERROR: Not all FDC files copied properly!");

  #check whether assemble only
  if (exists($args{assemble})) { 
    &showDebug("Assembling files only at $tempDir, no zip to take place",1);
    &printArrowSection("Moving files to assembly directory $outputDir");
    ((system "mv $tempDir/* $outputDir") == 0) || &printItem("ERROR: Could not move files from $tempDir!");
    rmdir "$tempDir" || &printItem("ERROR: Could not remove temp dir $tempDir!");
    &printArrowSection("MQ Log File Assembly is complete at $outputDir");
  } else { #proceed and zip

    &printArrowSection("Creating tar file $outputDir/$tarFile");
    my $curdir=$ENV{PWD};
    chdir dirname($tempDir);
    my $subdir=basename($tempDir);
    ((system "tar cpf $outputDir/$tarFile ./$subdir/*") == 0) || &printItem("ERROR: Could not create tar file $outputDir/$tarFile!");
    chomp(my $size=`ls -l $outputDir/$tarFile | awk '{print \$5}'`);
    &printItem("Uncompressed size is $size bytes.");
    chdir $curdir;
    ((system "rm -rf $tempDir") == 0) || &printItem("ERROR: Could not remove temp dir $tempDir!: $!");

    #get the compression program and suffix from Kline.cfg
    my $compressProg=&getPropertyValue("GLOBAL","compressionProgram");
    my $compressSuffix=&getPropertyValue("GLOBAL","compressionSuffix");
    &printArrowSection("Compressing $tarFile using $compressProg");
    !system("$compressProg -f $outputDir/$tarFile") || die "Error zipping $tarFile: $!\n";
    chomp(my $newsize=`ls -l $outputDir/$tarFile${compressSuffix} | awk '{print \$5}'`);
    &printItem("Compressed size is $newsize bytes.");
    &printArrowSection("MQ Log File Package is complete at $outputDir/$tarFile${compressSuffix}");
  } #end assembly check

} #end dumpLogs

#----------------------------------------------
sub getObjTypeList {
#returns a list of all the object types on this box
#returns undef if no qmgrs running
#----------------------------------------------
  my $this=&thisFunc();
  my $parent=&parentFunc();
  die "\nNo authority for $this ($parent); UID $UID not in mqm group\n" unless &isInMQMGroup();
  my %qmgrs=&getQmgrs;
  foreach my $qmgr (sort keys %qmgrs) { #loop through qmgrs until we find one that's running so we can query runmqsc
    next unless &isQmgrRunning($qmgr);
    my $cmd = "echo ${mqscquote}dis$mqscquote | $runmqsc $qmgr"; #force a syntax error so all attrs are listed
    &showDebug("command is $cmd",3);
    open RUNMQSC, "$cmd|" or die "Cannot execute runmqsc: $!\n";
    my @list;
    while (<RUNMQSC>) {
      &showDebug("MQSC LINE*****:\n$_\nEND MQSC LINE*****",4);
      push @list,$1 if /\s+DISPLAY (\w+)/;
    }
    close RUNMQSCATTRS;
    @list=&trim(@list); #dump spaces, if any
    @list=sort(@list);
    &showDebug("Found the following generic object types for this level of MQ: @list",2);
    return wantarray ? @list : \@list;
  }
  #if we get here, then no qmgrs were running
  $!="No running queue managers to query for object types!";
  &Kline::Display::showDebug($!,1);
  return undef;

} #end getObjTypeList

#----------------------------------------------
sub getAttrList {
#Queries a passed object and qmgr and returns a list of all the attributes
#that object possesses
# NOTICE: THIS FUNCTION IS NOT PERFECT AND CANNOT DO THINGS DIS QSTATUS TYPE(HANDLE)
# BUT IT CAN DO ALL NORMAL OBJECT TYPES AND DEFAULT STATUSES.
#----------------------------------------------
  my ($qmgr,$rawtype)=@_; #receive qmgr name and object type always

  #sanity check
  my $thisFunction=&thisFunc();
  unless ($qmgr && $rawtype) {
    warn "WARNING: $thisFunction did not receive proper parameters: qmgr, type\n";
    return 0;
  }
  unless (&isQmgrRunning($qmgr)) {
    warn "WARNING: $thisFunction: $qmgr is not running.  Operation Aborted.\n";
    return 0;
  }

  my (@attrs,@finalattrs);
  $rawtype=&convertObjType($rawtype); #will UC
  my $type=$rawtype; #need to have two vars since one is altered
  $type.="(\*)" unless ($type=~/qmgr/i); #qmgr object can't have wildcard
  #needs backslash for shell escape when command is issued
  my $cmd = "echo ${mqscquote}dis $type";
  $cmd .= " foo_bar$mqscquote | $runmqsc $qmgr"; #force a syntax error so all attrs are listed
  my ($rc,$out)=&Kline::Server::cmdLine({ command => $cmd,
                                       shift_rc => 1,
                                       });
  #die "Cannot execute runmqsc: $!\n" if $rc; of course it fails here every time!

  #now remove all the junk at the top of the array, until the 8427 code is found
  #file:///D:/perl_oreilly/cookbook/ch06_08.htm for splitting records
  my $olddelim=$/;
  undef $/;
  foreach (@$out) {
    &showDebug("MQSC LINE*****:\n$_\nEND MQSC LINE*****",4);
    @attrs=split(/\[([^\]]+)\]/, $_); #create array from line
    @attrs=&trim(@attrs); #dump spaces
    pop @attrs; #remove last element (not needed)
    shift @attrs; #remove first element (not needed)
    #now remove bad attrs like ALL, and clean
    foreach  my $attr (@attrs) {
      next if ($attr =~ /^$|^ALL$/); #purge empty lines or ALL
      next if $attr =~ /([^(]+)\(/; #if line has parens it's a search filter line and is a duplicate, so toss it
      &showDebug("Adding to attribute list for $rawtype: **--$attr--**",2);
      push @finalattrs,$attr; #load cleansed val
    }
  }

  #the name of the object isn't returned in the above, so need to add it manually
  #needs to be dynamic for all types except queues since the name is QUEUE regardless
  #of which type
  #OVER TIME, ADDITIONAL ITEMS WILL BE ADDED HERE AS THEY ARE NOTICED TO NOT SHOW UP IN ATTR SYNTAX OUTPUT
  #IF ATTRS ARE MISSING, ADD THEM HERE
  unless ($rawtype =~ /\bqmgr\b/i) { #skip this whole thing if qmgr; don't confuse with clusqmgr
    my @lastattrs="$rawtype"; #set the obj type as the objects name field by default
    @lastattrs="QUEUE" if ($rawtype =~ /q(s|u|r|mo|l|c|a)/i); #only clusqmgr and qmgr can't follow this rule
    @lastattrs=qw/CHANNEL CONNAME XMITQ CHLTYPE/ if ($rawtype =~ /chs/i); #chstatus has channel as its name
    &showDebug("FINAL Add to attr list for $rawtype (can be multiples): **--@lastattrs--**",2);
    push @finalattrs,@lastattrs;
  } 

  #I found that chstatus doesn't include xmitq, conname, or chltype in its attrlist, so add manually

  $/=$olddelim; #put the record delim back
  @finalattrs=sort(@finalattrs);
  return @finalattrs;

} #end getAttrList

#----------------------------------------------
sub convertObjType {
#receives a MQSC object type and converts it to long notation if short
#used by getAttrList and loadQmgrObjects
#1/2008 Updated for MQ6 object names and synonyms
#----------------------------------------------
local $_=shift(@_);
SWITCH: {
  if (/\bauthinfo\b/i) { return "AUTHINFO"; last SWITCH; }
  if (/\bchl\b/i) { return "CHANNEL"; last SWITCH; }
  if (/\bchs\b/i) { return "CHSTATUS"; last SWITCH; }
  if (/\bclusqmgr\b/i) { return "CLUSQMGR"; last SWITCH; }
  if (/\bnl\b/i) { return "NAMELIST"; last SWITCH; }
  if (/\bpro\b/i) { return "PROCESS"; last SWITCH; }
  if (/\bqa\b/i) { return "QALIAS"; last SWITCH; }
  if (/\bqc\b/i) { return "QCLUSTER"; last SWITCH; }
  if (/\bql\b/i) { return "QLOCAL"; last SWITCH; }
  if (/\bqmgr\b/i) { return "QMGR"; last SWITCH; }
  if (/\bqr\b/i) { return "QREMOTE"; last SWITCH; }
  if (/\bqm\b/i) { return "QMODEL"; last SWITCH; } #don't confuse with qmgr
  if (/\bq\b/i) { return "QUEUE"; last SWITCH; }
  if (/\bqs\b/i) { return "QSTATUS"; last SWITCH; }
  if (/\bconn\b/i) { return "CONN"; last SWITCH; }
  if (/\bservice\b/i) { return "SERVICE"; last SWITCH; }
  if (/\blstr\b/i) { return "LISTENER"; last SWITCH; }
  if (/\bsvstatus\b/i) { return "SVSTATUS"; last SWITCH; }
  if (/\blsstatus\b/i) { return "LSSTATUS"; last SWITCH; }
  if (/\bqmstatus\b/i) { return "QMSTATUS"; last SWITCH; }
  #if we get here, return what was passed, in UC
  } #end SWITCH
  return uc($_);

} #end convertObjType

{ #naked block
 my %sw;
 my $completed;
#----------------------------------------------
sub WMQConfig {
#check whether MQ is installed, returns hash with server and client as keys
#called at top of qmgrStatus and by Kline::Inventory::WMQ::main()
#call with any argument and it will return a statement about what is present
#shows basics of the install, not qmgrs; does not require mqm group membership
#calls getMQVersion to parse actual dspmqver output fully
#Supports AIX, Linux, Solaris, Win, with stub for HP
#----------------------------------------------
  my $mqsini="/var/mqm/mqs.ini"; #set default value
  if ($completed) {
    &showDebug("WMQConfig already ran, returning previous info:",2);
    return %sw;
  }
  
  $sw{server}=0;
  $sw{client}=0;

  ###########
  my $dspmqvercmd=&dspmqver(); #checks mult locations
  if (-x $dspmqvercmd) {
    my ($rc,$dspmqver)=&Kline::Server::cmdLine($dspmqvercmd); #aref, does not appear to require mqm group mem
    $rc=$rc >> 8;  #should be 0 for both unix and win
    die "$dspmqvercmd did not exit normally.  Cannot continue with inventory: $!\n" if $rc;
    $dspmqver=join "",@$dspmqver; #turn the aref back into a scalar string;
    $sw{complete}=$1 if ($dspmqver && $dspmqver=~/Version:\s+(.+)/); #gets the a.b.c.d version string (or 530.12  CSD12 on 5.3)
    $sw{dspmqver}=$dspmqver;
  }
  unless (exists $sw{complete} && $sw{complete}) { #either no cmd at all or didn't run normally
    &showDebug("dspmqver returned no version or cmd not found.  Returning empty assuming no MQ installed",1);
    return %sw;
 }
  
  ###########
  #now find specific stuff for each OS as to when the product was installed
  if ($WINDOWS) {
    #Windows can be checked completely by registry without shelling out at all
    #mq key will be completely missing if MQ is removed after install (checked on mq53 win2k)
    my $pointer=$Win32::TieRegistry::Registry->{$registryBaseKey}; #will not fail if key not found
    if (defined $pointer) { #MQ is found
      &Kline::Display::showDebug("The registry key \"$registryBaseKey\" was found; MQ is installed",1);
      
      #get version details
      if ($pointer->{"/VRMF"}=~/^(\d\.\d)\./) { #only works in 6.0+
        $sw{major}=$1;
        $sw{minor}=$pointer->{"/VRMF"};
      } else { #5.3 and older; pull from the MQServerVersion and MQServerRelease and fudge them around
        my ($version,$release);
        if ($version = $pointer->GetValue("MQServerVersion")) { #no slash needed here, dword, converts to 5
          $version=($Win32::TieRegistry::Registry->DWordsToHex) ? hex($version) : unpack("L",$version); #dword hex conversion
          &showDebug("converted REG_DWORD VERSION value: $version (DWordsToHex is ".$Win32::TieRegistry::Registry->DWordsToHex .")",1);
        }
        if ($release = $pointer->GetValue("MQServerRelease")) { #no slash needed here, dword, converts to 30
          $release=($Win32::TieRegistry::Registry->DWordsToHex) ? hex($release) : unpack("L",$release); #dword hex conversion
          &showDebug("converted REG_DWORD RELEASE value: $release (DWordsToHex is ".$Win32::TieRegistry::Registry->DWordsToHex .")",1);
          $release=substr($release,0,1); #keep only the first digit
        }
        if ($version && $release) {
          $sw{major}="$version.$release";
          $sw{minor}=$1 if $sw{complete}=~/(\S+)/; #we don't have anywhere else to grab this, so just pull the first segment off mqver output
        }
      } #env 5.3 vs 6.0 version differences
      $sw{packages}=join " ",$pointer->{"Components/"}->ValueNames();
      $sw{client}++ if $sw{packages}=~/client/i;
      $sw{type}.="Client" if $sw{client};
      $sw{server}++ if $sw{packages}=~/\bserver\b/i;
      $sw{type}.= " Server" if $sw{server};
      $sw{mqsini}="Windows Registry";
      #sw{mqsIniContents} #we'd have to dump the registry vals
      
      #GET INSTALL-DATE INFO FROM REGISTRY TIMESTAMPS
      #sw{origPkgName}not sure how to grab this
      #sw{lastPkgName}
      if ($sw{major}) {
        #this is a reasonable shot at a key that does not change over time (tested 5.3 application of CSD and observed no changes to any base keys or values)
        #I confirmed that parent keys are not timestamp updated when a child key is updated
        #Nothing in the base key changes when a CSD is applied (tested on 5.3), so we have to check the CSDS Applied key since a new value is written for each CSD
        my @t=split(/\//,$registryBaseKey); #get the "one-level higher" in the registry
        pop @t;
        my $toplvl=join("/",@t) ."/";
        $sw{origInstDate}=&Kline::Server::filetime($Win32::TieRegistry::Registry->{$toplvl}->Information("LastWrite")) ." GMT"; #reads the last write time on the registry key; comes in binary FILETIME and must be converted to something useful; tests show it's GMT returned
        my $subkey=($sw{major} < 6.0) ? "CSDS Applied" : "Maintenance Applied"; #key name where info is stored changed at 6.0; confirmed same at 7.0
        $sw{lastInstDate}=&Kline::Server::filetime($pointer->{$subkey}->Information("LastWrite")) ." GMT"; #reads the last write time on the registry key; comes in binary FILETIME and must be converted to something useful; tests show it's GMT returned
      }
    }
  } #end Windows check
  
  #PLATFORM-SPECIFIC UNIX CHECKS
  ###########
  if ($^O =~ /hpux/i) {
    die "WARNING: Cannot gather inventory for $^O; code not yet complete\n";
      
  ###########
  } elsif ($^O =~ /linux/) {
    ($sw{major},$sw{minor})=&getMQVersion();
    $sw{packages}=`rpm -qa --last| grep -i MQSeries 2>&1`;
    my $pkg="MQSeriesClient";
    if (system("rpm -q $pkg >/dev/null 2>&1")==0) { #look for client first
      $sw{client}++ ;
      $sw{type}.="Client ";
      foreach (`rpm -qa --last | grep -i $pkg 2>&1`) { #shows pkgs in descending date
        ($sw{origPkgName},$sw{origInstDate}) = /(\S+)\s+(.+)/; #do this each time, you'll wind up with the oldest
        unless ($sw{lastPkgName} && $sw{lastInstDate}) { #do this only once, you'll wind up with newest
          ($sw{lastPkgName},$sw{lastInstDate}) = /(\S+)\s+(.+)/; 
        }
      } #end foreach
    } #end client
    $pkg="MQSeriesServer";
    if (system("rpm -q $pkg >/dev/null 2>&1")==0) { #look for server last
      $sw{server}++ ;
      $sw{mqsini}=$mqsini;
      $sw{mqsiniContents}=&getFileContents($mqsini);
      $sw{type}.="Server ";
      undef $sw{lastPkgName}; #clear these so they load again
      undef $sw{lastInstDate};
      foreach (`rpm -qa --last | grep -i $pkg 2>&1`) { #shows pkgs in descending date
        ($sw{origPkgName},$sw{origInstDate}) = /(\S+)\s+(.+)/; #do this each time, you'll wind up with the oldest
        unless ($sw{lastPkgName} && $sw{lastInstDate}) { #do this only once, you'll wind up with newest
          ($sw{lastPkgName},$sw{lastInstDate}) = /(\S+)\s+(.+)/; 
        }
      } #end foreach
    } #end server
    
  ###########
  } elsif ($^O =~ /aix/) {
    ($sw{major},$sw{minor})=&getMQVersion();
    $sw{packages}=`lslpp -ha mqm* 2>&1`;
    my $pkg="mqm.client.rte";
    if (system("lslpp -i $pkg >/dev/null 2>&1")==0) { #look for client first
      $sw{client}++ ;
      $sw{type}.="Client ";
      foreach (`lslpp -hacq $pkg | grep APPLY 2>&1`) { #shows pkgs in descending date
        #AIX 4 and 5 both have format below, but ptf column is usually empty
        #Path:Fileset:level:ptf:action:status:date:time
        my @info=&trim(split /:/,$_);
        &showDebug("AIX Software info: @info",1);
        ($sw{lastPkgName},$sw{lastInstDate})=($info[1],"$info[6] $info[7]"); #do this each time, you'll wind up with newest
        unless ($sw{lastPkgName} && $sw{lastInstDate}) { #do this only once, you'll wind up with oldest
          ($sw{origPkgName},$sw{origInstDate})=($info[1],"$info[6] $info[7]");
        }
      } #end foreach
    } #end client
    $pkg="mqm.server.rte";
    if (system("lslpp -i $pkg >/dev/null 2>&1")==0) { #look for server last
      $sw{server}++ ;
      $sw{mqsini}=$mqsini;
      $sw{mqsiniContents}=&getFileContents($mqsini);
      $sw{type}.="Server ";
      undef $sw{lastPkgName}; #clear these so they load again
      undef $sw{lastInstDate};
      foreach (`lslpp -hacq $pkg | grep APPLY 2>&1`) { #shows pkgs in descending date
        #Path:Fileset:level:ptf:action:status:date:time
        my @info=&trim(split /:/,$_);
        ($sw{lastPkgName},$sw{lastInstDate})=($info[1],"$info[6] $info[7]"); #do this each time, you'll wind up with newest
        unless ($sw{origPkgName} && $sw{origInstDate}) { #do this only once, you'll wind up with oldest
          ($sw{origPkgName},$sw{origInstDate})=($info[1],"$info[6] $info[7]");
        }
      } #end foreach
    } #end server

    #convert the time from ; to : for each val
    $sw{lastInstDate}=~s/;/:/g;
    $sw{origInstDate}=~s/;/:/g;

  ###########
  } elsif ($^O =~ /solaris/) { #tested 05/02/08 with MQ 6.0.0.0 on Sparc v10
    ($sw{major},$sw{minor})=&getMQVersion();
    $sw{packages}=`pkginfo | grep -i mqm | cut -f2 -d ' ' | xargs -i pkgparam -v {} PKG NAME INSTDATE CLASSES 2>&1`;
    #pkginfo output looks like:
      #PKG='mqm'
      #NAME='WebSphere MQ for Sun Solaris'
      #INSTDATE='Apr 24 2008 14:40'
    my $pkg="client";
    if (system("pkgparam mqm CLASSES | grep $pkg >/dev/null 2>&1")==0) { #look for client first
      #pkgparam output looks like:
      #none base De_DE Es_ES Fr_FR It_IT Ja_JP java javasdk keyman Ko_KR man Pt_BR runtime samples server sol_client txclient Zh_CN Zh_TW
      &showDebug("MQ client is installed (found $pkg in pkgparam mqm CLASSES)",2);
      $sw{client}++ ;
      $sw{type}.="Client ";
    } #end client
    $pkg="server";
    if (system("pkgparam mqm CLASSES | grep $pkg >/dev/null 2>&1")==0) { #look for server last
      &showDebug("MQ Server is installed (found $pkg in pkgparam mqm CLASSES)",2);
      $sw{server}++ ;
      $sw{mqsini}=$mqsini;
      $sw{mqsiniContents}=&getFileContents($mqsini);
      $sw{type}.="Server ";
    } #end server
    foreach my $name (`pkginfo | grep mqm | cut -f2 -d ' ' | sort -r 2>&1`) { #shows pkgs in descending date
      chomp $name; #line feeds will kill the pkgparam cmd output
      &showDebug("Checking install date for package $name",2);
      $sw{origPkgName}=$name;
      my $cmd="pkgparam $name INSTDATE"; #will return just the date, no parsing needed
      chomp($sw{origInstDate} =(`$cmd`)); #do this each time, you'll wind up with the oldest
      &showDebug("pkg $name ORIGINALLY installed on $sw{origInstDate}",2);
      unless ($sw{lastPkgName} && $sw{lastInstDate}) { #do this only once, you'll wind up with newest
        $sw{lastPkgName}=$name;
        chomp($sw{lastInstDate} =(`$cmd`));
        &showDebug("pkg $name LAST installed on $sw{origInstDate}",2);
      }
    } #end foreach
  } #end platform branch
  ###########
  
  &showDebug("WMQ Server is installed",1) if $sw{server};
  &showDebug("WMQ Client is installed",1) if $sw{client};
  foreach (sort keys %sw) { &showDebug("SW KEYS:<$_>$sw{$_}</$_>",2); };
  &printItem("WebSphere MQ Server is installed") if $sw{server} && @_;
  &printItem("WebSphere MQ Client is installed") if $sw{client} && @_;
  $completed++; #signal that we already ran once
  
  return %sw;
} #end WMQConfig
} #end naked block

#----------------------------------------------
sub qmgrStatus {
#AKA qmgrConfig (show status and config of qmgrs)
#return status of qmgrs (doesn't display natively here)
#receive optional qmgr name return AoH w/info
#aborts if no mq server installed or if no qmgrs found
#href includes defqmgr, qmgr, running, status, pid, etime, dateaccurate, date,
#broker, brokerpid, brokertime, brokerstatus,
#installDate, installLocation, alterDate, qminiName, qminiContents (for inventory)
#installDate and alterDate are only grabbed if user is in mqm group; mqm group not required for exec
#qmgr status is not pinged if not in mqm group
#uptime and broker details are not collected on Win2k and previous due to limitations
#use qmgrStatusWin for Windows
#----------------------------------------------
  my @AoH;
  
  #do not proceed without MQ server, but don't use WMQConfig since it's slow
  #so just look for the strmqm file
  &Kline::Server::checkPlatform unless $mqBinaries;
  my $strmqm="strmqm";
  $strmqm.=".exe" if $WINDOWS;
  unless (-f catfile($mqBinaries,"bin",$strmqm)) { #don't bother if only client is installed
    warn "WMQ Server is not installed, so no queue managers can be installed.\n";
    return 0;
  }

  #if server install, then proceed
  my %qmgrs=&getQmgrs(@_); #does not require mqm group membership, will die if no qmgrs
  &showDebug("Found ".scalar(keys %qmgrs)." qmgrs configured on this system",1);
  unless (scalar(keys %qmgrs)) {  #no qmgrs exist
    &printItem("WebSphere MQ server is installed, but no queue managers have been created.");
    return 0;
  }
  #if qmgrs defined, proceed
  
  my $maxqmgrlen=0; #track for column widths
  my $defqmgr;
  ###########################
  if ($WINDOWS) {
    foreach my $qmgr (sort keys %qmgrs) {
      &showDebug("WORKING ON WINDOWS QMGR: $qmgr",1);
      my $rec={};
      $rec->{qmgr}=$qmgr;
      $maxqmgrlen=length($qmgr) if (length($qmgr) > $maxqmgrlen); #for stdout print
      if ($defqmgr && $qmgr eq $defqmgr) { $rec->{defqmgr}=1 } else { $rec->{defqmgr}=0; }
      ($rec->{installLocation},$rec->{installLocationLog})=($1,$2) if ($qmgrs{$qmgr}=~/"([^"]+)"\s+"([^"]+)"?/);
      my $pointer=$Win32::TieRegistry::Registry->{$registryBaseKey};
      $rec->{status}=$pointer->{"Status/QueueManager/$qmgr/QueueManagerStatus//CurrentStatus"}; #Running, EndedImmediately, etc.  Will handle manually killed off below
      $rec->{running}=($rec->{status}=~/running/i) ? 1 : 0;
      $rec->{statusDate}=Kline::Server::filetime($pointer->{"Status/QueueManager/$qmgr/QueueManagerStatus/"}->Information("LastWrite")) ." GMT"; #reads the last write time on the registry key; comes in binary FILETIME and must be converted to something useful; tests show it's GMT returned
      $rec->{date}=$rec->{statusDate}; #this is duplicate, it appears, and is not documented really very well anywhere in the code, but just a copy.
      $rec->{dateaccurate}=1; #assume that the date on the reg key is really the last time a change was made
      $rec->{qminiName}="Microsoft Windows Registry at $registryBaseKey/Configuration/QueueManager/$qmgr/";
      #$rec->{qminiContents} #leave it undef
      
      if ($rec->{running}) {
        my $wincheck; #need to check if win2k or earlier, in which case we'll never have the PID since no CommandLine, so need to skip the below
        if ($WINDOWS && ! $rec->{pid}) {
          $wincheck=0; #define it
          my ($major,$minor)=(&Win32::GetOSVersion())[1,2]; #5.0 is Win2k, 5.1 is XP, 5.2 is 2k3
          $wincheck++ if ($major && $major >= 5 && $minor && $minor >= 1); #if XP or newer, then not having a PID means something was killed; if 2k or older, then we'll never have pid
        }
        if ($wincheck) { #if a qmgr is killed off, the registry will still show running, but we'll find no procs
        
          #get pids
          my $pi=Win32::Process::Info->new();
          my $amqzxma0="amqzxma0.exe";
          my @qmgrProcs=grep {$_->{Name} =~ m/$amqzxma0/} $pi->GetProcInfo( { no_user_info => 1 },); #can take a few moments to complete
          foreach (@qmgrProcs) { #loop through all qmgr PIDs for ours
            foreach my $key (sort keys %$_) {
              &showDebug("found process-table $key: $_->{$key}",3);
            }
            next unless $_->{CommandLine}=~/$amqzxma0.+-m\s+\b$qmgr\b/; #skip all lines that ain't us
            $rec->{pid}=$_->{ProcessId};
            $rec->{etime}=localtime($_->{CreationDate}); #it's retrieved in epoch seconds
            last;
          }
          unless ($rec->{pid}) { #if a qmgr is killed off, the registry will still show running, but we'll find no procs
            &showDebug("\nWARNING: Queue manager $qmgr is listed in the registry as running, but no executing process could be found.  Assuming \"Ended Unexpectedly\"!",1);
            $rec->{status}="Ended Unexpectedly";
            $rec->{running}=0;
            $rec->{dateaccurate}=0; #we thought it was up, but found no pid; means it probably died, so the reg date isn't accurate afterall
            $rec->{pid}=0;
          }
        
          my $amqfcxba="amqfcxba.exe";
          my @brkrProcs=grep {$_->{Name} =~ m/$amqfcxba/} $pi->GetProcInfo( { no_user_info => 1 },);
          $rec->{broker}=0; #assume broker down; cannot check it in the registry
          foreach (@brkrProcs) { #loop through all broker pids for ours
           next unless $_->{CommandLine}=~/$amqfcxba.+-m\s+\b$qmgr\b/; #skip all lines that ain't us
            $rec->{broker}=1;
            $rec->{brokerstatus}="Running";
            $rec->{brokerpid}=$_->{ProcessId};
            $rec->{brokeretime}=localtime($_->{CreationDate}); #it's retrieved in epoch seconds
            last;
          }
          unless ($rec->{broker}) { #if broker not running, check if configured
            my @glob=glob "$rec->{installLocation}\\queues\\SYSTEM\!BROKER\!IQ*";
            if (@glob) { #check for required queues
              &showDebug("PubSub Broker queues found for $qmgr: @glob",2);
              $rec->{brokerstatus}="Down";
            } else {
              &showDebug("No PubSub Broker queues found for $qmgr",2);
              $rec->{brokerstatus}="Never Run";
            }
          } #end broker configured
        } else { #no wincheck, meaning Win2k or older
            #NOTE: CommandLine key is not present on Win2k, so there is no way to correlate PID's to qmgrs; use below debug to see what is available on different platforms
            &showDebug("Cannot correlate PID's to queue managers on Windows versions prior to XP/2003 since CommandLine key is not present",1);
        } #end wincheck
      
        #for inventory, get install date, install location, alterDate
        #installDate comes from CRDATE in dis qmgr
        #alterDate comes from ALTDATE in dis qmgr
        #installLocation would be in mqs.ini
        #qminiName is the name of the qm.ini file
        #qminiContents is the contents of the qm.ini file
        if ($rec->{running} && &Kline::WMQ::isInMQMGroup()) { #check running again in case of manual kill off above
          $rec->{installDate}=&getMQSCAttr($qmgr,"QMGR","CRDATE") || "";
          $rec->{alterDate}=&getMQSCAttr($qmgr,"QMGR","ALTDATE") || "";
      
          #ping the qmgr to assure response (requires mqm group membership)
          #used to grep here, but can't do that on windows; pull all output back and then grep in perl
          my $ping="echo ${mqscquote}ping qmgr$mqscquote | $runmqsc $qmgr";
          my ($rc,$out)=&Kline::Server::cmdLine($ping); #we're ignoring exit codes here
          $rec->{status}="NOT RESPONDING!!" unless grep /AMQ8415/,@$out; #8415 indicates ping complete
        } else {
          &showDebug("UID $UID is not in the mqm group or the qmgr was killed off; cannot get install date from runmqsc",1);
        } #end qmgr CRDATE and ALTDATE
        
      } else { #qmgr down; instead of proc info, see if we can get down info; won't get here if killed off manually
        $rec->{pid}=0;
        &showDebug("qmgr $qmgr down after $rec->{statusDate}",1);
        $rec->{date}=$rec->{statusDate}; #appears to be duplicate to statusDate; not documented      
      }
      
      #load it all
      push @AoH,$rec;
      $rec={};
    } #end foreach qmgr 
  
  ###########################
  } else { #UNIX/Linux
    &showDebug("Loading process table",1);
    my $ps=&Kline::Server::cmdLine("ps -eo pid,etime,comm,args | grep amqzxma0 | grep -v grep"); #aref
    my $dspmq=&Kline::Server::cmdLine("dspmq"); #should not need mqm group to run this #aref
    my $datadir; #qmgr data directory

    #check if pubsub is installed
    my $pubsub;
    my $pubsubFound=0;
    if (-f "/opt/mqm/bin/amqfcxba" || -f "/usr/mqm/bin/amqfcxba") {
      &showDebug("Loading pubsub broker table",1);
      $pubsubFound=1;
      $pubsub=&Kline::Server::cmdLine("ps -eo pid,etime,comm,args | grep amqfcxba | grep -v grep"); #aref
    } else { #it won't be found previous to 5.3.0.8, and never on zLinux
      &showDebug("Pubsub executable not found on this system (amqfcxba)",1);
    }

    #get the default qmgr, if any, and add to data
    $defqmgr=&getDefQmgr(); #does not require mqm group membership

    foreach my $qmgr (sort keys %qmgrs) {
      &showDebug("WORKING ON UNIX QMGR: $qmgr",1);
      my $rec={};
      if ($defqmgr && "$qmgr" eq "$defqmgr") { $rec->{defqmgr}=1 } else { $rec->{defqmgr}=0; }
      $rec->{qmgr}="$qmgr";
      $rec->{running}=0; #assume not running
      ($datadir,undef)=(split / /,$qmgrs{$qmgr});
      $maxqmgrlen=length($qmgr) if (length($qmgr) > $maxqmgrlen); #for stdout print
    
      #check qmgr status according to dspmq
      foreach (@$dspmq) {
        chomp;
       next unless /QMNAME\($qmgr\)/;
        if (/STATUS\(([^)]+)\)/) {
          &showDebug("qmgr $qmgr status is $1",1);
          $rec->{status}=$1;
          $rec->{running}=1 if ($rec->{status} =~ /running/i);
          last; #end the loop, we found it
        } else { #should never get here
          warn "Line $_ has incorrect syntax and was not expected";
        }
      } #end dspmq loop

      #get the install location
      ($rec->{installLocation},$rec->{installLocationLog})=($1,$3) if ($qmgrs{$qmgr}=~/(\S+)(\s+)?(\S+)?/);

      #get the file-stat date from status file (we'll check for accuracy if qmgr down)
      my $qmstatus_file="$datadir/qmstatus.ini";
      $rec->{statusDate}=&fileModDate($qmstatus_file);
      $rec->{dateaccurate}=1; #if down, we'll change assumption to 0 below
      
      #extract etime and pid
      if ($rec->{running}) {
        foreach (@$ps) { #-m QMGR will always show, even if default qmgr
          next unless (/(\d+)\s+(\S+)\s+(\S+)\s+.+-m \b$qmgr\b/); #skip all lines that ain't us
          $rec->{pid}=$1;
          $rec->{etime}=$2;
          &showDebug("qmgr $qmgr has pid $rec->{pid} and uptime $rec->{etime}",1);
          last; #end the loop, we found it
        } #end ps loop

        #for inventory, get install date, install location, alterDate
        #installDate comes from CRDATE in dis qmgr
        #alterDate comes from ALTDATE in dis qmgr
        #installLocation would be in mqs.ini
        #qminiName is the name of the qm.ini file
        #qminiContents is the contents of the qm.ini file
        if (&Kline::WMQ::isInMQMGroup()) {
          $rec->{installDate}=&getMQSCAttr($qmgr,"QMGR","CRDATE") || "";
          $rec->{alterDate}=&getMQSCAttr($qmgr,"QMGR","ALTDATE") || "";
      
          #ping the qmgr to assure response (requires mqm group membership)
          $rec->{status}="NOT RESPONDING!!" if (system("echo ${mqscquote}ping qmgr$mqscquote | $runmqsc $qmgr | grep AMQ8415 >/dev/null") != 0);

        } else {
          &showDebug("UID $UID is not in the mqm group; cannot get install date from runmqsc",1);
        }

      } else { #qmgr not running, get file mod time down
        $rec->{pid}=0;
        $rec->{dateaccurate}=0; #assume the status date is not accurate
        $rec->{dateaccurate}=1 if (-f $qmstatus_file && !(system("grep \"CurrentStatus=Ended\" $qmstatus_file >/dev/null")));
        &showDebug("qmgr $qmgr down after $rec->{statusDate}",1);
        $rec->{date}=$rec->{statusDate}; #appears to be duplicate to statusDate; not documented
      } #end qmgr run check

      if ($rec->{installLocation}) {
        my $qmini=catfile($rec->{installLocation},"qm.ini");
        $rec->{qminiName}=$qmini;
        if (-f $qmini) {
          $rec->{qminiContents}=&getFileContents("$qmini");
        } else {
          &showDebug("Not including contents of $qmini; file not found",1);
        }
      } else { #no install location (should never get here)
        $rec->{installLocation}="N/A";
        $rec->{qminiContents}="N/A";
      }

      #check for a broker
      if ($pubsubFound) { #binaries installed
        $rec->{broker}=0; #assume not running
        foreach (@$pubsub) { #-m QMGR will always show, even if default qmgr
          next unless (/(\d+)\s+(\S+)\s+(\S+)\s+.+-m \b$qmgr\b/); #skip all lines that ain't us
          $rec->{brokerpid}=$1;
          $rec->{brokeretime}=$2;
          $rec->{broker}=1;
          $rec->{brokerstatus}="Running";
          &showDebug("pubsub broker $qmgr has pid $rec->{brokerpid} and uptime $rec->{brokeretime}",1);
          last; #end the loop, we found it
        } #end pubsub running loop
        unless ($rec->{broker}) { #if not running, check if configured
          my @glob=glob "$datadir/queues/SYSTEM\!BROKER\!IQ*";
          if (@glob) { #check for required queues
            &showDebug("PubSub Broker queues found for $qmgr: @glob",2);
            $rec->{brokerstatus}="Down";
          } else {
            &showDebug("No PubSub Broker queues found for $qmgr",2);
            $rec->{brokerstatus}="Never Run";
          }
        } #end check whether down pubsub is configured

      } else { #pubsub not installed
        $rec->{broker}=0;
        $rec->{brokerstatus}="N/A";
      }
    
      #load it all
      push @AoH,$rec;
      $rec={};
    } #end qmgr loop
  } #end Unix vs Win

  return @AoH if (wantarray); #don't print if array desired

  #if we get here, then just print the output

  #don't print the header if NOHEAD is found in env (for mqhealth only)
  $maxqmgrlen+=2; #pad the qmgr space
  my $remaininglen=($Text::Wrap::columns - $maxqmgrlen - 36);
  my $format="%-${maxqmgrlen}s%-${remaininglen}s%-10s%-13s  %-10s\n";
  my $defsymbol="#";
  $remaininglen=($Text::Wrap::columns - $maxqmgrlen - 13);
  &showDebug("Remaining string length is $remaininglen",2);
  my $errformat="%-${maxqmgrlen}s%-${remaininglen}s  %-10s\n";
  print "NOTE: Default (or single) qmgr identified with trailing \"$defsymbol\" on qmgr name\n\n" if $defqmgr;
  printf $format,"","","MAIN","UPTIME","PUBSUB";
  printf $format,"QMGR","STATUS ".&showTime(),"PID","(DD-HH:MM:SS)","BROKER";
  &drawLine();
  foreach my $href (@AoH) {
    $href->{qmgr}.="$defsymbol" if $href->{defqmgr};
    if ($href->{running}) {
      printf $format,"$href->{qmgr}","$href->{status}","$href->{pid}","$href->{etime}","$href->{brokerstatus}";
    } else {
      if ("$href->{date}" eq "N/A") { #not currently mounted
        printf $errformat,"$href->{qmgr}","**$href->{status} (Not Mounted: HA)**","N/A";
      } else {
        if ($href->{dateaccurate}) {
          printf $errformat,"$href->{qmgr}","**$href->{status} at $href->{date}**","$href->{brokerstatus}";
        } else { #down date is not accurate
          printf $errformat,"$href->{qmgr}","**$href->{status} after $href->{date}**","$href->{brokerstatus}";
        }
      }
    }
  }

} #end qmgrStatus

#----------------------------------------------
sub alterGetPut {
#receives a qmgr, qname, and get/put.  It will toggle current
#setting to the opposite to either enable or disable the get/put
#for the given queue.  No direct interaction here, so any menuing
#needs to be built outside.
#----------------------------------------------
  my ($qmgr,$queue,$attr)=@_ ;
  unless ($qmgr && $queue && $attr) {
    &showDebug("WARNING: DID NOT RECEIVE PROPER ARGUMENTS",0);
    return 0;
  }
  
  #obtain the current status
  my $status=&trim(&getMQSCAttr("$qmgr","QL","$queue","$attr")) || return 0;
  &showDebug("Queue $queue is currently $attr $status",1);
  if ($status=~/ENABLED/i) {
    $status="DISABLED";
  } else {
    $status="ENABLED";
  }
  my $cmd="echo ${mqscquote}alt ql($queue) $attr($status)$mqscquote | $runmqsc $qmgr";
  &showDebug("Toggle command: $cmd",2);
  my $output=&Kline::Serve::cmdLine($cmd); #aref
  my $success=0; #assume failed
  foreach (@$output) { #check to make sure it worked
    $success++ if (/AMQ8008/);
  }

  if ($success) { #report action to log file for audit
    &showDebug("$queue toggled to $attr $status successfully",1);
  } else {
    &showDebug("$queue toggled to $attr $status FAILED",1);
    warn "The command $cmd failed:\n@$output\n";
    $status.=" FAILED";
  }
  return $status;

} #end alterGetPut

#----------------------------------------------
sub alterGetPutMenu {
#interactive menu-based function to select and toggle
#GET/PUT ENABLE/DISABLE for given queues
#To be run within MQMenu
#----------------------------------------------
  my $qmgr=&selectQmgr();
  &printTitleC("Local Queue Get/Put Toggle Utility for $qmgr");
  print "Enter a search filter for local queues to alter (RUNMQSC syntax), or \n<ENTER> for full queue list.\n\nNOTICE: Queue depths in the following table are not refreshed.\n\n";
  print "Search Filter: ";
  chomp (my $filter=<STDIN>);

  #setup logging
  #get the login user and effective user
  my $loginUser=getlogin; #this is the login user, not necessarily the effective ID
  #keep in mind that the whoami cmd may not work since sun is /usr/ucb and it
  #may not be in path, but since this is for fallback only, no biggie BT38230
  my $effectiveUser=getpwuid($<) ||`whoami 2>/dev/null`; #this is the current id, such as mqm
  chomp($effectiveUser);
  &showDebug("Effective id is $effectiveUser",1);
  &showDebug("Login user is $loginUser",1);

  my $logDir=&getLogDir(); #from Kline.pl; reads Kline.cfg global stanza
  my $getPutLog=&getPropertyValue("MQMENU","alterGetPutLog");
  $getPutLog="$logDir/$getPutLog" unless ($getPutLog =~ m!^/!);
  if ($getPutLog =~ /none/i) { #turn off logging
    undef $getPutLog;
    &showDebug("Logging is disabled",1);
  }
  if ($getPutLog) {
    &showDebug("Logging GetPut activities to $getPutLog",1);
    unless (-f $getPutLog) {
      #make sure dir exists
      unless (-d dirname($getPutLog)) {
        print STDERR "Get/Put log file directory ".dirname($getPutLog)." does not exist.  Creating...\n";
        mkdir dirname($getPutLog),0777 || die "Cannot create trace file directory for $getPutLog (try again as mqm user if not already): $!\n";
        sleep 2;
      }
      print STDERR "Get/Put log file $getPutLog does not exist.  Creating...\n";
      system ("touch $getPutLog")==0 || die "Cannot create trace file $getPutLog (try again as mqm user if not already): $!\n";
      chmod 0664,"$getPutLog";
    }
    open TRACEFILE, ">>$getPutLog" or die "Cannot open trace file ($getPutLog): $!\n";
    #print a comment line if the file is empty
    print TRACEFILE "# Get/Put Enable/Disable Toggle Trace File.  This file contains a listing of all\n# commands issued via the alterGetPut function.  Format is timestamp, login-id, \n# effective-id (in case of \"su\"), qmgr, queue, and toggle status issued.\n\n\n" if (-z $getPutLog);
    &flushBuffer(\*TRACEFILE);
  }

  #load objects and prepare the menu table
  my $href=&loadQmgrObjects("$qmgr","QL","$filter");
  my $resp; #final queue name to pass; need here for scope
  my $type; #get or put
  TOPQUEUES: { #whenever table needs to reprint, come here
  undef $type;
  undef $resp;
  my $i=0;
  my $remaining=keys %{ $href->{$qmgr}{QLOCAL} }; #need to know how many total in order to know paging
  &showDebug("$remaining queues will be used for screen paging",1);
  my $count=$remaining;
  my $format="%4s  %-48s %-8s %-8s %6s\n";
  my $pagerows=15;
  my $header=1;
  my @queues;
  for my $queue (sort keys %{ $href->{$qmgr}{QLOCAL} } ) {
      push @queues,$queue;
      if ($header) {
        print "\n";
        printf "$format","","QUEUE","GET","PUT","DEPTH";
        &drawLine();
        $header=0;
      }
      $remaining--;
      $i++;
      my $pause=0;
      $pause=1 if (($i % $pagerows)==0);
      &showDebug("Object $i, $remaining objects remaining",2);

      my $get=$href->{$qmgr}{QLOCAL}{$queue}{GET};
      my $put=$href->{$qmgr}{QLOCAL}{$queue}{PUT};
      my $curdepth=$href->{$qmgr}{QLOCAL}{$queue}{CURDEPTH};
      printf "$format","$i","$queue","$get","$put","$curdepth";

      print "\n**END OF QUEUE LIST**\n" unless $remaining;
      #pause if page is full or display complete
      if ($pause || !$remaining) {
        print "\nSelect a queue by number (or by name if desired, case-sensitive).\n";
        print "For faster syntax, enter a queue name or number followed by a space and either \n\"G\" or \"P\" to toggle GET or PUT respectively.  <CTRL-C> to exit.\n";
        print "Or press <ENTER> to display more queues ($remaining)...\n" if ($remaining);
        print "Choice: ";
        chomp($resp=<STDIN>);
        while ($resp=~/\d+(p|put|g|get)/) { #no space, catch it and do over
          print "ERROR: Shorthand must be entered with the queue number, followed by a space, \nand then the G or P operation.\n";
          print "Or press <ENTER> to display more queues...\n" if ($remaining);
          print "Choice: ";
          chomp($resp=<STDIN>);
        }
        if ($resp=~/p|put|g|get\b/i) { #needs trailing anchor, but leading anchor wil cause "1G" to fail
          $type="PUT" if ($resp=~/(\S+)\s*\bp|put\b/i);
          $type="GET" if ($resp=~/(\S+)\s*\bg|get\b/i);
          &showDebug("$type shorthanded to $type",1);
          $resp=$1; #so only queue name is in response now
          &showDebug("Response has been truncated to $resp",1);
        }
        $header=1;
      }

      #check whether anything was entered
      next if (defined($resp) && $resp=~/^\s*$/ && $remaining);
      if (defined($resp) && $resp=~/\d+/) { #numbers entered
        if ($resp <= $count && $count > 0) {
          $resp=$queues[$resp - 1] ;
          last;
        } else { #wrong number
          print "\nERROR: Valid queue number or name not selected\n\n";
          sleep 1;
          goto TOPQUEUES;
        }
      } elsif (defined($resp) && exists($href->{$qmgr}{QLOCAL}{$resp})) { #direct enter
        &showDebug("Direct selection of $resp",1);
        last;
      } elsif (defined($resp)) { #bad entry
        print "\nERROR: Valid queue number or name not selected\n\n";
        sleep 1;
        goto TOPQUEUES;
      }

  } #end queue loop
  } #end TOPQUEUES
  &showDebug("Queue $resp was selected",1);

  #find out whether get/put shorthand, else present menu
  my $typego=0;
  until ($typego) {
    if ($type) {
      $typego++;
    } else { #no shorthand
      &showDebug("Shorthand type selection not found; prompting user for GET or PUT",1);
      print "Do you wish to toggle 1) GET or 2) PUT for $resp?\n";
      print "Choice 1|2): ";
      chomp ($type=uc(<STDIN>));
      $type="GET" if ($type && $type eq 1);
      $type="PUT" if ($type && $type eq 2);
      $typego++ if ($type=~/get|put/i);
      print "\nERROR: Enter 1) GET or 2) PUT\n" unless $typego;
    }
  }

  #do the toggle
  my $status=&alterGetPut("$qmgr","$resp","$type");

  #update the HoHObjects href
  $href->{$qmgr}{QLOCAL}{$resp}{$type}=$status;

  #log the activity
  my $now=&showTime();
  printf TRACEFILE "%-15s | %-8s | %-8s | %-10s | %-45s | %-3s | %-10s\n","$now","$loginUser","$effectiveUser","$qmgr","$resp","$type","$status" if $getPutLog;

  #repaint the table and do it again
  &printArrowSection("Altered $qmgr: $resp $type $status");
  &pressEnter(" or <CTRL-C> to exit");
  goto TOPQUEUES;

} #end alterGetPutMenu

#----------------------------------------------
sub selectQmgr {
#for menu use, user is presented a list of qmgrs with status
#and selects one for operation.  Qmgr is return value
#----------------------------------------------
  my %qmgrs=&getQmgrs();
  my @qmgrlist;
  my $resp;
  my $count=0;
  my $format="  %2d - %-25s%-15s%-15s\n";

  if (scalar keys %qmgrs == 0) { #no qmgrs
    print "\nNo queue managers were found on this system.\n";
    return 0;
  }

  if (scalar keys %qmgrs == 1) { #1 qmgr, so skip it all and just return that
    my $only=(keys %qmgrs)[0];
    print "Found queue manager $only\n";
    return $only;
  }

  &printTitleC("Select A Queue Manager For This Operation");
  print "The following qmgrs are configured on this system.  Select a queue manager for \nthe requested operation (or <ENTER> to select default queue manager, if any):\n\n";
  my $defqmgr=&getDefQmgr();
  foreach my $qmgr (sort keys %qmgrs) {
    push @qmgrlist,$qmgr;
    my $default="";
    $default="*DEFAULT QMGR*" if ($defqmgr && "$defqmgr" eq "$qmgr");
    my $status="";
    $status="*NOT RUNNING*" unless (&isQmgrRunning($qmgr));
    $count++;
    printf "$format","$count","$qmgr","$status","$default";
  }

  #check whether onscreen confirmation is necessary
  my $verify=0;
  $verify=1 if ((&getPropertyValue("GLOBAL","verifyQmgrName"))=~/enabled/i); 

  my $go=0;
  until ($go) {
    print "\nPlease select a qmgr by number, or enter the name directly (case-sensitive)\n";
    print "Selection (or qmgr name): ";
    chomp ($resp=<STDIN>);
    if ($resp=~/\d+/) {#only numbers, convert
      if ($resp <= $count && $count > 0) {
        $resp=$qmgrlist[$resp - 1] ;
        $go++;
      } else { #wrong number
        print "\nERROR: Valid queue manager number or name not selected\n\n";
      }
    } elsif ($defqmgr && ($resp=~/^\s*$/)) { #blanks
      $resp=$defqmgr;
      $go++;
    } else { #name typed directly 
      if (exists($qmgrs{$resp})) {
        $go++;
      } else { #doesn't exist
        print "\nERROR: Valid queue manager number or name not selected\n\n";
      }
    } #end response check

    if ($go && $verify) { #ask are you sure
      print "\nYou selected: <$resp>,  Is this correct?\n";
      print "\n  Choice <y>/n: ";
      $go=0 if (<STDIN>=~/\bn\b|\bno\b/i);
    }
  } #end go loop

  return $resp;

} #end selectQmgr

#----------------------------------------------
sub dumpMQClusterRepos {
#receives name of qmgr and dumps the cluster repos
#----------------------------------------------
  my $qmgr=shift(@_) || &showDebug("NOT CALLED CORRECTLY!!",0);
  my $file=shift(@_) || &showDebug("NOT CALLED CORRECTLY!!",0);
  &showDebug("Dumping $qmgr repos to $file",1);
  unless ($qmgr && $file) {
    &showDebug("NOT CALLED CORRECTLY!!",0);
    return 0;
  }
  &checkPlatform() unless $mqBinaries; #set mqBinaries;
  return 0 unless &isQmgrRunning($qmgr);
  open RFDM,"> $file" or die "Cannot open $file: $!";
  my $tools=&toolsVersion();
  print RFDM "#Cluster repository dump for $qmgr at ". localtime() ." created by Kline v$tools.\n\n";
  close RFDM;
  return 1 if (system("$mqBinaries/bin/amqrfdm -m $qmgr < $Kline::KlineHome/clusrepos.resp >>$file")==0);
  &showDebug("amqrfdm failed to complete properly: $!",0); #if we get here, prob
  return 0;

} #end dumpMQClusterRepos

#----------------------------------------------
sub loadQueues {
#uses loadQmgrObjects to load the HoH with all queues, as an object that can be used for various purposes
#receives only a qmgr name
#1/15/06, NOT CURRENTLY USED ANYWHERE, ALL JUST CALL loadQmgrObjects directly
#----------------------------------------------
  my $qmgr=shift @_;
  my $href; #hashref of object list
  #sanity check
  my $thisFunction=&thisFunc();
  unless ($qmgr) {
    warn "WARNING: $thisFunction did not receive proper parameters: qmgr\n";
    return 0;
  }

  foreach my $type (qw/QL QR QA/) {
    &showDebug("Calling object load for qmgr $qmgr object type $type",0);
    $href=&loadQmgrObjects($qmgr,$type); #remembers its previous vals always so don't need to pass href in
  }

  &dumpObjectHash($href,3);
  return $href;

} #end loadQueues

#----------------------------------------------
sub dumpObjectHash {
#internal test function that will just print out the contents of the hash
#for verification before you try and use its values.  Not meant to use in
#a user function.
#Optionally receives debug value at which point data should be printed
#----------------------------------------------

  my $href=shift(@_) || &showDebug("CALLED INCORRECTLY!",0);
  my $this=&thisFunc();
  my $parent=&parentFunc();
  if (ref($href) ne "HASH") { #make sure the passed val is a href
    die "$this expected a hash reference from $parent, not $href";
  }

  my $level=shift(@_) || 0; #show dump only if debug is this or higher
  if ($Kline::Display::debug ge $level) { #proceed
    &showDebug("CALLED AT DEBUG LEVEL $level",$level);

    #this will be unsorted output, but is way faster than recursing
    use Data::Dumper;
    print Dumper(\$href);

    #this just tallies how many queues and shows them in an easier format
    for my $qmgr (sort keys %$href) {
      my $i;
      for my $type (sort keys %{ $href->{$qmgr} } ) {
        &printArrowSection("$qmgr type: $type");
        for my $obj (sort keys %{ $href->{$qmgr}{$type} } ) {
          $i++;
          print "  $type ($i): $obj\n";
        }
        undef $i;
        print "\n";
      }
    }
  } else { #debug not high enough
    &showDebug("DEBUG LEVEL NOT HIGH ENOUGH ($Kline::Display::debug) TO ENABLE PRINTOUT ($level)!",$Kline::Display::debug) if $Kline::Display::debug;
  }

} #end dumpObjectHash

#----------------------------------------------
{ #naked block to scope objects as static across calls
my $href; #persistent hashref for objects, shared for loadQmgrObjects and getMQSCAttr

sub loadQmgrObjects {

#Load a hashref with all the objects $->{qmgr}{type}{name}{attrs}{vals}
#of a particular type, passed as second arg. Type abbreviations are
#acceptable and case-insensitive. If no arg is passed, queue is assumed.
#ARGS: QMGR, OBJTYPE, NAME_FILTER
#First arg should be qmgr name.
#Second arg should be object type, but is optional and assume QL if null.
#Third arg is specific object name or search pattern (perl regexpr, not runmqsc), assume all if null.
#This code structure was taken directly from lsmqdp (but had to be modified).
#
#Returns hashref of all loaded objects, which is scoped outside this function in order
#to be static.  So each time it is called, $href(ret val) should be bigger
#
#This function attempts to do something already done quite well by public
#tools such as the MQSeries PERL modules.  However, since they require
#installation on each server and are not owned by IBM, a localized attempt
#needed to be made.
#
#All debug statements should be 2 or higher so they don't show (tons of
#output) unless desired when running through other functions
#
#Use the AMQ codes hash below to identify record delimiters
#
#Updated 12/10/07 for Gnosis so that query of Kline.cfg is not necessary (all info was here already)
#Updated 01/11/08 for Gnosis so that name filter was regexpr instead of runmqsc direct; also updated for WMQv6 object handling
#----------------------------------------------
#Sample runmqsc output
#dis ql(*) all
#AMQ8409: Display Queue details.
#   DESCR(WebSphere MQ Default Local Queue)
#   PROCESS( )                              BOQNAME( )
#   INITQ( )                                TRIGDATA( )
#   CLUSTER( )                              CLUSNL( )
#   QUEUE(A.REALLY.REALLY.REALLY.LONG.QUEUE.NAME.REALLY)
#   CRDATE(2004-08-18)                      CRTIME(20.48.07)
#   ALTDATE(2004-08-18)                     ALTTIME(20.48.07)
#   GET(ENABLED)                            PUT(ENABLED)
#   DEFPRTY(0)                              DEFPSIST(NO)
#   MAXDEPTH(5000)                          MAXMSGL(4194304)
#   BOTHRESH(0)                             SHARE
#   DEFSOPT(SHARED)                         HARDENBO
#   MSGDLVSQ(PRIORITY)                      RETINTVL(999999999)
#   USAGE(NORMAL)                           NOTRIGGER
#   TRIGTYPE(FIRST)                         TRIGDPTH(1)
#   TRIGMPRI(0)                             QDEPTHHI(80)
#   QDEPTHLO(20)                            QDPMAXEV(ENABLED)
#   QDPHIEV(DISABLED)                       QDPLOEV(DISABLED)
#   QSVCINT(999999999)                      QSVCIEV(NONE)
#   DISTL(NO)                               NPMCLASS(NORMAL)
#   DEFTYPE(PREDEFINED)                     TYPE(QLOCAL)
#   SCOPE(QMGR)                             DEFBIND(OPEN)
#   IPPROCS(0)                              OPPROCS(0)
#   CURDEPTH(0)
#AMQ8409: Display Queue details.  #(NEXT QUEUE FOLLOWS)

  my $objinfo= { #store each available object type and the AMQ code that separates records; used for looping on obj types and replaces deprecated MQObjectTypes from Kline.cfg
  #updated 01/2008 for MQ6 object types
    "AUTHINFO" => { "AMQ" => 8566,
                    "nameattr"=>"AUTHINFO",
                  },
    "CHANNEL"  => { "AMQ" =>8414, #name attr CHANNEL
                    "nameattr"=>"CHANNEL",
                  },
    "CHSTATUS" => { "AMQ" =>8417, #name attr CHANNEL
                    "nameattr"=>"CHANNEL",
                  },
    "CLUSQMGR" => { "AMQ" =>8441, #name attr QMNAME
                    "nameattr"=>"QMNAME",
                  },
    "PROCESS"  => { "AMQ" =>8407, #name attr PROCESS
                    "nameattr"=>"PROCESS",
                  },
    "NAMELIST" => { "AMQ" =>8550, #name attr NAMELIST
                    "nameattr"=>"NAMELIST",
                  },
    "QALIAS"   => { "AMQ" =>8409, #name attr QUEUE
                    "nameattr"=>"QUEUE",
                  },
    "QCLUSTER" => { "AMQ" =>8409, #name attr QUEUE
                    nameattr=>"QUEUE",
                  },
    "QLOCAL"   => { "AMQ" =>8409, #name attr QUEUE
                    "nameattr"=>"QUEUE",
                  },
    "QMGR"     => { "AMQ" =>8408, #name attr QMNAME (no (*))
                    "nameattr"=>"QMNAME",
                  },
    "QMODEL"   => { "AMQ" =>8409, #name attr QUEUE
                    "nameattr"=>"QUEUE",
                  },
    "QREMOTE"  => { "AMQ" =>8409, #name attr QUEUE
                    "nameattr"=>"QUEUE",
                  },
    "QUEUE"    => { "AMQ" =>8409, #name attr QUEUE
                    "nameattr"=>"QUEUE",
                  },
    "QSTATUS"  => { "AMQ" =>8450, #name attr QUEUE
                    "nameattr"=>"QUEUE",
                  },
    "LISTENER" => { "AMQ" =>8630, #name attr LISTENER
                    "nameattr"=>"LISTENER",
                  },
    "CONN"     => { "AMQ" =>8276, #name attr CONN (hex val)
                    "nameattr"=>"CONN",
                  },
    "LSSTATUS" => { "AMQ" =>8631, #name attr LISTENER
                    "nameattr"=>"LISTENER",
                  },
    "QMSTATUS" => { "AMQ" =>8705, #name attr QMNAME (no (*))
                    "nameattr"=>"QMNAME",
                  },
    "SERVICE"  => { "AMQ" =>8629, #name attr SERVICE
                    "nameattr"=>"SERVICE",
                  },
    "SVSTATUS" => { "AMQ" =>8632, #name attr SERVICE
                    "nameattr"=>"SERVICE",
                  },
  }; #end $objinfo
  
  my $ldebug=2; #easily change all debug statements in here, when needed
  my $objname; #this will be the object name that is used for hash key
  
  my $qmgr=shift || die "no qmgr provided"; #$1, required
  my $type=&convertObjType(shift(@_)) || "QUEUE";  #$2, optional; will also uc; *default object type* if none passed
  my $filtname=shift || ""; #optional $3, search filter pattern, applied after (*) results are gathered from runmqsc
  &showDebug("Searching Qmgr $qmgr",$ldebug);
  &showDebug("Searching for $type objects",$ldebug);
  &showDebug("Limiting search to names with $filtname",1) if $filtname;

  #sanity check
  my $this=&thisFunc();
  my $parent=&parentFunc();
  die "\nNo authority for $this ($parent); UID $UID not in mqm group\n" unless &isInMQMGroup();
  unless (&isQmgrRunning($qmgr)) {
    warn "WARNING: $this: $qmgr is not running.  Operation Aborted.\n";
    return undef;
  }
  unless (exists($objinfo->{$type})) {
    warn "WARNING: $this: $type does not have a valid AMQ delimiter code.\n";
    return undef;
  }

  &showDebug("AMQ Delimiter is set to $objinfo->{$type}{AMQ}",$ldebug);
  
  #get a list of attrs we're looking for on this type
  my @attrs=&getAttrList($qmgr,$type);
  
  #needs backslash for shell escape when command is issued
  my $cmd = "echo ${mqscquote}dis $type";
  $cmd.="(\*)" unless ($type=~/\bqmgr\b/i); #qmgr object can't have wildcard, don't confuse with clusqmgr; filter applied on results, not RUNMQSC
  $cmd .= " all$mqscquote | $runmqsc -e $qmgr"; #-e eliminates cmd itself from showing in output
  &showDebug("command is $cmd",$ldebug);
  open RUNMQSC, "$cmd|" or die "Cannot execute runmqsc: $!\n";
  my $rec = {}; #anon hash

  #see file:///D:/perl_oreilly/cookbook/ch06_08.htm for discussion on what is
  #happening here.  The AMQ code is being used as a record delim.
  #It is NOT being included in the results.  Results include garbage
  #in the first and last element, so they are discarded.  Now each array
  #element is a record; don't need to search for all attrs found
  my @chunks; #records from runmqsc
  { #naked block to scope messing with record separator
    local $/=undef; #get rid of record separator
    @chunks= split (/^AMQ$objinfo->{$type}{AMQ}: .+|One MQSC command/m, <RUNMQSC>);
  }
  
  #though rare, we need to make sure that we really have auth on the command before proceeding (or else you get empty stuff in results, causing other subs to get really mad)
  my $secErr=$1 if $chunks[0]=~/(AMQ\d+: Not authorized)/i;
  die "FATAL: Even though UID $UID is technically in the mqm group, runmqsc returned \"$1\".\nThis is likely due to security not having been refreshed on queue manager $qmgr.\n" if $secErr;
  
  #drop first and last
  pop @chunks;
  shift @chunks;
  my $totalChunks=scalar(@chunks);
  @chunks=&trim(@chunks); #drop whitespaces
  &showDebug("found $totalChunks $type records in runmqsc",$ldebug); #each one will print below in debug 3

  #now loop through each record and read the attrs and values, then load
  my $i;
  foreach (@chunks) {
    $i++;
    &showDebug("New record ($i) encountered****************************\n$_",$ldebug+1);
    # if not AMQ8409/8450 code, must be an error
    die "No matching objects were found\n" if ((/^(AMQ.*)\s+/) && ((index $1, "AMQ8147") == 0));
    die "$1\n" if ((/^(AMQ.*)\s+/) && ((index $1, "AMQ$objinfo->{$type}{AMQ}") < 0));

    $rec->{QMGR}=$qmgr;
    foreach my $attr (@attrs) {
      &showDebug("Checking for $attr in record $i",3);
      my $value = $1 if /\b$attr\b\(([^)]+)\)/; #this line pulls the actual val
      #match anything from an open paren until a closing paren

      #select attributes don't have () on them and need to be handled separately
      #this program will use the postive values for the attr name and store 1 or 0
      #in the property, to thus make it true/false.  ie: TRIGGER will be TRIGGER(1) and true
      #QMGR: SYNCPT WE ARE NOT HANDLING THIS SINCE IT WILL NOT LIKELY EVER BE USED
      #QL/QM: TRIGGER/NOTRIGGER, SHARE/NOSHARE, HARDENBO/NOHARDENBO (verified complete list in MQ6 too, all objects)
      if ($attr =~ /\btrigger\b|\bshare\b|\bhardenbo\b/i && $type=~/ql|qmo|queue/i) {
        &showDebug("Checking for special attr $attr on $type record $i",$ldebug);
        $value = 1 if /\b$attr\b/; #this line pulls the actual val
        $value = 0 if /\bNO$attr\b/; #this line pulls the actual val
        &showDebug("Marking special attr $attr:\"$value\" ($i)",$ldebug) if defined($value);
      }
      #CHS: SAVED/CURRENT ATTR NAME IS STORED AS "CURRENT | SAVED"
      if ($attr =~ /\bcurrent\b/i && $type=~/chs/i) {
        &showDebug("Checking for special attr $attr on $type record $i",$ldebug);
        $value = 1 if /\bcurrent\b/i; #this line pulls the actual val
        $value = 0 if /\bsaved\b/i; #this line pulls the actual val
        &showDebug("Marking special attr $attr:\"$value\" ($i)",$ldebug);
      }

      if (defined $value) {
        &showDebug("loading $attr:\"$value\" ($i)",3);
        $rec->{$attr} = $value;
      } else { #not found
        &showDebug("WARNING: $attr not found in $type record $i",3);
      } #end if defined val
    } #end foreach attr

    $objname=$rec->{$objinfo->{$type}{nameattr}} || warn "\nERROR: Cannot find proper object name for type $type.  Check hard-coded types in $this sub";
    if ($type=~/clusqmgr/i) {
      &showDebug("Object type is clusqmgr, which may have duplicates, so using anonymous/generic name",1);
      $objname={};
    }

    #load what we've found
    if ($filtname && $objname!~/$filtname/i) { #filter active and this doesn't match; skip it
      #from perl cookbook 6.18, the pattern that would be passed as var to negate a string: ^(?:(?!string).)*$
      &showDebug("NOTICE: Object $type($objname) does not match filter $filtname; object will not be loaded",0);
    } else { #no filter or filter doesn't match
      &showDebug("Loading completed $type($objname) object $i ($rec)",$ldebug);
      $href->{$qmgr}{$type}{$objname}=$rec;
    }
    $rec={};  #load a new anon hash since old one is done
    undef $objname;

  } #end while
  close RUNMQSC;

  return $href;

} #end loadQmgrObjects

#----------------------------------------------
sub getMQSCAttr {
#Query the Qmgr Object hash for a specific object
#Uses memory-resident HoH and only queries runmqsc if object type not loaded
#Receives QMGR, OBJTYPE, NAME, ATTR
#For QMGR object, no NAME is required (only 3 args then)
#----------------------------------------------
  my @out;
  my $qmgr=shift;
  my $type=shift;
  my ($property,$name);

  #sanity check
  my $thisFunction=&thisFunc();
  my $parent=&parentFunc();
  die "\nNo authority for $thisFunction ($parent); UID $UID not in mqm group\n" unless &isInMQMGroup();
  if ($qmgr && $type) { #regardless of the type, need these
    if ($type=~/QMGR/i) {
      $property=shift;
      unless ($property) {
        warn "WARNING: $thisFunction did not receive proper parameters: QMGR, TYPE, PROPERTY\n";
        return 0;
      }
    } else { #not QMGR object, get the name and property
      $name=shift;
      $property=shift;
      unless ($name && $property) {
        warn "WARNING: $thisFunction did not receive proper parameters: QMGR, TYPE, OBJ_NAME, PROPERTY\n";
        return 0;
      }
    }
  } else {
    warn "WARNING: $thisFunction did not receive proper parameters: QMGR, TYPE, PROPERTY\n";
    return 0;
  }

  $type=&convertObjType($type); #converts to long name and UC
  $property=uc($property);
  $name=uc($name);

  #if objtype not yet in hash, load it now (so only when needed)
  unless (exists($href->{$qmgr}{$type})) {
    &showDebug("Qmgr $qmgr object type $type not yet loaded.  Loading . . .",1);
    $href=&loadQmgrObjects($qmgr,$type);
  }

  if ($type=~/QMGR/) { #no name needed, so alter checks
    #error checking
    unless (exists ($href->{$qmgr}{$type}{$qmgr}{$property})) { #QMGR object stores the qmgr name as the object name
      warn "\nERROR ($thisFunction): $qmgr $type object exists, but property $property was not found!!! Loaded object types follow:\n";
      &dumpObjectHash($href);
      return 0;
    }
    @out=$href->{$qmgr}{$type}{$qmgr}{$property};

  } else { #normal object, do validations
    #error checking
    unless (exists ($href->{$qmgr}{$type}{$name})) {
      warn "\nERROR ($thisFunction): $qmgr $type object $name does not exist!!! Loaded object types follow:\n";
      &dumpObjectHash($href);
      return 0;
    }

    unless (exists ($href->{$qmgr}{$type}{$name}{$property})) {
      warn "\nERROR ($thisFunction): $qmgr $type object $name exists, but property $property was not found!!! \nLoaded attributes follow:\n";
      &dumpObjectHash($href);
      return 0;
    }

    #if we get here the object exists
    @out=$href->{$qmgr}{$type}{$name}{$property};
  } #end QMGR object type check

  return wantarray ? @out : $out[0];

} #end getMQSCAttr
} #end naked block for HoHObjects persistence

#----------------------------------------------
sub triggerCheck {
#check status of all triggered queues for a given qmgr
#stdout will be all triggered queues with their status. Also checks
#initqs for ipprocs to make sure trigmons running.
#currently setup for mqhealth use, which means err output goes to files to be
#picked up by ksh (can't pass structures directly); in future change output files
#to variables for use elsewhere in perl
#----------------------------------------------
  my $qmgr=shift(@_);
  my $errfile=shift(@_) || "/tmp/triggerCheck.err";
  my $summaryfile=shift(@_) || "/tmp/triggerCheck.out";
  &showDebug("Error output will be copied to $errfile",1);
  &showDebug("Summary output will be copied to $summaryfile",1);
  &showDebug("Loading QL objects for $qmgr",1);
  my $href=&loadQmgrObjects("$qmgr","QL"); #will persist over each call (so href gets bigger)
  &dumpObjectHash($href,2);

  my @explain;
  my @AoH; #need to store output while finding lengths so we can print pretty
  my $rec = {};
  my %max=(queue => 0, usage => 0, trigger => 0, initq => 0, process => 0, trigtype => 0, trigdepth => 0, trigmpri => 0, curdepth => 0, status => 0); #for column headings
  my ($totalTrig,$totalQ,$totalProbs)=(0,0,0);
  for my $queue (sort keys %{ $href->{$qmgr}{QLOCAL} } ) {
    $rec->{status}="OK";  #assume it's ok
    $totalQ++;
    my $localdebug=3; #easy way to change whether all the following debug statements print
    &showDebug("Checking trigger status for queue $queue",1);
    $rec->{queue}=$queue;
    $rec->{usage}=&getMQSCAttr("$qmgr","QL","$queue","USAGE");
    &showDebug("usage status: -$rec->{usage}-",$localdebug);
    $rec->{trigger}=&getMQSCAttr("$qmgr","QL","$queue","TRIGGER");
    &showDebug("trigger status: -$rec->{trigger}-",$localdebug);
    $rec->{initq}=&trim(&getMQSCAttr("$qmgr","QL","$queue","INITQ"));
    &showDebug("initq status: -$rec->{initq}-",$localdebug);
    $rec->{process}=&trim(&getMQSCAttr("$qmgr","QL","$queue","PROCESS"));
    &showDebug("process status: -$rec->{process}-",$localdebug);
    $rec->{trigtype}=&trim(&getMQSCAttr("$qmgr","QL","$queue","TRIGTYPE"));
    &showDebug("trigtype status: -$rec->{trigtype}-",$localdebug);
    $rec->{curdepth}=&trim(&getMQSCAttr("$qmgr","QL","$queue","CURDEPTH"));
    &showDebug("curdepth status: -$rec->{curdepth}-",$localdebug);

    #don't look at ximtq's by definition (even though they are triggered, they don't have to be config'd)
    if (($rec->{usage}=~/XMITQ/) || (!$rec->{trigger} && !$rec->{initq} && !$rec->{process})) { #if any of these have vals, then consider it a trig queue
      &showDebug("$queue has no trigger characteristics.  Skipping...",1);
      $rec={}; #clear all stored fields
      next;
    }
    &showDebug("     $queue HAS TRIGGER CHARACTERISTICS; Processing...",1);
    $totalTrig++;
    if (!$rec->{trigger} || !$rec->{initq} || !$rec->{process}) { #missing a setting
      &showDebug("$queue is missing a required trigger parameter",1);
      push @explain,"-Queue $queue is missing a required parameter or \"trigger\" is disabled (NOTRIGGER).\n";
      $rec->{status}="WARNING!!"; #set above
      $totalProbs++;
    } elsif ($rec->{trigtype}=~/depth/i) { #triggering on depth, find out more
      $rec->{trigdepth}=&trim(&getMQSCAttr("$qmgr","QL","$queue","TRIGDPTH"));
      &showDebug("trigdepth status: -$rec->{trigdepth}-",$localdebug);
      $max{trigdepth}=length($rec->{trigdepth}) if (length($rec->{trigdepth}) > $max{trigdepth});
      $rec->{trigmpri}=&trim(&getMQSCAttr("$qmgr","QL","$queue","TRIGMPRI"));
      &showDebug("trigmpri status: -$rec->{trigmpri}-",$localdebug);
      $max{trigmpri}=length($rec->{trigmpri}) if (length($rec->{trigmpri}) > $max{trigmpri});
      if ($rec->{curdepth} > $rec->{trigdepth}) {
        &showDebug("$queue is triggering on depth $rec->{trigdepth} but has not properly triggered yet (curdepth $rec->{curdepth}, trigmpri $rec->{trigmpri})",1);
        push @explain,"-Queue $queue is triggering on depth $rec->{trigdepth} but has not properly triggered yet (even though curdepth is $rec->{curdepth}; but could be related to trigmpri $rec->{trigmpri}).\n";
        $rec->{status}="WARNING!!";
        $totalProbs++;
      }
    } else { #not triggering on depth, so if there's a depth, something is wrong
      if ($rec->{curdepth}) {
        &showDebug("Triggering is enabled, but not on depth (TRIGTYPE: $rec->{trigtype}), and current depth ($rec->{curdepth}) is greater than zero",1);
        push @explain,"-Queue $queue triggering is enabled, but not on depth (TRIGTYPE: $rec->{trigtype}), and current depth ($rec->{curdepth}) is greater than zero.\n";
        $rec->{status}="WARNING!!";
        $totalProbs++;
      } else { #all must be well
        #$rec->{status}="OK"; #set above
      }
    } #end check for trigger probs

    #check the initq for ipprocs to indicate a trigmon is attached
    my $initq=$rec->{initq};
    &showDebug("Checking INITQ status for $queue ($initq)",1);
    if (exists($href->{$qmgr}{QLOCAL}{$initq})) { #make sure initq exists
      my $initprocs=&trim(&getMQSCAttr("$qmgr","QL","$rec->{initq}","IPPROCS"));
      &showDebug("queue $queue initq $rec->{initq} has $initprocs IPPROCS",2);
      unless ($initprocs) {
        &showDebug("Queue $queue initq $rec->{initq} has no IPPROCS",1);
        $rec->{status}="WARNING!!";
        $totalProbs++;
        push @explain,"-Queue $queue has initq \"$rec->{initq}\" specified, but the initq currently has no IPPROCS, indicating that a trigger monitor is not currently active.  Triggering will not work unless a trigger monitor (local or client) is running.\n";
      }
    } else { #initq doesn't exist
      &showDebug("Queue $queue initq $rec->{initq} does not exist.",1);
      $rec->{status}="WARNING!!";
      $totalProbs++;
      push @explain,"-Queue $queue has initq \"$rec->{initq}\" specified, but the initiation queue does not exist!\n";
    } #end check for initq setup OK

    $max{process}=length($rec->{process}) if (length($rec->{process}) > $max{process});
    $max{curdepth}=length($rec->{curdepth}) if (length($rec->{curdepth}) > $max{curdepth});
    $max{trigtype}=length($rec->{trigtype}) if (length($rec->{trigtype}) > $max{trigtype});
    $max{initq}=length($rec->{initq}) if (length($rec->{initq}) > $max{initq});
    $max{trigger}=length($rec->{trigger}) if (length($rec->{trigger}) > $max{trigger});
    $max{queue}=length($rec->{queue}) if (length($rec->{queue}) > $max{queue});
    $max{usage}=length($rec->{usage}) if (length($rec->{usage}) > $max{usage});
    push @AoH,$rec;
    $rec={};

  } #end loop through all local queues
  #ALL OUTPUT NOW LOADED

  #print all output
  my @header=("QUEUE","INITQ","PROCESS","TRIGTYPE","TRIGDEPTH","TRIGGER","CURDEPTH","STATUS");
  foreach ( @header) { #if max length isn't long enough for header, increase
    &showDebug("HEADER $_ max length started as $max{&lowerCase($_)}",2);
    $max{&lowerCase($_)}=length($_) if ($max{&lowerCase($_)} < length($_));
    &showDebug("HEADER $_ max length is now $max{&lowerCase($_)}",2);
  }
  my $format="%-$max{queue}s  %-$max{initq}s  %-$max{process}s  %-$max{trigtype}s  %-$max{trigdepth}s  %-$max{trigger}s  %$max{curdepth}s  %-$max{status}s\n";
  my @headerout;
  push @headerout,sprintf "$format",@header;
  push @headerout,&drawLine();
  print @headerout;
  my @errout; #error output buffer
  foreach my $href (@AoH) {
    my $init=$href->{initq} || "-";
    my $process=$href->{process} || "-";
    my $trigdepth=$href->{trigdepth} || "-";
    my $trigger=($href->{trigger}) ? "ON" : "OFF";
    my $out=sprintf "$format","$href->{queue}","$init","$process","$href->{trigtype}","$trigdepth","$trigger","$href->{curdepth}","$href->{status}";
    print $out;
    push @errout,$out unless ($href->{status}=~/OK/); #copy err lines to file for mqhealth
  }

  if (@errout) {
    #print all we know about probs to a file to be picked up by mqhealth
    open ERROUT, ">$errfile" || warn "Cannot open $errfile: $!";
    select ERROUT;
    print "Qmgr $qmgr has $totalQ local queues, $totalTrig of which are triggered, and within which $totalProbs problems were identified.  Triggered queues with problems follow:\n";
    print "\n";
    print @headerout;
    print @errout;
    print "\n";
    print "Detailed Explanation:\n";
    print @explain;
    select STDOUT;
    close ERROUT;
  }
  #for mqhealth, write file with total trig queues and total errors
  system("echo \"$totalQ $totalTrig $totalProbs\" > $summaryfile");

} #end triggerCheck

#----------------------------------------------
sub loadChanSecurity {
#Check security settings of all channels (default SVRCONN only) for a given qmgr.
#loads necessary channel security info to be used externally or by
#checkChanSecurity, which dumps it all to files for ksh usage (mqhealth)
#receives $1 as qmgr, $2 as mcx switches, and $3 (optional) as filters
#Filters, if used, can be for name or chltype, and are regexpr, else use all SVRCONNs
#Returns aref of AoH (undef if errors) and href of maxlengte
#If switches are null, then it is assumed to check SYSTEM.DEF.SVRCONN for 
#mcauser of nobody.  This is referred to as the default or $d check. Use 
#filter->{name} to specify additional name expressions for this check.
#To override the desired username for for the $d check, pass filter->{nobody}
#with desired name.
#----------------------------------------------

#Channel parms are checked according to the switches as follows.

#No Switches ___ [000] (-d)
#Assume default check of SYSTEM.DEF.SVRCONN for a restricted MCAUSER(nobody) to completely lock out the channel(s)

#Switches cmx [111]:
#  SSLCIPH(' ') & MCAUSER('mqm') & SCYEXIT(' '): runs under mqm, WARNING.
#  SSLCIPH(' ') & MCAUSER('mqm') & SCYEXIT(not ' '): still runs under mqm
#                 but exit should not allow client to connect, OK with NOTE.
#  SSLCIPH(' ') & MCAUSER(' ') & SCYEXIT(' '): non-java client, OK;
#                 java client runs under listener id (mqm), WARNING.
#  SSLCIPH(' ') & MCAUSER(' ') & SCYEXIT(not ' '): non-java client, OK;
#                 java client exit would reject MCAUSER(' '), OK with NOTE.
#  SSLCIPH(' ') & MCAUSER('nobody') & SCYEXIT(' '): OK.
#  SSLCIPH(' ') & MCAUSER('nobody') & SCYEXIT(not ' '): OK.
#  SSLCIPH(' ') & MCAUSER(not 'mqm' nor ' ' nor 'nobody) & SCYEXIT(' '): OK.
#  SSLCIPH(' ') & MCAUSER(not 'mqm' nor ' ' nor 'nobody) & SCYEXIT(not ' '): OK.
#  SSLCIPH(not ' ') & MCAUSER(any value) & SCYEXIT(any value): OK.

#Switches cm_ [110]:
#  SSLCIPH(' ') & MCAUSER('mqm'): runs under mqm, WARNING.
#  SSLCIPH(' ') & MCAUSER(' '): non-java client, OK;
#                 java client runs under listener id (mqm), WARNING.
#  SSLCIPH(' ') & MCAUSER('nobody'): OK.
#  SSLCIPH(' ') & MCAUSER(not 'mqm' nor ' ' nor 'nobody'): OK.
#  SSLCIPH(not ' ') & MCAUSER(any value): OK.

#Switches c_x [101]:
#  SSLCIPH(' ') & SCYEXIT(' '): WARNING.
#  SSLCIPH(' ') & SCYEXIT(not ' '): OK with NOTE.
#  SSLCIPH(not ' ') & SCYEXIT(any value): OK.

#Switches c__ [100]:
#  SSLCIPH(' '): WARNING.
#  SSLCIPH(not ' '): OK.

#Switches _mx [011]:
#  MCAUSER('mqm') & SCYEXIT(' '): WARNING.
#  MCAUSER('mqm') & SCYEXIT(not ' '): OK with NOTE.
#  MCAUSER(' ') & SCYEXIT(' '): non-java client, OK;
#                 java client runs under listener id (mqm), WARNING.
#  MCAUSER(' ') & SCYEXIT(not ' '): non-java client, OK;
#                 java client exit would reject MCAUSER(' '), OK with NOTE.
#  MCAUSER('nobody') & SCYEXIT(' '): OK.
#  MCAUSER('nobody') & SCYEXIT(not ' '): OK.
#  MCAUSER(not 'mqm' nor ' ' nor 'nobody') & SCYEXIT(' '): OK.
#  MCAUSER(not 'mqm' nor ' ' nor 'nobody') & SCYEXIT(not ' '): OK.

#Switches _m_ [010]:
#  MCAUSER('mqm'): runs under mqm, WARNING.
#  MCAUSER(' '):non-java client, OK;
#                 java client runs under listener id (mqm), WARNING.
#  MCAUSER('nobody'): OK.
#  MCAUSER(not 'mqm' nor ' ' nor 'nobody'): OK.

#Switches __x [001]:
#  SCYEXIT(' '): WARNING.
#  SCYEXIT(not ' '): NOTE.
#-------------

  my $qmgr=shift(@_);
  my $switches=lc(shift(@_)); #if no switches, use default check (below)
  my $filter=shift if @_;  #href to hold many types of filters like chltype, name, etc
  my $baduser="mqm"; #if mcauser is ever mqm, that's bad (handle blank manually)

  #sanity checks
  my $thisFunction=&thisFunc();
  unless ($qmgr) {
    warn "WARNING: $thisFunction did not receive proper parameter: qmgr, operation aborted.\n";
    return undef;
  }
  unless (&isQmgrRunning($qmgr)) {
    warn "WARNING: $thisFunction: $qmgr is not running (should not have been called!), operation aborted.\n";
    return undef;
  }

  #set the filters
  #first check for CHLTYPE()
  if (exists $filter->{chltype}) { #user wants a specific type
    &showDebug("User filtering CHLTYPE to $filter->{chltype}",1);
  } else { #use the default type, SVRCONN
    $filter->{chltype}="SVRCONN";
    &showDebug("Using default CHLTYPE filter $filter->{chltype}",1);
  }
  #now check for chl name
  if (exists $filter->{name}) { #user wants a specific name regexpr
    &showDebug("User filtering CHANNEL name to $filter->{name}",1);
  } else { #use default name filter, which is all
    $filter->{name}="";
    &showDebug("Using default CHANNEL filter $filter->{name}",1);
  }

  #Test switch characters and set switch variables:
  my ($c, $m, $x, $d)=(0,0,0,0);
  for (my $i=0; length($_=substr($switches,$i,1)); $i++) {
    if (/c/i) { $c++; &showDebug("Running with \"c\" switch",1); next; }
    if (/m/i) { $m++; &showDebug("Running with \"m\" switch",1); next; }
    if (/x/i) { $x++; &showDebug("Running with \"x\" switch",1); next; }
    if (/d/i) { $d++; &showDebug("Running with \"d\" switch",1); next; }
    warn "WARNING: $thisFunction received invalid switches: \"$switches\", operation aborted.\n";
    return undef;
  }
  unless ($c || $x || $m) { #setup the default check
    $filter->{nobody}="nobody" unless exists $filter->{nobody};
    $filter->{name}="SYSTEM.DEF.SVRCONN" unless exists $filter->{name};
    $d++;
    &showDebug("Running default check of MCAUSER($filter->{nobody})",1);
  }

  &showDebug("Loading all CHL objects for $qmgr",1);
  my $href=&loadQmgrObjects("$qmgr","CHL"); #will persist over each call (so href gets bigger)
  &dumpObjectHash($href,2);

  #now we know what we're going to do, start checking channels and loading results into AoH
  my @AoH;              #array of hashes to store output while finding lengths so we can print pretty
  my %max=(channel=>0, chltype=>0, sslciph=>0, mcauser=>0, scyexit=>0, status=>0); #for column headings to have pretty lengths

  for my $channel (sort keys %{ $href->{$qmgr}{CHANNEL} } ) { #loop through each channel we found
    my $rec = {}; #hash work record for channel parameters

    my $localdebug=3; #easy way to change whether all the following debug statements print
    &showDebug("Checking security for channel $channel",1);

    #set the basics in the rec
    $rec->{channel}=$channel;
    $rec->{qmgr}=$qmgr;
    $rec->{chltype}=$href->{$qmgr}{CHANNEL}{$channel}{CHLTYPE};
    $rec->{status}=1;  #tentatively assume it's OK unless we find otherwise

    #filter on name
    if ($filter->{name} && $channel!~/$filter->{name}/) {  #yes filter but no match
      &showDebug("User is filtering channel names on $filter->{name} and $channel DOES NOT MATCH, skipping",1);
      next;
    }

    #filter on type
    &showDebug("CHLTYPE: -$rec->{chltype}-",$localdebug);
    if ($filter->{chltype} && $rec->{chltype}!~/$filter->{chltype}/) {  #yes filter but no match
      &showDebug("User is filtering channel types on $filter->{chltype} and $channel $rec->{chltype} DOES NOT MATCH, skipping",1);
      next;
    }

    if ($c) { #check SSLCIPH
      $rec->{sslciph}=$href->{$qmgr}{CHANNEL}{$channel}{SSLCIPH};
      #If SSLCIPH property not found, set to blank:
      $rec->{sslciph}="" unless defined $rec->{sslciph};
      &showDebug("SSLCIPH: -$rec->{sslciph}-",$localdebug);
    }
    if ($m||$d) { #check MCAUSER
      $rec->{mcauser}=$href->{$qmgr}{CHANNEL}{$channel}{MCAUSER};
      &showDebug("MCAUSER: -$rec->{mcauser}-",$localdebug);
    }
    if ($x) { #check SCYEXIT
      $rec->{scyexit}=$href->{$qmgr}{CHANNEL}{$channel}{SCYEXIT};
      &showDebug("SCYEXIT: -$rec->{scyexit}-",$localdebug);
    }

    #start the checks
    #------
    if ($d) { #no switches, run default check for mcauser(nobody)
      &showDebug("DEFAULT MCAUSER($filter->{nobody}) check against \"$rec->{mcauser}\"",1);
      if ($rec->{mcauser}!~/$filter->{nobody}/i) { #mcauser isn't set to whatever we wanted (nobody)
        $rec->{status}=0;
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with MCAUSER set to \"$rec->{mcauser}\".  Security guidelines require MCAUSER to be set to \"$filter->{nobody}\", which restricts access so that this channel is rendered inoperative.\n";
      }
    } elsif ($c && $m && $x) {      #Switches are: -cmx
      &showDebug("Switches -cmx: Processing $qmgr CHANNEL($channel) CHLTYPE($rec->{chltype}) SSLCIPH($rec->{sslciph}) MCAUSER($rec->{mcauser}) SCYEXIT($rec->{scyexit})",1);
      if ($rec->{mcauser}=~/^$baduser$/i && $rec->{sslciph}=~/^\s*$/ && $rec->{scyexit}=~/^\s*$/) { #baduser, no ssl,  and no exit
        $rec->{status}=0; #bad channel
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined in an unsafe manner because MCAUSER is set to \"$rec->{mcauser}\" and SSLCIPH and SCYEXIT are not set.  Anyone wishing to access the queue manager with full authority need only know the name of this channel.  If you must have MCAUSER set to \"$rec->{mcauser}\", consider improving the channel's security by using SSL encryption and/or by adding a security exit.\n";

      } elsif ($rec->{sslciph}=~/^\s*$/ && $rec->{mcauser}=~/^\s*$/ && $rec->{scyexit}!~/^\s*$/) { #blank mcauser and no ssl, but yes exit
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with MCAUSER set to \"$rec->{mcauser}\" and no SSL encryption; however since SCYEXIT specifies an exit ($rec->{scyexit}), the channel is secure as long as the exit rejects MCAUSER($baduser).\n";

      } elsif ($rec->{sslciph}=~/^\s*$/ && $rec->{mcauser}=~/^\s*$/ && $rec->{scyexit}=~/^\s*$/) { #no mcauser, no ssl, no exit
        $rec->{status}=0;
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined in a potentially unsafe manner because MCAUSER, SSLCIPH and SCYEXIT are not set.  A java application that connects to this channel with a blank MCAUSER has access to the queue manager with the ID the listener runs under.  As this ID is normally \"$baduser\", the application would have full authority.  If you must have MCAUSER set to \"$rec->{mcauser}\", consider improving security by using SSL encryption and/or by adding a security exit.\n";
      
      } elsif ($rec->{sslciph}=~/^\s*$/ && $rec->{mcauser}=~/^\s*$/ && $rec->{scyexit}!~/^\s*$/) { #no ssl, blank mcauser, yes exit
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with MCAUSER is set to \"$rec->{mcauser}\"; however since SCYEXIT specifies an exit, the channel is secure as long as the exit rejects MCAUSER(' ').\n";

      } else { #All other combinations of SSLCIPH, MCAUSER & SCYEXIT settings are OK so no more tests are needed.
        #$rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with MCAUSER($rec->{mcauser}), SCYEXIT($rec->{scyexit}), SSLCIPH($rec->{sslciph}), which is a secure setting.\n";

      } #end $c $m $x checks

    #------
    } elsif ($c && $m && !$x) {  #Switches are: -cm
      &showDebug("Switches -cm_: Processing CHANNEL($channel) CHLTYPE($rec->{chltype}) SSLCIPH($rec->{sslciph}) MCAUSER($rec->{mcauser})",1);
      if ($rec->{sslciph}=~/^\s*$/ && $rec->{mcauser}=~/^$baduser$/) { #no ssl, mcauser bad
        $rec->{status}=0;
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined in an unsafe manner because MCAUSER is set to \"$baduser\" and SSLCIPH is not set.  Anyone wishing to access the queue manager with full authority need only know the name of this channel.  If you must have MCAUSER set to \"$baduser\", consider improving the channel's security by using SSL encryption and/or by adding a security exit.\n";
      }
      elsif ($rec->{sslciph}=~/^\s*$/ && $rec->{mcauser}=~/^\s*$/) { #no ssl, mcauser blank
        $rec->{status}=0;
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined in a potentially unsafe manner because MCAUSER and SSLCIPH are not set.  A java application that connects to this channel with a blank MCAUSER has access to the queue manager with the ID the listener runs under.  As this ID is normally \"$baduser\", the application would have full authority.  If you must have MCAUSER set to \"$rec->{mcauser}\", consider improving security by using SSL encryption and/or by adding a security exit.\n";
      } else { #All other combinations of SSLCIPH, MCAUSER & SCYEXIT settings are OK so no more tests are needed.
        #$rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with MCAUSER($rec->{mcauser}), SSLCIPH($rec->{sslciph}), which is a secure setting.\n";
      }#end if $c $m

    #------
    } elsif (!$c && !$m && $x) {     #Switches are: -x
      &showDebug("Switches -__x: Processing CHANNEL($channel) CHLTYPE($rec->{chltype}) SCYEXIT($rec->{scyexit})",1);
      if ($rec->{scyexit}=~/^\s*$/) { #no scyexit
        $rec->{status}=0;
        $rec->{explain}="QMGR $qmgr may have $rec->{chltype} channel $channel defined in an unsafe manner because SSLCIPH and SCYEXIT are not set.  Consider improving the channel's security by using SSL encryption and/or by adding a security exit.\n";
      } else { #yes scyexit
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with SCYEXIT specifying an exit \"$rec->{scyexit}\", so the channel is secure as long as the exit rejects unauthorized MCAUSERs.\n";
      }#end $x

    #------
    } elsif ($c && !$m && !$x) {    #Switches are: -c
      &showDebug("Switches -c__: Processing CHANNEL($channel) CHLTYPE($rec->{chltype}) SSLCIPH($rec->{sslciph})",1);
      if ($rec->{sslciph}=~/^\s*$/) { #no ssl
        $rec->{status}=0;
        $rec->{explain}="QMGR $qmgr may have $rec->{chltype} channel $channel defined in an unsafe manner because SSLCIPH is not set.  Consider improving the channel's security by using SSL encryption and/or by adding a security exit.\n";
      } else { #yes ssl
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with SSLCIPH \"$rec->{sslciph}\", so the channel is secure.\n";
      } #end $c

    #------
    } elsif (!$c && $m && !$x) {    #Switches are: -m
      &showDebug("-Switches _m_: Processing CHANNEL($channel) CHLTYPE($rec->{chltype}) MCAUSER($rec->{mcauser})",1);
      if ($rec->{mcauser}=~/^$baduser$/) { #bad mcauser
        $rec->{status}=0;
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined in an unsafe manner because MCAUSER is set to \"$rec->{mcauser}\".  Anyone wishing to access the queue manager with full authority need only know the name of this channel.  If you must have MCAUSER set to $baduser, consider improving the channel's security by using SSL encryption and/or by adding a security exit.\n";
      } elsif ($rec->{mcauser}=~/^\s*$/) { #blank mcauser
        $rec->{status}=0;
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined in a potentially unsafe manner because MCAUSER is not set.  A java application that connects to this channel with a blank MCAUSER has access to the queue manager with the ID the listener runs under.  As this ID is normally \"$baduser\", the application could have full authority.  If you must have MCAUSER set to \"$rec->{mcauser}\", consider improving security by using SSL encryption and/or by adding a security exit.\n";
      } else { #mcauser ok
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with MCAUSER \"$rec->{mcauser}\", so the channel is secure (as far as permissions for that user are secured).\n";
      } #end $m

    } elsif ($c && !$m && $x) {    #Switches are: -cx
      &showDebug("Switches -c_x: Processing CHANNEL($channel) CHLTYPE($rec->{chltype}) SSLCIPH($rec->{sslciph}) SCYEXIT($rec->{scyexit})",1);
      if ($rec->{sslciph}=~/^\s*$/ && $rec->{scyexit}=~/^\s*$/) { #no ssl, no exit
        $rec->{status}=0;
        $rec->{explain}="QMGR $qmgr may have $rec->{chltype} channel $channel defined in an unsafe manner because SSLCIPH and SCYEXIT are not set.  Consider improving the channel's security by using SSL encryption and/or by adding a security exit.\n";
      } elsif ($rec->{sslciph}=~/^\s*$/ && $rec->{scyexit}!~/^\s*$/) { #no ssl, yes exit
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with SCYEXIT specifying an exit, so the channel is secure as long as the exit rejects unauthorized MCAUSERs.\n";
      } else { #ssl yes
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with SSLCIPH \"$rec->{sslciph}\" and SCYEXIT \"$rec->{scyexit}\", so the channel is secure.\n";
      }#end $c $x

    #------
    } elsif (!$c && $m && $x) {     #Switches are: -mx
      &showDebug("Switches -_mx: Processing CHANNEL($channel) CHLTYPE($rec->{chltype}) MCAUSER($rec->{mcauser}) SCYEXIT($rec->{scyexit})",1);
      if ($rec->{mcauser}=~/^$baduser$/ && $rec->{scyexit}=~/^\s*$/) { #mcauser mqm, no exit
        $rec->{status}=0;
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined in an unsafe manner because MCAUSER is set to \"$rec->{mcauser}\" and SCYEXIT IS not set.  Anyone wishing to access the queue manager with full authority need only know the name of this channel.  If you must have MCAUSER set to \"$rec->{mcauser}\", consider improving the channel's security by using SSL encryption and/or by adding a security exit.\n";
      } elsif ($rec->{mcauser}=~/^$baduser$/ && $rec->{scyexit}=~/^\s*$/) { #mcauser mqm, yes exit
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with MCAUSER is set to \"$rec->{mcauser}\"; however since SCYEXIT specifies an exit, the channel is secure as long as the exit rejects MCAUSER(\"$baduser\").\n";
      } elsif ($rec->{mcauser}=~/^\s*$/ && $rec->{scyexit}=~/^\s*$/) { #blank mcauser, no exit
        $rec->{status}=0;
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined in a potentially unsafe manner because MCAUSER and SCYEXIT are not set.  A java application that connects to this channel with a blank MCAUSER has access to the queue manager with the ID the listener runs under.  As this ID is normally \"$baduser\", the application would have full authority.  If you must have MCAUSER set to \"$rec->{mcauser}\", consider improving security by using SSL encryption and/or by adding a security exit.\n";
      } elsif ($rec->{mcauser}=~/^\s*$/ && $rec->{scyexit}!~/^\s*$/) { #mcauser blank, yes exit
        $rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with MCAUSER is set to \"$rec->{mcauser}\"; however since SCYEXIT specifies an exit, the channel is secure as long as the exit rejects MCAUSER(' ').\n";
      } else { #All other combinations of SSLCIPH, MCAUSER & SCYEXIT settings are OK so no more tests are needed.
        #$rec->{explain}="QMGR $qmgr has $rec->{chltype} channel $channel defined with MCAUSER($rec->{mcauser}), SCYEXIT($rec->{scyexit}), which is a secure setting.\n";
      }#end $m $x

    } #end mcxd checks

    #set all the max lengths
    $max{qmgr}=length($rec->{qmgr}); #only one qmgr
    $max{channel}=length($rec->{channel}) if (length($rec->{channel}) > $max{channel});
    $max{chltype}=length($rec->{chltype}) if (length($rec->{chltype}) > $max{chltype});
    $max{sslciph}=length($rec->{sslciph}) if ($c && length($rec->{sslciph}) > $max{sslciph});
    $max{mcauser}=length($rec->{mcauser}) if (($m||$d) && length($rec->{mcauser}) > $max{mcauser});
    $max{scyexit}=length($rec->{scyexit}) if ($x && length($rec->{scyexit}) > $max{scyexit});
    #don't need length for status or explain

    push @AoH,$rec; #load the href onto the array

  } #end loop through all channels

  return (\@AoH,\%max);

} #end loadChanSecurity

#----------------------------------------------
sub checkChanSecurity {
#Take results of loadChanSecurity and print in a way that ksh (mqhealth) can use
#STDOUT will be all (SVRCONN) channels with their status.
#Currently set up for mqhealth use, which means err output goes to files to be
#picked up by ksh (can't pass structures directly); in future change output files
#to variables for use elsewhere in perl.
#----------------------------------------------
  my $qmgr=shift || die "No queue manager name provided";
  my $switches=lc(shift(@_)); #if no switches, it'll default to d for mcauser(nobody)
  my $errfile=shift(@_) || "/tmp/checkChanSecurity.err"; #contains errors
  my $summaryfile=shift(@_) || "/tmp/checkChanSecurity.out"; #contains integer vals for num chls, num probs, num notes
  &showDebug("Error output will be copied to $errfile",1);
  &showDebug("Summary output will be copied to $summaryfile",1);

  my ($AoH,$max)=&loadChanSecurity($qmgr); #load all the chan stuff up into structures
  #Test switch characters and set switch variables:
  my $thisFunction=&thisFunc();
  my ($c, $m, $x);
  for (my $i=0; length($_=substr($switches,$i,1)); $i++) {
    if (/c/i) { $c++; next; }
    if (/m/i) { $m++; next; }
    if (/x/i) { $x++; next; }
    warn "WARNING: $thisFunction received invalid switches: \"$switches\", operation aborted.\n";
    return undef;
  }

  #print all output, starting with a formatted header row based on included cols and needed lengths
  my @header=("CHANNEL","CHLTYPE");
  push @header,"SSLCIPH" if $c;
  push @header,"MCAUSER" if $m;
  push @header,"SCYEXIT" if $x;
  push @header,"STATUS";
  foreach ( @header ) { #if max length isn't long enough for header, increase
    &showDebug("$_ max length BEFORE is $max->{lc($_)}",2);
    $max->{lc($_)}=length($_) if $max->{lc($_ )} < length($_);
    &showDebug("$_ max length AFTER  is $max->{lc($_)}",2);
  }
  my $format="%-$max->{channel}s  %-$max->{chltype}s";
  $format.="  %-$max->{sslciph}s" if $c;
  $format.="  %-$max->{mcauser}s" if $m;
  $format.="  %-$max->{scyexit}s" if $x;
  $format.="  %-$max->{status}s\n";
  my @headerout;
  push @headerout,sprintf "$format",@header;
  push @headerout,&Kline::Display::drawLine();
  print @headerout unless defined wantarray; #print to STDOUT if we weren't expecting any results returned

  my @errout; #error output buffer
  my @explain; #store up the comments
  my ($total, $probs, $notes);#count up the numbers to dump to the temp files
  foreach my $href (@{ $AoH }) {
    $total++;
    ($href->{status}) ? $notes++ : $probs++;
    my @outparms;
    push @outparms,$href->{sslciph} if $c;
    push @outparms,$href->{mcauser} if $m;
    push @outparms,$href->{scyexit} if $x;
    my $out=sprintf "$format","$href->{channel}","$href->{chltype}",@outparms,($href->{status}) ? "OK" : "WARNING!!";
    print $out;
    push @errout,$out unless ($href->{status}); #copy err lines to file for mqhealth if failures
    push @explain,$href->{explain}; #includes both good and bad comments
  }

  if (@errout) {
    #print all we know about probs & notes to a file to be picked up by mqhealth
    open ERROUT, ">$errfile" || warn "Cannot open $errfile: $!";
    select ERROUT;
    print "Qmgr $qmgr has $total channels within which $probs problems and $notes notes were identified. These channels follow:\n\n";
    print @headerout;
    print @errout;
    print "\n";
    print "Detailed Explanation:\n";
    print @explain;
    select STDOUT;
    close ERROUT;
  }
  #for mqhealth, write file with total channels and total errors
  system("echo \"$total $probs $notes\" > $summaryfile");

} #end checkChanSecurity

#----------------------------------------------
sub DLQSummary {
#parse MA01 Q supportPac output and provide a summary of codes, times, etc. to STDOUT
#Receives $1 as QMGR, $2 as DLQ name
#If a $3 is passed, then don't return any value
#----------------------------------------------

  my $href;
  my ($qmgr,$dlq)=@_;
  #sanity check
  my $thisFunction=&thisFunc();
  unless ($qmgr && $dlq) {
    warn "WARNING: $thisFunction did not receive proper parameters: QMGR, DLQ\n";
    return 0;
  }
  my ($reason,$destq,$destqmgr,$app,$putdate,$puttime);
  my $cmd="$Kline::KlineHome/q -m $qmgr -i $dlq -dnf2 2>&1| egrep \"Reason|Dest. Queue|Dest. QMgr|PutApplName|Put Date|Put Time|========\"";
  &showDebug("Executing command $cmd",1);
  my $output=&Kline::Server::cmdLine($cmd);
  my $count=0;
  foreach (@$output) {
    chomp;
    &showDebug("line is $_",3);
    $reason=&trim($1) if (/Reason\s+:(.+)/);
    $destq=&trim($1) if (/Dest\. Queue\s+:\'(\S+)\s+\'/);
    $destqmgr=&trim($1) if (/Dest\. QMgr\s+:\'(\S+)\s+\'/);
    $app=&trim($1) if (/PutApplName\s+:\'([^\']+)\'/);
    $putdate=&trim($1) if (/Put Date\s+:\'(\d+)\'/);
    $puttime=&trim($1) if (/Put Time\s+:\'(\d+)\'/);
    &showDebug("Found reason \"$reason\"",2) if $reason;
    &showDebug("Found destq \"$destq\"",2) if $destq;
    &showDebug("Found destqmgr \"$destqmgr\"",2) if $destqmgr;
    &showDebug("Found app \"$app\"",2) if $app;
    &showDebug("Found putdate \"$putdate\"",2) if $putdate;
    &showDebug("Found puttime \"$puttime\"",2) if $puttime;
    if ($reason && $destq && $destqmgr && $app && $putdate && $puttime) { #full record
      $count++;
      &showDebug("Found record $count",1);
      $href->{$reason}{$destqmgr}{$destq}{$app}{count}++;
      $href->{$reason}{$destqmgr}{$destq}{$app}{min}="$putdate/$puttime" if ((exists $href->{$reason}{$destqmgr}{$destq}{$app}{min} && $href->{$reason}{$destqmgr}{$destq}{$app}{min} gt "$putdate/$puttime") || !exists($href->{$reason}{$destqmgr}{$destq}{$app}{min}));
      $href->{$reason}{$destqmgr}{$destq}{$app}{max}="$putdate/$puttime" if ((exists $href->{$reason}{$destqmgr}{$destq}{$app}{max} && $href->{$reason}{$destqmgr}{$destq}{$app}{max} lt "$putdate/$puttime") || !exists($href->{$reason}{$destqmgr}{$destq}{$app}{max}));
      undef $reason;
      undef $destq;
      undef $destqmgr;
      undef $app;
      undef $putdate;
      undef $puttime;
    }
  } #end lines

  print "Successfully parsed $count dead messages on $dlq for $qmgr\n";
  my $format="%5s  %-30s%-15s%-40s%-10s%-17s  %-17s\n";
  printf "$format","COUNT","REASON","DEST_QMGR","DEST_QUEUE","PUT_APP","OLDEST","NEWEST";
  drawLine("=");
  foreach my $reason (sort keys %{ $href } ) {
    foreach my $destqmgr (sort keys %{ $href->{$reason} } ) {
      foreach my $destq (sort keys %{$href->{$reason}{$destqmgr} } ) {
        foreach my $app (sort keys %{$href->{$reason}{$destqmgr}{$destq} } ) {
          my $count=$href->{$reason}{$destqmgr}{$destq}{$app}{count} || 0;
          my $min=$href->{$reason}{$destqmgr}{$destq}{$app}{min} || "N/A";
          my $max=$href->{$reason}{$destqmgr}{$destq}{$app}{max} || "N/A";
          printf "$format","$count","$reason","$destqmgr","$destq","$app","$min","$max";
        }
      }
    }
  } #end looping

  $_[2] ? return 0 : return $href; #only return the href if we don't pass $3

} #end DLQSummary

#----------------------------------------------
sub UNIXQmgrList {
#simple sub that calls getQmgrs and takes the return hash,
#pulling only the keys into an array for return.  Used by gnosis
#remember that this looks for mqs.ini, so no work on Win32
#----------------------------------------------
  my %mq=&getQmgrs();
  my @final=sort keys %mq;
  &showDebug("Returning the qmgr list @final",1);
  return @final;

} #end qmgrList

#----------------------------------------------
sub _qmgrListenerServices {
#Internal sub used by qmgrRunmqlsrPorts to query the MQ6+ service object(s) and return any reserved ports
#Does not check whether the ports are listening, only that they are reserved. 
#Receives the qmgr name, returns href with svcnames, ports;
#NOTE: If the service is running, then the cmdline will have -p and don't need to query here directly, so this is only called when they're down
#Cross platform, will die if not it mqm group, so only call if you're supposed to
#return undef if qmgr is not running or user not in mqm group
#----------------------------------------------
  my $qmgr=shift || die "Need qmgr";
  unless (&isQmgrRunning($qmgr) && &isInMQMGroup()) { #must be running and in group for us to continue
    &showDebug("queue manager $qmgr is either not running or user not in mqm group; no MQSC listener services can be scanned",1);
    return undef;
  }
   
  #make sure this is 6.0 or higher or it'll err
  my $href=&loadQmgrObjects($qmgr,"QMGR"); #this is platform mobile compared to shelling out dspmqver, and we don't need minors
  unless ($href->{$qmgr}{QMGR}{$qmgr}{CMDLEVEL} ge "600") {
    &showDebug("Cannot check MQSC listener services unless MQ version 6.0 or higher",1);
    return undef;
  }
  
  $href=&loadQmgrObjects($qmgr,"LISTENER");
  my %out;
  foreach my $svc (sort keys %{ $href->{$qmgr}{LISTENER} } ) { #ignore default and non-TCP svcs
    $svc=$href->{$qmgr}{LISTENER}{$svc};
    next if $svc->{LISTENER}=~/DEFAULT/; 
    next unless $svc->{TRPTYPE}=~/TCP/;
    #get here we have a valid service; pull the port
    my $port=$svc->{PORT};
    #will be port 0 if using defaults, we'll check that stuff elsewhere
    &showDebug("Found qmgr $qmgr listener service $svc->{LISTENER} with port $port",1);
    $out{$svc->{LISTENER}}=$port;
  }
  return \%out;

} #_qmgrListenerServices

#----------------------------------------------
sub qmgrRunmqlsrPorts {
# receive a qmgr hash and return a AoH with all current runmqlsrs
# mqm group membership is not required, though if listener is down, group membership will allow it to query reserved ports in the service
# 06/06/08 Updated for Win32 support
#----------------------------------------------
  my %qmgrs=@_;
  my @AoH;
  my $defqmgr=&getDefQmgr() || "";

  foreach my $qmgr (sort keys %qmgrs) {
    my $found=0;
    &showDebug("QMGR: $qmgr",1);
    my @out;
    my $mqsc_svcs; #href output from checking MQSC listeners, with {service}=port
    my %pids;

    if ($WINDOWS) {
      #In MQ6+, if you select to create a listener using the GUI, it will add a listener service object and not add anything to the registry
      #Registry vals in the Configuration/qmgr/TCP key are only added via QMGR "properties" updates in Eclipse.  If you add a port on the qmgr props,
      #then the listener-service with port 0 will use the default in the registry.  If no reg value, it will use 1414 if not in use, otherwise will not start.
      my $regport=$Win32::TieRegistry::Registry->{$registryBaseKey}->{"Configuration/QueueManager/$qmgr/TCP//Port"}; #if it exists, then defaults are using this port
      my $pi=Win32::Process::Info->new();
      my $runmqlsr="runmqlsr.exe";
      my @qmgrLsrProcs=grep {$_->{Name} =~ m/$runmqlsr/} $pi->GetProcInfo( { no_user_info => 1 },); #can take a few moments to complete
      foreach (@qmgrLsrProcs) { #loop through all qmgr PIDs for ours
        next unless $_->{CommandLine}=~/$runmqlsr.+-m\s+\b$qmgr\b/; #skip all lines that ain't us
        my $pid=$_->{ProcessId};
        my $aport;
        if ($_->{CommandLine}=~/-p (\d+)/) {
          $aport=$1;
          &showDebug("$qmgr runmqlsr process explicitly has port $aport",2);
        } else { #default port of some type, registry or system default
          #check the registry (the service must be using port 0, which is default: points first to reg then to 1414)
          if ($regport) { #if it exists, then we must be listening there, else it's default
            $aport=$regport;
            &showDebug("$qmgr is using a runmqlsr service with default port 0 specified, but the registry shows port $aport as specifically reserved",2);
          } else { #use system default 1414 if no -p is found
            $aport=1414;
            &showDebug("$qmgr is using runmqlsr service with default port 0 specified and no reg values; 1414 selected as default",2);
          }
        }
        push @out,$aport;
        
        #load the pids too
        &showDebug("Adding pid $pid to port $aport",2);
        push @{ $pids{$aport} },$pid;
        
        #we will check whether actually listening later, though multiple defaults
        #(therefore pointing to 1414) with only a single qmgr listening will likely show up as all listening
      }
      
      #if nothing presently running, we can still check for port reservation in registry and then in the service
      unless (@out) {
        #check registry
        if ($regport) {
          &showDebug("$qmgr has nothing listening but has port $regport reserved in registry",2);
          push @out,$regport;
        }
        #check the service object, if any (shared sub since unix needs the same)
        my $mqsc_svcs=&_qmgrListenerServices($qmgr);
        push @out,values %$mqsc_svcs if defined $mqsc_svcs;
      }
      
   } else { #UNIX/Linux
    # needs to handle default qmgr with no port specified
    #looks like this: mqm      22456 15865  1 14:04 pts/8    00:00:00 runmqlsr -t tcp

      my $port=&Kline::Server::cmdLine("ps -eo pid,args | grep -v grep | grep runmqlsr"); #aref
      foreach (@$port) {
        chomp;
        if (/-m \b$qmgr\b/ || (($qmgr eq $defqmgr) && !/-m/)) {
          &showDebug("Found runmqlsr for $qmgr: $_",2);
          my $aport;
          if (/-p (\d+)/) {
            $aport=$1;
          } else {
            #use 1414 if no -p is found
            $aport=1414;
            &showDebug("$qmgr is using runmqlsr but no specified port; 1414 selected as default",2);
          }
          push @out,$aport;
          #load the pids too
          if (/^\s*(\d+)\s+/) {
            &showDebug("Adding pid $1 to port $aport",2);
            push @{ $pids{$aport} },$1;
          }
        } else {
          &showDebug("no match for runmqlsr $qmgr: $_",3);
        }
      } #end ps output loop
      
      #if nothing presently running, we can still check for port reservation in the service
      unless (@out) {
        #check the service object, if any (shared sub since unix needs the same)
        $mqsc_svcs=&_qmgrListenerServices($qmgr);
        push @out,values %$mqsc_svcs if defined $mqsc_svcs;
      }
    } #end Win vs UNIX
    
    @out=sort(&getUnique(&trim(@out)));
    if (@out) {
      &showDebug("$qmgr is using runmqlsr on ports: -@out-",1);
      my $count;
      foreach my $port (@out) {
        my $rec={};
        $count++;
        $rec->{qmgr}=$qmgr;
        $rec->{type}="runmqlsr";
        $rec->{port}=$port;
        if (keys %pids) {
          my @pids=@{ $pids{$port} };
          &showDebug("port $port has pids @pids",2);
          $rec->{pids}=[ @pids ];
        
        } else {
          &showDebug("port $port has NO pids",2);
        }
        if (keys %$mqsc_svcs) { #provide the MQSC service names, if any exist
          $rec->{service}="Applicable MQSC Listener Object Names (from which port assignments were queried): " .join ", ",keys %$mqsc_svcs;  
        }
        push @AoH,$rec;
        #we warn again after combining with inet, so not needed here
        #warn "WARNING: $qmgr is listening on multiple ports (@out) for runmqlsr.  \nThis is acceptable but is not common.  Make sure this configuration is \nindeed correct.\n\n" if ($count > 1);
      }
    } else {
      &showDebug("$qmgr is NOT using runmqlsr ports",1);
    }
  } #end qmgr loop
  return @AoH;

} #end qmgrRunmqlsrPorts

#----------------------------------------------
sub qmgrInetdPorts {
#read through all (x)inetd configs for amqcrsta and return AoH
#with configured info
#mqm group membership not required
#----------------------------------------------
  die "cannot run on $^O" if $WINDOWS;
  my %qmgrs=@_; #need to only look for specified qmgrs
  my @AoH;
  my $inet=&Kline::Server::cmdLine("ps -eo comm,args | grep -v grep | grep inetd"); #aref
  if (@$inet) { #one of the daemons is running

    #xinetd--------------------------------------------------
    if (grep /xinetd/,@$inet) { #xinet is running here
      &showDebug("xinetd detected running on this machine",1);
      #this line throws away auth errors when grepping; may cause issues in future
      #when can't read file and cannot find out why, but makes 99% more easy.
      #find all files for xinetd with amqcrsta
      my $files=&Kline::Server::cmdLine({
                                                 command=>"grep amqcrsta /etc/xinetd.d/* 2>/dev/null | cut -f1 -d':' | uniq",
                                                 shift_rc=>1,
                                                 no_redirect=>1,
                                                }); #aref
      if (@$files) {
        for my $file (@$files) {

          #xinetd entries look like this:
          #service ibm-mqseries-MISHA
          #{
            #flags           = REUSE
            #socket_type     = stream
            #wait            = no
            #user            = mqm
            #server          = /opt/mqm/bin/amqcrsta
            #server_args     = -m MISHA
            #log_on_failure  += USERID
            #disable         = no
          #}

          chomp($file);
          &showDebug("$file is configured for xinetd",1);
          #file can have multiple entries; load each entry into hash
          if (open XINETD, "<$file" or warn "Cannot open $file: $!\n") {
            my $rec={};
            while (<XINETD>) {
              next if (/(^#)|^\s+$/);
              chomp;
              if (/^service/ .. /\s*\}/) {
                &showDebug("line is $_",4);
                if (/^service\s+(\S+)/) {
                  my $service=$1;
                  &showDebug("Loading xinetd service $service",1);
                  $rec->{service}="$service";
                }
                if (/\s*(\S+)\s+(=|\+=)\s+(.+)/) {
                  my $parm=&trim($1);
                  my $val=&trim($3);
                  &showDebug("loading parm \"$parm\" -$val-",3);
                  $rec->{$parm}="$val";
                  if (($parm=~/server_args/) && ($val=~/-m (\S+)/)) {
                    my $qmgr=$1;
                    &showDebug("Identified xinetd listener \"$rec->{service}\" for qmgr \"$qmgr\"",1);
                    $rec->{qmgr}="$qmgr";
                  }
                }

                if (/\}/) { #if line is close curly, end of rec
                  &showDebug("Loading completed record for $rec->{service}",1);
                  $rec->{file}="$file";
                  $rec->{type}="xinetd";
                  if (exists $qmgrs{$rec->{qmgr}}) { #only load if qmgr in scope
                    push @AoH,$rec;
                  } else { #qmgr not in scope
                    &showDebug("qmgr $rec->{qmgr} is not in scope; RECORD LOAD ABORTED.",1);
                  }
                  $rec={};
                }
              }
            } #end while
            close XINETD;
          } #end check successful file open
        } #end loop through xinetd files
      } else { #no files; not configured
        &showDebug("No qmgrs configured in xinetd",1);
        #do not return here; it will do later
      }
    #inetd--------------------------------------------------
    } else { #normal inet detected
      &showDebug("inetd detected running on this machine",1);
      my $file="/etc/inetd.conf";

      #/etc/inetd.conf looks like this on Sun and AIX
      #mq101   stream  tcp     nowait  mqm     /usr/mqm/bin/amqcrsta amqcrsta -m D03MQ101

      my $crstas=&Kline::Server::cmdLine("cat $file | grep -i amqcrsta | grep -v \"^#\""); #aref
      my $rec={};
      foreach my $line (@$crstas) {
        if ($line=~/(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.+)/) {
          ($rec->{service},$rec->{socket_type},$rec->{wait},$rec->{user},$rec->{server},$rec->{server_args})=($1,$2,$4,$5,$6,$7);
          if (($rec->{server_args}) && ($rec->{server_args}=~/-m (\S+)/)) {
            my $qmgr=$1;
            &showDebug("Identified inetd service \"$rec->{service}\" for qmgr \"$qmgr\"",1);
            $rec->{qmgr}="$qmgr";
          }
          $rec->{file}="$file";
          $rec->{type}="inetd";
          if (exists $qmgrs{$rec->{qmgr}}) { #only load if qmgr in scope
            push @AoH,$rec;
          } else {
            &showDebug("qmgr $rec->{qmgr} is not in scope; RECORD LOAD ABORTED.",1);
          }
          $rec={};
        }
      }


    } #end check for inetd vs xinetd

    #ports--------------------------------------------------
    &showDebug("SERVICE LOCATION COMPLETE: CROSS-REFERENCING PORTS NOW",1);
    #now go through AoH and add ports; current status can be obtained elsewhere
    foreach my $href (@AoH) {
      &showDebug("FINDING PORT for $href->{qmgr} (service $href->{service})",1);
      if ($href->{disable} && $href->{disable}=~/yes/i) {
        warn "WARNING: $href->{qmgr} is setup for xinetd listeners in $href->{file} \n(service $href->{service}), but is currently disabled.\n\n";
      }

      #/etc/services looks like this:
      #ibm-mqseries-MISHA      1416/tcp        # IBM MQSeries
      #ibm-mqseries-MISHA      1416/udp        # IBM MQSeries
      my $services=&Kline::Server::cmdLine("cat /etc/services | grep -w ^$href->{service}");
      if (@$services) {
        my $count;
        foreach my $service (@$services) {
          chomp($service);
          if ($service=~/(\d+)\/tcp/) {
              &showDebug("Valid service: \"$service\"",2);
              $href->{port}=$1;
              $count++;
          } #check services line
        } #each services line
        if ($count >1) {
          warn "WARNING: The service $href->{service} (for qmgr $href->{qmgr}) is not unique \nin /etc/services.  Listeners will not function correctly until all service \nnames are unique. Entries follow:\n";
          foreach (@$services) { print STDERR "$_\n"; }
          print STDERR "\n";
       }
      } else { #no /etc/services entry found
        warn "WARNING: $href->{qmgr} is setup for (x)inetd listeners in $href->{file} \n(service $href->{service}), but no valid entry in /etc/services could be found.\n\n";
        $href->{port}=0;
      }

    } #end AoH loop through services

  } else { #neither inetd nor xinetd is running
    &showDebug("Neither inetd nor xinetd are running",1);
  }

  return @AoH;

} #end qmgrInetdPorts

#----------------------------------------------
sub debugInetAoH {
#Dumps the inet AoH in debug for verification
#----------------------------------------------
  my @AoH=@_;
  my $thisFunc=&thisFunc();
  my $parent=&parentFunc();
  &showDebug("ERROR: AoH into $thisFunc from $parent was empty.",1) unless @AoH;

  &showDebug("==========================DUMPING INETD/SERVICES HASH FROM $parent:",2);
  foreach my $href (@AoH) {
    &showDebug("--------------------------",2);
    foreach my $prop (sort keys %$href) {
      &showDebug("($parent) $prop = \"$href->{$prop}\"",2);
    }
  }

} #end debugInetAoH

#----------------------------------------------
sub listenerInfo {
#Primary function for listeners; checks runmqlsr and inetd
#then checks status; does not show misconfigs (that's in mngmqlsr)
#returns AoH with all the info for use elsewhere
#input args would be qmgr name(s)
#mqm group membership is not required
#06/06/08 updated for use on Windows; now also checks MQSC listener service for port assignments if not lsr running
#----------------------------------------------
  my @AoH;
  my %qmgrs;
  if (@_) {
    &showDebug("Received args \"@_\"",2);
    %qmgrs=&getQmgrs(@_);
  } else { 
    %qmgrs=&getQmgrs();
  }

  #gather all runmqlsr information for specified qmgrs
  push @AoH,&qmgrRunmqlsrPorts(%qmgrs); #no mqm group membership required, though useful when needed to query LISTENER service objects
  &debugInetAoH(@AoH); #dump the hash for debug 2

  #check (x)inetd and cross-check /etc/services for all qmgrs
  unless ($WINDOWS) {
    push @AoH,&qmgrInetdPorts(%qmgrs); #no mqm group membership required
    &debugInetAoH(@AoH); #dump the hash for debug 2
  }

  #loop check all ports if active and append to AoH
  @AoH=&Kline::Server::portIsListening(@AoH); #no mqm group membership required
  &debugInetAoH(@AoH); #dump the hash for debug 2

  return @AoH;

} #end listenerInfo

{ #naked block for OAM persistence
  my $oam; #hashref so we keep all the oam junk between calls
#----------------------------------------------
sub getOAMPermissions {
#Query OAM for permissions and load them into structures that can be 
#used by health checks.  Based initially upon the mqauth52 (MS0I, withdrawn)
#supportPac for amqoamd command query.  However, now uses dmpmqaut command
#query, which is an easier, more complete output format to parse.  Group mqm is
#filtered by default from all results since it always has auth.
#Takes $1 as qmgr, $2 as optional flag to cause mqm permissions to be ignored
#$3 (optional) can be set to "amqoamd" to force amqoamd parsing (deprecated) 
#instead of dmpmqaut, which appeared in MQ 5.3 (now used by default)
#Returns hashref of objects (persistent across calls) $href->{\@cmd} includes the name of the command the data was populated with.
#Return format $oam->{object type}{profile}{entity type}{entity}=authority
#Will try dmpmqaut and switch to amqoamd if dmpmqaut is not found.
#used by Kline::HealthCheck::AllObjectAuthorities (OAMAdmin and OAMAltusr)
#----------------------------------------------

  my $qmgr=shift;
  my $no_mqm=shift(@_) if @_; #if $2 is passed with anything, ignore mqm entries (true false)
  my $cmdname=shift(@_) if @_; #if $3 passed, it's the name of the cmd to use (persumably forcing to amqoamd)
  return $oam if exists $oam->{$qmgr}; #already loaded, so don't get it again
  
  # verify command is valid
  $cmdname="dmpmqaut" unless $cmdname;
  &checkPlatform unless ($mqBinaries); #find binary location
  my $cmd = $oam->{"\@cmd"} || catfile($mqBinaries,"bin","$cmdname"); #if we ran once already, just use the command from before, regardless
  $cmd.=".exe" if $WINDOWS && $cmd!~/exe$/i;
  if ($cmd=~/dmpmqaut/ && !-e $cmd || !-x $cmd) { #switch to amqoamd if dmpmqaut not available (5.2)
    &showDebug("NOTICE: $cmd is not available, therefore, trying fallback to amqoamd",1);
    $cmd="$mqBinaries/bin/amqoamd";
  }
  my $thisFunction=&thisFunc();
  my $parent=&parentFunc();
  die "\nNo authority for $thisFunction ($parent); UID $UID not in mqm group\n" unless &isInMQMGroup();
  $cmd.=".exe" if $WINDOWS && $cmd!~/exe$/i;
  if (!-e $cmd || !-x $cmd) {
    die "$cmd either doesn't exist or isn't executable\n";
  }
  $cmdname=$cmd;
  $cmd="$winquote$cmd$winquote" if $cmd=~/\s/ && $WINDOWS; #need to add quotes, but only after file tests and before adding args
  $cmd .= " -m $qmgr"; # run cmd and read from pipe
  my ($rc,$output)=&Kline::Server::cmdLine({
    command=>$cmd,
    shift_rc=>1,
  });
  
  #BT99819 workaround for MQ bug with RC=71 all the time on some levels
  if ($rc) {
    if ($rc == 71 ) { #Problem occurs in 6.0.2.0 and is fixed in 6.0.2.2, but I don't have a test box for 6.0.2.1 to tell if it was fixed there or not
      &Kline::Display::showDebug("WORKAROUND: NON-ZERO RETURN CODE.  Command \"$cmd\" has rc=$rc, but this is a known WMQ issue in versions up to 6.0.2.2; therefore, ignoring the RC",1);
    } else {
      die "Could not complete command {{ $cmd }} (rc=$rc): ($!)\n@$output\n";
    }
  }
    
  my $obj_count = 0;
  my $output_count = 0;
  my ($obj_type,$obj_name,$principal,$entity_type,$authorities); #prin_type only populated with dmpmqaut
    
  $oam->{"\@cmd"}=$cmdname unless $oam->{"\@cmd"}; #allow the user to see the command used in the results.
  $oam->{$qmgr}={}; #if a qmgr has only default perms, it would otherwise not show in output at all
      
  ######dmpmqaut
  if ($cmdname=~/dmpmqaut/) {
    #first line should start with profile in this format
    #profile:     self
    #object type: qmgr
    #entity:      mqm
    #entity type: group
    #authority:   allmqi crt dlt chg dsp
    #- - - - - - - -

    #--------
    # loop through input
    foreach (@$output) {
      if (/No matching authority records/) { #nothing found, so return empty (really shouldn't get here)
        &showDebug("$qmgr: $_",1);
        return $oam;
      }
      
      if (/^- - - - /) { #clear out the rest; it's a new object (won't encounter at very beginning or very end)
        unless ($obj_type && $obj_name && $principal && $entity_type && $authorities) { 
          # something is wrong - missing type &| name &| principal &| unknown type
          print STDERR "object type: ",
            ((defined $obj_type) ? $obj_type : "undef"), "\n";
          print STDERR "object name: ",
            ((defined $obj_name) ? $obj_name : "undef"), "\n";
          print STDERR "entity: ",
            ((defined $principal) ? $principal : "undef"), "\n";
          print STDERR "entity type: ",
            ((defined $principal) ? $principal : "undef"), "\n";
          die "error: parse failed at input line ", $. - 1, "\n";
        }
        $obj_name = $obj_type = $entity_type = $principal = $authorities = undef;
        next;
      }

      if (/^\s*profile:\s+(.+)/) {
        ($obj_name = $1) =~ s/\s*$//g;  # right-trim object name
        $obj_name=lc($obj_name) if $obj_name=~/\b(SELF)|(\@CLASS)\b/; #on windows, these are uppercase, though lower on unix and in our compares
      } elsif (/^\s*object type:\s+(.+)/) {
        ($obj_type = $1) =~ s/\s*$//g;  # right-trim object name
      } elsif (/^\s*entity:\s+(.+)/) {
        ($principal = $1) =~ s/\s*$//g;  # right-trim object name
      } elsif (/^\s*entity type:\s+(.+)/) { #not actually stuffed anywhere just yet; amqoamd doesn't have it
        ($entity_type = $1) =~ s/\s*$//g;  # right-trim object name
      } elsif (/^\s*authority:\s+(.+)/) { 
        ($authorities = $1) =~ s/\s*$//g;  # right-trim object name
      }

      #if we have all 5 fields, load the data, but only if it's a principal we care about
      if ($obj_type && $obj_name && $principal && $entity_type && $authorities) { 
        $obj_count++;
        #store the object unless principal is mqm
        if ($principal=~/\bmqm(\@)?/i && $no_mqm) {  #group  may someteimes have domain attached as \@
          &showDebug("Skipping object $obj_type $obj_name principal $principal",2);
          next;
        }
        #store the info onto a nice fat hash ref
        &showDebug("Loading $qmgr $obj_type $obj_name $principal $authorities",2);
        $oam->{$qmgr}{$obj_type}{$obj_name}{$entity_type}{$principal}=$authorities; 
        $output_count++;
      } #end completeness check

    } #end foreach OAM lines dmpmqaut
    
  ######amqoamd (5.2 and previous mostly)
  } elsif ($cmdname=~/amqoamd/) { #deprecated, but works for 5.2 (and 5.3, 6.0, present in 7.0 though no doc)
    die "Cannot parse amqoamd on windows yet" if $WINDOWS;
  
    #format of amqoamd output
    # AMQOAMD
    #  ********
    #  Queue Manager - PAMPANGA

    # ObjectType:  qmgr
    # ObjectName:  self
    #
    #     Principal:          mqm (ON UNIX)
    #     \nUser name:        cnkline@KLINE-T61 (ON WINDOWS) OR test group@KLINE-T61
    #     Authorities:        altusr connect inq set setall setid chg crt dlt dsp (0x00ff07ff)
    #     SID:                S-1-5-21-1850897905-3423105737-3125664128-1008 (ON WINDOWS)
    #     Principal:          badgroup
    #     Authorities:        altusr connect inq (0x00000411)
    #     Principal:          goodgroup
    #     Authorities:        altusr connect inq (0x00000411)
    #
    # ObjectType:  qmgr
    # ObjectName:  @class

    # hash to map pre-5.2csd04 amqoamd object types to setmqaut object types
    my %name_map = (
      'Queue Manager' => 'qmgr',
      'Queue'         => 'queue',
      'Process'       => 'process',
      'Namelist'      => 'namelist',
    );
  
    # check for error - amqoamd seems to return 0 for both success and failure
    # read 1st line - blank is ok, otherwise it's an error
    my $line1 = @$output[0];
  
    if ($line1 =~ /^\w+.*$/) {
      die "error executing $cmd: $line1\n";
    }
  
    #--------
    # loop through input
    foreach (@$output) {
      if (/^\s+ObjectType:\s+(\S+)/) {
        $obj_type = $name_map{$1} || $1;
        $obj_type=~s/\s*$//g;
        $obj_count++;
        $obj_name = $principal = $authorities = undef; #clear out the rest; it's a new object
      } elsif (/^\s+ObjectName:\s+(.+)/) {
        ($obj_name = $1) =~ s/\s*$//g;  # right-trim object name
      } elsif (/^\s+Principal:\s+(\w+)/) {
        ($principal = $1) =~ s/\s*$//g;  # right-trim object name
      } elsif (/^\s+Authorities:\s+(.+)\(\w+\)/) { #skips the hex vals on end
        ($authorities = $1) =~ s/\s*$//g;  # right-trim object name
  
        # write object data
        if (defined $obj_type && defined $obj_name && defined $principal && grep $obj_type, values(%name_map) ) {
          #store the object unless principal is mqm
          if ($principal eq "mqm" && $no_mqm) { 
            &showDebug("Skipping object $obj_type $obj_name principal $principal",2);
            next;
          }
          #store the info onto a nice fat hash ref
          &showDebug("Loading $qmgr $obj_type $obj_name $principal $authorities",2);
          $oam->{$qmgr}{$obj_type}{$obj_name}{$principal}=$authorities; #remember that last auth value is hex val and can be dropped
          $output_count++;
        } else {
          # something is wrong - missing type &| name &| principal &| unknown type
          print STDERR "obj_type : ",
            ((defined $obj_type) ? $obj_type : "undef"), "\n";
          print STDERR "obj_name : ",
            ((defined $obj_name) ? $obj_name : "undef"), "\n";
          print STDERR "principal: ",
            ((defined $principal) ? $principal : "undef"), "\n";
          die "error: parse failed at input line ", $. - 1, "\n";
        }
  
      # if uncommented this will print non-blank lines not captured above
      #} else { unless (/^\s*$/) { print STDERR $_;}  }
      } #end line check
    } #end foreach OAMD
  } #end amqoamd/dmpmqaut
  
  &showDebug("$qmgr found $obj_count objects, processed $output_count entries.",1);
  return $oam;

} #end getOAMPermissions

#----------------------------------------------
sub getOAMProfileMatches {
#Receives qmgr ($1), type ($2), object name ($3), name of a principal ($4)
#If that principal has any auths, using both direct match to oam auths
#and profile-pattern match, appropriate perms are returned (scalar)
#else return undef
#If qmgr object, name should always be self or you'll get infinite loop
#Uses $oam defined by getOAMPermissions (still in scope)
#used by Kline::HealthCheck::AllObjectAuthorities (OAMAdmin and OAMAltusr)
#uses dspmqaut to extract actual perms on the object if wildcard profiles are detected on the object type
#----------------------------------------------
  my $parms=shift || die "no parms";
  my $qmgr=$parms->{qmgr};
  my $objtype=lc($parms->{objtype});
  my $objname=$parms->{objname};
  my $enttype=$parms->{enttype};
  my $entity=$parms->{entity};
  die "Need qmgr, objtype, object name, entity type and entity to continue" unless ($qmgr && $objtype && $entity && $objname && $enttype);
  my %class; #track who has @class permissions on this type; @class always seems to only have crt perms in it
  my $switch=($enttype=~/group/i) ? "-g" : "-p"; #setmqaut either -g or -p
  
  #combine @class into others if it exists so we're only looking in one place
  if (exists $oam->{$qmgr}{$objtype}{"\@class"}) { #make a hash of entities and auths, then we'll add them to any matching entities
    &showDebug("Combining \@class permissions into main permissions",2);
    foreach my $ent (sort keys %{ $oam->{$qmgr}{$objtype}{"\@class"}{$enttype} } ) { #should look at @class perms and loop on groups
      my $newperms=$oam->{$qmgr}{$objtype}{"\@class"}{$enttype}{$ent};
      next if $newperms=~/none/i; #no need to add none
      $class{$ent}=$newperms;
    }
  }

  #in MQ6 (not yet tested in 7), dmpmqaut output will show a profile for a specific user even if it was applied with a wildcard, so long as there was also a specific auth for that user given in another command.  In other words, the wildcard profile will show as a dmpmqaut entry, but there will also be another entry with an object's actual name and complete actual permissions if that object was ever directly given any permissions.  However, if only the wildcard is applied and no direct permissions, then dspmqaut will only show the wildcard profile and nothing on the object itself.  Tested with QL NEW1 and NEW2.  Apply NEW* put and NEW1 get.  Found dmpmqaut output for NEW1 with get/put, plus profiles with NEW* put and NEW2 (though it showed no profile besdies the wildcard).
  
  if (exists $oam->{$qmgr}{$objtype}{$objname}{$enttype}{$entity}) {
    #if we get here, then we don't need to look for profiles since they won't be applied
    &showDebug("Found exact $qmgr permissions match on $objname: $oam->{$qmgr}{$objtype}{$objname}{$enttype}{$entity}",1);
    
    #add @class before returning
    if (exists $class{$entity} && $oam->{$qmgr}{$objtype}{$objname}{$enttype}{$entity}!~/$class{$entity}/) { #don't add more than once
      &showDebug("$qmgr $objtype $entity: (previous) $oam->{$qmgr}{$objtype}{$objname}{$enttype}{$entity}; now also has \@class $class{$entity}",1);
      $oam->{$qmgr}{$objtype}{$objname}{$enttype}{$entity}.=" $class{$entity}";
    }
    return $oam->{$qmgr}{$objtype}{$objname}{$enttype}{$entity};
  } 
  
  #now look for profile matches (means no direct permissions were ever applied to this object)
  my $do_dspmqaut;
  foreach my $profile ( sort keys %{ $oam->{$qmgr}{$objtype} } ) {
    #can search for ?, *, and **
    $do_dspmqaut++ if $profile=~/\?|\*/;
  }
  return $class{$entity} || undef unless $do_dspmqaut;
  &showDebug("$do_dspmqaut $objtype profiles have wildcards that may apply to $objname; checking with dspmqaut",1);
    
  #if we get here, then it has profiles with metachars.  Rather than try to keep up with the regexprs supported by the product, we'll shell out and call dspmqaut for the entity in question.  Then WMQ will be forced to determined the auth level and we'll simply parse it back.  We only do this if we know there are wildcard profiles in effect, so won't be too expensive.
    
  # verify command is valid
  &checkPlatform unless ($mqBinaries); #find binary location
  my $cmd = "$mqBinaries/bin/dspmqaut";
  $cmd.=".exe" if $WINDOWS && $cmd!~/exe$/;
  if (!-e $cmd || !-x $cmd) {
    die "$cmd either doesn't exist or isn't executable\n";
  }
  $cmd="$winquote$cmd$winquote" if $cmd=~/\s/ && $WINDOWS; #need to add quotes, but only after file tests and before adding args
  $entity="$winquote$entity$winquote" if $entity=~/\s/ && $WINDOWS;
  $cmd .= " -m $qmgr -t $objtype -n $objname $switch $entity"; # run cmd and read from pipe
  my ($rc,$out)=&Kline::Server::cmdLine({
                        command => $cmd,
                        shift_rc => 1,
              });
  croak "could not execute $cmd ($rc): @$out$!\n" if $rc;
  shift @$out; #drop the first line, which is header; remaining lines, if any, are the perms
  if (@$out) {
    foreach (@$out) { chomp; $_=&Kline::Utils::trim($_) };
    my $perms=join " ",@$out;
    &showDebug("dspmqaut returned the following perms for $objname: $perms",1);
    return $perms; #these are exact perms so no need to add @class
  }
  &showDebug("no permissions found for $objname",1);
  return $class{$entity} || undef; #no perms found
  
} #getOAMProfileMatches 
} #end naked block $oam

#----------------------------------------------
sub isInMQMGroup {
#checks the current real UID to see if in the mqm group.  
#Optionally receives name of user to check instead of current
#Returns true/false, or undef if mqm group doesn't exist
#On windows, if in the Administrator group, also returns true
#See Kline::Server::isInGroup for actual code
#----------------------------------------------

  if ($WINDOWS) { #if windows, administrator group should also provide needed mq access
    #(tested this on a W2k/mq5.3 box using a new user id added to admin group and it was able to access WMQ things fine)
    &showDebug("Checking for administrator group membership since that allows access to WMQ",1);
    return 1 if &Win32::IsAdminUser() || &isInGroup("Administrators",@_); #though IsAdminUser is more appropriate, it's only for the current user, not the one passed in.
  }
  return &isInGroup("mqm",@_);

} #end isInMQMGroup

#----------------------------------------------
1; #tell the sourcing program we are true
