########################################################################
# Introscope EPAgent Config File Reader Perl Library
#                                                                      
# CA Wily Introscope(R) Version 9.1 Release 9.1.1.0
# Copyright (c) 2012 CA. All Rights Reserved.
# Introscope(R) is a registered trademark of CA.
#
########################################################################
# Config File Reader Module
# -------------------------
# Reads a flexible configuration/properties file.
# Because we like Perl, we allow arbitrary expressions on the
# right-hand side of the equal sign which we simply eval-in
# So quote your regexes, etc.
#
########################################################################

package Wily::Config;

use FindBin;
use lib ("$FindBin::Bin", "$FindBin::Bin/../lib/perl");

sub parseConfig {
    my ($cfgFilename, $defaultValues) = @_;

    my %parsedResults = (); # Return results will contain the parsed 
                            # config file in a nice hash.

    my $buffer = undef;
    my $numLeft = 0, $numRight = 0; # used to count braces/brackets to
                                    # support multi-line syntax.

    open(CFG, $cfgFilename) || die "Could not open file: <$cfgFilename>\n";

    while (my $line = <CFG>) {

        ###################
        # Clean-up the line
        ###################
        chomp;
        $line =~ s/^\s*//;            # Trim whitespace before and after
        $line =~ s/\s*$//;
        $line =~ s/([^\\])#.*$/$1/;   # Delete trailing comments

        ########################
        # Skip if a trivial line
        ########################
        next if ( $line =~ /^\s*#/ ); # Skip comment lines
        next if ( $line =~ /^\s*$/ ); # Skip blank lines

        ###################################
        # Concatenate to our current buffer
        ###################################
        $buffer .= $line; 
        # print "### buffer=<$buffer>\n";

        ###################################
        # Support multi-line syntax 
        # i.e. if the line continues, keep
        #      looping and reading 
        ###################################
        
        # If line ends with trailing back-slash '\', it continues
        # on the next line.
        #
        next if ($buffer =~ s/\\\s*$//);

        # If braces/brackets are not closed i.e. the current buffer
        # is unbalanced i.e. has more left brackets/braces than right
        # ones, then the line continues on. This is part of our flexible
        # array/hash syntax support within the properties file.
        #
        $numLeft  = ($buffer =~ s/[{\[]/$&/g);
        $numRight = ($buffer =~ s/[}\]]/$&/g);
        next if ($numLeft > $numRight);

        #################################################################
        # We now have a balanced and complete line, so time to parse it.
        # Let's make sure it has the following syntax: <name>=<value>
        # We ignore it if it doesn't...
        #################################################################
        if ($buffer =~ /(^[^=]+)=(.+)/) { 

            my ($name, $value) = ($1, $2);
            #print "### name=<$name>, value=<$value>\n";

            # Trim whitespace before/after $name,$value.
            $name  =~ s/^\s*//; 
            $name  =~ s/\s*$//;
            $value =~ s/^\s*//;
            $value =~ s/\s*$//;
            
            ################################################################
            # We do an 'eval' on the value so that it can contain
            # perl variables and expressions. 
            #
            # We provide some syntactic sugar for simple lists
            # (don't require square brackets, can just be comma-delimited).
            # This is useful for small lists that fit on one line.
            # So if it contains commas and doesn't have surrounding
            # braces or brackets, then we create an anonymous array,
            # double-quote it, then eval it.
            ################################################################

            if ($value =~ /,/ && $value !~ /^[\[\{]/) {
                $value = eval qq([$value]);

            } else {
                # Otherwise just eval the value 
                #
                $value = eval $value;
            }

            ########################################################
            # $@ contains any errors from above eval's. If there are
            # errors, there's a syntax error in the config file,
            # and we exit...
            ########################################################
            die "ERROR: reading config file <$cfgFilename>: $@\n" if $@;

            $parsedResults{$name} = $value;

        } # if ($buffer =~ /(^[^=]+)=(.+)/) # name=value syntax

        $buffer = undef; # Done processing this buffer, so clear it.

    } # while (my $line = <CFG>)

    # if buffer is non-empty, we could have imbalanced braces/brackets
    #
    die "ERROR: reading config file <$cfgFilename>: " .
        "possible imbalance of braces/brackets\n" if ($buffer); 

    # Override parsedResults with defaults, if specifed.
    #
    if (ref $defaultValues) {
        foreach my $key (keys %{$defaultValues}) {
            my $keyType = ref($parsedResults{$key});
            if (!$keyType) {
                # Scalar
                $parsedResults{$key} = $defaultValues->{$key} 
                    if ! exists $parsedResults{$key};
            } elsif ($keyType eq 'ARRAY') {
                # Array
                $parsedResults{$key} = $defaultValues->{$key} 
                    if ! exists $parsedResults{$key};
            } elsif ($keyType eq 'HASH') {
                # Hash
                foreach my $subkey (keys %{$defaultValues->{$key}}) {
                    $parsedResults{$key}->{$subkey} = 
                        $defaultValues->{$key}->{$subkey} 
                            if ! exists ($parsedResults{$key}->{$subkey});
                }
            } # if ($keyType ... ) 
        } # foreach my $key (keys %{$defaultValues})
    } # if (ref $defaultValues) 

    close(CFG);                        # Done, close config file
    bless (\%parsedResults, 'Config'); # For OO-semantics
    return \%parsedResults;            # Return results
}

1;

