########################################################################
# Introscope EPAgent Perl Library
#                                                                      
# CA Wily Introscope(R) Version 9.1 Release 9.1.1.0
# Copyright (c) 2012 CA. All Rights Reserved.
# Introscope(R) is a registered trademark of CA.
#
########################################################################
# HTTP Module
# -----------
# This Perl Module supports gathering of metrics based on the availability
# and performance of the http service at a specified URL.
#
########################################################################

package Wily::HTTP;

##########
# Imports
##########

use strict;

use FindBin;
use lib ("$FindBin::Bin", "$FindBin::Bin/../lib/perl", "$FindBin::Bin/..");

use Getopt::Long;

use Wily::PrintMetric;

############
# Constants
############
my $kAV  = 'av'; # Availability
my $kRT  = 'rt'; # Round-Trip Response Time (for connect)
my $kTP  = 'tp'; # Throughput 
my $kALL = 'all'; # Throughput 

my $kMethod = 'method';
my $kURL    = 'url';
my $kArg    = 'arg';

my $kDefaultTimeout = 10; # seconds
my $kDefaultUserAgent = 'Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0)';

my $kDefaultType     = 'IntCounter';
my $kDefaultResource = 'Remote Machine';

my $kService         = 'Service';

my $kAvailableMetricName         = 'Available';
my $kResponseTimeMetricName      = 'Response Time';
my $kAverageThroughputMetricName = 'Average Throughput';

my $kHTTPCookies       = 'HTTP::Cookies';
my $kHTTPRequestCommon = 'HTTP::Request::Common';
my $kLWPUserAgent      = 'LWP::UserAgent';
my $kCryptSSLeay       = 'Crypt::SSLeay';
my $kNetSSLeay         = 'Net::SSLeay';

##########
# Globals
##########
my ($opt_help, $opt_debug, $opt_url, $opt_get, $opt_request, 
    $opt_metric, $opt_timeout) = ();
my %metric = ();
my @urls = ();

my ($response, %reply_headers); # Solaris, Net::SSLeay
my ($ua, $res);                 # Windows, Crypt::SSLeay
my $delim;

my %INIT = ();                    # to track initialization state

##############
# Subroutines
##############

sub usage {

#-----------------------------------------------------------------------------
# 2002 10 24 jenko FUTURE: 
# The 'timeout' works on Solaris, not Windows. Appears that the Perl Windows 
# Sockets lib or interaction is buggy. Suspect that Windows Sockets interface
# is non-standard and doesn't allow timeout to be set in select() calls as in
# Unix. As this functionalty wasn't a requirement, let's take it out of the
# usage for now and re-address when it is required.
#    -timeout <sec>    Optional: timeout for connection in seconds
#-----------------------------------------------------------------------------

    print <<EOF;

Usage: httpSvcAvailability.pl  [-h] [-metric av|rt|tp|all] 
                               -request GET|HEAD
                               -url <URL1,URL2> 

    -h                Show this help message.

    -d                Debug mode (verbose messages).

    -get              Optional: Specify whether to perform an HTTP GET after
                        connection. Same as -request HEAD,GET.

    -request          Optional: comma-delimited list of HTTP requests to
                        issue to the server (no spaces).
                        Some web sites may not allow HEAD requests or some
                        appservers may not support GET.
                        Must specify at least one of: GET,HEAD
                        Default: -request HEAD

    -metric <m1,m2>   Optional: comma-delimited listing of metrics to return
                         av  = Availability / Can Connect (0=no, 1=yes)
                         rt  = Round-Trip Response Time (for connect)
                         tp  = Throughput (only returned if get specified)
                         all = all statistics

    -url <URL1,URL2>  Required: A comma-delimited set of URLs 
                      with optional port numbers. The URLs can include
                      port#s (if they are different from defaults) as
                      well as http: or https: to differentiate HTTP vs HTTPS.

EOF

}

sub debug {
    my @args = @_;
    while (my $arg = shift(@args)) {
        print $arg if $opt_debug;
    }
}


sub parseCommandLine {
    GetOptions( 'url=s'     => \$opt_url,
                'get'       => \$opt_get,
                'request=s' => \$opt_request,
                'metric=s'  => \$opt_metric,
                'timeout=i' => \$opt_timeout,
                'h'         => \$opt_help,
                'd'         => \$opt_debug,
              );

    if ( $opt_help || !$opt_url ) {
        &usage();
        exit;
    }

    $opt_request= 'HEAD,GET' if ($opt_get  && !$opt_request);
    $opt_request= 'HEAD'     if (!$opt_get && !$opt_request);

    # Allow for different delimiters
    #
    my ($firstChar, undef, $lastChar) = $opt_url =~ /^(.)(.*)(.)$/;
    if ( ($firstChar eq $lastChar) && ($firstChar !~ /\w/) ) {
        $delim = $firstChar;
        $opt_url =~ s/^$delim(.*)$delim$/$1/;
    } else {
        $delim = ',';
    }

    # Parse URLs
    # 
    @urls = split(/$delim/, $opt_url);

    # Parse Metrics
    #
    $opt_metric = $kALL if (!$opt_metric); # default is 'all'
    my @tmp = split(/,/, $opt_metric);
    while ( my $opt = shift(@tmp) ) {
        $metric{$opt}++;
    }

    # Set Timeout
    #
    $opt_timeout = $kDefaultTimeout if (!$opt_timeout);
}

sub useNetSSLeay {
    my ($protocol, $os) = @_;
    $os = $^O if !$os;

    my $rv = 0;
    if ( $os =~ /solaris/i && $protocol =~ /^https$/i ) {
        $rv = 1;
    } 

    return $rv;
}

sub getHTTPLibs {

    my $protocol = shift;
    my $os = $^O;

    my @libs = ();  # return value - an array of library string names 
                    # that are required by this OS and protocol

    # Required libs: HTTP::Cookies
    #                HTTP::Request::Common
    #                LWP::UserAgent
    #
    push(@libs, $kHTTPCookies, $kHTTPRequestCommon, $kLWPUserAgent); 

    if ( $os =~ /MSWin/i && $protocol =~ /^https$/i ) {
        push(@libs, $kCryptSSLeay); # Windows HTTPS requires Crypt::SSLeay
    } 

    if ( $os =~ /solaris/i && $protocol =~ /^https$/i ) {
        push(@libs, $kNetSSLeay);   # Solaris HTTPS requires Net::SSLeay
    }

    return @libs;
}

sub initHTTP {
    my $protocol = shift;

    # 2002 10 09 jenko FIXME: ugly switch stmts until we figure out
    # better cross-platform set of libraries for http/s.
    #
    my @libs = &getHTTPLibs($protocol);
    foreach my $lib (@libs) {
        return if $INIT{$lib};
        $INIT{$lib}++;

        eval "require $lib";
        if ($@) { # error, can't find library

            # if perl executable does not contain path, make msg more
            # descriptive
            my $msg = "found in your PATH: $ENV{PATH}" if ($^X !~ m@^\.*/@);

            die <<EOF;
ERROR: The Perl library, $lib, must be installed with the
       Perl installation on this machine to support the $protocol protocol.
       Please check the Perl libraries installed and used by:
           $^X
       $msg
EOF
        }

        if ($opt_debug && $lib eq $kNetSSLeay) {
            $Net::SSLeay::trace = 2; 
        }

        # 2002 10 23 jenko FIXME: timeout settings don't apply to initial
        # connect problems (e.g. invalid hostname). Need to find out how
        # to set TCP timeout settings, in addition to HTTP connect...
        # Appears that LWP timeout only applies to valid connections and
        # subsequent HTTP requests.
        #
        if ($lib eq $kLWPUserAgent) {
            $ua = new LWP::UserAgent;
            # $ua = LWP::UserAgent->new(timeout => $opt_timeout);
            $ua->agent($kDefaultUserAgent);
            $ua->timeout($opt_timeout);
        }
    }
}

sub checkURLs {

    while ( my $url = shift(@urls) ) {

        debug "### processing URL=<$url>\n";

        next if $url =~ /^\s*$/; # in case of custom delims, may have empty elems
        $url =~ s/^\s*//; # trim preceding and
        $url =~ s/\s*$//; # trailing whitespace

        my ($success, $statusCode, $statusMessage) = ();

        my ($protocol, $hostname, $port, $path) = $url =~ m@^(https?)://([\w.\-]+)(:\d+)?(.*)@i;

        $protocol = uc($protocol); # uppercase
        $port = 443 if (!$port && ($protocol =~ /https/i) );

        # Initialize HTTP libraries (differs based on OS and protocol)
        #
        initHTTP($protocol);

        debug "### protocol=<$protocol>,hostname=<$hostname>,port=<$port>,path=<$path>\n";

        # 2002 10 09 jenko FIXME: find msec precision time func and use it instead
        # of 'time'
        #
        my $start_time = time();
        if ($opt_debug) {
            my $start_time_str = localtime();
            debug "### Start: " . $start_time_str . "\n";
        }

        my $byteCount = 0;

        # Perform HEAD if $opt_request contains HEAD
        #
        my $contentHead;
        if ( $opt_request =~ /HEAD/i ) {
            if (&useNetSSLeay($protocol)) {
                ($contentHead, $response, %reply_headers)
                  = Net::SSLeay::get_https($hostname, $port, $path, 'HEAD',
                     Net::SSLeay::make_headers(
                             'User-Agent' => $kDefaultUserAgent,
                     ));
                ($success, $statusCode, $statusMessage) = 
                    &parseNetSSLeayResponse($response);
            } else {
                $res = $ua->request( HTTP::Request::Common::HEAD $url ); 
                $contentHead = $res->content();
                $success = $res->is_success;
                $statusCode = $res->code();
                $statusMessage = $res->message();
            }
            $byteCount = length($contentHead);
        } 

        # 2002 10 09 jenko FIXME: need to review char vs. byte counts for
        # multi-byte encodings
        my $contentGet;
        # do GET *only* if HEAD was successful (or if no HEAD and GET)
        if ( ($opt_request =~ /HEAD/i && $success && $opt_request =~ /GET/i) ||
             ($opt_request !~ /HEAD/i && $opt_request =~ /GET/i) ) { 

            if (&useNetSSLeay($protocol)) {
                ($contentGet, $response, %reply_headers)
                  = Net::SSLeay::get_https($hostname, $port, $path,
                         Net::SSLeay::make_headers(
                                 'User-Agent' => $kDefaultUserAgent,
                         ));
                ($success, $statusCode, $statusMessage) = 
                    &parseNetSSLeayResponse($response);
            } else {
                $res = $ua->request( HTTP::Request::Common::GET $url ); 
                $contentGet = $res->content();
                $success = $res->is_success;
                $statusCode = $res->code();
                $statusMessage = $res->message();
            }
            $byteCount += length($contentGet);
        }

        my $end_time = time();
        if ($opt_debug) {
            my $end_time_str = localtime();
            debug "### End: " . $end_time_str . "\n";
        }

        # Capture results if debug mode
        #
        if ($opt_debug) {
            open(RESP, "> response.html");
            print RESP $contentHead if $contentHead;
            print RESP $contentGet if $contentGet;
            close(RESP);
            if ($success) {
                debug "### SUCCESS!\n";
            } else { 
                debug "### FAILURE: success=$success, code=$statusCode\n    msg=<$statusMessage>\n";
            }
        }

        # Set Subresource
        #
        my $subresource = "$hostname|$kService|$protocol";

        if ($success) {
            # Availability 
            #
            if ( $metric{$kAV} || $metric{$kALL} ) {
                Wily::PrintMetric::printMetric( type => $kDefaultType,
                                                resource => $kDefaultResource,
                                                subresource => $subresource, 
                                                name => $kAvailableMetricName,
                                                value => 1,
                                              );
            }

            # Response Time
            #
            if ( $metric{$kRT} || $metric{$kALL} ) {
                my $seconds = ($end_time - $start_time);
                Wily::PrintMetric::printMetric( type => $kDefaultType,
                                                resource => $kDefaultResource,
                                                subresource => $subresource, 
                                                name => $kResponseTimeMetricName,
                                                value => $seconds,
                                              );
            }

            # Throughput
            #
            if ( $metric{$kTP} || $metric{$kALL} ) {
                my $KBTransferred = ($byteCount / 1024);
                my $seconds = ($end_time - $start_time);
                my $throughput = 0;
                if ($seconds) {
                    $throughput = int ($KBTransferred / $seconds);
                } 
                Wily::PrintMetric::printMetric( type => $kDefaultType,
                                                resource => $kDefaultResource,
                                                subresource => $subresource, 
                                                name => $kAverageThroughputMetricName,
                                                value => $throughput,
                                              );
            }

        } else {
            # Availability 
            #
            if ( $metric{$kAV} || $metric{$kALL} ) {
                Wily::PrintMetric::printMetric( type => $kDefaultType,
                                                resource => $kDefaultResource,
                                                subresource => $subresource, 
                                                name => $kAvailableMetricName,
                                                value => 0,
                                              );
            }

            my $errorMsg;
            if (&useNetSSLeay($protocol)) {
                # $errorMsg = $contentHead . $contentGet;
                $errorMsg = $response;
            } else {
                # $errorMsg = $res->as_string();
                $errorMsg = "HTTP Error Code=$statusCode, Message=$statusMessage\n";
                print STDERR ("ERROR (httpSvcAvailability.pl): $errorMsg" );
            }
            # print STDERR ("ERROR: http status code=$statusCode\n    msg=$statusMessage\n");
        } 
    }
}

sub parseNetSSLeayResponse {
    my $response = shift;
	my ($success, $code, $message) = (1, -1, "UNDEF");

	if ($response =~ m@HTTP/[\d.]+\s+(\d+)(.*)@) {
		$code = $1;
		$message = $2;
		$success = 0;
		print "### parseNetSSLeayResponse(): response=<$response>\n" 
			if $opt_debug;
	} else {
		#
	}

    return($success, $code, $message);
}

1;

