#//////////////////////////////////////////////////////////////////////////////
#//
#//  MemGroup.pl
#//  Win32::AdminMisc Perl extension test script to test the GetGroups() 
#//  function for memory leaks. 
#//  support.
#//
#//  Copyright (c) 1998-2003 Dave Roth
#//  Courtesy of Roth Consulting
#//  http://www.roth.net/
#//
#//  This file may be copied or modified only under the terms of either 
#//  the Artistic License or the GNU General Public License, which may 
#//  be found in the Perl 5.0 source kit.
#//
#//  2003.07.14  :Date
#//  20030714    :Version
#//////////////////////////////////////////////////////////////////////////////
#   Test the GetGroups() function for memory leaks...
#

use strict;
use Win32::AdminMisc;

my $ITERATION = 500 ;
my $Server    = '';

print << 'EOT';

=============================================================================
                        E x p l a n a t i o n

This will execute the GetMachines(), GetGroups() and GetUsers() functions
a total of $ITERATION times each. The amount of memory that the system
allocates for function is monitored.
Ideally a relatively small chunk of memory is allocated for the first call
to the first function. This memory is freed up after the call but the OS will
leave it allocated to this process. Therefore...

It is normal for a small amount to be allocated for the first function call
and the remaining functions to not allocate any additional memory.

If the amount allocated consistantly changes and/or if the changes are
radically large then there may indeed be a memory leak.

An example of a normal run is:
Testing GetMachines() for memory leaks...
  Iteration:  500/ 500 ( 100% Done)  Memory allocated: 28,672 bytes

Testing GetUsers() for memory leaks...
  Iteration:  500/ 500 ( 100% Done)  Memory allocated: 0 bytes

Testing GetGroups() for memory leaks...
  Iteration:  500/ 500 ( 100% Done)  Memory allocated: 0 bytes
=============================================================================
EOT


print "\nTesting GetMachines() for memory leaks...\n";
RunMemoryTest( $ITERATION, \&TestGetMachines );

print "\nTesting GetUsers() for memory leaks...\n";
RunMemoryTest( $ITERATION, \&TestGetUsers );

print "\nTesting GetGroups() for memory leaks...\n";
RunMemoryTest( $ITERATION, \&TestGetGroups );

print "\n\nFinished.\n";

################################################
# Our test functions
#
sub TestGetGroups
{
    my $Result = 1;
    my %List;

    if( ! Win32::AdminMisc::GetGroups( $::Server, GROUP_TYPE_ALL, \%List, "" ) )
    {
        print "\nERROR! GetGroups() returned undef!\n";
        $Result = 0;
    }
    return( $Result );
}

sub TestGetMachines
{
    my $Result = 1;
    my %List;

    if( ! Win32::AdminMisc::GetMachines( $::Server, UF_SERVER_TRUST_ACCOUNT  | UF_INTERDOMAIN_TRUST_ACCOUNT | UF_WORKSTATION_TRUST_ACCOUNT, \%List, "" ) )
    {
        print "\nERROR! GetGroups() returned undef!\n";
        $Result = 0;
    }
    return( $Result );
}

sub TestGetUsers
{
    my $Result = 1;
    my %List;

    if( ! Win32::AdminMisc::GetUsers( $::Server, "", \%List ) )
    {
        print "\nERROR! GetGroups() returned undef!\n";
        $Result = 0;
    }
    return( $Result );
}


################################################
#   Format a number to look pretty
#
sub FormatNumber
{
    my( $Number ) = @_;

    {} while ($Number =~ s/^(-?\d+)(\d{3})/$1,$2/);

    return( $Number );
}

################################################
#   Execute a memory test on a specified function
#
sub RunMemoryTest
{
    my( $iTotalCount, $SubRoutine ) = @_;
    my( $iCount ) = 0;

    my $Mem = GetMemoryLevel();

    while( $iCount++ < $iTotalCount )
    {
        if( ! &$SubRoutine )
        {
            print "\nFAILURE: Function failed. Aborting test.\n";
            return;
        }

        print "\r" . " " x 70 . "\r";

        printf( "  Iteration: % 4d/% 4d (% 3d%% Done)  Memory allocated: %s bytes",
                $iCount,
                $iTotalCount,
                int( $iCount * 100 / $iTotalCount ),
                FormatNumber( GetMemoryLevel() - $Mem ) );
    }

    print "\n";

    return;
}

################################################
#   Retrieve the current memory usage
#
sub GetMemoryLevel
{
    my( %Memory ) = Win32::AdminMisc::GetMemoryInfo();
    return( $Memory{'VirtTotal'} - $Memory{'VirtAvail'} );
}
